#include "FWIIntegralOperator.hpp"

integralOperator::integralOperator(const RKOperator& _RKOp) : RKOp(_RKOp),opt(RKOp.getOpt()) {

}

void integralOperator::assembleInteRHS(Vectors &rhsWF, bool fwd)
{
  RKOp.calculateRHSRK(rhsWF, fwd);
  integrateFWD(rhsWF);
}
void integralOperator::run(const Vectors& WaveFieldPoint, Vectors& WaveFieldEvaluation) const {
    if(!opt.fwd && opt.discrete) {
      Exit("CHECK THIS AGAIN")
//        integrateBWD(WaveFieldPoint);
    }
//    RKOp.runLoop(WaveFieldPoint,WaveFieldEvaluation);
    if(opt.fwd){
      Vector tmp(WaveFieldEvaluation[0]);
      tmp.Clear();
      for (int i =0 ; i<WaveFieldPoint.size(); ++i) {
        RKOp.loopFwdOneStep(WaveFieldPoint,tmp,i);
        WaveFieldEvaluation[i] += tmp;
      }
    }
    if(!opt.discrete && !opt.fwd) {
      Vector tmp(WaveFieldEvaluation[0]);
      tmp.Clear();
      for (int i = WaveFieldPoint.size() - 1; i > 0; --i) {
        RKOp.loopBwdOneStep(WaveFieldPoint, tmp, i);
        WaveFieldEvaluation[i] += tmp;
      }
    }
}

void integralOperator::integrateFWD(Vectors& WaveFieldEvaluation) const{

    for (size_t timeStep = 1;timeStep < WaveFieldEvaluation.size(); timeStep++ ){
      WaveFieldEvaluation[timeStep] += WaveFieldEvaluation[timeStep-1];
    }

}


void integralOperator::integrateBWD(Vectors& WaveFieldEvaluation) const {
    for (int timeStep = WaveFieldEvaluation.size()-2;timeStep  >= 0; timeStep-- ){
        WaveFieldEvaluation[timeStep] += WaveFieldEvaluation[timeStep+1];
    }
}
