#ifndef _FWIRKOPERATOR_H_
#define _FWIRKOPERATOR_H_

#include "lib5_solve/solver/preconditioner/Preconditioner.hpp"
#include "m++.hpp"
//#include "../TimeStepping/src/DGAssemble.hpp"
#include "DGAssemble.hpp"

class inteOptions {
public:
  bool fwd = false;
  bool discrete = false;
  bool debug = false;
  bool includeu0 = false;

  inteOptions(bool _fwd) : fwd(_fwd) {}

  explicit inteOptions(bool _fwd, bool _discrete, bool _debug, bool _includeu0) :
      fwd(_fwd), discrete(_discrete), debug(_debug), includeu0(_includeu0) {}
};

class RKOperator {
    double dt = -1.0;
    Matrix *MKappa = nullptr;
    std::unique_ptr<Matrix> M = nullptr;
    std::unique_ptr<Matrix> A = nullptr;
    bool initialized = false;
    bool cleanUpMatrizes = false;
    std::shared_ptr<DGAssemble> assemble = nullptr;
    RVector b = RVector(1);
    RVector c = RVector(1);
    Matrix B;
    Matrix C;
    const inteOptions opt;
    std::unique_ptr<Preconditioner> PC = nullptr;
    double thresholdIncrease = 0.25;
    double threshold = 0.25;
private:
    void assembleMatrices(DGAssemble &assemble, const Vector &U);

public:
//  RKOperator(const RKOperator &) = delete;

    RKOperator(Matrix *_M, Matrix *_A, const Vector &U, const inteOptions &_opt);

    RKOperator(std::shared_ptr<DGAssemble> _assemble, const Vector &U, const inteOptions &_opt);

    RKOperator(std::shared_ptr<DGAssemble> _assemble,
               const Vector &U, const inteOptions &_opt, int rkorder, double dt);

  void initialize(int rkorder, double dt);

  const Matrix &GetB() const { return B; };

  const Matrix &GetC() const { return C; }

    const Matrix &GetM() const { return *M; }

    const Matrix &GetMKappa() const { return *MKappa; }

    const Matrix &GetA() const { return *A; }

  double getDt() const { return dt; }

  const DGAssemble &GetAssemble() {return *assemble;}

  void runLoop(const Vectors &WaveFieldPoint, Vectors &WaveFieldEvaluation) const;

    void loopFwd(const Vectors &WaveFieldPoint, Vectors &WaveFieldEvaluation) const;

  void loopBwd(const Vectors &WaveFieldPoint, Vectors &WaveFieldEvaluation) const;

  void loopBwdDiscrete(const Vectors &WaveFieldPoint, Vectors &WaveFieldEvaluation) const;

  inteOptions getOpt() const { return opt; }

    void applyM(Vectors &rhsWF) const;

    void calculateRHSRK(Vectors &rhsWF, bool fwd) const;

    ~RKOperator() {
        if (cleanUpMatrizes) {
            delete MKappa;
        }
    }


  void
  loopBwdOneStep(const Vectors &WaveFieldPoint, Vector &WaveFieldEvaluation, int timeStep) const;

  void
  loopFwdOneStep(const Vectors &WaveFieldPoint, Vector &WaveFieldEvaluation, int timeStep) const;
};

#endif
