//
// Created by cr on 13.10.21.
//

#include "FWILinearizedAssembler.hpp"

void FWILinearizedAssembler::assembleSecondLinearizedMGalerkin(const Vector &materialDirection1,
                                                               const Vector &materialDirection2,
                                                               Matrix &mSecond) const {
    mSecond = 0.0;
    //This throws an error for TestSecondDerivative... numL = number of "Damping Dimension"
    if(numL != 0 ){
        Exit("Damping not implemented here");
    }
    for (const auto &elemIt: eleCache) {
        const DGAcousticElement *elem = elemIt.second;
        const Cell &C = elem->GetCell();
        DGRowEntries mSecond_c(mSecond, C, C);

        double rhoPoint = materialPoint(C(),0);
        double rhoDirec1 = materialDirection1(C(), 0);
        double rhoDirec2 = materialDirection2(C(),0);

        double VpPoint = materialPoint(C(),1);
        double VpDirec1 = materialDirection1(C(), 1);
        double VpDirec2 = materialDirection2(C(),1);

        double secondLinFac1 = 2.0*rhoDirec1*rhoDirec2/( rhoPoint*rhoPoint*rhoPoint*VpPoint*VpPoint);
        double secondLinFac2 = 2.0*rhoDirec1*VpDirec2/(VpPoint*VpPoint*VpPoint*rhoPoint*rhoPoint);
        double secondLinFac3 = 2.0*VpDirec1*rhoDirec2/(rhoPoint*rhoPoint*VpPoint*VpPoint*VpPoint);
        double secondLinFac4 = 6.0*VpDirec1*VpDirec2/(VpPoint*VpPoint*VpPoint*VpPoint*rhoPoint);
        double resultingsecondLinFac = secondLinFac1 + secondLinFac2
                                        + secondLinFac3 + secondLinFac4;//Removed Minus (Derivative itself has none)
        int nV = elem->V_dimension();
        int nP = elem->shape_size();
        int p_dim = elem->p_dimension();
        for (int i = 0; i < nP; ++i) {
            for (int q = 0; q < elem->nQ(); ++q) {
                double w = elem->QWeight(q);
                double P_i = elem->Pressure(q, i);
                for (int j = 0; j < nP; ++j) {
                    double P_j = elem->Pressure(q, j);
                    mSecond_c(nV + i, nV + j) += w * P_i * P_j * resultingsecondLinFac;
                }
            }
        }
        //Velocity part is zero
        for (int q = 0; q < elem->nQ(); ++q) {
            double w = elem->QWeight(q);
            for (int i = 0; i < nV; ++i) {
                for (int j = 0; j < nV; ++j) {
                    mSecond_c(i, j) = 0.0;
                }
            }
        }
    }
    mSecond.CreateSparse();
}

void FWILinearizedAssembler::assembleLinearizedMGalerkin(const Vector &materialDirection,
                                                         Matrix &mPrime) const {
  mPrime = 0.0;
  for (const auto &elemIt: eleCache) {
    const DGAcousticElement *elem = elemIt.second;
    const Cell &C = elem->GetCell();
    DGRowEntries mPrime_c(mPrime, C, C);
    int Row = materialPoint.find_row(C()).Id();
    double rhoFac= getRhoFac(Row);
    double rhoDirec = materialDirection(Row, 0);
    std::vector<double> pFac = getpFacs(materialDirection, Row, C);
    int nV = elem->V_dimension();
    int nP = elem->shape_size();
    int p_dim = elem->p_dimension();
    for (int q = 0; q < elem->nQ(); ++q) {
      double w = elem->QWeight(q);
      for (int i = 0; i < nV; ++i) {
        VectorField V_i = elem->Velocity(q, i);
        for (int j = 0; j < nV; ++j) {
          VectorField V_j = elem->Velocity(q, j);
          mPrime_c(i, j) += w * V_i * V_j * rhoDirec*rhoFac;
        }
      }
    }

    for (int i = 0; i < nP; ++i) {
      for (int q = 0; q < elem->nQ(); ++q) {
        double w = elem->QWeight(q);
        double P_i = elem->Pressure(q, i);
        for (int j = 0; j < nP; ++j) {
          double P_j = elem->Pressure(q, j);
          mPrime_c(nV + i, nV + j) += w * P_i * P_j * pFac[0];
          for (int n = 1; n < p_dim; ++n) {
            mPrime_c(nV + n * nP + i, nV + n * nP + j) += w * P_i * P_j * pFac[n];
          }
        }
      }
    }
  }
//  mPrime.CreateSparse();
}

void FWILinearizedAssembler::assembleDGalerkin2(const Vector &materialDirection,
                                                Matrix &diag) const {
  diag = 0.0;
  for (const auto &elemIt: eleCache) {
    auto elem = elemIt.second;
    const Cell &C = elem->GetCell();
    DGRowEntries diag_c(diag, C, C);
    int Row = materialPoint.find_row(elemIt.first).Id();
    std::vector<double> pFac = getpFacs(materialDirection, Row, C);
    int nV = elem->V_dimension();
    int nP = elem->shape_size();
    int p_dim = elem->p_dimension();
    for (int q = 0; q < elem->nQ(); ++q) {
      double w = elem->QWeight(q);
      for (int i = 0; i < nP; ++i) {
        double P_i = elem->Pressure(q, i);
        for (int j = 0; j < nP; ++j) {
          double P_j = elem->Pressure(q, j);
          for (int n = 1; n < p_dim; ++n) {
            diag_c(nV + n * nP + i, nV + n * nP + j) +=
                w * P_i * P_j * pFac[n] * 1.0 / getTauI(C, n);
          }
        }
      }
    }
  }
}

void FWILinearizedAssembler::MPrimeMatVecOneElemDiagonal(const DGAcousticElement &elem,
                                                         const Vector &materialDirection,
                                                         const Vector &JVec,
                                                         const Vector &tmp,
                                                         Vector &Vec) const {
  const Cell &C = elem.GetCell();
  int Row = materialPoint.find_row(C()).Id();
  double rhoFac = getRhoFac(Row);
  double rhoDirec = materialDirection(Row, 0);

  std::vector<double> pFac = getpFacs(materialDirection, Row, C);
  int nV = elem.V_dimension();
  int nP = elem.shape_size();
  int p_dim = elem.p_dimension();
  int Row2 = elem.Row().Id();
//  int Row2 = Vec.find_row(C()).Id();
  for (int j = 0; j < nV; ++j) {
    Vec(Row2, j) = rhoDirec * rhoFac * tmp(Row2, j);
  }
  for (int j = 0; j < nP; ++j) {
    Vec(Row2, j + nV) = tmp(Row2, j + nV) * pFac[0];
    for (int n = 1; n < p_dim; ++n) {
      Vec(Row2, j + nV + n * nP) = pFac[n] * (tmp(Row2, j + nV + n * nP) +
                                              1.0 / problem_fwd->Tau_i(C, C(), n) *
                                              JVec(Row2, j + nV + n * nP));
    }
  }

}

void FWILinearizedAssembler::MPrimeMatVecOneElem(const DGAcousticElement &elem,
                                                 const Vector &materialDirection,
                                                 const Vector &JVec,
                                                 const Vector &tmp,
                                                 Vector &Vec) const {
  const Cell &C = elem.GetCell();
  int Row = materialPoint.find_row(C()).Id();
  double rhoFac = getRhoFac(Row);
  double rhoDirec = materialDirection(Row, 0);
  std::vector<double> pFac = getpFacs(materialDirection, Row, C);
  int nV = elem.V_dimension();
  int nP = elem.shape_size();
  int p_dim = elem.p_dimension();
  int Row2 = elem.Row().Id();
  for (int q = 0; q < elem.nQ(); ++q) {
    double w = elem.QWeight(q);
    VectorField VVec = elem.Velocity(q, tmp);
    for (int j = 0; j < nV; ++j) {
      VectorField V_j = elem.Velocity(q, j);
      Vec(Row2, j) += w * VVec * V_j * rhoDirec * rhoFac;
    }
  }
  for (int q = 0; q < elem.nQ(); ++q) {
    double w = elem.QWeight(q);
    for (int j = 0; j < nP; ++j) {
      double PVec = elem.Pressure_i(q, tmp, 0);
      double P_j = elem.Pressure(q, j);
      Vec(Row2, j + nV) += w * PVec * P_j * pFac[0];
      for (int n = 1; n < p_dim; ++n) {
        double PVecn = elem.Pressure_i(q, tmp, n);
        double PJVec = elem.Pressure_i(q, JVec, n);
        Vec(Row2, j + nV + n * nP) += w * P_j * pFac[n]
                                      * (PVecn + 1.0 / problem_fwd->Tau_i(C, C(), n) * PJVec);
      }
    }
  }
}

void FWILinearizedAssembler::MPrimeMatVec(const Vector &materialDirection,
                                          const Vector &JVec, Vector &Vec) const {
  Vector tmp(Vec);
  Vec.Clear();
  for (const auto &elemIt: eleCache) {
    MPrimeMatVecOneElem(*elemIt.second, materialDirection, JVec, tmp, Vec);
  }
}

void FWILinearizedAssembler::MPrimeMatVecDiagonal(const Vector &materialDirection,
                                                  const Vector &JVec, Vector &Vec) const {
  Vector tmp(Vec);
  Vec.Clear();
  for (const auto &elemIt: eleCache) {
    MPrimeMatVecOneElemDiagonal(*elemIt.second, materialDirection, JVec, tmp, Vec);
  }
}

void
FWILinearizedAssembler::MPrimeMatVecDiagonal(const Vector &materialDirection, const Vector &JVec,
                                             const Vector &Vec_fd,
                                             Vector &result) const {
  for (const auto &elemIt: eleCache) {
    MPrimeMatVecOneElemDiagonal(*elemIt.second, materialDirection, JVec, Vec_fd, result);
  }
}


void FWILinearizedAssembler::assembleM(const Vector &materialDirection,
                                       Matrix &diag) const {
  for (const auto& elemIt : eleCache){
    const DGAcousticElement *elem = elemIt.second;
    const Cell &C = elem->GetCell();
    DGRowEntries mPrime2_c(diag, C, C);
    int Row = materialPoint.find_row(C()).Id();
    double rhoFac = getRhoFac(Row);
    double rhoDirec = materialDirection(Row, 0);
    std::vector<double> pFac = getpFacs(materialDirection, Row, C);
    int nV = elem->V_dimension();
    int nP = elem->shape_size();
    int p_dim = elem->p_dimension();
    for (int i = 0; i < nV; ++i) {
      mPrime2_c(i, i) = rhoDirec * rhoFac;
    }
    for (int i = 0; i < nP; ++i) {
      mPrime2_c(i + nV, i + nV) = pFac[0];
      for (int n = 1; n < p_dim; ++n) {
        mPrime2_c(i + nV + n * nP, i + nV + n * nP) = pFac[n];
      }
    }
  }
    diag.CreateSparse();
}

void
FWILinearizedAssembler::FluxMatrixOneCell(const cell &c, Matrix &A, const Vector &materialDirection,
                                          bool dampingFlux) const {
  double alpha = (numL == 0) ? 1 : problem_fwd->alpha(*c, c());
  const Cell &C = *c;
  const DGAcousticElement *elem = eleCache.at(C());
  int p_dim = elem->p_dimension();
  DGRowEntries A_c(A, C, C);
  double zK = GetZK(C);
  double zKPrime = GetZKPrime(C,materialDirection);
  for (int f = 0; f < c.Faces(); ++f) {
    DGAcousticFaceElement felem(A, C, f, numL);
    int nV = felem.V_dimension();
    int nP = felem.shape_size();
    if (materialPoint.GetMesh().OnBoundary(C, f)) {
      if (!dampingFlux) {
        BFParts bnd(A.GetMesh(), C);
        for (int q = 0; q < felem.nQ(); ++q) {
          double w = felem.QWeight(q);
          VectorField N = felem.QNormal(q);
          Point z = felem.QPoint(q);
          if (bnd[f] == 2 || bnd[f] == 0) {
            for (int j = 0; j < nV; ++j) {
              Scalar VN_j = felem.Velocity(q, j) * N;
              for (int i = 0; i < nV; ++i) {
                Scalar VN_i = felem.Velocity(q, i) * N;
                A_c(i, j) -= w * zKPrime * VN_i * VN_j;
              }
            }
          } else if (bnd[f] == 1) {
            for (int j = 0; j < nP; ++j) {
              double P_j = felem.Pressure(q, j);
              for (int i = 0; i < nP; ++i) {
                double P_i = felem.Pressure(q, i);
                for (int n = 0; n < p_dim; ++n) {
                  for (int k = 0; k < p_dim; ++k) {
                    A_c(nV + n * nP + i, nV + k * nP + j) +=
                        w * 1.0 / squared(zK) * zKPrime * P_i * P_j;
                  }
                }
              }
            }
          }
        }
      }
    } else {

      cell cf = materialPoint.GetMesh().find_neighbour_cell(c, f);
      const Cell &CF = *cf;

      int f1 = materialPoint.GetMesh().find_neighbour_face_id(c.Face(f), cf);
      DGAcousticFaceElement felem_1(A, CF, f1, numL);
      int nV_1 = felem_1.V_dimension();
      int nP_1 = felem_1.shape_size();
      DGRowEntries A_cf(A, C, CF);
      double zKf = GetZK(CF);
      double zKfPrime = GetZKPrime(CF,materialDirection);
            double alpha1 = 1.0 / (zK + zKf);
            double alpha1Prime = -alpha1 * alpha1 * (zKPrime + zKfPrime);
            double alpha2Prime = zKPrime * zKf * alpha1 + zK * zKfPrime * alpha1
                                 + zK * zKf * alpha1Prime;
            double alpha3Prime = zKfPrime * alpha1 + zKf * alpha1Prime;
            double alpha4Prime = zKPrime * alpha1 + alpha1Prime * zK;
            for (int q = 0; q < felem.nQ(); ++q) {
                double w = felem.QWeight(q);
                VectorField N = felem.QNormal(q);
        for (int i = 0; i < nV; ++i) {
          Scalar VN_i = felem.Velocity(q, i) * N;
          for (int j = 0; j < nV; ++j) {
            Scalar VN_j = felem.Velocity(q, j) * N;
            A_c(i, j) -= w * alpha2Prime * VN_i * VN_j;
          }
          for (int j = 0; j < nV_1; ++j) {
            Scalar V1N_j = felem_1.Velocity(q, j) * N;
            A_cf(i, j) += w * alpha2Prime * VN_i * V1N_j;
          }
          for (int j = 0; j < nP; ++j) {
            double P_j = felem.Pressure(q, j);
            for (int n = 0; n < p_dim; ++n) {
              A_c(i, nV + n * nP + j) -= w * alpha4Prime * P_j * VN_i;
            }
          }
          for (int j = 0; j < nP_1; ++j) {
            double P1_j = felem_1.Pressure(q, j);
            for (int n = 0; n < p_dim; ++n) {
              A_cf(i, nV_1 + n * nP_1 + j) += w * alpha4Prime * P1_j * VN_i;
            }
          }
        }
        for (int i = 0; i < nP; ++i) {
          double P_i = felem.Pressure(q, i);
          for (int j = 0; j < nP; ++j) {
            double P_j = felem.Pressure(q, j);
            for (int n = 0; n < p_dim; ++n) {
              for (int k = 0; k < p_dim; ++k) {
                A_c(nV + n * nP + i, nV + k * nP + j) -= w * alpha1Prime * P_i * P_j;
              }
            }
          }
          for (int j = 0; j < nP_1; ++j) {
            double P1_j = felem_1.Pressure(q, j);
            for (int n = 0; n < p_dim; ++n) {
              for (int k = 0; k < p_dim; ++k) {
                A_cf(nV + n * nP + i, nV_1 + k * nP + j) += w * alpha1Prime * P_i * P1_j;
              }
            }
          }
          for (int j = 0; j < nV; ++j) {
            Scalar VN_j = felem.Velocity(q, j) * N;
            for (int n = 0; n < p_dim; ++n) {
              A_c(nV + n * nP + i, j) -= w * alpha3Prime * VN_j * P_i;
            }
          }
          for (int j = 0; j < nV_1; ++j) {
            Scalar V1N_j = felem_1.Velocity(q, j) * N;
            for (int n = 0; n < p_dim; ++n) {
              A_cf(nV + n * nP + i, j) += w * alpha3Prime * V1N_j * P_i;
            }
          }
        }
      }
    }
  }
  return;
}

void FWILinearizedAssembler::assembleLinearizedFluxMatrix(const Vector &materialDirection,
                                                          Matrix &A) const {
  A = 0.0;
  bool dampingFlux = false;
  Config::Get("dampingFlux", dampingFlux);
  const Mesh &M = A.GetMesh();
  for (cell c = A.cells(); c != A.cells_end(); ++c) {
    FluxMatrixOneCell(c, A, materialDirection, dampingFlux);
  }
  A.CreateSparse();
}

void FWILinearizedAssembler::assembleD(Matrix &diag) {
  diag = 0;
  for (const auto& elemIt : eleCache){
    const DGAcousticElement *elem = elemIt.second;
    const Cell &C = elem->GetCell();
    DGRowEntries diag_c(diag, C,C);
    int nP = elem->shape_size();
    int p_dim = elem->p_dimension();
    int nV = elem->V_dimension();
    for (int i = 0; i < nP; ++i) {
      for (int n = 1; n < p_dim; ++n) {
        diag_c(nV + n * nP + i, nV + n * nP + i) = 1.0 / problem_fwd->Tau_i(C, C(), n);
      }
    }
  }
}

std::vector<double> FWILinearizedAssembler::getpFacs(const Vector &materialDirection,
                                                     int Row, const Cell &C) const {

    if (problem_fwd->getParametrization() == "taup_positive") {
        double taup_direc = (numL > 0) ? materialDirection(Row, 2) : 0;
        return pFacsTaupPositive(getRho(Row), getVP(Row), getTauP(Row),
                                 materialDirection(Row, 0), materialDirection(Row, 1),
                                 taup_direc,
                                 getAlpha(C), numL);
    } else if (problem_fwd->getParametrization() == "taup") {
        double taup_direc = (numL > 0) ? materialDirection(Row, 2) : 0;
        return pFacsTaup(getRho(Row), getVP(Row), getTauP(Row),
                         materialDirection(Row, 0),
                         materialDirection(Row, 1),
                         taup_direc,
                         getAlpha(C), numL);
    } else if (problem_fwd->getParametrization() == "native") {
        std::vector<double> vec;
        for (int i = 0; i < numL + 1; ++i) {
            double kappa_i = materialPoint(Row, 1 + i);
            double kappa_iDirec = materialDirection(Row, 1 + i);
            vec.push_back(-(1.0/(kappa_i * kappa_i)) * kappa_iDirec);
        }
        return vec;
        Exit("not implemented!");
    } else {
        Exit("wrong parametrization");
    }
}

double FWILinearizedAssembler::getAlpha(const Cell &C) const {
  return (numL == 0) ? 0.0 : problem_fwd->alpha(C, C());
}

double FWILinearizedAssembler::getTauI(const Cell &C, int n) const {
  return problem_fwd->Tau_i(C, C(), n);
}
double FWILinearizedAssembler::getRhoFac(int Row) const {
  if (problem_fwd->getParametrization() == "taup_positive") {
    return exp(materialPoint(Row, 0));
  } else if (problem_fwd->getParametrization() == "taup" ||
             problem_fwd->getParametrization() == "native") {
    return 1.0;
  } else {
    Exit("wrong parametrization");
  }
}

double FWILinearizedAssembler::getRho(int Row) const {
  if (problem_fwd->getParametrization() == "taup_positive") {
    return exp(materialPoint(Row, 0));
  } else if (problem_fwd->getParametrization() == "taup") {
    return materialPoint(Row, 0);
  } else if (problem_fwd->getParametrization() == "native") {
    return materialPoint(Row, 0);
  } else {
    Exit("wrong parametrization");
  }

}

double FWILinearizedAssembler::getVP(int Row) const {
  if (problem_fwd->getParametrization() == "taup_positive") {
    return exp(materialPoint(Row, 1));
  } else if (problem_fwd->getParametrization() == "taup") {
    return materialPoint(Row, 1);
  } else {
    Exit("wrong parametrization");
  }
}

double FWILinearizedAssembler::getTauP(int Row) const {
  if (problem_fwd->getParametrization() == "taup_positive") {
    return (numL == 0) ? 0.0 : exp(materialPoint(Row, 2));
  } else if (problem_fwd->getParametrization() == "taup") {
    return (numL == 0) ? 0.0 : materialPoint(Row, 2);
  } else {
    Exit("wrong parametrization");
  }
}

std::vector<double>
pFacsTaupPositive(double rho, double vP, double tauP, double rhoDirec, double vPDirec,
                  double tauPDirec,
                  double alpha, int numL) {
  double tauPFac = (numL == 0) ? 0.0 : alpha * tauP / (squared(vP) * rho) * tauPDirec;
  double rhoFac = (numL == 0) ? (-1.0 / (squared(vP) * rho) * rhoDirec)
                              : -(1.0 + alpha * tauP) / (squared(vP) * rho) * rhoDirec;

  double vPFac = (numL == 0) ? -2.0 / (squared(vP) * rho) * vPDirec
                             : -2.0 * (1.0 + alpha * tauP) / (squared(vP) * rho) * vPDirec;
  double tau_pfac2 = (numL == 0) ? 0.0 : -tauPDirec / (squared(vP) * rho * tauP);
  std::vector<double> vec;
  vec.push_back(rhoFac + vPFac + tauPFac);
  if (numL == 0) {
    vec.push_back(0.0);
  } else {
    for (int i = 0; i < numL; ++i) {
      vec.push_back(rhoFac / tauP + vPFac / tauP + tau_pfac2);
    }
  }
  return vec;
}

std::vector<double>
pFacsTaup(double rho, double vP, double tauP, double rhoDirec, double vPDirec, double tauPDirec,
          double alpha, int numL) {
  double tauPFac = (numL == 0) ? 0.0 : alpha / (squared(vP) * rho) * tauPDirec;
  double rhoFac = (numL == 0) ? (-1.0 / (squared(vP * rho)) * rhoDirec)
                              : -(1 + alpha * tauP) / (squared(rho * vP)) * rhoDirec;

  double vPFac = (numL == 0) ? -2.0 / (vP * vP * vP * rho) * vPDirec
                             : -2.0 * (1.0 + alpha * tauP) / (vP * vP * vP * rho) * vPDirec;
  double tau_pfac2 = (numL == 0) ? 0.0 : -tauPDirec / (squared(vP * tauP) * rho);
  std::vector<double> vec;
  vec.push_back(rhoFac + vPFac + tauPFac);
  if (numL == 0) {
    vec.push_back(0.0);
  } else {
    for (int i = 0; i < numL; ++i) {
      vec.push_back(rhoFac / tauP + vPFac / tauP + tau_pfac2);
    }
  }
  return vec;
}