//
// Created by cr on 24.08.22.
//

#include "FWIRhsAssemblerAdjoint.hpp"

DGVAcousticRHSAssemblerAdjoint::DGVAcousticRHSAssemblerAdjoint(const FWIObservationOperatorNEW &Obs) :
        Spec(Obs.GetSpaceOp().GetSpecification()), Obs(Obs) {
    dt = Spec.timeDelta();
    T = Spec.endTime();
    int rkorder = -2;
    Config::Get("rkorder", rkorder);
    if (!(rkorder == -2 || rkorder == -302 || rkorder == -1002 || rkorder == 1002 || rkorder == -2002)) {
        Exit("Debug this method for your rkorder"); //rkorder = -102 does NOT(!!!) work
    }
}

void DGVAcousticRHSAssemblerAdjoint::Initialize(Vector &u) const {}

void DGVAcousticRHSAssemblerAdjoint::setSeis(const FWISeismogram *s) const {
    DGVAcousticRHSAssemblerAdjoint::Seis = s;
}

void DGVAcousticRHSAssemblerAdjoint::RHS(double t_, Vector &b) const {
    b.Clear();
    double t = T - t_;
    double ratio = t / dt;
    int maxstep = static_cast<int>(round(T / dt));
    int findstep = static_cast<int>(round(ratio));
    for (const Point &receiver: Spec) {
        if (!Obs.GetSpaceOp().GetSpaceSupport().contains(receiver)) {
            continue;
        }
        const FWIReceiverTimeseries &timeseries = Seis->getTimeSeries(receiver);
        FWIState meas(Seis->getStateComponents());
        for (int step = 0; step < 5; ++step) {
            if (findstep - 2 + step < 0 || findstep - 2 + step > maxstep) { continue; }
            const std::pair<double, FWIState *> &ts_entry = timeseries[findstep - 2 + step];
            double norm = abs(ts_entry.first - t) / dt;
            if (norm > 1) continue;
            meas += (1 - norm) * (*ts_entry.second) * Obs.GetTimeOp()(ts_entry.first);
        }
        Obs.GetMask().adjoint(meas);
        Obs.GetSpaceOp().SetSpaceValue(meas, receiver, Obs.GetSpaceOp().GetSpaceSupport().at(receiver), b);
    }
}
