//
// Created by cr on 24.08.22.
//

#include "FWIRhsAssemblerLinearized.hpp"

DGVAcousticRHSAssemblerLinearizedFirstOrder::DGVAcousticRHSAssemblerLinearizedFirstOrder(const Vectors &U,
                                                                                         std::shared_ptr<WaveProblemDGTSVA> problem_fwd_,
                                                                                         const std::unordered_map<Point, DGAcousticElement *> &eleCache_,
                                                                                         const Vector &materialPoint_, const Vector &m_test,
                                                                                         double dt, bool discrete,
                                                                                         PMLHandler *handler)
        : DGVAcousticRHSAssembler(), U_m(U), PMLH(handler) {
    Config::Get("rkorder", rkorder);
    if (!(rkorder == -102 || rkorder == -302 || rkorder == -2 || rkorder == -1002 || rkorder == 1002 ||
          rkorder == -2002)) {
        Exit("make sure this works for your chosen rkorder");
    }
    FWILinearizedAssembler linearizedAssembler(problem_fwd_, eleCache_, materialPoint_);
    Matrix Mprime(U_m[0]);
    Matrix Dprime(U_m[0]);

    linearizedAssembler.assembleLinearizedMGalerkin(m_test, Mprime);
    if (problem_fwd_->getNumL() > 0)linearizedAssembler.assembleDGalerkin2(m_test, Dprime);
    else {
        Dprime = 0.0;
    }
    Mprime *= 1.0 / dt;
    if (discrete) {
        Matrix APrime(U_m[0]);
        APrime = 0.0;
        Vector lel(m_test);
        lel.Accumulate();
        linearizedAssembler.assembleLinearizedFluxMatrix(lel, APrime);
        Dprime += (-1.0) * APrime;
    }
    Dprime *= -0.5;
    A = std::make_unique<Matrix>(Dprime);
    B = std::make_unique<Matrix>(Dprime);
    (*A) += (-1.0) * Mprime;
    (*B) += (1.0) * Mprime;
    A->CreateSparse();
    B->CreateSparse();
    if (handler) {
        if (handler->isPML()) {
            VecTmp = std::make_unique<Vector>(U[0]);
            VecTmp->Clear();
        }
    }
}

void DGVAcousticRHSAssemblerLinearizedFirstOrder::RHS(double t, Vector &b) const {
    if (PMLH->isPML()) {
        *VecTmp = (*A) * U_m[step + 1] + (*B) * U_m[step];
        b.Clear();
        PMLWaveVectorFWI::transfer(PMLH->GetNonPMLToPML(), *VecTmp, b);
    } else {
        b = (*A) * U_m[step + 1] + (*B) * U_m[step];
    }
    if (rkorder == -102 || rkorder == -302) {
        if (increaseStep) {
            step++;
            increaseStep = false;
        } else {
            increaseStep = true;
        }
    } else if (rkorder == -2 || rkorder == -1002 || rkorder == 1002 || rkorder == -2002) {
        step++;
    } else {
        Exit("check this for this rkorder")
    }
}


DGVAcousticRHSAssemblerLinearizedSecondOrder::DGVAcousticRHSAssemblerLinearizedSecondOrder(
        const Vectors &U, std::shared_ptr<WaveProblemDGTSVA> problem_fwd_,
        const std::unordered_map<Point, DGAcousticElement *> &eleCache_, const Vector &materialPoint_,
        const Vector &m_test1, const Vector &m_test2, double dt_, bool discrete,
        FWILinOpRedFirstOrder &LinOp): DGVAcousticRHSAssembler(), U_m(U), dt(dt_) {
    // Do I need: bool discrete; Removed argument int *handler and PMLH(handler)
    EleCache = &eleCache_;
    Config::Get("rkorder", rkorder);
    if (!(rkorder == -102 || rkorder == -302 || rkorder == -2 || rkorder == -1002 || rkorder == 1002 ||
          rkorder == -2002)) {
        Exit("make sure this works for your chosen rkorder");
    }
    //!!!Is initialization of u_prime_direction correct? Need correct dimension (I think)
    u_prime_direction1 = std::make_unique<Vectors>(U_m);
    u_prime_direction2 = std::make_shared<Vectors>(U_m);

    LinOp.Solve(m_test1, *u_prime_direction1);
    LinOp.Solve(m_test2, *u_prime_direction2);//reused for Psi in CG_Reginn Method!

    //Material Point (integer!!) does what?
    FWILinearizedAssembler linearizedAssembler(problem_fwd_, eleCache_, materialPoint_);
    //!!!??? For size of Matrix do I need to initialize members here? And How
    MPrime1 = std::make_unique<Matrix>(U_m[0]);
    MPrime2 = std::make_unique<Matrix>(U_m[0]);
    MDoublePrime = std::make_unique<Matrix>(U_m[0]);
    //Matrices will be cleared / = 0.0; in methods

    linearizedAssembler.assembleLinearizedMGalerkin(m_test1, *MPrime1);
    linearizedAssembler.assembleLinearizedMGalerkin(m_test2, *MPrime2);
    linearizedAssembler.assembleSecondLinearizedMGalerkin(m_test1, m_test2, *MDoublePrime);
    //!!Make Sparse??
    MPrime1->CreateSparse();
    MPrime2->CreateSparse();
    MDoublePrime->CreateSparse();
}

void DGVAcousticRHSAssemblerLinearizedSecondOrder::RHS(double t, Vector &b) const {
    double dt_inv = 1.0/dt;
    b = -dt_inv * (*MDoublePrime * ((U_m[step+1] - U_m[step]) )
            + (*MPrime1) * ((*u_prime_direction2)[step+1] - (*u_prime_direction2)[step])
            + (*MPrime2) * ((*u_prime_direction1)[step+1] - (*u_prime_direction1)[step]));

    //Internal step organise system from First Order
    if (rkorder == -102 || rkorder == -302) {
        if (increaseStep) {
            step++;
            increaseStep = false;
        } else {
            increaseStep = true;
        }
    } else if (rkorder == -2 || rkorder == -1002 || rkorder == 1002 || rkorder == -2002) {
        step++;
    } else {
        Exit("check this for this rkorder")
    }

}

//Automatic generated one: const std::shared_ptr<Vectors> &DGVAcousticRHSAssemblerLinearizedSecondOrder::getUPrimeDirection2() const {
std::shared_ptr<Vectors> DGVAcousticRHSAssemblerLinearizedSecondOrder::getUPrimeDirection2() const {
    return u_prime_direction2;
}
std::shared_ptr<Vectors> &DGVAcousticRHSAssemblerLinearizedSecondOrder::getUPrimeDirection2Ref() {
    return u_prime_direction2;
}
