//
// Created by cr on 24.08.22.
//

#ifndef FWI_FWIRHSASSEMBLERLINEARIZED_HPP
#define FWI_FWIRHSASSEMBLERLINEARIZED_HPP

#include "m++.hpp"
#include "FWIAbstractInterfaceDGTSVA.hpp"
#include "DGAcousticElement.hpp"
#include "DGVAcousticRHSAssembler.hpp"
#include "FWILinearizedAssembler.hpp"
#include "Reduced/FWILinOpRedFirstOrder.hpp"
#include "../Reduced/FWILinOpRedFirstOrder.hpp"
#include "Reduced/FWILinOpRed.hpp"

class FWILinOpRedFirstOrder;

class DGVAcousticRHSAssemblerLinearizedFirstOrder : public DGVAcousticRHSAssembler {
    const Vectors &U_m;
    std::unique_ptr<PMLWaveVectorFWI> PMLVec;
    std::unique_ptr<Matrix> A;
    std::unique_ptr<Matrix> B;
    mutable int step = 0;
    mutable bool increaseStep = false;
    PMLHandler *PMLH;
    std::unique_ptr<Vector> VecTmp;

public:
    int rkorder = -1;

    DGVAcousticRHSAssemblerLinearizedFirstOrder(const Vectors &U, std::shared_ptr<WaveProblemDGTSVA> problem_fwd_,
                                                const std::unordered_map<Point, DGAcousticElement *> &eleCache_,
                                                const Vector &materialPoint_, const Vector &m_test, double dt, bool discrete,
                                                PMLHandler *handler);

    void RHS(double t, Vector &b) const override;

    void RHSSecondOrder(double t, Vector &b) const;

    void Initialize(Vector &) const override {}

    ~DGVAcousticRHSAssemblerLinearizedFirstOrder() = default;

};


/*
 * Erbt von class DGVAcousticRHSAssembler {
protected:
    const std::unordered_map<Point, DGAcousticElement *> *EleCache{};
public:
    DGVAcousticRHSAssembler() = default;
    void SetElementCache(const std::unordered_map<Point, DGAcousticElement *> *_EleCache);
    virtual void RHS(double t, Vector &b) const = 0;
    virtual void Initialize(Vector &u) const = 0;
 */

/*
 * This class can ultimately return a RightHandSide (RHS) with its method RHS.
 * Second Derivative \Phi''(Material)[MaterialDirection1, MatDir2] for \Phi = F(V(.))
 * has right hand side
 * V''(Mat)[MatDir1, MatDir2] U_m' + V'(Mat)[MatDir1] r1' + V'(Mat)[MatDir1] r2'
 * for r1/2 = \Phi'(Mat)[MatDir1/2]
 */
class DGVAcousticRHSAssemblerLinearizedSecondOrder : public DGVAcousticRHSAssembler {
    const Vectors &U_m;
    //Naive Save everything, but the finite Differences. (Then in RHS do logik).
    std::unique_ptr<Vectors> u_prime_direction1;
    std::shared_ptr<Vectors> u_prime_direction2;//MUss abgespeichert werden, für Psi, kann also auch nicht gemerged werden mit u_prime_direction1
    std::unique_ptr<Matrix> MPrime1;
    std::unique_ptr<Matrix> MPrime2;
    std::unique_ptr<Matrix> MDoublePrime;
    double dt;
    mutable int step = 0;
    mutable bool increaseStep = false;
    PMLHandler *PMLH;
    std::unique_ptr<PMLWaveVectorFWI> PMLVec;

public:
    int rkorder = -1;

    DGVAcousticRHSAssemblerLinearizedSecondOrder(
            const Vectors &U, std::shared_ptr<WaveProblemDGTSVA> problem_fwd_,
            const std::unordered_map<Point, DGAcousticElement *> &eleCache_, const Vector &materialPoint_,
            const Vector &m_test1, const Vector &m_test2, double dt, bool discrete,
            FWILinOpRedFirstOrder &LinOp);

    void RHS(double t, Vector &b) const override;

    void Initialize(Vector &) const override {}

    ~DGVAcousticRHSAssemblerLinearizedSecondOrder() = default;

    /*
     * Used for Psi'(Mat)[Dir2] to not have to calculate it twice.
     */
    std::shared_ptr<Vectors> getUPrimeDirection2() const;
    //The following one was automatically generated
    std::shared_ptr<Vectors> &getUPrimeDirection2Ref();
};

#endif //FWI_FWIRHSASSEMBLERLINEARIZED_HPP
