#ifndef _FWI_ABSTRACTINTERFACEDGTSVA_
#define _FWI_ABSTRACTINTERFACEDGTSVA_

#include "DGViscoAcousticProblem.hpp"
#include "ForwardModels/FWIForwardModel.hpp"

class FWIMaterialVAcousticDGTS : public FWIMaterial {
    int numL;
    int paramComps;
    int para;
    std::string parametrization;
protected:
    bool initial_value;
public:
    FWIMaterialVAcousticDGTS(int _numL = 1)
            : numL(_numL), initial_value(false), parametrization(""), paramComps(0) {}

    virtual void setInitialValueFlag(bool ivf) { initial_value = ivf; }

    int getNumL() const { return numL; }

    int getParamComps() const {
        if (parametrization == "") {
            Exit("make sure you set the parametrization in the problem!!")
        }
        return paramComps;
    }

    virtual double Kappa_i(const Point &, int) const {
        Exit("not specified");
        return 0.0;
    }

    virtual double Rho(const Point &) const {
        Exit("not specified");
        return 0.0;
    }

    virtual double Kappa(const Cell &C, const Point &x) const {
        double kappa = 0.0;
        for (int i = 0; i <= numL; ++i) {
            kappa += Kappa_i(C, x, i);
        }
        return kappa;
    }

    virtual double Kappa_i(const Cell &, const Point &x, int i) const {
        return Kappa_i(x,
                       i);
    }

    virtual double invKappa_i(const Cell &,
                              const Point &x,
                              int i) const { return invKappa_i(x, i); }

    virtual double invKappa_i(const Point &x, int i) const { return 1.0 / Kappa_i(x, i); }

    virtual double Tau_i(const Cell &C, const Point &x, int i) const {
        Exit("not specified");
        return 0.0;
    }

    virtual double invTau_i(const Cell &C, const Point &x, int i) const {
        return 1.0 / Tau_i(C, x, i);
    }

    virtual double Rho(const Cell &, const Point &x) const { return Rho(x); }

    virtual double alpha(const Cell &,
                         const Point &) const {
        Exit("not implemented here!");
        return 0.;
    }
    virtual double alpha1() const {
        Exit("not implemented here!");
        return 0.;
    }
    virtual double alpha2() const {
        Exit("not implemented here!");
        return 0.;
    }

    virtual double V_p(const Cell &, const Point &) const {
        Exit("not implemented here!");
        return 0.;
    }

    virtual double Tau_p(const Cell &,
                         const Point &) const {
        Exit("not implemented here!");
        return 0.;
    }

    void setParametrization(string s);

    virtual const std::string &getParametrization() const { return parametrization; }

    ~FWIMaterialVAcousticDGTS() {}
};

class FWIBndConditionsVAcoustic : public FWIBndConditions {
    int dim;
    int numL;
public:
    FWIBndConditionsVAcoustic(int _dim, int _numL)
        : FWIBndConditions(), dim(_dim), numL(_numL) {
    }

    int getStateComps() const { return dim + 1 + numL; }

    int getDim() const { return dim; }

    double Dirichlet_i(double t, const Point &z, int i) const {
        return ut(t,
                  z,
                  dim + i);
    }

    double Neumann(double t, const Cell &C, const Point z, const VectorField &N) const;

    virtual double ut(double, const Point &, int) const {
        Exit("not implemented");
        return 0.0;
    }

    virtual double ut(double t, const Cell &, const Point &z, int i) const {
        return ut(t,z,i);
    }

    virtual bool exact() const {
        Exit("not implemented");
        return false;
    }
};

class FWIRhsVAcoustic : public FWIRhs {
public:
    FWIRhsVAcoustic() : FWIRhs() {}

    virtual int getNrPointSrc() const {
        Exit("CIAO");
        return 0;
    }

    virtual Point getPointSrcLoc(int /*nr*/) const {
        Exit("Not implemented");
        return Origin;
    }

    std::vector<Point> &GetPointSrcLocOfShot(int) const {Exit("implement this")};

    virtual bool RHS(double t) const {
        Exit("not specified if there's a RHS");
        return false;
    }

    virtual VectorField ForceV(double, const Cell &, const Point &, int = -1) const {
        Exit("Not implemented for this RHS");
        return zero;
    }

    virtual double ForceP_i(double, const Cell &, const Point &, int, int = -1) const {
        Exit("Not implemented for this RHS");
        return 0.0;
    }

    virtual void Force(VectorField &V_rhs,
                       RVector &Force,
                       double t,
                       const Cell &C,
                       const Point &z,
                       int q = -1) const {
        V_rhs = ForceV(t, C, z, q);
        for (int n = 0; n < Force.size(); ++n) {
            Force[n] = ForceP_i(t, C, z, n, q);
        }
    }
};

class WaveProblemDGTSVA
        : public FWIBndConditionsVAcoustic, public FWIRhsVAcoustic, public FWIMaterialVAcousticDGTS {
protected:
public:
    WaveProblemDGTSVA(int _numL, int _dim) : FWIBndConditionsVAcoustic(_dim, _numL),
                                             FWIRhsVAcoustic(),
                                             FWIMaterialVAcousticDGTS(_numL) {}

    int GetStateComponents() {
        return getDim() + 1 + getNumL();
    }

    using FWIRhsVAcoustic::getNrPointSrc;

    virtual const char *Name() const {
        Exit("NO NAME");
        return "";
    }
};


#endif
