//
// Created by cr on 15.10.21.
//

#ifndef FWI_FWIADJOINTCALCULATOR_HPP
#define FWI_FWIADJOINTCALCULATOR_HPP


#include <Vector.hpp>
#include <utility>
#include "DGElement.hpp"
#include "FWIHelpers.hpp"
#include "FWIAbstractInterfaceDGTSVA.hpp"
#include "DGAcousticElement.hpp"
#include "Assembler/FWILinearizedAssembler.hpp"

class FWIAdjointCalculator {
    void updateNative(int Row, double integral_rho, double integral_kappa_0,
                      const RVector &integral_M,
                      const RVector &integral_D, Vector &Adjoint_m_mv) const;

  void updateTaup(int Row, double integral_rho,
                  double integral_kappa_0, const RVector &integral_M,
                  const RVector &integral_D, Vector &Adjoint_m_mv,
                  const Cell &C) const;

  void updateTaupPositive(int Row, double integral_rho,
                          double integral_kappa_0, const RVector &integral_M,
                          const RVector &integral_D, Vector &Adjoint_m_mv,
                          const Cell &C) const;

    const Vector &Material;
    const std::unordered_map<Point, DGAcousticElement *> &eleCache;
    reconstructionContainer recon;
  const std::string &param = "";
  int numL = -1;
    double dt;
    const std::shared_ptr<WaveProblemDGTSVA> &problem_fwd;
public:
    FWIAdjointCalculator(const Vector &Material_,
                         const std::string &param_,
                         const std::unordered_map<Point, DGAcousticElement *> &eleCache_,
                         reconstructionContainer recon_,
                         int numL_, double dt_,
                         const std::shared_ptr<WaveProblemDGTSVA> &
                         problem_fwd_) :
            Material(Material_), eleCache(eleCache_), recon(recon_), param(param_),
            numL(numL_), dt(dt_), problem_fwd(problem_fwd_) {}

    FWIAdjointCalculator(const Vector &Material_,
                         const std::unordered_map<Point, DGAcousticElement *> &eleCache_,
                         const std::shared_ptr<WaveProblemDGTSVA> &problem_fwd_) :
            param(problem_fwd_->getParametrization()),
            Material(Material_),
            eleCache(eleCache_),
            problem_fwd(problem_fwd_) {
        Config::Get("dt", dt);
        Config::Get("numL", numL);
        if (problem_fwd->getParametrization() == "taup_positive") {
            Config::Get("FWIReconstructRho", recon.Rho);
            Config::Get("FWIReconstructV_P", recon.Vp);
            Config::Get("FWIReconstructTauP", recon.Taup);
        }
    }

  void calculateL2AdjointOneStep(const Vector &U_m_bwd,
                                 const Vector &U_m_fd,
                                 const Vector &LambdaStep, Vector &Adjoint_m_mv) const;
  void PseudoHessianHelper(const Vector &U_m_bwd, const Vector &U_m_fd,
                           const Vector &LambdaStep,
                           Vector &Adjoint_m_mv) const;
  void calculateL2AdjointOneStepSO(const Vector &u_m_fd, const Vector &u_prime_fd,
                                   const Vector &LambdaStep, const Vector &zStep, const Vector &MaterialDirection,
                                   Vector &Adjoint_m_mv, const bool Psi) const;
  void MaterialFactorsSO(int Row, const Vector &MaterialDirection, std::vector<double> &factors) const;

};

#endif //FWI_FWIADJOINTCALCULATOR_HPP
