//
// Created by lionbart on 13.03.23.
//

#include "FWIAdjointCalculatorSO.h"

void FWIAdjointCalculatorSO::updateTaup(int Row, double integral_rho,
                                      double integral_kappa_0, const RVector &integral_M,
                                      const RVector &integral_D, Vector &Adjoint_m_mv,
                                      const Cell &C) const {
//    double alpha = (numL == 0) ? 0 : problem_fwd->alpha(C, C());
    double alpha = 0.0;
    double rho = Material(Row, 0);
    double V_p = Material(Row, 1);
    double grad_taup = 0.0;
    double tau_p = 0.0;

    double rhofac = -(1 + alpha * tau_p) / (squared(rho * V_p));
    double V_pfac = (-2.0) * (1.0 + alpha * tau_p) / (V_p * V_p * V_p * rho);
    double grad_rho = integral_rho;
    grad_rho += rhofac * integral_kappa_0;
    double grad_V_P = V_pfac * integral_kappa_0;

    if (recon.Rho) {
        Adjoint_m_mv(Row, 0) += grad_rho;
    }
    if (recon.Vp == 1) {
        Adjoint_m_mv(Row, 1) += grad_V_P;
    }
    if (recon.Taup == 1 && numL > 0) {
        Adjoint_m_mv(Row, 2) += grad_taup;
    }
}


void FWIAdjointCalculatorSO::calculateL2AdjointOneStep(const Vector &U_m_bwd, const Vector &U_m_fd,
                                                     const Vector &LambdaStep,
                                                     Vector &Adjoint_m_mv) const {
    RVector integral_M(0.0, numL);
    RVector integral_D(0.0, numL);
    for (const auto& elemIt : eleCache){
        integral_M = 0.0;
        integral_D = 0.0;
        const DGAcousticElement &elem = *elemIt.second;
        const Cell &C = elem.GetCell();
        int Row = Adjoint_m_mv.find_row(C()).Id();

        double a_inv = 1.0 / elem.Area();
        double cc_rho = 0.0;
        double cc_kappa_0 = 0.0;
        RVector cc_M(0.0, numL + 1); // with partial_t
        RVector cc_D(0.0, numL + 1); // without partial_t
        for (int q = 0; q < elem.nQ(); ++q) {
            double w = elem.QWeight(q);
            const VectorField &V_fd = elem.Velocity(q, U_m_fd);
            const VectorField &V_lamb = elem.Velocity(q, LambdaStep);
            cc_rho += w * V_fd * V_lamb;
            double p_fd_0 = elem.Pressure_i(q, U_m_fd, 0);
            double p_lamb_0 = elem.Pressure_i(q, LambdaStep, 0);
            cc_kappa_0 += w * p_fd_0 * p_lamb_0;
            for (int i = 1; i <= numL; ++i) {
                double p_lamb_i = elem.Pressure_i(q, LambdaStep, i);
                cc_M[i - 1] += w * elem.Pressure_i(q, U_m_fd, i) * p_lamb_i;
                cc_D[i - 1] += w * elem.Pressure_i(q, U_m_bwd, i) * p_lamb_i;
            }
        }
        double integral_rho = cc_rho * a_inv * dt;
        double integral_kappa_0 = cc_kappa_0 * a_inv * dt;
        for (int i = 1; i <= numL; ++i) {
            integral_M[i - 1] += cc_M[i - 1] * a_inv * dt;
            integral_D[i - 1] += cc_D[i - 1] * a_inv * dt;
        }
        if (param == "taup") {
            updateTaup(Row, integral_rho, integral_kappa_0, integral_M, integral_D, Adjoint_m_mv, C);
        } else {
            Exit("Invalid parameterization");
        }
    }
}