//
// Created by lionbart on 13.03.23.
//

#ifndef FWI_FWIADJOINTCALCULATORSO_H
#define FWI_FWIADJOINTCALCULATORSO_H


#include <Vector.hpp>
#include <utility>
#include "DGElement.hpp"
#include "FWIHelpers.hpp"
#include "FWIAbstractInterfaceDGTSVA.hpp"
#include "DGAcousticElement.hpp"

class FWIAdjointCalculatorSO {
    const Vector &Material;
    const Vector &MaterialDirection;
    const std::unordered_map<Point, DGAcousticElement *> &eleCache;
    reconstructionContainer recon;
    const std::string &param = "";
    int numL = -1;
    double dt;
    const std::shared_ptr<WaveProblemDGTSVA> &problem_fwd;

    void updateTaup(int Row, double integral_rho,
                    double integral_kappa_0, const RVector &integral_M,
                    const RVector &integral_D, Vector &Adjoint_m_mv,
                    const Cell &C) const;
public:
    FWIAdjointCalculatorSO(const Vector &Material_, const Vector &MaterialDirection_,
                         const std::string &param_,
                         const std::unordered_map<Point, DGAcousticElement *> &eleCache_,
                         reconstructionContainer recon_,
                         int numL_, double dt_,
                         const std::shared_ptr<WaveProblemDGTSVA> &
                         problem_fwd_ = nullptr) :
            Material(Material_), MaterialDirection(MaterialDirection_), eleCache(eleCache_), recon(recon_), param(param_),
            numL(numL_), dt(dt_), problem_fwd(problem_fwd_) {}

    FWIAdjointCalculatorSO(const Vector &Material_, const Vector &MaterialDirection_,
                         const std::unordered_map<Point, DGAcousticElement *> &eleCache_,
                         const std::shared_ptr<WaveProblemDGTSVA> &problem_fwd_) :
            param(problem_fwd_->getParametrization()),
            Material(Material_), MaterialDirection(MaterialDirection_),
            eleCache(eleCache_),
            problem_fwd(problem_fwd_) {
        Config::Get("dt", dt);
        Config::Get("numL", numL);
        if (problem_fwd->getParametrization() == "taup_positive") {
            Config::Get("FWIReconstructRho", recon.Rho);
            Config::Get("FWIReconstructV_P", recon.Vp);
            Config::Get("FWIReconstructTauP", recon.Taup);
        }
    }

    void calculateL2AdjointOneStep(const Vector &U_m_bwd,
                                   const Vector &U_m_fd,
                                   const Vector &LambdaStep, Vector &Adjoint_m_mv) const;
//    void PseudoHessianHelper(const Vector &U_m_bwd, const Vector &U_m_fd,
//                             const Vector &LambdaStep,
//                             Vector &Adjoint_m_mv) const;
//    void updateNative(int Row, double integral_rho, double integral_kappa_0,
//                      const RVector &integral_M,
//                      const RVector &integral_D, Vector &Adjoint_m_mv) const;
//    void updateTaupPositive(int Row, double integral_rho,
//                            double integral_kappa_0, const RVector &integral_M,
//                            const RVector &integral_D, Vector &Adjoint_m_mv,
//                            const Cell &C) const;
};

#endif //FWI_FWIADJOINTCALCULATORSO_H