//
// Created by cr on 23.08.22.
//

#include "FWIDGVAcousticMaterialFromVector.hpp"

void FWIDGVAcousticMaterialFromVector::fillVector(const WaveProblemDGTSVA &problem_fwd, const Vector &Mat) {
  MatRepara->Clear();
  if (problem_fwd.getParametrization() == "native") {
      *MatRepara = Mat;
  } else if (problem_fwd.getParametrization() == "taup") {
      for (cell c = MatRepara->cells(); c != MatRepara->cells_end(); ++c) {
          double Rho = Mat(c(), 0);
          double V_p = Mat(c(), 1);
          double Tau_p = (problem_fwd.getNumL() > 0) ? Mat(c(), 2) : 0;
          double alpha = (problem_fwd.getNumL() > 0) ? problem_fwd.alpha(*c, c()) : 0;
          (*MatRepara)(c(), 0) = Rho;
          (*MatRepara)(c(), 1) = (V_p * V_p * Rho) / (1 + alpha * Tau_p);
          if (problem_fwd.getNumL() > 0) {
              for (int i = 0; i < problem_fwd.getNumL(); ++i) {
                  (*MatRepara)(c(), 2 + i) = (V_p * V_p * Rho * Tau_p) / (1 + alpha * Tau_p);
              }
          }

      }
    } else if (problem_fwd.getParametrization() == "taup_positive") {
      for (cell c = MatRepara->cells(); c != MatRepara->cells_end(); ++c) {
          double Rho = exp(Mat(c(), 0));
          double V_p = exp(Mat(c(), 1));
          double Tau_p = (problem_fwd.getNumL() > 0) ? exp(Mat(c(), 2)) : 0;
          double alpha = (problem_fwd.getNumL() > 0) ? problem_fwd.alpha(*c, c()) : 0;
          (*MatRepara)(c(), 0) = Rho;
          (*MatRepara)(c(), 1) = (V_p * V_p * Rho) / (1 + alpha * Tau_p);
          if (problem_fwd.getNumL() > 0) {
              for (int i = 0; i < problem_fwd.getNumL(); ++i) {
                  (*MatRepara)(c(), 2 + i) = (V_p * V_p * Rho * Tau_p) / (1 + alpha * Tau_p);
              }
          }
      }
    } else {
        Exit("Unknown parameterization");
    }
}

FWIDGVAcousticMaterialFromVector::FWIDGVAcousticMaterialFromVector(const WaveProblemDGTSVA &prob_, const Vector &Mat,
                                                                   PMLHandler *handle_,PMLHandlerAAO* handleAAO_)
        : DGVAcousticMaterialFromVector(Mat),prob(prob_),handle(handle_),handleAAO(handleAAO_) {
  NumDamping = prob.getNumL();
    fillVector(prob, Mat);
    if (handle) {
        if (handle->isPML()) {
          bool increaseFlag = handle->GetConfig().getTaperStyle() != 1;
          std::unique_ptr<PMLMatVectorFWI> Mater = std::make_unique<PMLMatVectorFWI>(*handle,
                                                                                     *MatRepara,
                                                                                     increaseFlag);
          MatRepara = std::move(Mater);
          if (handle->GetConfig().getTaperStyle() == 1) {
            TauP = std::make_unique<Vector>(*MatRepara);
            TauP->Clear();
            for (cell c = MatRepara->cells(); c != MatRepara->cells_end(); ++c) {
              double Kappa0 = (*MatRepara)(c(), 1);
              double Kappa1 = (*MatRepara)(c(), 2);
              (*TauP)(c(), 0) = Kappa1 / Kappa0;
            }
          }
        }
    }
  if (handleAAO){
    handleAAO->SetPMLZero(*MatRepara);
    handleAAO->CommunicateBoundaryToPML(*MatRepara);
    handleAAO->increaseImpedance(*MatRepara,NumDamping);
  }
    MatRepara->Accumulate();
}

double FWIDGVAcousticMaterialFromVector::Kappa_i(const Cell &C, const Point &x, int n) const {
    if(handle && handle->GetConfig().getTaperStyle() == 1){
      if(handle->PMLToBoundary.contains(C())){
        const auto &Borders = handle->GetMeshBorders();
        double distX = -infty;
        double distY = -infty;
        double relDistX = -infty;
        double relDistY = -infty;
        if (x[0] < Borders[0]) {
          distX = abs(x[0] - Borders[0]);
          relDistX = distX / handle->GetConfig().GetPMLPaddingLength(0);
        } else if (x[0] > Borders[1]) {
          distX = abs(x[0] - Borders[1]);
          relDistX = distX / handle->GetConfig().GetPMLPaddingLength(1);
        } else if (x[0] == Borders[0] || x[0] == Borders[1]) {
          distX = 0.0;
        }
        if (x[1] < Borders[2]) {
          distY = abs(x[1] - Borders[2]);
          relDistY = distY / handle->GetConfig().GetPMLPaddingLength(2);
        } else if (x[1] > Borders[3]) {
          distY = abs(x[1] - Borders[3]);

          relDistY = distY / handle->GetConfig().GetPMLPaddingLength(3);
        } else if (x[1] == Borders[2] || x[1] == Borders[3]) {
          distY = 0.0;
        }
        if (distX < Eps && distY < Eps) return (*MatRepara)(C(), 1 + n);
        double relDist = max(relDistX, relDistY);
      double deltad = handle->GetConfig().GetCosConfig().deltad;
      double vQ = handle->GetConfig().GetCosConfig().vQ;
      double vB = handle->GetConfig().GetCosConfig().vB;
      double Fac = relDist < deltad ? vQ +  (1-vQ) * cos(relDist/deltad *M_PI_2)
                                  : vQ+(vB-vQ)*(relDist-deltad)/(1-deltad);
        double taup = (*TauP)(C(),0);
        double alpha1 = handle->GetConfig().GetCosConfig().alpha1;
        double alpha2 = handle->GetConfig().GetCosConfig().alpha2;
        double QM = (1.0 + taup * alpha1) / (taup*alpha2);
        double Qb = handle->GetConfig().GetCosConfig().Qb;
        double QT = pow(QM,1.0 - relDist) * pow(Qb, relDist); // Exponential Taper from Q Value at
        double TauPML = 1.0/(alpha2*QT-alpha1);
        double quot1 = (1 + alpha1 * TauPML) / (1 + alpha1 * taup);
        double quot2 = (1.0 + handle->GetNumL() * TauPML) / (1.0 + handle->GetNumL() * taup);
        double KappaFac =1.0/ (  sqrt(quot1*quot2));
        double Kappa0 = (*MatRepara)(C(),1);
        if (n == 0){
          return Kappa0 *  KappaFac * sqrt(Fac);

        }else {
          return Kappa0 * KappaFac * sqrt(Fac) * TauPML;
        }
      }
    }
    return (*MatRepara)(C(), 1 + n);
    Exit("wrong call")
}

double FWIDGVAcousticMaterialFromVector::invKappa_i(const Cell &C, const Point &x, int n) const {
    return 1.0 / (*this).Kappa_i(C,x,n);
    Exit("wrong call")
}

double FWIDGVAcousticMaterialFromVector::invTau_i(const Cell &C, const Point &x, int n) const {
    return prob.invTau_i(C,x,n);
}

double FWIDGVAcousticMaterialFromVector::Kappa(const Cell &C, const Point &x) const {
  double kappa = 0;
  for (int i = 0; i <= GetNumDamping(); ++i) {
    kappa += Kappa_i(C, x, i);
  }
  return kappa;
}

double FWIDGVAcousticMaterialFromVector::Rho(const Cell &C, const Point &x) const {
  if(handle && handle->GetConfig().getTaperStyle() == 1){
    if(handle->PMLToBoundary.contains(C())){
      const auto &Borders = handle->GetMeshBorders();
      double distX = -infty;
      double distY = -infty;
      double relDistX = -infty;
      double relDistY = -infty;
      if (x[0] < Borders[0]) {
        distX = abs(x[0] - Borders[0]);
        relDistX = distX / handle->GetConfig().GetPMLPaddingLength(0);
      } else if (x[0] > Borders[1]) {
        distX = abs(x[0] - Borders[1]);
        relDistX = distX / handle->GetConfig().GetPMLPaddingLength(1);
      } else if (x[0] == Borders[0] || x[0] == Borders[1]) {
        distX = 0.0;
      }
      if (x[1] < Borders[2]) {
        distY = abs(x[1] - Borders[2]);
        relDistY = distY / handle->GetConfig().GetPMLPaddingLength(2);
      } else if (x[1] > Borders[3]) {
        distY = abs(x[1] - Borders[3]);

        relDistY = distY / handle->GetConfig().GetPMLPaddingLength(3);
      } else if (x[1] == Borders[2] || x[1] == Borders[3]) {
        distY = 0.0;
      }
      if (distX < Eps && distY < Eps) return (*MatRepara)(C(), 0);
      double relDist = max(relDistX, relDistY);
      double deltad = handle->GetConfig().GetCosConfig().deltad;
      double vQ = handle->GetConfig().GetCosConfig().vQ;
      double vB = handle->GetConfig().GetCosConfig().vB;
      double Fac = relDist < deltad ? vQ +  (1 - vQ) * cos(relDist / deltad * M_PI_2)
                                    : vQ + (vB - vQ) * (relDist - deltad)/(1 - deltad);
      double taup = (*TauP)(C(),0);
      double alpha1 = handle->GetConfig().GetCosConfig().alpha1;
      double alpha2 = handle->GetConfig().GetCosConfig().alpha2;
      double QM = (1.0 + taup * alpha1) / (taup * alpha2);
      double Qb = handle->GetConfig().GetCosConfig().Qb;
      double QT = pow(QM,1.0 - relDist) * pow(Qb, relDist); // Exponential Taper from Q Value at
      double TauPML = 1.0/(alpha2 * QT - alpha1);
      double quot1 = (1 + alpha1 * TauPML) / (1 + alpha1 * taup);
      double quot2 = (1.0 + handle->GetNumL() * TauPML) / (1.0 + handle->GetNumL() * taup);
      double RhoFac =  sqrt(quot1 / quot2);
      double Rho = (*MatRepara)(C(), 0);
      return Rho * RhoFac * 1.0/sqrt(Fac);
    }
  }
    return (*MatRepara)(C(), 0);
}
