#include "FWIDataContainerDomain.hpp"

FWIDataContainerDomain::FWIDataContainerDomain(const Vector &_mat) {
    Mat = std::make_unique<Vector>(_mat);
}

FWIDataContainerDomain::FWIDataContainerDomain(const Vectors &_vecs) {
    Vecs = std::make_unique<Vectors>(_vecs);
}

FWIDataContainerDomain::FWIDataContainerDomain(const Vector &_mat, const Vectors &_vecs) {
    Mat = std::make_unique<Vector>(_mat);
    Vecs = std::make_unique<Vectors>(_vecs);
}

void FWIDataContainerDomain::setComp(int i, double b) {
    for (cell c = (*Mat).cells(); c != (*Mat).cells_end(); ++c) {
        (*Mat)(c(), i) = b;
    }
    Mat->SetAccumulateFlag(true);
}

void FWIDataContainerDomain::setCompToZero(int i) {
    setComp(i, 0.0);
}

FWIDataContainerDomain::~FWIDataContainerDomain() {
    Mat.reset();
    Vecs.reset();
}

void FWIDataContainerDomain::resetWaveData() {
    Vecs.reset();
}

FWIDataContainerDomain &FWIDataContainerDomain::operator-=(const FWIDataContainerDomain &s) {
    *Mat -= s.getMat();
    if (Vecs != nullptr) { *Vecs -= s.getVectors(); }
    return *this;
}

FWIDataContainerDomain &FWIDataContainerDomain::operator+=(const FWIDataContainerDomain &s) {
    *Mat += s.getMat();
    if (Vecs != nullptr) { *Vecs += s.getVectors(); }
    return *this;
}

FWIDataContainerDomain &FWIDataContainerDomain::operator=(const FWIDataContainerDomain &s) {
    *Mat = s.getMat();
    if (Vecs != nullptr) { *Vecs = s.getVectors(); }
    return *this;
}

FWIDataContainerDomain &FWIDataContainerDomain::operator=(double val) {
    *Mat = val;
    if (Vecs != nullptr) {
        *Vecs = val;
    }
    return *this;
}

FWIDataContainerDomain &FWIDataContainerDomain::operator*=(double val) {
    *Mat *= val;
    if (Vecs != nullptr) {
        *Vecs *= val;
    }
    return *this;
}

void FWIDataContainerDomain::operator*=(const RMatrix &vals) {
    Exit("implement this for vector")
//    (*Mat) *= vals;
    if (Vecs != nullptr) {
        Exit("This operator is not defined for AAO ");
    }
}

void FWIDataContainerDomain::operator*=(const RVector &vals) {
    Exit("implement this for vector")

//        (*Mat) *= vals;
    if (Vecs != nullptr) {
        Exit("This operator is not defined for AAO ");
    }
}

FWIDataContainerDomain::FWIDataContainerDomain(const string &type, const Vector &MatVec, const Vector &WaveVec,
                                               int timeSteps) {
    if (type == "Classical" || type == "FullWave") {
        Mat = std::make_unique<Vector>(MatVec);
        (*Mat).Clear();
    } else if (type == "AllAtOnce") {
        Vecs.reset();
        Vecs = std::make_unique<Vectors>(timeSteps + 1, WaveVec);
        Vecs->Clear();
        Mat = std::make_unique<Vector>(MatVec);
        (*Mat).Clear();
    }
}

FWIDataContainerDomain::FWIDataContainerDomain(const FWIDataContainerDomain &cont) {
    if (cont.isMaterial()) Mat = std::make_unique<Vector>(cont.getMat());
    if (cont.isWaveField()) Vecs = std::make_unique<Vectors>(cont.getVectors());
}

constAB<double, const FWIDataContainerDomain &> operator*(double d, const FWIDataContainerDomain &M) {
    return {d, M};

}