//
// Created by cr on 14.11.22.
//

#ifndef FWI_FWIDATACONTAINERDOMAIN_HPP
#define FWI_FWIDATACONTAINERDOMAIN_HPP

#include "Vector.hpp"
#include "RMatrix.hpp"

class FWIDataContainerDomain {
    std::unique_ptr<Vectors> Vecs = nullptr;
    std::unique_ptr<Vector> Mat = nullptr;

public:
    FWIDataContainerDomain() = default;

    FWIDataContainerDomain(const FWIDataContainerDomain &cont);

    FWIDataContainerDomain(const std::string &type, const Vector &MatVec, const Vector &WaveVec, int timeSteps);

    explicit FWIDataContainerDomain(const Vector &_mat);

    explicit FWIDataContainerDomain(const Vectors &_vecs);

    explicit FWIDataContainerDomain(const Vector &_mat,
                                    const Vectors &_vecs);

    ~FWIDataContainerDomain();

    void setComp(int i, double b);

    void resetWaveData();

    void setCompToZero(int i);

    bool isWaveField() const { return !(Vecs == nullptr); }

    bool isMaterial() const { return !(Mat == nullptr); }

    const Vectors &getVectors() const { return *Vecs; }

    Vectors &getVectorsWrite() { return *Vecs; }

    const Vector &getMat() const { return *Mat; }

    Vector &getMatWrite() { return *Mat; }

    void Clear() {
        (*Mat).Clear();
        if (Vecs != nullptr) { (*Vecs).Clear(); }
    }

    FWIDataContainerDomain &operator-=(const FWIDataContainerDomain &s);

    FWIDataContainerDomain &operator+=(const FWIDataContainerDomain &s);

    FWIDataContainerDomain &operator=(const FWIDataContainerDomain &s);

    FWIDataContainerDomain &operator=(double val);

    FWIDataContainerDomain &operator*=(double val);

    void operator*=(const RMatrix &vals);

    void operator*=(const RVector &vals);
};

constAB<double, const FWIDataContainerDomain &> operator*(double, const FWIDataContainerDomain &);


#endif //FWI_FWIDATACONTAINERDOMAIN_HPP
