#include "FWIDataContainerRange.hpp"

FWIDataContainerRange::FWIDataContainerRange(const std::string &type, const Vector &Vec, int timeSteps) {

    if (type == "Classical") {
        seismo = std::make_unique<FWISeismogram>();
    } else if (type == "AllAtOnce") {
        seismo = std::make_unique<FWISeismogram>();
        Vecs = std::make_unique<Vectors>(timeSteps + 1, Vec);
        Vecs->Clear();
    } else if (type == "FullWave") {
        Vecs = std::make_unique<Vectors>(timeSteps + 1, Vec);
    } else {
        Exit("unkown reconstruction type!")
    }
}

FWIDataContainerRange::FWIDataContainerRange(const Vectors &_vecs) {
    Vecs = std::make_unique<Vectors>(_vecs);
}

FWIDataContainerRange::FWIDataContainerRange(const FWISeismogram &s, const Vectors &_vecs) {
    seismo = std::make_unique<FWISeismogram>(s);
    Vecs = std::make_unique<Vectors>(_vecs);
}

FWIDataContainerRange::FWIDataContainerRange(const FWISeismogram &s) {
    seismo = std::make_unique<FWISeismogram>(s);
}

FWIDataContainerRange::FWIDataContainerRange(const FWIDataContainerRange &cont) {
    if (cont.isSeismo()) seismo = std::make_unique<FWISeismogram>(cont.getSeismogram());
    if (cont.isWaveField()) Vecs = std::make_unique<Vectors>(cont.getVectors());
}

FWIDataContainerRange &FWIDataContainerRange::operator-=(const FWIDataContainerRange &s) {
    if (isSeismo()) { *seismo -= s.getSeismogram(); }
    if (isWaveField()) { *Vecs -= s.getVectors(); }
    return *this;
}

FWIDataContainerRange &FWIDataContainerRange::operator+=(const FWIDataContainerRange &s) {
    if (isSeismo()) { *seismo += s.getSeismogram(); }
    if (isWaveField()) { *Vecs += s.getVectors(); }
    return *this;
}

FWIDataContainerRange &FWIDataContainerRange::operator=(const FWIDataContainerRange &s) {
    if (isSeismo()) { *seismo = s.getSeismogram(); }
    if (isWaveField()) { *Vecs = s.getVectors(); }
    return *this;
}

FWIDataContainerRange &FWIDataContainerRange::operator=(double val) {
    if (isSeismo()) { *seismo = val; }
    if (isWaveField()) { *Vecs = val; }
    return *this;
}

FWIDataContainerRange &FWIDataContainerRange::operator*=(double val) {
    if (isSeismo()) { *seismo *= val; }
    if (isWaveField()) { *Vecs *= val; }
    return *this;

}

FWIDataContainerRange::~FWIDataContainerRange() {
    seismo.reset();
    Vecs.reset();
}

void FWIDataContainerRange::resetWaveData() {
    Vecs.reset();
    Vecs = nullptr;
}