//
// Created by cr on 14.11.22.
//

#ifndef FWI_FWIDATACONTAINERRANGE_HPP
#define FWI_FWIDATACONTAINERRANGE_HPP

#include "Vector.hpp"
#include "Observation/FWISeismogram.hpp"

class FWIDataContainerRange {
    std::unique_ptr<FWISeismogram> seismo = nullptr;
    std::unique_ptr<Vectors> Vecs = nullptr;
public:
    FWIDataContainerRange() = default;

    FWIDataContainerRange(const std::string &type, const Vector &Vec, int nSteps);

    FWIDataContainerRange(const FWIDataContainerRange &cont);

    explicit FWIDataContainerRange(const FWISeismogram &s);

    explicit FWIDataContainerRange(const FWISeismogram &s, const Vectors &_vecs);

    explicit FWIDataContainerRange(const Vectors &_vecs);

    ~FWIDataContainerRange();

    bool isSeismo() const { return !(seismo == nullptr); }

    bool isWaveField() const { return !(Vecs == nullptr); }

    const FWISeismogram &getSeismogram() const { return *seismo; }

    const Vectors &getVectors() const { return *Vecs; }

    FWISeismogram &getSeismogramWrite() {
        if (!isSeismo()) {
            Exit("Seismogram not initialized!");
        } else {
            return *seismo;
        }
    }

    Vectors &getVectorsWrite() { return *Vecs; }

    std::unique_ptr<Vectors> &getVectorsPointer() { return Vecs; }

    FWIDataContainerRange &operator-=(const FWIDataContainerRange &s);

    FWIDataContainerRange &operator+=(const FWIDataContainerRange &s);

    FWIDataContainerRange &operator=(const FWIDataContainerRange &s);

    FWIDataContainerRange &operator=(double val);

    FWIDataContainerRange &operator*=(double val);

    void resetWaveData();


};


#endif //FWI_FWIDATACONTAINERRANGE_HPP
