//
// Created by cr on 24/03/23.
//

#ifndef FWI_FWINONLINEARSOLVERSO_HPP
#define FWI_FWINONLINEARSOLVERSO_HPP
#include "FWINonlinearSolvers.hpp"
#include "ForwardModels/FWIForwardModelDGTSVASO.hpp"
#include "RegularisationStrategy.hpp"

class FWINonlinearSolverSO : public FWINonlinearSolver {
public:
protected:
    int update_strategy;
    bool setSKP = false;
    std::vector<vector<int>> subproblems;
    ReginnOptions opt;
    FWIForwardModelDGTSVASO &FWDSO;

    std::vector<RegularisationStrategyWinkler> regsFO;
    std::vector<RegularisationStrategyWinkler> regsSO;
    bool FirstOrderLoop;
    size_t prob;
public:
    FWINonlinearSolverSO(FWIForwardModel &FWD, const std::string& name, ReginnOptions _opt) :
            FWINonlinearSolver(name.c_str(), FWD), opt(_opt),FWDSO(dynamic_cast<FWIForwardModelDGTSVASO&>(FWD)) {
        verbose = 5;
        iter = 0;
    }

    ~FWINonlinearSolverSO() {}

    void initializeSubproblems(const std::vector<FWIDataContainerRange> &data_obs,
                               std::vector<FWISubproblem> &nonlinearEvaluations) {
        subproblems.resize(data_obs.size());
        nonlinearEvaluations.resize(data_obs.size());
        for (int i = 0; i < data_obs.size(); ++i) {
            subproblems[i].resize(1);
            nonlinearEvaluations[i].emplace_back(FWDSO.getReconstructionType(),
                                                 FWDSO.GetWaveVec(),
                                                 FWDSO.TConf.timeSteps);
            subproblems[i][0] = i;

        }
    }

    void innerCG(const FWIDataContainerDomain &iterateNonlinear,
                 FWISubproblem &rkm,
                 FWIDataContainerDomain &skm)  {
        FWIDataContainerDomain d(iterateNonlinear);
        d.Clear();
        FWIDataContainerDomain um(iterateNonlinear);
        um.Clear();
        FWISubproblem q(rkm);
//    {
//      FWISeismogram tmp = q[0].getSeismogram();
//      tmp.collect();
//      tmp.writeToFile("ABC");
//    }
        double beta = 0.0;
        /* Abbruchkriterien */
        unsigned int mmax = (FirstOrderLoop) ? regsFO[prob].calcMaxInner() : regsSO[prob].calcMaxInner();
        int m = 0;
        if (FirstOrderLoop) {
            regsFO[prob].calcTheta();
        } else {
            regsSO[prob].calcTheta();
        }
        double condition = (FirstOrderLoop) ? regsFO[prob].calcInnerStopp() : regsSO[prob].calcInnerStopp();
        mout.StartBlock("Inner Loop: ");

        if (!FirstOrderLoop) {
            vout(2) << "PSI Inner Loop with Abbruchkriterium Residuum zu erreichen " << condition << " und m_max " << mmax << std::endl;
        } else {
            vout(2) << "FIRST ORDER Inner Loop with Abbruchkriterium Residuum zu erreichen " << condition << " und m_max " << mmax << std::endl;
        }

        Date innerLoop;
        double resNorm = 0.0;//Ein Schritt wird immer berechnet.Date innerLoop;
        do {
            m++;
            //d = \Ph/si_k(p_k)*r_{k,m-1}
            d.Clear();
            FWDSO.AdjointLinearized(iterateNonlinear, rkm, d);
            //u_m = d + \beta ||d||^2 u_{m-1}
            double dnormsquared = FWDSO.scalarProductL2(d, d);
            um *= beta * dnormsquared;
            um += d;

            //q = \Ph/si_k(p_k)u_m
            FWDSO.Linearized(iterateNonlinear,um,q);

            //\alpha = dnormsquare / qnormsquare
            double qnormsquared = FWDSO.scalarProductL2(q, q);
            double alpha = dnormsquared/qnormsquared;
            //s_km = s_{k,m-1} + \alpha u_m
            um *= alpha;
            skm += um;
            um *= (1.0/alpha);
            //r_km = r_{k,m-1} - \alpha q
            q *= alpha;
            rkm -= q;
            //\beta = 1/dnormsquared
            beta = 1.0/dnormsquared;
            /* Abbruchkriterium */
            resNorm = sqrt(FWDSO.scalarProductL2(rkm, rkm));
            vout(2) << "InnerCG resNorm : " << resNorm << std::endl;
        } while (m < mmax && resNorm > condition);
        vout(2) << "Time needed for inner loop: " <<  Date() - innerLoop << endl;
        vout(2) << "Innere Schleife No. 0/1 : " << FirstOrderLoop << " brauchte " << m << " = m innere Iterationen" << endl;
        mout.EndBlock();
        /* Erfasse die benötigte Schritteanzahl */
        if(FirstOrderLoop) {
            regsFO[prob].addInnerIter(m);
//      regsFO[prob].addResNorm(resNorm);
//      regsFO[prob].calcTheta();//Wird schon vor Beginn der Schleife gemacht.
        } else {
            regsSO[prob].addInnerIter(m);
//      regsSO[prob].addResNorm(resNorm);
//      regsSO[prob].calcTheta();
        }
        //Result of method is skm
    }


    void operator()(const IAssemble &, Vector &) override {
        Exit("not implemented");
    }

    NonlinearOutput operator()(const std::vector<FWIDataContainerRange> &dataObs,
                               FWIDataContainerDomain &iterateNonlinear, std::string nameSuff) override {
        //!!Achtung: Eventuell(!) muss hier noch irgendwo (oder woanders)
        //! FWIForwardModelDGTS::initializeRecon() oder so aufgerufen werden. Falls andere Config verwendet wird.

        mout << "Memory Usage at start" << PPM->TotalMemoryUsage() << endl;
        Start = Date();
        std::vector<FWISubproblem> nonlinearEvaluations;
        std::vector<FWIDataContainerDomain> iterateList;
        FWISubproblem bk;
        int writeSeismoToFileEvery = INT_MAX;
        Config::Get("WriteSeismoToFileEvery", writeSeismoToFileEvery);
        initializeSubproblems(dataObs, nonlinearEvaluations);
        mout << "Initialize Subproblems " << PPM->TotalMemoryUsage() << endl;
        size_t num_probs = subproblems.size();
        FWIDataContainerDomain skm(FWDSO.getReconstructionType(), FWDSO.GetMatVec(),
                                   FWDSO.GetWaveVec(), FWDSO.TConf.timeSteps);
        FWIDataContainerDomain hk(FWDSO.getReconstructionType(), FWDSO.GetMatVec(),
                                  FWDSO.GetWaveVec(), FWDSO.TConf.timeSteps);

        regsFO.resize(subproblems.size(),
                      RegularisationStrategyWinkler(opt.gamma, opt.thetaMax, opt.maxInnerLoopSteps));
        regsSO.resize(subproblems.size(),
                      RegularisationStrategyWinkler(opt.gamma, opt.thetaMax, opt.maxInnerLoopSteps));

        //Can be taken out (-> resNorm takes its use)
        //regsFO[0].setBkNorm(std::vector<double>(num_probs));//Initialize static BkNorm Array Variable


        d_0 = 0.0;
//        max_iter = 10;//Wird ultimativ über Config gesteuert!!
        std::cout << std::setprecision(15);
        /* Main for loop, whole CG-REGINN */
        vout(2) << "Starting the Loops with max_iter: " << max_iter << " and num_problems " << num_probs << std::endl;
        for (; iter < max_iter; ++iter) {
            vout(2) << std::endl << std::endl << "Iteration Main Looop Counting Variable iter: " << iter << std::endl;

            /* Aktuell Fehler besonders mit prob != 0, aber auch mit prob = 0 in iteration > 1
            *
            * Z.B. bei SolveAndMeasure, OneTimeStep, StageFunctions, getStageMatrix, initializeStageMatrices
            * LinearSolver (), PointBlockJacobi, invertSmallMatrix daa sind dann einträge zu klein.
            */
            /* Loop to treat each subproblems "iter"ations calculation step */
            for (prob = 0; prob < num_probs; ++prob) {
                vout(2) << std::endl << "iteration: " << iter << "  PROBLEM Number: " << prob << std::endl;
                FWISubproblem &nonlinearEvaluation = nonlinearEvaluations[prob];
                vout(2) << Name() + " problem " << prob << ": d(" << iter << ")" << endl;
                //Get Nonlinear Evaluation \Phi(iterateNonlin) = nonlinEval
                FWDSO.NonlinearEvaluation(iterateNonlinear,nonlinearEvaluation,subproblems[prob]);
                mout << "after NonlinearEvaluation " << PPM->TotalMemoryUsage() << endl;
                if( iter % writeSeismoToFileEvery  == 0) {
                    FWISeismogram seis(nonlinearEvaluation[0].getSeismogram());
                    seis.collect();
                    seis.writeToFile(Config::GetDataPath()+ "FWI/" + "Seis" + std::to_string(prob) + "_" +
                                     std::to_string(iter));
                }

                //Calculate Residuum bk = d_obs - nonlinEval\Phi(itnonlin)
                FWDSO.calculateDefect(dataObs, nonlinearEvaluation, bk, subproblems[prob]);
                //norm bk used for both CG-Loops
                regsFO[prob].addResNorm(sqrt(FWDSO.scalarProductL2(bk, bk)));
                regsSO[prob].addResNorm(sqrt(FWDSO.scalarProductL2(bk, bk)));
                if (bk.size() != nonlinearEvaluation.size()) {
                    Exit(" DataContainer sizes do not match")
                }
                //        vout(1) << "*************************************" << endl
                //                << "  " + Name() + " Subproblem " << prob << ": d(" << iter
                //                << "): Starting inner loop "
                //                << "( ||r||_2 = " << regs[prob].getResNorm(iter)
                //                << ", theta*||r||_2 = " << regs[prob].calcInnerStopp()
                //                << ", theta = " << regs[prob].getTheta(iter)
                //                << ", m_max = " << regs[prob].calcMaxInner() << " )" << endl;
                FWISubproblem rkm(bk);
                skm.Clear();
                /* CG method with First Order Operator and regFO Tolerance (Abbruchkriterium) */
                vout(1) << "Prä-Residuumnorm: " << sqrt(rkm[0].getSeismogram() * rkm[0].getSeismogram())  << std::endl;
                FirstOrderLoop = true;
                innerCG(iterateNonlinear,rkm,skm);
                vout(1) << "Residuumnorm 1: " << sqrt(rkm[0].getSeismogram() * rkm[0].getSeismogram())  << std::endl;
//            vout(1) << "Skm 1 * Skm 1: " << skm.getMat() * skm.getMat()  << std::endl;

                /* Prepare second CG-inner-loop */
                hk = skm;
                vout(3) << "----- Intermezzo: Change Operators and Recalculate Residuum -----" << endl;
                FWDSO.ChangeLinearized(hk, subproblems[prob]);
                /* r_k,0 = b_k - \Psi(p_k)s_k,0 */
                FWDSO.Linearized(iterateNonlinear,hk,rkm);
                rkm *= -1.0;
                rkm += bk;
                vout(1) << "Intermezzo Residuum: " << sqrt(rkm[0].getSeismogram() * rkm[0].getSeismogram()) << endl;
                /* CG method with Psi, so 2. and 1. Order Operator and regSO Tolerance (Abbruchkriterium) */
                FirstOrderLoop = false;
                vout(3) << "----- Intermezzo Ende -----" << endl;

                innerCG(iterateNonlinear,rkm,skm);
                vout(1) << "  " + Name() + " Subproblem: " << prob << ": d(" << iter
                        << "): Finished first / second inner loop "
                        << "( " << regsFO[prob].getInnerIter(iter) << ", " << regsSO[prob].getInnerIter(iter) << " steps )" << endl
                        << "*************************************" << endl;
                vout(1) << "Residuumnorm 2: " << sqrt(rkm[0].getSeismogram() * rkm[0].getSeismogram())  << endl;
//            vout(1) << "Skm 2: " << skm.getMat() * skm.getMat()  << std::endl;

                /* Update nonlinear Iterator */
                iterateNonlinear += skm;

                /* Plotsachen */
                int plotIter = 1;
                Config::Get("PlotFactorIterates", plotIter);
                int plotShot = -1;
                Config::Get("PlotShot", plotShot);

                if (iter % plotIter == 0 && (plotShot < 0 || prob == plotShot)) {
                    string name = "_fwi_" + Name() + nameSuff + "_iterate_m_allshots_" + to_string(prob) + "_"
                                  + to_string(iter);
                    FWDSO.plot(iterateNonlinear.getMat(),name.c_str());
                }
            }
        }

        vout(1) << " " + Name() + ": d(" << iter << ")= " << d << "\n";
        NonlinearOutput output;
        for (size_t prob = 0; prob < num_probs; ++prob) {
            output.push_back(regsFO[prob].getResNormVector(0));
        }
        return output;
//        vout(0) << " " << IterationEnd();
    }
};

#endif //FWI_FWINONLINEARSOLVERSO_HPP
