//
// Created by cr on 02.12.21.
//

#include "FWITaper.hpp"
#include "Observation/FWIObservationSpecification.hpp"

FWITaper::FWITaper(const Vector &Vec, const FWIObservationSpecification &spec, const TaperConf &Conf_,
                   WaveProblemDGTSVA &prob) : TaperVector(0.0, Vec), Conf(Conf_) {
    TaperVector = 1.0;
    paramComps = prob.getParamComps();
    for (cell c = TaperVector.cells(); c != TaperVector.cells_end(); ++c) {
        if (c()[0] < Conf.boxLeft || c()[0] > Conf.boxRight || c()[1] < Conf.boxLower || c()[1] > Conf.boxUpper) {
            TaperVector(c(), 0) = 0.0;
            continue;
        }
        bool found = false;
        if (Conf.taperIgnoreReceiverCells) {
            for (const auto &re: spec) {
                if (PointInCellDGTS_old(re, *c)) {
                    found = true;
                    break;
                }
            }
            if (found) {
                TaperVector(c(), 0) = 0.0;
                continue;
            }
        }
        if (Conf.taperIgnoreSourceCells) {
            found = false;
            for (int shot = 0; shot < prob.getShotCount(); ++shot) {
                prob.setShot(shot);
                int nrPointSrc = prob.getNrPointSrc();
                for (int i = 0; i < nrPointSrc; ++i) {
                    Point src = prob.getPointSrcLoc(i);
                    if (PointInCellDGTS_old(src, *c))
                        found = true;
                    break;
                }
            }
            if (found) {
                TaperVector(c(), 0) = 0.0;
                continue;
            }
        }
        double tapeval = 0.0;
        int insupp = 0;
        for (int shot = 0; shot < prob.getShotCount(); ++shot) {
            prob.setShot(shot);
            int nrPointSrc = prob.getNrPointSrc();
            for (int i = 0; i < nrPointSrc; ++i) {
                Point src = prob.getPointSrcLoc(i);
                double distance = dist(c(), src) / Conf.taperRadiusSource;
                if (distance > 1.0) continue;
                else {
                    tapeval += taperFunc(distance);
                    insupp++;
                }
            }
        }
        for (const auto &re: spec) {
            double distance = dist(c(), re) / Conf.taperRadiusReceiver;
            if (distance > 1.0) continue;
            else {
                tapeval += taperFunc(distance);
                insupp++;
            }
        }
        if (insupp > 0) TaperVector(c(), 0) = 1 - tapeval / insupp;
    }
}

void FWITaper::ApplyTaperVector(Vector &Mat) {
    for (cell c = Mat.cells(); c != Mat.cells_end(); ++c) {
        for (int i = 0; i < paramComps; ++i) {
            Mat(c(), i) *= TaperVector(c(), 0);
        }
    }
}

double FWITaper::taperFunc(double distance) const {
    return exp(Conf.taperExponent * (1 / (1 - pow(distance, 4)) - 1));
}
