//
// Created by cr on 09.11.22.
//

#include "FWITimeSolver.hpp"
#include "PMLHandler.hpp"
void FWITimeSolver::SolveAndMeasure(const std::shared_ptr<ILinearTimeAssemble> &assemble,
                                    const FWIObservationOperatorNEW &Obs,
                                    FWISeismogram &s, Vectors *U_m) {
    Date Start;
    std::unique_ptr<Vector> U_t;
    if (PMLH->isPML()) {
        U_t = std::make_unique<PMLWaveVectorFWI>(*PMLH);
    } else {
        U_t = std::make_unique<Vector>(WaveVectorUnpadded);
    }
    U_t->Clear();
    s.ts_clear();
    s.setSpecification(Obs.GetSpaceOp().GetSpecification());
    s.setStateComponents(Obs.GetMask().GetSeismoComps());
    std::unique_ptr<TimeIntegrator> timeIntegrator = GetFWITI(rkorder);
    mout.StartBlock("TI");
    mout.StartBlock("Initialization");
    timeIntegrator->Initialize(assemble.get(), *U_t);
    mout.EndBlock();
    mout << "Starting " << timeIntegrator->Name() << " (Order=" << timeIntegrator->GetOrder() << ")" << endl;
    if (U_m) {
        if (PMLH->isPML()) {
            PMLWaveVectorFWI::transfer(PMLH->GetPMLToNonPML(), *U_t, (*U_m)[assemble->Step()]);
        } else {
            (*U_m)[assemble->Step()] = *U_t;
        }
    }
    Obs.MeasureSingleWaveField(*U_t, s, assemble->Step());
    timeIntegrator->SetDivergenceCheck(false);
    while (!assemble->IsFinished()) {
        timeIntegrator->OneTimeStep(assemble.get(), *U_t);
        if (U_m) {
            if (PMLH->isPML()) {
                PMLWaveVectorFWI::transfer(PMLH->GetPMLToNonPML(), *U_t, (*U_m)[assemble->Step()]);
            } else {
                (*U_m)[assemble->Step()] = *U_t;
            }
        }
        Obs.MeasureSingleWaveField(*U_t, s, assemble->Step());
    }
    timeIntegrator->Destruct();
    mout.EndBlock();
}

void FWITimeSolver::Solve(const std::shared_ptr<ILinearTimeAssemble> &assemble, Vectors &U_m) {
    std::unique_ptr<Vector> U_t;
    if (PMLH->isPML()) {
        U_t = std::make_unique<PMLWaveVectorFWI>(*PMLH);
    } else {
        U_t = std::make_unique<Vector>(WaveVectorUnpadded);
    }
    U_t->Clear();
    std::unique_ptr<TimeIntegrator> timeIntegrator = GetFWITI(rkorder);
        mout.StartBlock("TI");
    timeIntegrator->Initialize(assemble.get(), *U_t);
    mout << "Starting " << timeIntegrator->Name() << " (Order=" << timeIntegrator->GetOrder() << ")" << endl;
    if (PMLH->isPML()) {
        PMLWaveVectorFWI::transfer(PMLH->GetPMLToNonPML(), *U_t, U_m[assemble->Step()]);
        } else {
        U_m[assemble->Step()] = *U_t;
    }
    timeIntegrator->SetDivergenceCheck(false);
    while (!assemble->IsFinished()) {
        timeIntegrator->OneTimeStep(assemble.get(), *U_t);
            if (PMLH->isPML()) {
                PMLWaveVectorFWI::transfer(PMLH->GetPMLToNonPML(), *U_t, U_m[assemble->Step()]);
            } else {
                U_m[assemble->Step()] = *U_t;
            }
    }
    timeIntegrator->Destruct();
    mout.EndBlock();

}
