#ifndef _FWIFORWARDMODELDGTSVAFULLWAVE_H_
#define _FWIFORWARDMODELDGTSVAFULLWAVE_H_

#include "FWIForwardModelDGTSVA.hpp"
#include "FWIDataContainerRange.hpp"
#include "FWIDataContainerDomain.hpp"

class FWIForwardModelDGTSVAFullWave : public FWIForwardModelDGTSVA {

protected:
    std::vector<const Vectors *> uMFull;
public:
    FWIForwardModelDGTSVAFullWave(std::shared_ptr<Meshes> meshes, LevelPair lev, int deg,
                                  std::shared_ptr<WaveProblemDGTSVA> problem_fwd_,
                                  bool _forwardOnly = false) : FWIForwardModelDGTSVA(meshes, lev,
                                                                                     deg, nullptr) {
        if (problem_fwd_ != nullptr) {
            setProblem(problem_fwd_);
        }

    }

    void AdjointLinearizedParameter2SolutionMap(const Vector &m, const Vectors &u_m, const FWIRhs &v,
                                                Vectors *lambda,
                                                Vector &adjoint_m_mv);

    std::string getReconstructionType() const override { return "FullWave"; }

    void createRealData(const FWIDataContainerDomain &m, std::vector<FWIDataContainerRange> &obs) override {}

    void LinearizedParameter2SolutionMap(const Vector &, const Vector &, const Vectors &, Vectors *);

    const char *Name() const override { return "FWIForwardModelDGTSVAFullWave"; }

    void NonlinearEvaluation(const FWIDataContainerDomain &, FWISubproblem &, std::vector<int>) override;

    void Parameter2SolutionMap(const Vector &m, Vectors &u_m);

    void calculateDefect(const std::vector<FWIDataContainerRange> &, const FWISubproblem &, FWISubproblem &,
                         std::vector<int> subproblem) override;

    void AdjointLinearized(const FWIDataContainerDomain &, const FWISubproblem &, FWIDataContainerDomain &) override;

    void Linearized(const FWIDataContainerDomain &, const FWIDataContainerDomain &, FWISubproblem &) override;

    void calculateAdjointWithoutSaving(const Vector &Material, const std::shared_ptr<ILinearTimeAssemble> &assemble,
                                       Vector &Adjoint_m_mv) const;

    void calculateL2AdjointStep(const Vector &Material, const Vectors &U_m, int step,
                                const Vector &LambdaStep,
                                Vector &Adjoint_m_mv) const;

    double scalarProductL2(const FWIDataContainerRange &, const FWIDataContainerRange &) const override;

    double scalarProductL2(const FWIDataContainerDomain &, const FWIDataContainerDomain &) const override;

};
#endif 
