//
// Created by cr on 14/03/23.
//

#ifndef FWI_FWIFORWARDMODELDGTSVANORMALIZED_HPP
#define FWI_FWIFORWARDMODELDGTSVANORMALIZED_HPP


#include "FWIForwardModelDGTSVA.hpp"
#include "Reduced/FWILinOpRedFirstOrderNormalized.hpp"


class FWIForwardModelDGTSVANormalized : public FWIForwardModelDGTSVA {
  FWISeismogram SeisMNormalized;
  std::unordered_map<Point,double> SeisMNorms;
public:
  FWIForwardModelDGTSVANormalized(std::shared_ptr<Meshes> meshes, LevelPair lev, int deg,
                                  const std::shared_ptr<WaveProblemDGTSVA> &problem_fwd) :
      FWIForwardModelDGTSVA(std::move(meshes), lev, deg, problem_fwd) {}
  void Nonlinear(const Vector& m, Vectors& uM,FWISeismogram &sM,int shot) override {
    Parameter2SeismogramMap(m,sM,shot,&uM);
    SeisMNormalized = sM;
    SeisMNorms.clear();
    for (auto  rec : SeisMNormalized.getSpecification()) {
      auto &ts = SeisMNormalized.getTimeSeries(rec);
      double norm = sqrt(ts * ts);
      SeisMNorms[rec] = norm;
    }
    DivideByFactorTracewise(sM,SeisMNorms);
  }
  void MakeLinearAndAdjointOperator(const Vector& m, Vectors& uM) override{
    mout << "Assembling Linearized Operators" << endl;
    AdjOp = std::make_unique<FWILinOpAdjRedNormalized>(uM,SeisMNormalized, SeisMNorms,
                                                       m, *Obs, problem_fwd, PMLH, TConf, recon,
                                             eleCacheContainer->Get());
    LinOp = std::make_unique<FWILinOpRedFirstOrderNormalized>(uM, SeisMNormalized, SeisMNorms,
                                                    m, problem_fwd, PMLH, TConf,
                                                    eleCacheContainer->Get(),false);
    mout << "Done!" << endl;
  }
    void AdjointLinearized(const FWIDataContainerDomain &nonlinearPoint,
                           const FWISubproblem &adjointLinearizedPoint,
    FWIDataContainerDomain &adjointLinearizedEvaluation) override;

    void Linearized(const FWIDataContainerDomain &nonlinearPoint,
                    const FWIDataContainerDomain &linearizedPoint,
    FWISubproblem &linearizedEvaluation) override;

    void calculateDefect(const std::vector<FWIDataContainerRange> &data_obs,
                         const FWISubproblem &nonlinearEvaluation,
                         FWISubproblem &defect,
    std::vector<int> subproblem) override;
};


#endif //FWI_FWIFORWARDMODELDGTSVANORMALIZED_HPP
