//
// Created by cr on 14.10.21.
//

#include "GramianBlockInverse.hpp"

#include <utility>

RMatrix getRMatrixFromCell(const Matrix &A, const Cell &C) {
    std::vector<Point> z = A.GetDoF().GetStoragePoints(C);
//    A.NodalPoints(C, z);
    std::vector<row> R(z.size());
    int n = 0;
    for (int i = 0; i < R.size(); ++i) {
        R[i] = A.find_row(z[i]);
        n += R[i].n();
    }
    RMatrix a(0.0, n, n);
    std::vector<double> &data = a.Data();
    for (int i = 0, j = 0; i < R.size(); ++i)
        for (int k = 0; k < R[i].n(); ++k, ++j)
      for (int i1 = 0, j1 = 0; i1 < R.size(); ++i1)
        for (int k1 = 0; k1 < R[i1].n(); ++k1, ++j1)
          data[n * j + j1] = A(R[i], R[i1])[k * R[i1].n() + k1];

  return a;
}

void
assembleScalarProductMatrix(Matrix &M,
                            const std::unordered_map<Point, DGAcousticElement *> &eleCache) {
    M = 0.0;
  for (const auto& elemIt : eleCache){
        const DGAcousticElement *elem = elemIt.second;
        const Cell &C = elem->GetCell();
        DGRowEntries M_c(M, C, C);
        int nV = elem->V_dimension();
        int nP = elem->shape_size();
        int p_dim = elem->p_dimension();
        for (int q = 0; q < elem->nQ(); ++q) {
            double w = elem->QWeight(q);
          for (int i = 0; i < nV; ++i) {
              VectorField V_i = elem->Velocity(q, i);
              for (int j = 0; j < nV; ++j) {
                  VectorField V_j = elem->Velocity(q, j);
          M_c(i, j) += w * (V_i * V_j);
        }
      }
      for (int i = 0; i < nP; ++i) {
        double P_i = elem->Pressure(q, i);
        for (int j = 0; j < nP; ++j) {
          double P_j = elem->Pressure(q, j);
          for (int n = 0; n < p_dim; ++n) {
            M_c(i + nV + n * nP, j + nV + n * nP) += w * P_i * P_j;
          }
        }
      }
    }
  }
  M.CreateSparse();
  return;
}

GramianBlockInverse::GramianBlockInverse(std::shared_ptr<Matrix> scalarProductMatrix_,
                                         const std::unordered_map<Point, DGAcousticElement *> &eleCache)
        :
        scalarProductMatrix(std::move(scalarProductMatrix_)),
        eleCache(eleCache) {
  for (const auto &elemIt : eleCache){
        const DGAcousticElement *elem = elemIt.second;
        const Cell &C = elem->GetCell();
        double a = elem->Area();
        if (ScalarMInverse.find(a) == ScalarMInverse.end()) {
            RMatrix Sm = getRMatrixFromCell(*scalarProductMatrix, C);
            Sm.Invert();
            ScalarMInverse.emplace(a, SparseRMatrix(Sm));
        }
    }
}

void GramianBlockInverse::scalarProductInverse2(const Vector &u, Vector &b) const {
  b.Clear();
  for (const auto& elemIt : eleCache){
    productOneCellScalarProductInverse(u, *elemIt.second, b);
  }
}

void GramianBlockInverse::scalarProductInverse2(Vector &b) const {
  Vector u(b);
  b.Clear();
  for (const auto& elemIt : eleCache){
   productOneCellScalarProductInverse(u, *elemIt.second, b);
  }
}

void GramianBlockInverse::productOneCellScalarProductInverse(const Vector &u, const DGAcousticElement &elem,
                                                             Vector &b) const {
    double a = elem.Area();
  const SparseRMatrix &SpSm = ScalarMInverse.at(a);
  int Row = u.find_row(elem.GetCell()()).Id();
  productOneCellSparse(u, SpSm, Row, b);
}

void GramianBlockInverse::productOneCellSparse(const Vector &u, const SparseRMatrix &SpSm, int Row,
                                               Vector &b) const {
  for (int i = 0; i < SpSm.rows(); ++i) {
    for (int k = SpSm.rowind(i); k < SpSm.rowind(i + 1); ++k) {
      b(Row, i) += SpSm.nzval(k) * u(Row, SpSm.colptr(k));
    }
  }
}
