#ifndef _DELTAKERNELS_
#define _DELTAKERNELS_

#include <unordered_set>
#include "Point.hpp"
#include "Cell.hpp"


class FWIDeltaKernel {
protected:
    const std::vector<Point> *midPoints;
    Point c;
    double r;
    double normconst = -1;
    int exponent;
    int dim;
public:
    FWIDeltaKernel(int _dim, const Point &center, double radius,
                   const std::vector<Point> *_midPoints = nullptr, int _exponent = 1) :
            midPoints(_midPoints),
            c(center),
            r(radius),
            exponent(_exponent),
            dim(_dim) {}

    virtual double operator()(const Point &, const Cell &) const {
        Exit("not implemented");
        return 0.0;
    }

    virtual ~FWIDeltaKernel() = default;
};

class FWIDeltaFunctionGauss : public FWIDeltaKernel {
public:
    FWIDeltaFunctionGauss(int _dim,
                          const Point &center,
                          double radius,
                          int _exponent = 1);

    double operator()(const Point &x, const Cell &ce) const override;

    double centeredAndNormalizedValue(double t) const {
        return exp(-exponent / (1 - t * t)) / normconst;
    }

    const Point &getCenter() const { return c; }
};

class FWIDeltaFunctionConstant : public FWIDeltaKernel {
public:
    FWIDeltaFunctionConstant(int _dim, const Point &center, double radius,
                             const vector<Point> *_midPoints, int _exponent = 1);

    double operator()(const Point &x, const Cell &ce) const override;
};


#endif
