//
// Created by cr on 11.11.22.
//

#include "FWIObservationCreator.hpp"
#include "Observation/FWIObservationSpecification.hpp"

auto FWIObservationCreator::CreateShared() -> std::shared_ptr<FWIObservationOperatorNEW> {
    return std::shared_ptr<FWIObservationOperatorNEW>(Create());
}

auto FWIObservationCreator::CreateUnique() -> std::unique_ptr<FWIObservationOperatorNEW> {
    return std::unique_ptr<FWIObservationOperatorNEW>(Create());
}

FWIObservationOperatorNEW *FWIObservationCreator::Create() {
    if (!Mask) {
        Mask = GetMeasureMaskNEW(mesh.dim(), numL, MeasureComp);
    }
    if (!spec) {
        getSpec();
    }
    spaceOp = std::make_unique<FWIObservationOperatorSpacePoint>(eleCache, spec, mesh,
                                                                 MeasureComp); // SpacePoint Hardcoded
    timeOp = std::make_unique<FWIObservationOperatorTimePoint>(spec);                        // TimePoint Hardcoded
    return new FWIObservationOperatorNEW(std::move(timeOp), std::move(spaceOp), std::move(Mask));
}

FWIObservationCreator &FWIObservationCreator::WithSpec(std::shared_ptr<FWIObservationSpecification> spec_) {
    spec = std::move(spec_);
    return *this;
}

FWIObservationCreator &FWIObservationCreator::WithMask(std::unique_ptr<MeasureMask> Mask_) {
    Mask = std::move(Mask_);
    return *this;
}

FWIObservationCreator &FWIObservationCreator::WithSpaceOp(std::unique_ptr<FWIObservationOperatorSpace> spaceOp_) {
    spaceOp = std::move(spaceOp_);
    return *this;
}

FWIObservationCreator &FWIObservationCreator::WithTimeOp(std::unique_ptr<FWIObservationOperatorTime> timeOp_) {
    timeOp = std::move(timeOp_);
    return *this;
}

FWIObservationCreator::FWIObservationCreator(const EleCacheType &eleCache_, const Mesh &mesh_, int numL_) :
        eleCache(eleCache_), mesh(mesh_) {
    numL = numL_;
    readConfig();
}

void FWIObservationCreator::readConfig() {
    Config::Get("MeasureComp", MeasureComp);
    if (!(MeasureComp == 0 || MeasureComp == 1)) THROW("Measurement Configuration unknown");
}

void FWIObservationCreator::getSpec() {
    spec = FWIObsSpecWPCreator().CreateShared();
}
