//
// Created by cr on 06/02/23.
//
#include "Observation/FWIObservationSpecification.hpp"

FWIObservationSpecification::FWIObservationSpecification() : t_start(0), t_end(1), t_delta(0.001) { set_n_steps(); }

FWIObservationSpecification::FWIObservationSpecification(const FWIObservationSpecification& spec)
    : list<Point>(spec), t_start(spec.t_start), t_end(spec.t_end), t_delta(spec.t_delta) { set_n_steps(); }

FWIObservationSpecification::FWIObservationSpecification(double start_time, double end_time, double delta_t,
                                                         const list<Point>& receiver_positions)
    : list<Point>(receiver_positions), t_start(start_time), t_end(end_time), t_delta(delta_t) { set_n_steps(); }

FWIObservationSpecification::FWIObservationSpecification(double start_time, double end_time,
                                                         double delta_t,const char* fn) {
  readFromFile(fn);
  t_start = start_time;
  t_end = end_time;
  t_delta = delta_t;
  set_n_steps();
}

void FWIObservationSpecification::refine(double fac) {
  t_delta = t_delta *fac;
  set_n_steps();
}

double FWIObservationSpecification::startTime() const { return t_start; }

double FWIObservationSpecification::endTime() const { return t_end; }

double FWIObservationSpecification::timeDelta() const { return t_delta; }

int FWIObservationSpecification::nSteps() const { return n_steps; }

double FWIObservationSpecification::step(int i) const { return t_start + i*t_delta; }

// -- IO stuff --
void FWIObservationSpecification::printInfo(const char* offset) const {
  mout << offset << "FWIObservationSpecification:" << endl;
  mout << offset << "  times: [" << t_start << ", " << t_end << "],  delta_t: " << t_delta << endl;
  mout << offset << "  receiver count: " << size() << endl;
  mout << offset << "  receiver positions:" << endl;
  for (auto r=begin(); r!=end(); ++r)
    mout << offset << "    " << *r << endl;
}

void FWIObservationSpecification::readFromFile(const std::string& filename) {
  std::ifstream file;
  file.open(filename.c_str());
  file >> *this;
  file.close();
}

void FWIObservationSpecification::writeToFile(const std::string& filename) {
  if (PPM->proc() != 0) return;
  std::ofstream file;
  file.open(filename.c_str());
  file << *this;
  file.close();
}