//
// Created by cr on 06/02/23.
//

#ifndef FWI_FWIOBSERVATIONSPECIFICATION_HPP
#define FWI_FWIOBSERVATIONSPECIFICATION_HPP

#include <memory>
#include <algorithm>
#include "../../../DSPFilters/Dsp.h"
#include "../../../DSPFilters/Filter.h"
#include "IFOSToFWI.hpp"
#include "FFT.hpp"
#include "CMatrix.hpp"
#include "Parallel.hpp"
#include "RVector.hpp"
#include "Point.hpp"
#include "m++.hpp"
#include "TimeDate.hpp"
#include <iomanip>
#include <fstream>
#include <random>

class FWIObservationSpecification : public std::__cxx11::list<Point> {
    double t_start;
    double t_end;
    double t_delta;
    int n_steps;
    int SamplingFactor = 1;
    void set_n_steps() { n_steps = static_cast<int>(round((t_end-t_start)/t_delta)) + 1; }
public:
    FWIObservationSpecification();
    FWIObservationSpecification(const FWIObservationSpecification& spec);
    FWIObservationSpecification(double start_time, double end_time, double delta_t,
                                const list<Point>& receiver_positions);

    explicit FWIObservationSpecification(double start_time, double end_time,
                                         double delta_t,const char* fn);

    void refine(double fac);
    int GetSamplingFactor(){
      return SamplingFactor;
    }
  void SetSamplingFactor(int SamplingFactor_){
     SamplingFactor = SamplingFactor_;
  }
    double startTime() const;

    double endTime() const;

    double timeDelta() const;

    int nSteps() const;

    double step(int i) const;

    // -- IO stuff --
    void printInfo(const char* offset="") const;

    void readFromFile(const std::string& filename);

    void writeToFile(const std::string& filename);

    friend std::ofstream& operator << (std::ofstream& o, const FWIObservationSpecification& spec) {
        o << std::to_string(spec.t_start) << endl;
        o << spec.t_end << endl;
        o << spec.t_delta << endl;
        o << spec.size() << endl;
        for (auto i=spec.begin(); i!=spec.end(); ++i) {
            for (int d = 0; d < (*i).SpaceDim(); ++d)
                o << (*i)[d] << "  ";
            o << endl;
        }
        return o;
    }
    friend std::istream& operator >> (std::istream& in, FWIObservationSpecification& spec) {
        spec.clear();
        double d1 = 0.0;
        double d2 = 0.0;
        double d3 = 0.0;
        in >> d1;
        spec.t_start =  d1;
//        in >> spec.t_start;
        in >> d2;
         spec.t_end= d2;
        in >> d3;
         spec.t_delta =d3 * spec.GetSamplingFactor();
        int receiver_count = 0;
        in >> receiver_count;
        for (int i=0; i<receiver_count; ++i) {
            Point p = Origin;
            for (int d = 0; d < p.SpaceDim(); ++d)
                in >> p[d];
            spec.push_back(p);
        }
        spec.set_n_steps();
        return in;
    }
};



#endif //FWI_FWIOBSERVATIONSPECIFICATION_HPP
