//
// Created by cr on 06/02/23.
//

#ifndef FWI_FWIRECEIVERTIMESERIES_HPP
#define FWI_FWIRECEIVERTIMESERIES_HPP

#include "RVector.hpp"
#include <iomanip>
#include <fstream>

typedef RVector FWIState;


class FWIReceiverTimeseries : public std::vector<std::pair<double,FWIState*>> {
int state_comps;
public:
explicit FWIReceiverTimeseries(int state_comps);
friend double operator*(const FWIReceiverTimeseries &s1, const FWIReceiverTimeseries &s2){
  if(s1.empty() || s2.empty()) return 0.0;
  double res = 0.0;
  double dt = s1[1].first - s1[0].first;
  res += 0.5*(*s1[0].second) * (*s2[0].second);
  for (int i = 1; i < s1.size()-1; ++i) {
    res += (*s1[i].second) * (*s2[i].second);
  }
    res += 0.5*(*s1[s1.size()-1].second) * (*s2[s1.size()-1].second);
  return res*dt;
}
  void operator *= (double val) {
      for (auto & m : *this)
        *m.second *= val;
}

FWIReceiverTimeseries(const FWIReceiverTimeseries &ts);
~FWIReceiverTimeseries();

friend std::ofstream& operator << (std::ofstream& o, const FWIReceiverTimeseries& ts)
{
  o << std::to_string(static_cast<int>(ts.size())).c_str() << endl;
//  o << ts.size() << endl;
  for (auto m=ts.begin(); m!=ts.end(); ++m) {
    o << "  " << m->first << '\t';
    const FWIState& state = *m->second;
    for (int i=0; i<ts.state_comps; ++i)
      o << state[i] << " ";
////            o << *m->second;
    o << endl;
  }
  return o;
}

friend std::istream& operator >> (std::istream& in, FWIReceiverTimeseries& ts) {
  ts.clear();
  int ts_size = 0;
  in >> ts_size;
  for (int i=0; i<ts_size; ++i) {
    double t;
    auto *state = new FWIState(0.0, ts.state_comps);
    in >> t;
    for (int j=0; j<ts.state_comps; ++j)
      in >> (*state)[j];
//            in >> *state;
    ts.push_back(std::make_pair(t, state));
  }
  return in;
}

void setStateComponents(int comps);
};


#endif //FWI_FWIRECEIVERTIMESERIES_HPP
