//
// Created by cr on 15.11.22.
//

#include "MeasureMask.hpp"
#include <memory>
void MaskPressureSumVA::mask(FWIState &v) const {
    double meas = 0;
    for (int l = 0; l <= numL; ++l) {
        meas += v[dim + l];
    }
    v.resize(1);
    v[0] = meas;
}

void MaskPressureSumVA::adjoint(FWIState &v) const {
    double meas = v[0];
    v.resize(dim + numL + 1);
    for (int d = 0; d < dim; ++d) {
        v[d] = 0.0;
    }

    for (int l = 0; l <= numL; ++l) {
        v[dim + l] = meas;
    }
}

void MaskPressureSumVA::maskAssemble(FWIState &v) const {
    int vecSize = v.size();
    int shapeSize = vecSize / (dim + numL + 1);
    for (int d = 0; d < dim; ++d) {
        for (int s = 0; s < shapeSize; ++s) {
            v[d * shapeSize + s] = 0.0;
        }
    }

}

std::unique_ptr<MeasureMask> GetMeasureMaskNEW(int dim, int numL, int MeasureComp) {
    if (MeasureComp == 0) {
        mout << "Measure overall pressure p = p_0 + ... + p_L" << endl;
        return std::make_unique<MaskPressureSumVA>(numL, dim);
    } else if (MeasureComp == 1) {
        mout << "Measure full wavefield measure (v_1, v_2, p_0, ..., p_L)" << endl;
        return std::make_unique<MaskPressureSumVA>(numL, dim);
    } else {
        Exit("This type of Measurement Mask is not implemented Measurecomp: " + std::to_string(MeasureComp));
        return nullptr;
    }
}