//
// Created by cr on 21.11.22.
//

#include "FWIInversionProblems.hpp"
#include "Observation/DeltaKernels.hpp"
#include "Problems/Sources/FWISourceWavelet.hpp"
#include "WaveProblemFromImages2D.hpp"
#include "Problems/Sources/WaypointSources.hpp"
#include "Problems/FWIAcousticInversionProb.hpp"



class FWIVAcousticCamembert : public FWIVAcousticInversionProb {
    double v_bg = 4000.0;
    double rho_bg = 2000.0;
    double impedance_bg = 1;
    double v_perturb = 1.15;
    double radiusPerturb = 1600.0;
    Point midPerturb = Point(2400.0, 3000.0);
public:
    explicit FWIVAcousticCamembert(std::unique_ptr<SourceWaveletFromF> TimeSource_) :
    FWIVAcousticInversionProb(0, 2) {
        std::string s = "native";
        Config::Get("parametrization", s);
        setParametrization(s);
        Sources = std::make_unique<WaypointSources>(std::vector<Point>{{100, 200},
                                                                       {100, 5800}}, std::vector<int>{14}, true);
      int size = Sources->GetShotCount();
      for (int i = 0; i<size;++i){
        srcSignals.emplace_back(std::make_unique<SourceWaveletBase>(*TimeSource_));
      }
        Config::Get("source_factor", factor);
        Config::Get("v_perturb", v_perturb);
        Config::Get("rho_bg", rho_bg);
        Config::Get("v_bg", v_bg);
        Config::Get("impedance_bg", impedance_bg);
    }

    const char *Name() const override { return "Camembert"; }


    bool isInPerturb(const Point &x) const {
        if (dist(x, midPerturb) < radiusPerturb && !initial_value) {
            return true;
        }
        return false;
    }

    double alpha(const Cell &c, const Point &z) const override {
        if (getNumL() == 0) Exit("wrong call!!");
        return -1.0;
    }


    double Rho(const Cell &c, const Point &z) const override {
        return rho_bg * impedance_bg;
    }

    double Tau_p(const Cell &c, const Point &z) const override {
        Exit ("wrong caw");
        return -10.0;
    }

    double V_p(const Cell &c, const Point &z) const override {
        if (isInPerturb(z)) {
            return v_bg * v_perturb;
        } else return v_bg;
        return v_bg;
    }

    ~FWIVAcousticCamembert() override = default;
};

class FWIVAcousticWaveC2Benchmark : public FWIVAcousticInversionProb {
    double v_bg = 3500.0;
    double rho_bg = 2000.0;
    double tau_p_bg = 0.02;
    double impedance_bg = 1;
    double v_perturb = 1.05;
    double rho_perturb = 1.05;
    double tau_p_perturb = 4.0;
    double size_perturb = 200.0;
    double x_perturb = 900.0;
    double z_perturb_0 = 600.0;
    double z_perturb_1 = 900.0;
    double z_perturb_2 = 1200.0;
    double perturb_0 = 1.05;
    double perturb_1 = 1.10;
    double perturb_2 = 0.95;
    std::string monoparam;
    int Qfactor = 15;

public:
    FWIVAcousticWaveC2Benchmark(std::unique_ptr<SourceWaveletFromF> TimeSource_, int _numL) :
            FWIVAcousticInversionProb(_numL, 2) {
        std::string s = "native";
        Config::Get("parametrization", s);
        setParametrization(s);
        Sources = std::make_unique<WaypointSources>(std::vector<Point>{{600, 370},
                                                                       {600, 1620}}, std::vector<int>{6}, true);
      int size = Sources->GetShotCount();
      for (int i = 0; i<size;++i){
        srcSignals.emplace_back(std::make_unique<SourceWaveletBase>(*TimeSource_));
      }
            Config::Get("source_factor", factor);
        wRef = 2 * TimeSource_->GetSignalFreq() * Pi;
        Config::Get("Monoparam", monoparam);
        MOUT(monoparam)
        trim(monoparam);
        if (monoparam == "multi" || monoparam.empty()) {
            Config::Get("v_perturb", v_perturb);
            Config::Get("rho_perturb", rho_perturb);
            Config::Get("tau_p_perturb", tau_p_perturb);
        } else if (monoparam != "Vp" && monoparam != "Tau_p" && monoparam != "Rho") {
            Exit("Unknown Monoparameter choice!")
        } else {
            Config::Get("z_perturb_0", z_perturb_0);
            Config::Get("z_perturb_1", z_perturb_1);
            Config::Get("z_perturb_2", z_perturb_2);
            Config::Get("z_perturb_0_value", perturb_0);
            Config::Get("z_perturb_1_value", perturb_1);
            Config::Get("z_perturb_2_value", perturb_2);
            Config::Get("x_perturb", x_perturb);
        }
        Config::Get("size_perturb", size_perturb);
        Config::Get("rho_bg", rho_bg);
        Config::Get("v_bg", v_bg);
        Config::Get("tau_p_bg", tau_p_bg);
        Config::Get("impedance_bg", impedance_bg);
        Config::Get("Qfactor", Qfactor);
        std::vector<double> frequencies;
        if (getNumL() == 1) {
            frequencies = std::vector<double>{TimeSource_->GetSignalFreq()};
        } else if (getNumL() == 5) {
            if (Qfactor == 15) {
//                frequencies = std::vector<double>{ 0.28, 1.15, 5.72, 26.2, 76.5}; //bohlens method
                frequencies = std::vector<double>{0.4971, 2.1266, 10.3723, 46.693, 120.0193};
                tau_p_bg = 0.0767;
            } else if (Qfactor == 20) {
                frequencies = std::vector<double>{0.72, 2.80, 13.00, 56.0, 150};
                tau_p_bg = 0.0534;
            } else if (Qfactor == 100) {
                frequencies = std::vector<double>{0.72, 2.80, 13.00, 56.0, 150};
                tau_p_bg = 0.0534 / 5.6;
            } else {
                THROW("This QFactor is not implemented! Allowed Choices are 15,20,100");
            }
        }
//        tau_i.resize(frequencies.size());
        for (double f: frequencies) {
            tau_i.push_back(1.0 / (f * 2.0 * M_PI));
        }
        calculateAlphaValue();
        MOUT(alphaValue);
    }

    const char *Name() const override { return "VAcousticWaveC2Benchmark"; }

    bool isInPerturb(const Point &x, double mat_perturb) const {
        if (initial_value) {
            return false;
        }
        if (x[0] > x_perturb && x[0] < x_perturb + size_perturb) {
            if (x[1] > mat_perturb && x[1] < mat_perturb + size_perturb) {
                return true;
            } else {
                return false;
            }
        }
        return false;
    }

    double Rho(const Cell &c, const Point &z) const override {
        if (monoparam.empty() || monoparam == "multi") {
            if (isInPerturb(z, z_perturb_1)) {
                return rho_bg * rho_perturb;
            } else return rho_bg;
        }
        if ("Rho" == monoparam) {
            if (isInPerturb(z, z_perturb_0)) {
                return rho_bg * perturb_0;
            } else if (isInPerturb(z, z_perturb_1)) {
                return rho_bg * perturb_1;
            } else if (isInPerturb(z, z_perturb_2)) {
                return rho_bg * perturb_2;
            }
        }
        return rho_bg * impedance_bg;
    }

    double Tau_p(const Cell &c, const Point &z) const override {
        if (monoparam.empty() || monoparam == "multi") {
            if (isInPerturb(z, z_perturb_2)) {
                return tau_p_bg * tau_p_perturb;
            } else return tau_p_bg;
        }
        if (monoparam == "Tau_p") {
            if (isInPerturb(z, z_perturb_0)) {
                return tau_p_bg * perturb_0;
            } else if (isInPerturb(z, z_perturb_1)) {
                return tau_p_bg * perturb_1;
            } else if (isInPerturb(z, z_perturb_2)) {
                return tau_p_bg * perturb_2;
            }
        }
        return tau_p_bg;
    }

    double V_p(const Cell &c, const Point &z) const override {
        if (monoparam.empty() || monoparam == "multi") {
            if (isInPerturb(z, z_perturb_0)) {
                return v_bg * v_perturb;
            } else return v_bg;
        }
        if (monoparam == "Vp") {
            if (isInPerturb(z, z_perturb_0)) {
                return v_bg * perturb_0;
            } else if (isInPerturb(z, z_perturb_1)) {
                return v_bg * perturb_1;
            } else if (isInPerturb(z, z_perturb_2)) {
                return v_bg * perturb_2;
            }
        }
        return v_bg;
    }

    ~FWIVAcousticWaveC2Benchmark() override = default;
};

class FWIAcousticWave1DBenchmark : public FWIVAcousticInversionProb {
    double v_bg = 1.0;
    double rho_bg = 1.0;
    double v_perturb = 1.5;
    double rho_perturb = 1.05;
    double size_perturb = 0.5;
    double x_perturb = 2.5;
    double x_perturb_0 = 2.5;
    double perturb_0 = 1.05;
    std::string monoparam;
    double tau_p = 0.3;
public:
    FWIAcousticWave1DBenchmark(std::unique_ptr<SourceWaveletFromF> TimeSource_, int numL) :
            FWIVAcousticInversionProb(numL, 1) {
        std::string s = "native";
        Sources = std::make_unique<WaypointSources>(std::vector<Point>{{0.5, 0.0},
                                                                       {0.5, 0.0}}, std::vector<int>{6}, false);
      int size = Sources->GetShotCount();
      for (int i = 0; i<size;++i){
        srcSignals.emplace_back(std::make_unique<SourceWaveletBase>(*TimeSource_));
      }
        Config::Get("taup_bg", tau_p);
        Config::Get("parametrization", s);
        setParametrization(s);
        Config::Get("source_factor", factor);
        wRef = 2 * TimeSource_->GetSignalFreq() * Pi;
        if (monoparam.empty()) {
            Config::Get("v_perturb", v_perturb);
            Config::Get("rho_perturb", rho_perturb);
        } else if (monoparam != "Vp" && monoparam != "Tau_p" && monoparam != "Rho") {
            Exit("Unknown Monoparameter choice!")
        } else {
            Config::Get("x_perturb_0", x_perturb_0);
            Config::Get("x_perturb_0_value", perturb_0);
        }
        Config::Get("size_perturb", size_perturb);
        Config::Get("rho_bg", rho_bg);
        Config::Get("v_bg", v_bg);
        switch (numL) {
            case 0:
                break;
            case 1:
                tau_i.push_back(1.0 / wRef);
                break;
            case 2:
                tau_i.push_back(1.0 / wRef);
                tau_i.push_back(1.0 / wRef);
                break;
            default: Exit("number of mechanisms not implemented!")
        }
        calculateAlphaValue();
    }

    const char *Name() const override { return "FWIAcousticWave1DBenchmark"; }

    bool isInPerturb(const Point &x, double x_perturb_loc) const {
        if (initial_value) {
            return false;
        }
        if (x[0] > x_perturb_loc && x[0] < x_perturb_loc + size_perturb) {
            return true;
        } else {
            return false;
        }
    }

    VectorField ForceV(double t, const Cell &, const Point &, int = -1) const override {
        return -v_bg * (*srcSignals[Sources->GetShot()])(t) * factor;
    }

    double Rho(const Cell &c, const Point &z) const override {
        return rho_bg;
    }

    double Tau_p(const Cell &c, const Point &z) const override {
        return tau_p;
    }

    double V_p(const Cell &c, const Point &z) const override {
        if (isInPerturb(z, x_perturb_0)) {
            return v_bg * v_perturb;
        }
        return v_bg;
    }

    ~FWIAcousticWave1DBenchmark() override = default;
};


class Marmousi2_VA_rhs : public WaveProblemFromImages2D {
public:
    Marmousi2_VA_rhs(const Point &_minP, const Point &_maxP, std::unique_ptr<SourceWaveletFromF> Ricker_,
                     int numL) : WaveProblemFromImages2D(_minP, _maxP, numL) {
        Sources = std::make_unique<WaypointSources>(std::vector<Point>{{2000.5, 0.5},
                                                                       {7999.5, 0.5}}, std::vector<int>{12}, true);
      int size = Sources->GetShotCount();
      for (int i = 0; i<size;++i){
        srcSignals.emplace_back(std::make_unique<SourceWaveletBase>(*Ricker_));
      }
        Config::Get("source_factor", factor);
        wRef = Ricker_->GetSignalFreq() * 2 * M_PI;
        std::vector<double> frequencies = {0.151, 1.93, 18.9};
        tau_i = std::vector<double>{1.0 / (frequencies[0] * 2 * M_PI),
                                    1.0 / (frequencies[1] * 2 * M_PI),
                                    1.0 / (frequencies[2] * 2 * M_PI)};
        calculateAlphaValue();
    }
};

//Copied from FWIAcousticWave1DBenchmark. Not pretty, will hopefully do.
class FWIVAcousticTwoMedium : public FWIVAcousticInversionProb {
    double v_bg = 3500.0;
    double rho_bg = 2000.0;
    double tau_p_bg = 0.02;
    double impedance_bg = 1;
    double v_perturb = 1.05;
    double rho_perturb = 1.05;
    double tau_p_perturb = 4.0;
    double size_perturb = 200.0;
    double x_perturb = 900.0;//Will alles "rechts" von x=900 anders haben
//    double z_perturb_0 = 600.0;
//    double z_perturb_1 = 900.0;
//    double z_perturb_2 = 1200.0;
    double perturb_0 = 1.05;
    double perturb_1 = 1.10;
    double perturb_2 = 0.95;
    std::string monoparam;
    int Qfactor = 15;

public:
    FWIVAcousticTwoMedium(std::unique_ptr<SourceWaveletFromF> TimeSource_, int _numL) :
            FWIVAcousticInversionProb(_numL, 2) {
        std::string s = "native";
        Config::Get("parametrization", s);
        setParametrization(s);
        Sources = std::make_unique<WaypointSources>(std::vector<Point>{{600, 370},
                                                                       {600, 1620}}, std::vector<int>{6}, true);
        int size = Sources->GetShotCount();
        for (int i = 0; i<size;++i){
            srcSignals.emplace_back(std::make_unique<SourceWaveletBase>(*TimeSource_));
        }
        Config::Get("source_factor", factor);
        wRef = 2 * TimeSource_->GetSignalFreq() * Pi;
        Config::Get("Monoparam", monoparam);
        MOUT(monoparam)
        trim(monoparam);
        if (monoparam == "multi" || monoparam.empty()) {
            Config::Get("v_perturb", v_perturb);
            Config::Get("rho_perturb", rho_perturb);
            Config::Get("tau_p_perturb", tau_p_perturb);
        } else if (monoparam != "Vp" && monoparam != "Tau_p" && monoparam != "Rho") {
            Exit("Unknown Monoparameter choice!")
        } else {
//            Config::Get("z_perturb_0", z_perturb_0);
//            Config::Get("z_perturb_1", z_perturb_1);
//            Config::Get("z_perturb_2", z_perturb_2);
            Config::Get("z_perturb_0_value", perturb_0);
            Config::Get("z_perturb_1_value", perturb_1);
            Config::Get("z_perturb_2_value", perturb_2);
            Config::Get("x_perturb", x_perturb);
        }
        Config::Get("size_perturb", size_perturb);
        Config::Get("rho_bg", rho_bg);
        Config::Get("v_bg", v_bg);
        Config::Get("tau_p_bg", tau_p_bg);
        Config::Get("impedance_bg", impedance_bg);
        Config::Get("Qfactor", Qfactor);
        std::vector<double> frequencies;

        if (getNumL() != 0) {
            Exit("Not foreseen for other than NumL = 0 yet."
                 "Might work, try and delete me here. Greetings Lennart.")
        }
//        if (getNumL() == 1) {
//            frequencies = std::vector<double>{TimeSource_->GetSignalFreq()};
//        } else if (getNumL() == 5) {
//            if (Qfactor == 15) {
////                frequencies = std::vector<double>{ 0.28, 1.15, 5.72, 26.2, 76.5}; //bohlens method
//                frequencies = std::vector<double>{0.4971, 2.1266, 10.3723, 46.693, 120.0193};
//                tau_p_bg = 0.0767;
//            } else if (Qfactor == 20) {
//                frequencies = std::vector<double>{0.72, 2.80, 13.00, 56.0, 150};
//                tau_p_bg = 0.0534;
//            } else if (Qfactor == 100) {
//                frequencies = std::vector<double>{0.72, 2.80, 13.00, 56.0, 150};
//                tau_p_bg = 0.0534 / 5.6;
//            } else {
//                THROW("This QFactor is not implemented! Allowed Choices are 15,20,100");
//            }
//        }
//        tau_i.resize(frequencies.size());
        for (double f: frequencies) {
            tau_i.push_back(1.0 / (f * 2.0 * M_PI));
        }
//        calculateAlphaValue();
        MOUT(alphaValue);
    }//ende Konstruktor

    const char *Name() const override { return "FWIVAcousticTwoMedium"; }

    //Hier anpassen.
    bool isInPerturb(const Point &x, double mat_perturb) const {
        if (initial_value) {
            return false;
        }
//        if (x[0] > x_perturb && x[0] < x_perturb + size_perturb) {
        if (x[0] < x_perturb) {
            return true;
        } else {
            return false;
        }
    }


    double Rho(const Cell &c, const Point &z) const override {
        if (monoparam.empty() || monoparam == "multi") {
//            if (isInPerturb(z, z_perturb_1)) {
            if (isInPerturb(z, 0)) {
                return rho_bg * rho_perturb;
            } else return rho_bg;
        }
        if ("Rho" == monoparam) {
//            if (isInPerturb(z, z_perturb_0)) {
            if (isInPerturb(z, 0)) {
                return rho_bg * perturb_0;
//            } else if (isInPerturb(z, z_perturb_1)) {
            } else if (isInPerturb(z, 0)) {
                return rho_bg * perturb_1;
//            } else if (isInPerturb(z, z_perturb_2)) {
            } else if (isInPerturb(z, 0)) {
                return rho_bg * perturb_2;
            }
        }
        return rho_bg * impedance_bg;
    }

    double Tau_p(const Cell &c, const Point &z) const override {
        if (monoparam.empty() || monoparam == "multi") {
//            if (isInPerturb(z, z_perturb_2)) {
            if (isInPerturb(z, 0)) {
                return tau_p_bg * tau_p_perturb;
            } else return tau_p_bg;
        }
        if (monoparam == "Tau_p") {
//            if (isInPerturb(z, z_perturb_0)) {
            if (isInPerturb(z, 0)) {
                return tau_p_bg * perturb_0;
//            } else if (isInPerturb(z, z_perturb_1)) {
            } else if (isInPerturb(z, 0)) {
                return tau_p_bg * perturb_1;
//            } else if (isInPerturb(z, z_perturb_2)) {
            } else if (isInPerturb(z, 0)) {
                return tau_p_bg * perturb_2;
            }
        }
        return tau_p_bg;
    }

    double V_p(const Cell &c, const Point &z) const override {
        if (monoparam.empty() || monoparam == "multi") {
//            if (isInPerturb(z, z_perturb_0)) {
            if (isInPerturb(z, 0)) {
                return v_bg * v_perturb;
            } else return v_bg;
        }
        if (monoparam == "Vp") {
//            if (isInPerturb(z, z_perturb_0)) {
            if (isInPerturb(z, 0)) {
                return v_bg * perturb_0;
//            } else if (isInPerturb(z, z_perturb_1)) {
            } else if (isInPerturb(z, 0)) {
                return v_bg * perturb_1;
//            } else if (isInPerturb(z, z_perturb_2)) {
            } else if (isInPerturb(z, 0)) {
                return v_bg * perturb_2;
            }
        }
        return v_bg;
    }

    ~FWIVAcousticTwoMedium() override = default;
};


//const Point &minP, const Point &maxP, int _numL,
//    std::unique_ptr<SourceWaveletBase> TimeSource_
class ADDAX : public WaveProblemFromImages2D {
public:
  ADDAX(std::unique_ptr<SourceWaveletFromF> wav,Point Pmin,
        Point Pmax,int numL)  : WaveProblemFromImages2D(Pmin,Pmax,numL)  {
    setParametrization("taup_positive");
    int shotStart = 0;
    int shotEnd   = 56;
    Config::Get("ADDAXShotStart",shotStart);
    Config::Get("ADDAXShotEnd",shotEnd);
    Sources = std::make_unique<ReadSources>(Config::GetBuildPath()+ "/FWI/models/ADDAX/source_locations.txt",shotStart,shotEnd);
    int s = Sources->GetShotCount();
    Config::Get("source_factor", factor);
    SetTaui();
    for (int i = 0; i<s;++i){
      srcSignals.emplace_back(std::make_unique<SourceWaveletBase>(*wav));
    }
    for (int i = 0; i<s;++i){
      srcSignals[i]->WriteToFile(Config::GetDataPath()+ "SRC" + to_string(i));
    }
  }
  double ForceP_i(double t, const Cell &c, const Point &z, int n, int) const override {
    return (n == 0) ? (*srcSignals[Sources->GetShot()])(t) * factor : 0.0;
  }
  void SetTimeSource(std::unique_ptr<SourceWaveletBase> wav) override {
    srcSignals[Sources->GetShot()] = std::move(wav);
  }
  std::unique_ptr<SourceWaveletBase> MoveTimeSource() override {
    return std::move(srcSignals[Sources->GetShot()]);
  }
  void SetTaui (){
    if (getNumL() == 0) return;
    else if(getNumL() == 1){
      Exit("not implemented yet")
    }else if (getNumL() ==2){
      wRef = 7 * 2 * M_PI;
      std::vector<double> frequencies = {2.0, 20.0};
      tau_i = std::vector<double>{1.0 / (frequencies[0] * 2 * M_PI),
                                  1.0 / (frequencies[1] * 2 * M_PI)};
      calculateAlphaValue();
      //to attain the right Q factor, taup = 0.0350
    }
  }
};

std::shared_ptr<FWIVAcousticInversionProb> GetInversionProblem(const WPConf &conf) {
    auto name = conf.s_prob;
    auto numL = conf.numL;
    std::unique_ptr<SourceWaveletFromF> SrcWavlet = nullptr;
    SrcWavlet = GetSourceWavelet(conf.TConf, conf.SConf);
    if (name == "Camembert") return std::make_shared<FWIVAcousticCamembert>(std::move(SrcWavlet));
    else if (name == "VAcousticWaveC2Benchmark") {
        if (numL != 0 && numL != 1 && numL != 5) {
            Exit("Benchmark only for 0/1/5 damping mechanisms");
        }
        return std::make_shared<FWIVAcousticWaveC2Benchmark>(std::move(SrcWavlet), numL);
    } else if (name == "VAcousticWaveTwoMedium") {
        if (numL != 0) {
            Exit("Benchmark only for 0 damping mechanisms. Maybe completely fine for damping but "
                 "I really don't know what I am doing here so far. Lennart");
        }
        return std::make_shared<FWIVAcousticTwoMedium>(std::move(SrcWavlet), numL);
    } else if (name == "VAcousticWaveMarmousi2") {
        if (numL != 3 && numL != 0) Exit("Marmousi2 only for numL=3! or numL=0")
        Point Pmin(0.0, 0.0);
        Point Pmax(10000.0, 3000.0);
        Config::Get("Pmin", Pmin);
        Config::Get("Pmax", Pmax);
        return std::make_shared<Marmousi2_VA_rhs>(Pmin, Pmax, std::move(SrcWavlet), numL);
    } else if (name == "Benchmark1D") {
        return std::make_shared<FWIAcousticWave1DBenchmark>(std::move(SrcWavlet), numL);
    } else if (name == "ADDAX") {
      Point ADDAXPmin(0.0, 0.0);
      Point ADDAXPmax(12160.0,2200.0);
      Config::Get("ADDAXPmin", ADDAXPmin);
      Config::Get("ADDAXPmax", ADDAXPmax);
      return std::make_shared<ADDAX>(std::move(SrcWavlet),ADDAXPmin,ADDAXPmax,numL);
    } else Exit("Problem not found");
    return nullptr;
}


