#include "FWIProblemsDGTSVA.hpp"
#include "FWIAbstractInterfaceDGTSVA.hpp"
#include "WaveProblemFromImages2D.hpp"
#include "Problems/Sources/FWISourceWavelet.hpp"


class Acoustic_Riemann_1D : public WaveProblemDGTSVA {
    double c;
public:
    Acoustic_Riemann_1D() : WaveProblemDGTSVA(0, 1) {
        c = 1;
        setParametrization("native");
    }

    void setInitialValueFlag(bool) override {}

    bool RHS(double) const override { return false; }

    bool exact() const override { return true; }

    const char *Name() const override { return "Acoustic_Riemann_1D"; }

    double ut(double t, const Cell &, const Point &x, int i) const override {
        double s = x[0] - c * t - 1;
        double y = exp(-20 * s * s);
        switch (i) {
            case 0:
                return c * y;
            case 1:
                return 0;
            case 2:
                return -y;
            default:
                mout << "Wrong i: " << i << endl;
                Exit("Exception")
        }
        Exit("something went wrong here!");
        return -1.0;
    }

    double Rho(const Point &) const override { return 1; }

    double Kappa_i(const Cell &, const Point &, int) const override { return 1; }

    VectorField ForceV(double, const Cell &, const Point &, int = -1) const override { return zero; }

    double ForceP_i(double, const Cell &, const Point &,
                    int,
                    int = -1) const override { return 0; }
};

class Acoustic_Benchmark_Cosine : public WaveProblemDGTSVA {
    double rho, kappa;
    Point *n;
    double norm;
    double f;
    double c;

public:
    Acoustic_Benchmark_Cosine() : WaveProblemDGTSVA(1, 2),
                                  rho(1), kappa(1) {
        setParametrization("native");
        Config::Get("kappa", kappa);
        Config::Get("rho", rho);
        n = new Point(1, 0, 0, 0);
        c = sqrt(kappa / rho);
        //        Config::Get("Direc", (*n));
        (*n)[0] = 1.0;
        (*n)[1] = 0.0;
        norm = sqrt((*n)[0] * (*n)[0] + (*n)[1] * (*n)[1]);
        (*n) /= norm;
        f = 1;
    }

    bool exact() const { return true; }

    bool RHS(double) const { return false; }

    const char *Name() const { return "Cosine wave"; }

    double ut(double t, const Point &x, int i) const {
        double nx = (*n)[0] * x[0] + (*n)[1] * x[1];
        switch(i) {
            case 0: return c * (*n)[0] * cos(f * (nx + c * t));
            case 1: return c * (*n)[1] * cos(f * (nx + c * t));
            case 2: return kappa * cos(f * (nx + c * t));
            default: return 0;
        }
    }

    double Rho(const Point &) const { return rho; }

    double Kappa_i(const Point &, int) const { return kappa; }

    VectorField ForceV(double,
                       const Cell &,
                       const Point &,
                       int = -1) const { return zero; }

    double ForceP_i(double, const Cell &, const Point &, int n, int= -1) const {
        if (n != 0) Exit("Wrong call!");
        return 0.0;
    }
};

class VAcoustic_test : public WaveProblemDGTSVA {
    double kappa;
    double rho;
    double c;
    Point direc;
public:
    VAcoustic_test(int dim) : WaveProblemDGTSVA(0, dim), kappa(25.0), rho(1.0),
                              c(sqrt(kappa / rho)), direc(1.0, 0.0) {
//        setParametrization("native");
        setParametrization("taup_positive");
    }

    virtual void setInitialValueFlag(bool) override {}

    bool exact() const override { return false; }

    bool RHS(double) const override { return true; }

    const char *Name() const override { return "Cosine wave"; }

    double ut(double t, const Point &x, int i) const override {
        if (this->getDim() == 1) {
            double s = x[0] - c * t - 1;
            double y = exp(-20 * s * s);
            if (i == 0) {
                return c * y;
            } else if (i == 1) {
                return kappa * (-y);
            } else {
                Exit("ERROR");
            }
        } else if (this->getDim() == 2) {
            double s = x[0] * direc[0] + x[1] * direc[1] - c * t - 4;
            double s2 = x[0] * direc[0] + x[1] * direc[1] + c * t - 6;
            double y = exp(-20 * s * s);
            double y2 = exp(-20 * s2 * s2);
            switch(i) {
                case 0: return direc[0] * c * y - direc[0] * c * y2;
                case 1: return direc[1] * c * y - direc[1] * c * y2;
                case 2: return -kappa * y - kappa * y2;
                default: return 0;
            }
        } else {
            Exit("ERROR");
            return -infty;
        }
        return 0;
    }

    double Rho(const Point &z) const override {
        if (this->getDim() == 1) {
            return rho;
        } else if (this->getDim() == 2) {
            if (z[1] > 0.7 || z[1] < 0.3) {
                return rho * 1000.0;
            }
            if (z[0] > 4.5) {
                return rho * 1000.0;
            }
            return rho;
        } else{
            Exit("Error in Dim!");
        }
    }

    double Kappa_i(const Point &x, int) const override {
        if (this->getDim() == 1) {
            return kappa;
        } else if (this->getDim() == 2) {
            if (x[1] > 0.7 || x[1] < 0.3) {
                return kappa / 1000.0;
            }
            if (x[0] > 4.5) {
                return kappa / 1000.0;
            }
            return kappa;
        } else {
            Exit("Error in Dim!");
        }
    }

    VectorField ForceV(double t, const Cell &, const Point &z, int = -1) const override {
        if (this->getDim() == 1) return zero;
        if (t > 0.1) return zero;
        double s = abs(z[0] - 4);
        double y = 100 * exp(-20 * s * s);
        return VectorField(-y, 0);
        return zero;
    }

    double ForceP_i(double t,
                    const Cell &,
                    const Point &z,
                    int n,
                    int= -1) const override {
        if (this->getDim() == 1) return 0.0;
        if (t > 0.1) return 0;
        double s = abs(z[0] - 4);
        return 100 * exp(-20 * s * s);
        if (n != 0) Exit("Wrong call!");
        return 0;
    }

    double alpha(const Cell &, const Point &z) const override {
        Exit("acoustic problem, no alpha there!");
        return -infty;
    }

    double V_p(const Cell &, const Point &z) const override {
        return sqrt(kappa / rho);
        return infty;
    }
};

class FWIVAcoustic_Benchmark_ExpP1_2 : public WaveProblemDGTSVA {
public:
    FWIVAcoustic_Benchmark_ExpP1_2()
        : WaveProblemDGTSVA(2, 2) {
        setParametrization("native");
    }

    bool exact() const { return true; }

    bool RHS(double) const { return false; }

    VectorField ForceV(double,
                       const Cell &,
                       const Point &,
                       int = -1) const { return zero; }

    double ForceP_i(double,
                    const Cell &,
                    const Point &,
                    int,
                    int= -1) const { return 0.0; }

    const char *Name() const { return "Constant p0=0.0 ,p1=exp(-t/tau), v = (0,0)"; }

    double ut(double t, const Cell &, const Point &x, int i) const {
        switch(i) {
            case 0:return 0.0;
            case 1:return 0.0;
            case 2:return 0.0;
            case 3:return exp(-t / Tau1(x));
            default:return 0;
        }
    }

    double Tau1(const Point &) const { return 0.1; }

    double Tau_i(const Cell &C, const Point &x, int i) const {
        if (i > getNumL()) Exit("Wrong Number of damping mechanisms")
        return 0.1;
    }

    double Rho(const Point &) const { return 1.0; }

    double Kappa_i(const Point &, int i) const {
        switch(i) {
            case 0:return 0.9;
            case 1:return 0.1;
        }
        Exit("Too many kappas!");
        return 0.0;
    }
};

class FWIVAcousticSinCos2d_RHS : public WaveProblemDGTSVA {
    double omega;
    double rho;
    Point kappa_rel;
    Point taus;
public:
    FWIVAcousticSinCos2d_RHS(int _numL = 1) : WaveProblemDGTSVA(_numL, 2) {
        setParametrization("native");
        Config::Get("taus", taus);
        Config::Get("kappa_rel", kappa_rel);
        if (_numL == 0) {
            kappa_rel[0] = 1;
        } else {
            for (int i = 0; i < _numL; ++i) {
                if (abs(kappa_rel[i]) < Eps || abs(taus[i]) < Eps) {
                    Exit("not all kappas/ defined")
                }
            }
        }
        rho = 1.0;
        Config::Get("rho", rho);
        omega = 1.0 / 4.0;
        Config::Get("omega", omega);
        omega *= 2 * M_PI;
    }

    const char *Name() const { return "VAcousticSinCos2d_RHS p,q = TODO"; }

    bool exact() const { return true; }

    bool RHS(double) const { return true; }

    double ut(double t, const Point &x, int i) const {
        if (i < 2) {
            return Velocity(t, x)[i];
        } else {
            return Pressure(t, x, i - 2);
        }
    }

    double Pressure(double, const Point &z, int i) const {
        return kappa_rel[i] * sin(omega * z[0]) * sin(omega * z[1]);
    }

    VectorField Velocity(double t, const Point &x) const {
        return VectorField((t - 1) * omega * sin(omega * x[1]) * cos(omega * x[0]),
                           (t - 1) * omega * sin(omega * x[0]) * cos(omega * x[1]), 0.0);
    }

    int Deg() const { return 1000; }

    double ForceP_i(double t, const Cell &c, const Point &z, int n, int= -1) const {
        if (n == 0) {
            return 2 * (t - 1) * omega * omega * sin(omega * z[0]) * sin(omega * z[1]);
        } else {
            return 2 * (t - 1) * omega * omega * sin(omega * z[0]) * sin(omega * z[1])
                + 1.0 / (Tau_i(c, z, n) * Kappa_i(c, z, n)) * Pressure(t, z, n);
        }
    }

    VectorField ForceV(double,
                       const Cell &,
                       const Point &,
                       int = -1) const { return zero; }

    double Kappa_i(const Cell &, const Point &, int i) const {
        return kappa_rel[i];
    }

    double Tau_i(const Cell &C, const Point &x, int i) const {
        return taus[i - 1];
    }

    double Rho(const Cell &, const Point &) const {
        return rho;
    }
};

class WaveProblemDGTSVAZero : public WaveProblemDGTSVA {
public:
    WaveProblemDGTSVAZero() : WaveProblemDGTSVA(2, 2) {
        setParametrization("native");
    }

    const char *Name() const { return "AcousticWaveProblemZero p,q = 0,(0,0)"; }

    double ut(double, const Point &, int) const { return 0; }

    double Rho(const Point &) const { return 1.0; }

    double Kappa_i(const Point &, int) const { return 1.0; }

    bool exact() const { return false; }

    bool RHS(double) const { return false; }

    VectorField ForceV(double, const Cell &, const Point &, int = -1) const { return zero; }

    double ForceP_i(double, const Cell &, const Point &, int, int) const { return 0.0; }
};


std::shared_ptr<WaveProblemDGTSVA> GetWaveProblem(const WPConf &conf) {
    auto name = conf.s_prob;
    auto level = conf.level;
    auto dim = conf.dim;
    auto numL = conf.numL;
    if (name.empty()) {
        name = "Zero";
        Config::Get("Problem", name);
    }
    if (name == "Zero") return std::make_shared<WaveProblemDGTSVAZero>();
    else if (name == "Acoustic_Benchmark_Cosine")
        return std::make_shared<Acoustic_Benchmark_Cosine>();
    else if (name == "VAcoustic_test")
        return std::make_shared<VAcoustic_test>(dim);
    else if (name == "Benchmark_ExpP1_2")
        return std::make_shared<FWIVAcoustic_Benchmark_ExpP1_2>();
    else if (name == "VAcousticSinCos2d_RHSDGTS")
        return std::make_shared<FWIVAcousticSinCos2d_RHS>(numL);
    else if (name == "Acoustic_Riemann_1D") {
        return std::make_shared<Acoustic_Riemann_1D>();
    } else Exit("Problem not found");
    return nullptr;
}




