//
// Created by cr on 18.11.22.
//

#include "FWILinOpAdjRedFirstOrder.hpp"

void FWILinOpAdjRedFirstOrder::calculateAdjointIntern(Vector &AdjointVector, Vectors *LambdaFull) {
    assemble->ResetTime(TConf.t0, TConf.T, TConf.dt);
    std::unique_ptr<Vector> LambdaStep2;
    std::unique_ptr<Vector> LambdaStepUnpadded;
    if (PMLH->isPML()) {
        LambdaStep2 = std::make_unique<PMLWaveVectorFWI>(*PMLH);
        LambdaStepUnpadded = std::make_unique<Vector>(U_m[0]);
        LambdaStepUnpadded->Clear();
    } else {
        LambdaStep2 = std::make_unique<Vector>(U_m[0]);
    }
    LambdaStep2->Clear();
    AdjointVector.Clear();
    timeIntegrator->Initialize(assemble.get(), *LambdaStep2);
    timeIntegrator->SetDivergenceCheck(false);
    mout.StartBlock("TI");
    mout << "Starting " << timeIntegrator->Name() << " (Order=" << timeIntegrator->GetOrder() << ")" << endl;
    while (!assemble->IsFinished()) {
        timeIntegrator->OneTimeStep(assemble.get(), *LambdaStep2);
        if (PMLH->isPML()) {
            (*LambdaStepUnpadded).Clear();
            PMLWaveVectorFWI::transfer(PMLH->GetPMLToNonPML(), *LambdaStep2, *LambdaStepUnpadded);
            calculateL2AdjointStep(*LambdaStepUnpadded, AdjointVector, assemble->Step());
            if (LambdaFull) LambdaFull[assemble->Step()] = *LambdaStepUnpadded;
        } else {
            calculateL2AdjointStep(*LambdaStep2, AdjointVector, assemble->Step());
            if (LambdaFull) LambdaFull[assemble->Step()] = *LambdaStep2;
        }
    }
    timeIntegrator->Destruct();
    mout.EndBlock();
    AdjointVector *= -1.0;
}

FWILinOpAdjRedFirstOrder::FWILinOpAdjRedFirstOrder(const Vectors &U_m_, const Vector &MatVec_, const FWIObservationOperatorNEW &Obs,
                                                   const std::shared_ptr<WaveProblemDGTSVA> &prob_,
                                                   const std::shared_ptr<PMLHandler> &PMLH_, const TimeConf &TConf_,
                                                   const reconstructionContainer &recon, const EleCacheType &eleCache_) : PMLH(PMLH_),
                                                                                                      eleCache(
                                                                                                              eleCache_),
                                                                                                      U_m(U_m_),
                                                                                                      MatVec(MatVec_),
                                                                                                      TConf(TConf_) {
    AdjointRHS = std::make_shared<DGVAcousticRHSAssemblerAdjoint>(Obs);
    VAPlotter = std::make_shared<DGVAcousticPlotter>(prob_->getNumL(), nullptr);
    auto Mat = std::make_unique<FWIDGVAcousticMaterialFromVector>(*prob_, MatVec, PMLH.get());
    matrixAssembler = std::make_shared<DGVAcousticMatrixAssembler>(std::move(Mat), false);
    assemble = std::make_shared<DGVAcousticAssembleCombined>(VAPlotter, matrixAssembler, AdjointRHS,
                                                             TConf.t0,TConf.T,TConf.dt);
    if (PMLH->isPML()) {
        assemble->SetEleCache(&PMLH->GetEleCache());
    } else {
        assemble->SetEleCache(&eleCache);
    }
    fwiAdjointCalculator = std::make_unique<FWIAdjointCalculator>(MatVec, prob_->getParametrization(), eleCache,
                                                                  recon,
                                                                  prob_->getNumL(), TConf.dt, prob_);
    timeIntegrator = GetFWITI(TConf.rkorder);
}

void FWILinOpAdjRedFirstOrder::calculateL2AdjointStep(const Vector &LambdaStep, Vector &AdjointVector, int step) {
    int StepBwd = (U_m).size() - 1 - step;
    const Vector &U_m_bwd(U_m[StepBwd]);
    Vector U_m_fd(U_m[StepBwd + 1]);
    U_m_fd -= U_m_bwd;
    U_m_fd *= (1.0 / TConf.dt);
    fwiAdjointCalculator->calculateL2AdjointOneStep(U_m_bwd, U_m_fd, LambdaStep, AdjointVector);
}
