//
// Created by cr on 18.11.22.
//

#ifndef FWI_FWILINOPADJREDFIRSTORDER_HPP
#define FWI_FWILINOPADJREDFIRSTORDER_HPP

#include "Vector.hpp"
#include "PMLHandler.hpp"
#include "DGVAcousticMatrixAssembler.hpp"
#include "Assembler/FWIRhsAssemblerAdjoint.hpp"
#include "DGVAcousticAssembleCombined.hpp"
#include "Observation/FWIObservationDGTSNEW.hpp"
#include "FWIAbstractInterfaceDGTSVA.hpp"
#include "FWIDGVAcousticMaterialFromVector.hpp"
#include "FWIAdjointCalculator.hpp"
#include "NormalizedHelpers.hpp"
#include "FWILinOpAdjRed.h"

class FWILinOpAdjRedFirstOrder : public FWILinOpAdjRed {
    const std::shared_ptr<PMLHandler> &PMLH;
    const EleCacheType &eleCache;
    std::shared_ptr<DGVAcousticMatrixAssembler> matrixAssembler;
    std::shared_ptr<DGVAcousticAssembleCombined> assemble;
    const Vectors &U_m;
    const Vector &MatVec;
    std::shared_ptr<DGVAcousticPlotter> VAPlotter = nullptr;
    const TimeConf &TConf;
    std::unique_ptr<FWIAdjointCalculator> fwiAdjointCalculator = nullptr;
    std::unique_ptr<TimeIntegrator> timeIntegrator = nullptr;
protected:
  std::shared_ptr<DGVAcousticRHSAssemblerAdjoint> AdjointRHS;
  void calculateAdjointIntern(Vector &AdjointVector, Vectors *LambdaFull);
public:
    FWILinOpAdjRedFirstOrder(const Vectors &U_m_, const Vector &MatVec_, const FWIObservationOperatorNEW &Obs,
                             const std::shared_ptr<WaveProblemDGTSVA> &prob_,
                             const std::shared_ptr<PMLHandler> &PMLH_, const TimeConf &TConf_,
                             const reconstructionContainer &recon,
                             const EleCacheType &eleCache_);

    virtual void CalculateAdjoint(const FWISeismogram &s, Vector &AdjointVector, Vectors *LambdaFull){
      AdjointRHS->setSeis(&s);
      calculateAdjointIntern(AdjointVector,LambdaFull);
    }

    void calculateL2AdjointStep(const Vector &LambdaStep, Vector &AdjointVector, int step) override;
};
class FWILinOpAdjRedNormalized : public FWILinOpAdjRedFirstOrder {
  const FWISeismogram &sMNormed;
  const std::unordered_map<Point,double> &sMNorms;
  FWISeismogram TMP;
public:
  FWILinOpAdjRedNormalized(const Vectors &U_m_, const FWISeismogram& sMNormed_,
                           const std::unordered_map<Point,double> &sMNorms_,
                           const Vector &MatVec_, const FWIObservationOperatorNEW &Obs,
                   const std::shared_ptr<WaveProblemDGTSVA> &prob_,
                   const std::shared_ptr<PMLHandler> &PMLH_, const TimeConf &TConf_,
                   const reconstructionContainer &recon,
                   const EleCacheType &eleCache_) :
          FWILinOpAdjRedFirstOrder(U_m_, MatVec_, Obs, prob_, PMLH_, TConf_, recon, eleCache_), sMNormed(sMNormed_),
          sMNorms(sMNorms_), TMP(sMNormed_)
                   {

  }

    void CalculateAdjoint(const FWISeismogram &s, Vector &AdjointVector, Vectors *LambdaFull) override{
    TMP = s;
      PrepareLinearizedAndAdjointNormalized(sMNormed,sMNorms,TMP);
      AdjointRHS->setSeis(&TMP);
      calculateAdjointIntern(AdjointVector,LambdaFull);
  }

};


#endif //FWI_FWILINOPADJREDFIRSTORDER_HPP
