//
// Created by lionbart on 09.03.23.
//

#ifndef FWI_FWILINOPADJREDSECONDORDER_H
#define FWI_FWILINOPADJREDSECONDORDER_H

#include "Vector.hpp"
#include "PMLHandler.hpp"
#include "DGVAcousticMatrixAssembler.hpp"
#include "Assembler/FWIRhsAssemblerAdjoint.hpp"
#include "DGVAcousticAssembleCombined.hpp"
#include "Observation/FWIObservationDGTSNEW.hpp"
#include "FWIAbstractInterfaceDGTSVA.hpp"
#include "FWIDGVAcousticMaterialFromVector.hpp"
#include "FWIAdjointCalculator.hpp"
#include "FWILinOpRedFirstOrder.hpp"
#include "FWILinOpAdjRed.h"
//for debugging
#include "normCalculator.hpp"

class FWILinOpAdjRedSecondOrder : public FWILinOpAdjRed {
    const std::shared_ptr<PMLHandler> &PMLH;
    const EleCacheType &eleCache;
    std::shared_ptr<DGVAcousticRHSAssemblerAdjoint> AdjointRHS;
    std::shared_ptr<DGVAcousticMatrixAssembler> matrixAssembler;
    std::shared_ptr<DGVAcousticAssembleCombined> assemble;

    const Vectors &U_m;
    const Vector &Material;
    const Vector &MatDirection;
    std::shared_ptr<DGVAcousticPlotter> VAPlotter = nullptr;
    const TimeConf &TConf;
    std::unique_ptr<FWIAdjointCalculator> fwiAdjointCalculator = nullptr;
    std::unique_ptr<TimeIntegrator> timeIntegrator = nullptr;
    const bool Psi;
    bool verboseAdj = false;

    std::shared_ptr<FWILinOpRedFirstOrder> LinOp = nullptr; //new in comparison to Adjoint First Order
//    std::unique_ptr<FWILinearizedAssembler> LinAssembler = nullptr; //Might need some functions (Material') from here
public:
    FWILinOpAdjRedSecondOrder(const Vectors &U_m_, const Vector &MatVec_, const Vector &MatDirection_, const FWIObservationOperatorNEW &Obs,
                   const std::shared_ptr<WaveProblemDGTSVA> &prob_,
                   const std::shared_ptr<PMLHandler> &PMLH_, const TimeConf &TConf_,
                   const reconstructionContainer &recon,
                   const EleCacheType &eleCache_, const bool Psi_ = false);

    void CalculateAdjoint(const FWISeismogram &s, Vector &AdjointVector, Vectors *LambdaFull) override;

private:
    void calculateL2AdjointStep(const Vector &LambdaStep, Vector &AdjointVector, int step) override;

    void calculateL2AdjointStep(const Vectors &LambdaFull, const Vectors &zFull,
                                  Vectors &U_prime, Vector &AdjointVector, int step);
};

#endif //FWI_FWILINOPADJREDSECONDORDER_H
