//
// Created by cr on 18.11.22.
//

#include "FWILinOpRedFirstOrder.hpp"

void FWILinOpRedFirstOrder::Solve(const Vector &m_test, Vectors &U_linearized) {
    FWITimeSolver TSolveR(U_m[0], PMLH, TConf.rkorder);
    TSolveR.Solve(getAssembler(m_test), U_linearized);
}

void FWILinOpRedFirstOrder::SolveAndMeasure(const Vector &m_test,
                                            const FWIObservationOperatorNEW &Obs,
                                            FWISeismogram &s_linearized, Vectors *U_linearized) {
    FWITimeSolver TSolveR(U_m[0], PMLH, TConf.rkorder);
    TSolveR.SolveAndMeasure(getAssembler(m_test), Obs, s_linearized, U_linearized);
}

FWILinOpRedFirstOrder::FWILinOpRedFirstOrder(const Vectors &U_m_, const Vector &m_, const std::shared_ptr<WaveProblemDGTSVA> &prob_,
                                             const std::shared_ptr<PMLHandler> &PMLH_, const TimeConf &TConf_,
                                             const EleCacheType &eleCache_,
                                             bool discrete_, bool bwd_)
        : prob(prob_), U_m(U_m_), m(m_), PMLH(PMLH_), eleCache(eleCache_), TConf(TConf_), discrete(discrete_), bwd(bwd_) {
    VAPlotter = std::make_shared<DGVAcousticPlotter>(prob->getNumL(), nullptr);
    auto Mat = std::make_unique<FWIDGVAcousticMaterialFromVector>(*prob, m, PMLH.get());
    matrixAssembler = std::make_shared<DGVAcousticMatrixAssembler>(std::move(Mat), true);
}

std::shared_ptr<DGVAcousticAssembleCombined> FWILinOpRedFirstOrder::getAssembler(const Vector &m_test) {
    auto rhs = std::make_shared<DGVAcousticRHSAssemblerLinearizedFirstOrder>(U_m, prob, eleCache, m, m_test, TConf.dt,
                                                                             discrete, PMLH.get());

    std::shared_ptr<DGVAcousticAssembleCombined> assembleCombined;
    if (bwd) { /* Constructs assembler for left side of Evolution Eq. to discretize -A: Bu' + Au = rhs */
        std::shared_ptr<DGVAcousticMatrixAssembler> matrixAssemblerBackward;
        auto MaterialForAssembler = std::make_unique<FWIDGVAcousticMaterialFromVector>(*prob, m);
        matrixAssemblerBackward = std::make_shared<DGVAcousticMatrixAssembler>(std::move(MaterialForAssembler), false);
        assembleCombined = std::make_shared<DGVAcousticAssembleCombined>(VAPlotter, matrixAssemblerBackward, rhs,
                                                                         TConf.t0,TConf.T,TConf.dt);
    } else {
        assembleCombined = std::make_shared<DGVAcousticAssembleCombined>(VAPlotter, matrixAssembler, rhs,
                                                                         TConf.t0,TConf.T,TConf.dt);
    }
    assembleCombined->SetEleCache(PMLH->isPML() ? &PMLH->GetEleCache() : &eleCache);
    return std::move(assembleCombined);
}

const Vectors &FWILinOpRedFirstOrder::getUM() const {
    return U_m;
}

const std::shared_ptr<WaveProblemDGTSVA> &FWILinOpRedFirstOrder::getProb() const {
    return prob;
}

const Vector &FWILinOpRedFirstOrder::getM() const {
    return m;
}

const std::shared_ptr<PMLHandler> &FWILinOpRedFirstOrder::getPmlh() const {
    return PMLH;
}

bool FWILinOpRedFirstOrder::isDiscrete() const {
    return discrete;
}

const std::shared_ptr<DGVAcousticMatrixAssembler> &FWILinOpRedFirstOrder::getMatrixAssembler() const {
    return matrixAssembler;
}

const std::shared_ptr<DGVAcousticPlotter> &FWILinOpRedFirstOrder::getVaPlotter() const {
    return VAPlotter;
}

const EleCacheType &FWILinOpRedFirstOrder::getEleCache() const {
    return eleCache;
}

const TimeConf &FWILinOpRedFirstOrder::getTConf() const {
    return TConf;
}
