



#include "FWILinOpRedSecondOrder.h"

#include <utility>

/*

 const Vectors &U_m_, const Vector &MatVec_, const Vector &MatDirection_,
 const FWIObservationOperatorNEW &Obs,
                   const std::shared_ptr<WaveProblemDGTSVA> &prob_,
                   const std::shared_ptr<PMLHandler> &PMLH_, const TimeConf &TConf_,
                   const reconstructionContainer &recon,
                   const EleCacheType &eleCache_



 */

FWILinOpRedSecondOrder::FWILinOpRedSecondOrder(const Vectors &U_m_, const Vector &m_, Vector m_test1_, const std::shared_ptr<WaveProblemDGTSVA> &prob_,
                                               const std::shared_ptr<PMLHandler> &PMLH_, const TimeConf &TConf_, const EleCacheType &eleCache_,
                                               bool discrete, bool Psi_) :
                                               firstOrder(FWILinOpRedFirstOrder(U_m_, m_, prob_, PMLH_, TConf_,
                                                                                eleCache_, discrete)),
                                               m_test1(std::move(m_test1_)), Psi(Psi_) {}
//FWILinOpRedSecondOrder::FWILinOpRedSecondOrder(const Vectors &U_m_, const Vector &m_, const Vector &m_test1_,
//                                               const std::shared_ptr<WaveProblemDGTSVA> &prob_,
//                                               const std::shared_ptr<PMLHandler> &PMLH_, const TimeConf &TConf_,
//                                               const EleCacheType &eleCache_, bool discrete, bool Psi_) :
//                                               firstOrder(FWILinOpRedFirstOrder(U_m_, m_, prob_, PMLH_, TConf_,
//                                                                                eleCache_, discrete)),
//                                               m_test1(m_test1_), Psi(Psi_) {}


void FWILinOpRedSecondOrder::Solve(const Vector &m_test2, Vectors &U_linearizedSO) {
    //TOD0 (Copied from FirstOrder)
    FWITimeSolver TSolveR(firstOrder.getUM()[0], firstOrder.getPmlh(), firstOrder.getTConf().rkorder);
    std::shared_ptr<DGVAcousticAssembleCombined> assemblerCombine = getAssembler(m_test2);
    TSolveR.Solve(assemblerCombine, U_linearizedSO);
    if(Psi) {
        const std::shared_ptr<DGVAcousticRHSAssemblerLinearizedSecondOrder> &dataContainer =
                std::dynamic_pointer_cast<DGVAcousticRHSAssemblerLinearizedSecondOrder>(assemblerCombine->getRhsAssembler());
        std::shared_ptr<Vectors> Psi_u_prime2 = dataContainer->getUPrimeDirection2();
        U_linearizedSO *= 0.5;
        U_linearizedSO += *Psi_u_prime2;
    }
}


void FWILinOpRedSecondOrder::SolveAndMeasure(const Vector &m_test2, const FWIObservationOperatorNEW &Obs,
                                             FWISeismogram &s_linearizedSO, Vectors *U_linearizedSO) {
    FWITimeSolver TSolveR(firstOrder.getUM()[0], firstOrder.getPmlh(), firstOrder.getTConf().rkorder);
//    TSolveR.SolveAndMeasure(getAssembler(m_test2, Psi_u_prime2), Obs, s_linearizedSO, U_linearizedSO);

    //ReducedSO2D getAssembler schmeißt auch schon zweimal ImplicitMidpointRule an... (standardDauer)
    std::shared_ptr<DGVAcousticAssembleCombined> assemblerCombine = getAssembler(m_test2);

    TSolveR.SolveAndMeasure(assemblerCombine, Obs, s_linearizedSO, U_linearizedSO);

    if(Psi) {
        const std::shared_ptr<DGVAcousticRHSAssemblerLinearizedSecondOrder> &dataContainer =
                std::dynamic_pointer_cast<DGVAcousticRHSAssemblerLinearizedSecondOrder>(assemblerCombine->getRhsAssembler());
        std::shared_ptr<Vectors> Psi_u_prime2 = dataContainer->getUPrimeDirection2();
        //Falls U_linearizedSO als nullptr übergeben wurde, wird da auch nichts reingefüllt.
        if (U_linearizedSO) {
            *U_linearizedSO *= 0.5;
            *U_linearizedSO += *Psi_u_prime2;
        }
        //Same for seismogram:
        FWISeismogram s_linearized_u_prime2(s_linearizedSO);
//      std::string  name2 = "test2";
//      for(int i =0; i < Psi_u_prime2->size();++i) {
//        firstOrder.getVaPlotter()->plotWaveVector((*Psi_u_prime2)[i],i,name2.c_str());
//      }
        Obs.MeasureFullWave(*Psi_u_prime2, s_linearized_u_prime2);
        s_linearizedSO *= 0.5;
        s_linearizedSO += s_linearized_u_prime2;
    }
}

std::shared_ptr<DGVAcousticAssembleCombined> FWILinOpRedSecondOrder::getAssembler(const Vector &m_test2) {
//    std::cout << " ''''''''''' LinOpSO getAssembler just Started (rhs now)" << std::endl;
//  rhs used Implicit Midpoint Rule twice for building all the rhs
    auto rhs = std::make_shared<DGVAcousticRHSAssemblerLinearizedSecondOrder>(firstOrder.getUM(), firstOrder.getProb(),
                                                                             firstOrder.getEleCache(), firstOrder.getM(),
                                                                             m_test1, m_test2, firstOrder.getTConf().dt,
                                                                             firstOrder.isDiscrete(), firstOrder);
    auto assembleCombined = std::make_shared<DGVAcousticAssembleCombined>(firstOrder.getVaPlotter(),
                                                                          firstOrder.getMatrixAssembler(), rhs,
  firstOrder.getTConf().t0,firstOrder.getTConf().T,firstOrder.getTConf().dt);
    return std::move(assembleCombined);
}