#ifndef FWI_FWILINOPREDSECONDORDER_H
#define FWI_FWILINOPREDSECONDORDER_H

#include "Vector.hpp"
#include "FWITimeSolver.hpp"
#include "DGVAcousticAssembleCombined.hpp"
#include "DGVAcousticMatrixAssembler.hpp"
#include "FWIDGVAcousticMaterialFromVector.hpp"
#include "Assembler/FWIRhsAssemblerLinearized.hpp"
#include "FWILinOpRed.hpp"
#include "FWILinOpRedFirstOrder.hpp"

/**
 * @brief Class holds methods for calculating the second derivative (!?)
 *
 * With \Phi(m) = F(V(m)) we have
 * \Phi''(m)[m_test1, m_test2 ] = F''(V(m))[V'(m)[m_test1], V'(m)[m_test2])  +  F'(V(m))[V''(m)[m_test1, m_test2]]
 */

class FWILinOpRedSecondOrder : public FWILinOpRed {
    FWILinOpRedFirstOrder firstOrder;
    Vector m_test1; //Direction \Phi''(m)[m_test1, m_test2 (<- method "Solve" argument) ]
    const bool Psi;

public:
    FWILinOpRedSecondOrder(const Vectors &U_m_, const Vector &m_, Vector m_test1_, const std::shared_ptr<WaveProblemDGTSVA> &prob_,
                          const std::shared_ptr<PMLHandler> &PMLH_, const TimeConf &TConf_, const EleCacheType &eleCache_,
                          bool discrete, bool Psi_ = false);

    void Solve(const Vector &m_test2, Vectors &U_linearizedSO) override;
    // The first direction of \Phi''(Mat)[m_test1, m_test2] is a member

    void SolveAndMeasure(const Vector &m_test2, const FWIObservationOperatorNEW &Obs,
                         FWISeismogram &s_linearized, Vectors *U_linearizedSO) override;

    std::shared_ptr<DGVAcousticAssembleCombined> getAssembler(const Vector &m_test2);
};

#endif //FWI_FWILINOPREDSECONDORDER_H
