//
// Created by cr on 21.11.22.
//

#ifndef FWI_VTUMODELREADER_HPP
#define FWI_VTUMODELREADER_HPP

#include "Vector.hpp"
#include <iomanip>
#include <iostream>
#include <string>
#include <fstream>

class VTUModelReader {
    const std::vector<string> &filenames;
    std::unordered_map<Point, int> lookupMap;
    unsigned int currentIterate;

    cell *foundPoint(const VectorMatrixBase &mgr, double pp1, double pp2);

    void checkLines(const std::string &line_s,
                    std::vector<std::ifstream> &fileStreams,
                    size_t nrOfFiles);

    std::string reconstructParameter;

    void createLookuplistWithoutSave(const VectorMatrixBase &mgr);

    void readPointData(vector<double> &dataVTK, const std::string &filenameVTK, const std::string &) const;

public:
    VTUModelReader(const std::vector<string> &_filenames,
                   const VectorMatrixBase &mgr,
                   const std::string _reconstructParameter) :
            filenames(_filenames), currentIterate(0) {
        if (_reconstructParameter == "") Exit("specify inversion parameter!");
        createLookuplistWithoutSave(mgr);
    }

    void getData(Vector &M, std::string &DataStr, int param);


    const std::string &getCurrentIterateName() { return filenames[currentIterate]; }

    void nextIterate() {
        if (currentIterate > filenames.size()) {Exit("No next iterate!")}
        ++currentIterate;
    }
};

#endif //FWI_VTUMODELREADER_HPP
