//
// Created by cr on 17.02.22.
//

#ifndef FWI_NORMCALCULATOR_HPP
#define FWI_NORMCALCULATOR_HPP


#include "Vector.hpp"
#include "FWIMppHelpers.hpp"
#include "DGElement.hpp"
#include "DGAcousticElement.hpp"

class normCalculator {
    const std::list<unsigned int> &reconlist;
    const std::unordered_map<Point, DGAcousticElement *> &eleCache;
public:
    normCalculator(const std::list<unsigned int> &reconlist_,
                   const std::unordered_map<Point, DGAcousticElement *> &eleCache_
    ) : reconlist(reconlist_), eleCache(eleCache_) {}

    Vector calculateTVNormGradient(const Vector &Mat,
                                   Vector &GradX,
                                   Vector &GradY,
                                   int order = 1) const;

    void calculateModelGradient(const Vector &Mat,
                                const std::string &direc_s,
                                Vector &grad) const;

    static void calculateModelGradientCentral(const Vector &Mat,
                                              const std::string &direc_s,
                                              Vector &grad);

    double calculateTVNorm(Vector &Mat);

    double calculateTVNorm2(const Vector &Mat, int order, Vector *gradX,
                            Vector *gradY) const {
        Vector VecOne(Mat);
        VecOne = 1.0;
        Vector Grad(VecOne);
        if (gradX == gradY && gradX == nullptr) {
            Vector GradX(Mat);
            Vector GradY(Mat);
            GradX = 0.0;
            GradY = 0.0;
            Grad = calculateTVNormGradient(Mat, GradX, GradY, order);
        } else {
            Vector &GradX = *gradX;
            Vector &GradY = *gradY;
            Grad = calculateTVNormGradient(Mat, GradX, GradY, order);
        }
        double l2 = 0.0;
        for (const auto &elemIt : eleCache){
            auto elem = elemIt.second;
            double a = elem->Area();
            row ru = Grad.find_row(elem->GetCell()());
            for (unsigned int k: reconlist) {
                double para_u = Grad(ru, k);
                l2 += para_u * a;
            }
        }
        return PPM->Sum(l2);
    }

};

namespace nc {
    double CellWiseStepProduct(const Vector &U, const Vector &V, const DGAcousticElement &elem, double dt);

    double CellWiseTimeProduct(const Vectors &U, const Vectors &V, const DGAcousticElement &elem, double dt);

    double L2tScalar(const Vector &u, const Vector &v, const EleCacheType &eleCache, double dt);

    double L2TScalar(const Vectors &u, const Vectors &v, const EleCacheType &eleCache, double dt);

    double L2TNorm(const Vectors &u, const EleCacheType &eleCache, double dt);

    double L2MaterialScalar(const Vector &u, const Vector &v, const EleCacheType &eleCache,
                            const std::list<unsigned int> &reconlist);

    double L2MaterialNorm(const Vector &u, const EleCacheType &eleCache,
                          const std::list<unsigned int> &reconlist);
}

#endif //FWI_NORMCALCULATOR_HPP
