#include "FWISeismogramsmoothTest.hpp"
#include "Observation/FWISeismogram.hpp"
#include "FWIHelpers.hpp"
#include "Problems/Sources/FWISourceWavelet.hpp"

class SeismogramsmoothTest : public Test {
public:
    FWISeismogram loadData;

    SeismogramsmoothTest() {
        loadData.readFromFile(
                "s_ex__ex__point__point__levelex__12__2__dtex__0.0040000.002000__deg__2__rkorder__-1002__shot__0");
    }

    FWISeismogram LowPassSeismo(double freq) {
        return LowPass(loadData, freq);
    }

    FWISeismogram LowPass(FWISeismogram s, double freq, int order = 1) {
        s.LowPassFilter(freq,order);
        return s;
    }
    FWISeismogram BandPass(FWISeismogram s, double freq,double width, int order = 1) {
      s.BandPassFilter(freq,width, order);
      return s;
   }
    FWISeismogram SourceSeismo(double signalFreq) {
        FWIObservationSpecification spec = loadData.getSpecification();
        spec.resize(1);
        FWISeismogram seis(spec, 1);
        auto rick = RickerDerivativeWavelet(signalFreq, 0.2,spec.endTime(), spec.timeDelta(), spec.endTime(), 1);
        for (auto ts = seis.ts_begin(); ts != seis.ts_end(); ++ts) {
            auto ts1 = ts->second;
            for (auto &meas: *ts1) {
                meas.second[0] = rick(meas.first);
            }
        }
        return seis;
    }

    FWISeismogram LowPassSource(double signalFreq, double filterFreq) {
        return LowPass(SourceSeismo(signalFreq), filterFreq);
    }
    FWISeismogram BandPassSource(double signalFreq, double filterFreq, double width, int order = 1) {
    return BandPass(SourceSeismo(signalFreq), filterFreq, width, order);
  }
};

TEST_F(SeismogramsmoothTest, SeismogramsmoothTest) {
    LowPassSeismo(10).writeToFile("10Hz");
    LowPassSeismo(20).writeToFile("20Hz");
    LowPassSeismo(30).writeToFile("30Hz");
    LowPassSeismo(40).writeToFile("40Hz");
    LowPassSeismo(50).writeToFile("50Hz");
    SourceSeismo(10).writeToFile("10HzSourceUnfiltered");
    LowPassSource(50, 5).writeToFile("25HzSourceLowPass5Hz");
    LowPassSource(50, 10).writeToFile("25HzSourceLowPass10Hz");
    LowPassSource(50, 15).writeToFile("25HzSourceLowPass15Hz");
    LowPassSource(50, 20).writeToFile("25HzSourceLowPass20Hz");
    LowPassSource(50, 25).writeToFile("25HzSourceLowPass25Hz");
    LowPassSource(50, 30).writeToFile("25HzSourceLowPass30Hz");
    LowPassSource(50, 35).writeToFile("25HzSourceLowPass35Hz");
    LowPassSource(50, 40).writeToFile("25HzSourceLowPass40Hz");
    LowPassSource(50, 50).writeToFile("25HzSourceLowPass50Hz");
    LowPassSource(50, 60).writeToFile("25HzSourceLowPass60Hz");
    BandPassSource(50, 6,5).writeToFile("25HzSourceBandPass5Hz");
    BandPassSource(50, 10,5).writeToFile("25HzSourceBandPass10Hz");
    BandPassSource(50, 15,5).writeToFile("25HzSourceBandPass15Hz");
     BandPassSource(50, 15,10,2).writeToFile("25HzSourceBandPass15HzWidth10Order2");
  BandPassSource(50, 15,10,1).writeToFile("25HzSourceBandPass15HzWidth10Order1");

   BandPassSource(50, 20,5).writeToFile("25HzSourceBandPass20Hz");
    BandPassSource(50, 25,5).writeToFile("25HzSourceBandPass25Hz");
    BandPassSource(50, 30,5).writeToFile("25HzSourceBandPass30Hz");

   BandPassSource(50, 35,5).writeToFile("25HzSourceBandPass35Hz");
    BandPassSource(50, 40,5).writeToFile("25HzSourceBandPass40Hz");
    BandPassSource(50, 50,5).writeToFile("25HzSourceBandPass50Hz");
    BandPassSource(50, 60,5).writeToFile("25HzSourceBandPass60Hz");
    SourceSeismo(25).writeToFile("25HzSourceUnfiltered");
    SourceSeismo(20).writeToFile("20HzSourceUnfiltered");
    SourceSeismo(30).writeToFile("30HzSourceUnfiltered");
    SourceSeismo(40).writeToFile("40HzSourceUnfiltered");
    SourceSeismo(50).writeToFile("50HzSourceUnfiltered");

}

int main(int argc, char **argv) {
    return MppTest(
            MppTestBuilder(argc, argv).
                    WithoutDefaultConfig().
                    WithConfigEntry("TimeIntegratorVerbose", 0).
                    WithScreenLogging().
                    WithConfPath("../").
                    WithGeoPath("../conf/geo/").
                    WithPlotPath("data/vtu/").
                    WithPPM()
    ).RUN_ALL_MPP_TESTS();
}
