//
// Created by cr on 21/12/22.
//

#ifndef FWI_FWITESTLINEARIZEDMATRIX_H
#define FWI_FWITESTLINEARIZEDMATRIX_H
#include "TestEnvironment.hpp"
#include "Vector.hpp"
#include "FWIHelpers.hpp"
#include "FWIAbstractInterfaceDGTSVA.hpp"
#include "Problems/FWIInversionProblems.hpp"
#include "DGVAcousticPlotter.hpp"
#include "Assembler/FWILinearizedAssembler.hpp"
#include "DGAssemble.hpp"
#include "FWIVectorFillFunctions.hpp"
#include "FWIProblemWrapperVA.hpp"

using ConfigMap = std::map<std::string, std::string>;
const ConfigMap TestFWILinearized{
    {"ConfigVerbose",                  "2"},
    {"ForwardSolver",                  "VADGTS"},
    {"Distribution",                   "RCB"},
    {"Model",                          "FWIInverse_DGTS"},
    {"CountCellsOld",                  "0"},
    {"dim",                            "1"},
    {"taup_bg",                        "0.1"},
    {"Mesh",                           "oneDim"},
    {"Problem",                        "Benchmark1D"},
    {"startingValue_",                 "1"},
    {"Preconditioner_S",               "SuperLU"},
    {"Preconditioner_M",               "PointBlockJacobi"},
    {"Problem",                        "VAcousticWaveC2Benchmark"},
    {"ReconstructionType",             "Classical"},
    {"sourceSpace",                    "2"},
    {"useTaper",                       "0"},
    {"taperRadiusSource",              "0"},
    {"taperRadiusReceiver",            "100"},
    {"taperExponent",                  "-0.1"},
    {"taperIgnoreReceiverCells",       "1"},
    {"taperIgnoreSourceCells",         "0"},
    {"Monoparam",                      "Vp"},
    {"SignalName",                     "Ricker"},
    {"SignalFreq",                     "10"},
    {"SignalCutoff",                   "0"},
    {"SignalDelay",                    "0"},
    {"numberOfSourcesPerShot",         "1"},
    {"rho_bg",                         "2"  },
    {"v_bg",                           "5.0"},
    {"size_perturb",                   "0"},
    {"FWIReceiverWaypoints",           "2"},
    {"FWIReceiverWaypoint0",           "2.1 0"},
    {"FWIReceiverWaypoint1",           "3.5 0"},
    {"FWIReceiverIncludeLastWaypoint", "0"},
    {"FWIReceiverCount0",              "20"},
    {"UseSparseMatrix",                "1"},
    {"f_source",                       "10"},
    {"FWISourceWaypoints",             "2"},
    {"FWISourceWaypoint0",             "0.5, 0"},
    {"FWISourceWaypoint1",             "0.5, 0"},
    {"FWISourceCount0",                "1"},
    {"FWISourceIncludeLastWaypoint",   "0"},
    {"numL",                           "1"},
    {"compMeasure",                    "0"},
    {"source_component",               "2"},
    {"allowPlotting_M",                "1"},
    {"allowPlotting_WF",               "1"},
    {"plotFactor",                     "1"},
    {"parametrization",                "taup_positive"},
    {"T",                              "1.00"},
    {"dt",                             "0.005"},
    {"deg",                            "1"},
    {"level",                          "5"},
    {"plot_tStep",                     "-1"},
    {"rkorder",                        "-2"},
    {"Overlap",                        "dG1"},
    {"LinearSolver",                   "LS"},
    {"LinearVerbose",                  "-1"},
    {"FWIReconstructRho",              "0"},
    {"FWIReconstructV_P",              "1"},
    {"plotV",                          "1"},
    {"dampingFlux", "0"}
};
enum{
  MASS_MATRIX = 1,
  FLUX_MATRIX = 2,
  DAMPING_MATRIX =3
};
struct TestParams {
  std::string parametrization;
  int param;
  int matrix;
  std::vector<double> values = std::vector<double>(6, 0.0);
};

class FWITestLinearizedMatrix : public TestWithParam<TestParams> {
protected:
  std::unique_ptr<WPConf> Conf = nullptr;
  std::shared_ptr<Meshes> meshes;
  std::shared_ptr<WaveProblemDGTSVA> prob;
  std::unique_ptr<FWIDiscretizationStack> discStack;
  std::unique_ptr<EleCacheContainer> eleCache;
  std::unique_ptr<Vector> Mat;
  std::unique_ptr<DGVAcousticPlotter> plotter;
  const std::list<unsigned int> reconlist{0, 1, 2};
public:
  FWITestLinearizedMatrix() {
    Config::Initialize(TestFWILinearized);
    Conf = std::make_unique<WPConf>();
    meshes = getFWIMesh();
    prob = GetInversionProblem(*Conf);
    prob->setParametrization(GetParam().parametrization);
    int deg = 0;
    Config::Get("deg", deg);
    discStack = std::make_unique<FWIDiscretizationStack>(meshes, prob->getParamComps(), prob->GetStateComponents(),
                                                         deg, LevelPair{Conf->level, -1});
    eleCache = std::make_unique<EleCacheContainer>(*discStack->wV, prob->getNumL());
    plotter = std::make_unique<DGVAcousticPlotter>(prob->getNumL(), &eleCache->Get());
    Mat = std::make_unique<Vector>(*discStack->mV);
    Mat->Clear();
  }

  std::shared_ptr<DGAssemble> GetAssemble(const Vector &Material) {
    auto Mp = std::make_unique<Vector>(*discStack->mV);
    Mp->Clear();
    ReparametrizeMaterialVector(Material, *prob, *Mp);
    Mp->Accumulate();
    auto WrappedProblem = std::make_shared<FWIProblemWrapperVA>(*prob, *prob, std::move(Mp),
                                                                prob->getNumL(), prob->getDim(), prob);
    auto assemble = std::make_shared<DGAssemble>(WrappedProblem, true, -1, Conf->TConf.t0,
                                                 Conf->TConf.T,Conf->TConf.dt);
    assemble->SetElementCache(&eleCache->Get());
    return std::move(assemble);
  }

  void TearDown() override {
    Config::Close();
  }
  void GetMatrix(const std::shared_ptr<DGAssemble> &assemble, Matrix& M, int i){
    if (i == MASS_MATRIX){
      assemble->MassMatrix(M);
    }else if (i == FLUX_MATRIX){
      assemble->FluxMatrixOnly(M);
    } else if (i == DAMPING_MATRIX){
      assemble->DampingMatrix(M);
      M *= (-1.0);
    }
  }
  void GetLinearizedMatrix(const FWILinearizedAssembler &assemble,const Vector& Rand, Matrix& M, int i){
    if (i == MASS_MATRIX){
      assemble.assembleLinearizedMGalerkin(Rand,M);
    }else if (i == FLUX_MATRIX){
      assemble.assembleLinearizedFluxMatrix(Rand,M);
    }else if (i == DAMPING_MATRIX){
      assemble.assembleDGalerkin2(Rand,M);
    }
  }
};


#endif //FWI_FWITESTLINEARIZEDMATRIX_H
