//
// Created by lionbart on 27.02.23.
//

#include "FWITestLinearizedMatrixMaterialSecondOrder.h"

#include <experimental/random>
#include "FWITestLinearizedMatrix.h"
#include "normCalculator.hpp"


TEST_P(FWITestLinearizedMatrixSecondOrder, FWITestLinearizedMatrixSO) {
    Vector Rand1(*Mat);
    Vector Rand2(*Mat);
    Vector U(*discStack->wV);
    Rand1.Clear();
    Rand2.Clear();
    std::experimental::reseed(42);
    //Fill Random input into Material (type Vector)
    // WHY three values for each cell if material is constant??
    for (cell c = Rand1.cells(); c != Rand1.cells_end(); ++c) {
        auto r = 1 + 2 * double(std::experimental::randint(0, 10000) / 10000.0);
        auto r2 = 1 + 2 * double(std::experimental::randint(0, 10000) / 10000.0);
        auto r3 = 2 * double(std::experimental::randint(0, 10000) / 10000.0);
        (*Mat)(c(), 0) = r;
        (*Mat)(c(), 1) = r2;
        (*Mat)(c(), GetParam().param) = r;
    }
    Vector ShiftMat(*Mat);
    ShiftMat.Clear();
    FWILinearizedAssembler ass(prob, eleCache->Get(), *Mat);
    auto assnorm = GetAssemble(*Mat);

    //Setup Material Directions Rand1, Rand2 (type Vector)
    for (cell c = Rand1.cells(); c != Rand1.cells_end(); ++c) {
        auto r1 = 2 * double(std::experimental::randint(0, 10000) / 10000.0);
        auto r2 = 2 * double(std::experimental::randint(0, 10000) / 10000.0);
        Rand1(c(), GetParam().param) = r1;
        Rand2(c(), GetParam().param) = r2;
    }
    Rand1 *= 1.0 / nc::L2MaterialNorm(Rand1, eleCache->Get(), reconlist);
    Rand2 *= 1.0 / nc::L2MaterialNorm(Rand2, eleCache->Get(), reconlist);


    /*
     * Prepare each part of differential Quotient:
     * f'(Material + Rand2)[Rand1] - f'(Material)[Rand1] - f''(Material)[Rand1, Rand2]
     *       Prime                -       SubPrime       -      doublePrime
     *
     * Other Idea: (not implemented): f''(x)[h1, h2] =
     *      lim 1/||h1||||h2|| * (f(x + h1 + h2) + f(x) - f(x + h1) - f(x + h2))
     *
     * First uses first Derivative, second method not. Just in case this should be tested independently.
    */
//    Matrix M(U); //Not Necessary?!!
//    GetMatrix(assnorm,M,GetParam().matrix); //Brauche ich für ersten Ansatz nicht.

    Matrix Prime(U);
    Matrix SubPrime(U);
    Matrix doublePrime(U);
    GetLinearizedMatrix(ass, Rand1,SubPrime,GetParam().matrix);
    GetSecondLinearizedMatrix(ass, Rand1, Rand2, doublePrime, GetParam().matrix);

    std::vector<double> quotients(6, 0.0);
    Vector MatPlusScaleRand2(*Mat);
    std::vector<double> factors {1.0, 0.1, 0.01, 0.001, 0.0001, 0.00001};

    for (int i = 0; i < 6; ++i) {
        Matrix DiffQuot(U);
        //Regard Limes (Rand2 to 0) as fac*Rand2
        Scalar fac = factors[i];
//        Scalar fac = pow(10, -i); //For i=0 fac = 2.6e-322...; i=1 fac = 1... häää

        //Prime ~ f(Mat + Rand2)[Rand1] needs to be recalculated in every step since not linear in Rand2.
        Vector Rand2Scale = Rand2;
        Rand2Scale *= fac;
        MatPlusScaleRand2 = *Mat + Rand2Scale;

        FWILinearizedAssembler assShifted(prob, eleCache->Get(), MatPlusScaleRand2);
        Prime = 0.0;
        GetLinearizedMatrix(assShifted,Rand1,Prime,GetParam().matrix);
//        MOUT(Prime.norm());
        Prime += (-1.0) * SubPrime;
        Prime += (-fac) * doublePrime;

        quotients[i] = Prime.norm()/fac;
    }
    MOUT(quotients);
    EXPECT_VECTOR_NEAR(quotients, GetParam().values, max(quotients[0] * 1e-7,1e-10));
}
INSTANTIATE_TEST_SUITE_P(FWITestLinearizedMatrixSecondOrder, FWITestLinearizedMatrixSecondOrder, Values(
//        TestParamsSO{"taup",0,MASS_MATRIXSO, {0, 0, 0, 0, 0, 0}},
        TestParamsSO{"taup",0,MASS_MATRIXSO, {0.0022459954, 0.00031301238, 3.258099e-05, 3.2714731e-06, 3.2728167e-07, 3.2729536e-08}},
        TestParamsSO{"taup",1,MASS_MATRIXSO, {0.047689972, 0.0082647894, 0.00089025047, 8.9715015e-05, 8.9784576e-06, 8.9791601e-07}}));
//        TestParamsSO{"taup",1,MASS_MATRIXSO, {0, 0, 0, 0, 0, 0}}));

        /*TestParams{"native",1,MASS_MATRIXSO, {22.109882, 14.237205, 3.1700744, 0.36189663, 0.036710521,
                                            0.0036763444}},
        TestParams{"native",2,MASS_MATRIXSO, {61.136744, 49.591937, 17.289952, 2.3023656, 0.23813514,
                                            0.023895493}},
        TestParams{"native",0,FLUX_MATRIXSO, {58.707154, 45.391566, 21.893275, 4.384172,
                                            0.49746691,0.050449789}},
        TestParams{"native",1,FLUX_MATRIXSO, { 3.3941962, 0.92052305, 0.11551639, 0.011868747,
                                             0.0011901609,0.000119049}},
        TestParams{"native",2,FLUX_MATRIXSO, {3.3941962, 0.92052305, 0.11551639, 0.011868747,
                                            0.0011901609, 0.000119049}},
        TestParams{"native",2,DAMPING_MATRIXSO, {3841.3349, 3115.9533, 1086.3597, 144.66189, 14.962472, 1.5013981}},
        TestParams{"taup",0,MASS_MATRIXSO, {32954.718284568997, 32028.742615239782, 25094.295839286307,
                                          7928.8218185107989, 1011.2793, 103.99237}},
        TestParams{"taup",1,MASS_MATRIXSO, {82379.560659440976, 62740.695595155397, 17083.167078519928,
                                          2041.0486291673767, 208.11016, 20.851881}},
        TestParams{"taup",2,MASS_MATRIXSO, {19909.763289753446, 10353.917154006049, 1785.3430469132998,
                                          192.47927, 19.399476,1.9414763}},
        TestParams{"taup",0,FLUX_MATRIXSO, {117.52293, 44.333615, 6.1630617, 0.64139099,
                                          0.064401294,0.0064427628}},
        TestParams{"taup",1,FLUX_MATRIXSO, {48.059946, 13.649703, 1.6963577, 0.17391154, 0.017435162,
                                          0.0017439575}},
        TestParams{"taup",2,FLUX_MATRIXSO, {0.81954064, 0.10824454,0.01118995, 0.0011227966,0.00011231786,
                                          1.1232107e-05}},
        TestParams{"taup",0,DAMPING_MATRIXSO, {1165343.4, 1131428.9, 886437.03, 280079.45,  35722.651, 3673.4492}},
        TestParams{"taup",1,DAMPING_MATRIXSO, {5170541.1, 3937922.6, 1072229, 128106.97, 13062.091, 1308.774}},
        TestParams{"taup",2,DAMPING_MATRIXSO, {1250967.3, 650555.8, 112176.41, 12093.83, 1218.905, 121.98655}},
        TestParams{"taup_positive",0,MASS_MATRIXSO, {0.18871896, 0.015032206, 0.0014720252, 0.00014689676,
                                                   1.4686625e-05,1.4686341e-06}},
        TestParams{"taup_positive",1,MASS_MATRIX, {0.046756905, 0.006844487, 0.00071500229,
                                                   7.1817585e-05, 7.1849442e-06,7.1852584e-07}},
        TestParams{"taup_positive",2,MASS_MATRIX, {0.015060111, 0.0011947275, 0.00011700487,
                                                   1.1676372e-05, 1.1673966e-06, 1.1673685e-07}},
        TestParams{"taup_positive",0,FLUX_MATRIX, {14.670229, 1.2239617, 0.12035963, 0.012015936,
                                                   0.0012013936, 0.00012013764}},
        TestParams{"taup_positive",1,FLUX_MATRIX, {14.670229, 1.2239617, 0.12035963, 0.012015936,
                                                   0.0012013936,0.00012013796}},
        TestParams{"taup_positive",2,FLUX_MATRIX, {0.29466384, 0.027173636, 0.0026978722, 0.00026959734,
                                                   2.6957833e-05, 2.6957903e-06}},
        TestParams{"taup_positive",0,DAMPING_MATRIX, {0.27987723, 0.033652253, 0.0034325371, 0.00034393923, 3.4400788e-05, 3.4401457e-06}},
        TestParams{"taup_positive",1,DAMPING_MATRIX, {0.93587548, 0.13171988, 0.013699788, 0.0013754518, 0.00013760011, 1.3760545e-05}},
        TestParams{"taup_positive",2,DAMPING_MATRIX, {0.13805148, 0.016295158, 0.0016587577,
                                                      0.00016617322, 1.6620304e-05, 1.6620585e-06}}
));*/
int main(int argc, char **argv) {
    return MppTest(
            MppTestBuilder(argc, argv).
                    WithoutDefaultConfig().
                    WithScreenLogging().
                    WithConfPath("../").
                    WithGeoPath("../conf/geo/").
                    WithPlotPath("data/vtu/").
                    WithPPM()
    ).RUN_ALL_MPP_TESTS();
}
