#include "FWITestProcTransfer.hpp"

TEST_P(FWITestProcTransfer2, Test1) {
    PMLWaveVectorFWI PMLVec(PML);
    *Vec = infty;
    PMLVec.Clear();
    for (cell c = PMLVec.cells(); c != PMLVec.cells_end(); ++c) {
        int i = PMLVec.Id(c());
        for (int k = 0; k < PMLVec.Dof(i); ++k) {
            PMLVec(c(), k) = k;
        }
    }
    PMLVec.PaddedToUnpadded(*Vec);
    for (cell c = Vec->cells(); c != Vec->cells_end(); ++c) {
        int i = Vec->Id(c());
        for (int k = 0; k < Vec->Dof(i); ++k) {
            (*Vec)(c(), k) -= k;
        }
    }
    EXPECT_NE(Vec->ConsistentNorm(), 1e-16);
}

TEST_P(FWITestProcTransfer2, Test2) {
    VecMat->Clear();
    for (cell c = VecMat->cells(); c != VecMat->cells_end(); ++c) {
        CellBoundaryFaces BF(VecMat->GetMesh(), c);
        if (BF.size() == 0) continue;
        int i = VecMat->Id(c());
        for (int k = 0; k < VecMat->Dof(i); ++k) {
            (*VecMat)(c(), k) = sqrt(c()[0] * c()[0] + c()[1] + c()[1]) + 0.1 * k;
        }
    }
    PMLMatVectorFWI PMLVec(PML, *VecMat,true);
    Vector VecFullPML(0.0, PMLFull.GetMatVector());
    for (cell c = PMLVec.cells(); c != PMLVec.cells_end(); ++c) {
        int i = PMLVec.Id(c());
        for (int k = 0; k < PMLVec.Dof(i); ++k) {
            VecFullPML(c(), k) = PMLVec(c(), k);
        }
    }

    for (cell c = VecMat->cells(); c != VecMat->cells_end(); ++c) {
        CellBoundaryFaces BF(VecMat->GetMesh(), c);
        if (BF.size() == 0) continue;
        int i = VecMat->Id(c());
        for (int k = 0; k < VecMat->Dof(i); ++k) {
            VecFullPML(c(), k) = sqrt(c()[0] * c()[0] + c()[1] + c()[1]) + 0.1 * k;
        }
    }
    auto result = TestPMLTransfer(PML.GetNonPMLBoundaryToPML(), VecFullPML);
    EXPECT_NE(result.first + result.second, 1e-16);

}

INSTANTIATE_TEST_SUITE_P(FWITestProcTransfer2, FWITestProcTransfer2, Values(
        TestParams{2, 0, 0.5, 0, 1, 2, {1, 1, 1, 1}, {0, 0, 1, 1}, {0.5, 0.5, 0.5, 0.5}},
        TestParams{4, 0, 0.5, 0, 1, 2, {1, 1, 1, 1}, {0, 1, 1, 1}, {0.5, 0.5, 0.5, 0.5}},
        TestParams{4, 0, 0.5, 0, 1, 2, {1, 1, 1, 1}, {0, 1, 1, 1}, {1, 1, 1, 1}},
        TestParams{4, 0, 0.5, 0, 1, 2, {1, 1, 1, 1}, {0, 1, 1, 1}, {1.5, 1.5, 1.5, 1.5}},
        TestParams{4, 0, 0.5, 0, 1, 2, {1, 1, 1, 1}, {1, 1, 1, 1}, {1.5, 1.5, 1.5, 1.5}},
        TestParams{4, 0, 0.5, 0, 1, 2, {1, 1, 1, 1}, {1, 1, 1, 0}, {1.5, 1.5, 1.5, 1.5}},
        TestParams{4, 0, 0.5, 0, 1, 2, {1, 1, 1, 1}, {1, 1, 0, 1}, {1.5, 1.5, 1.5, 1.5}},
        TestParams{4, 0, 0.5, 0, 1, 2, {1, 1, 1, 1}, {1, 1, 0, 0}, {1.5, 1.5, 1.5, 1.5}},
        TestParams{4, 0, 0.5, 0, 1, 2, {1, 1, 1, 1}, {1, 0, 1, 1}, {1.5, 1.5, 1.5, 1.5}},
        TestParams{4, 0, 0.5, 0, 1, 2, {1, 1, 1, 1}, {1, 0, 1, 0}, {1.5, 1.5, 1.5, 1.5}},
        TestParams{4, 0, 0.5, 0, 1, 2, {1, 1, 1, 1}, {1, 0, 0, 0}, {1.5, 1.5, 1.5, 1.5}},
        TestParams{4, 0, 0.5, 0, 1, 2, {1, 1, 1, 1}, {0, 1, 1, 1}, {1.5, 1.5, 1.5, 1.5}},
        TestParams{4, 0, 0.5, 0, 1, 2, {1, 1, 1, 1}, {0, 0, 1, 1}, {1.5, 1.5, 1.5, 1.5}},
        TestParams{4, 0, 0.5, 0, 1, 2, {1, 1, 1, 1}, {0, 0, 0, 1}, {1.5, 1.5, 1.5, 1.5}},
        TestParams{4, 0, 0.5, 0, 1, 2, {1, 1, 1, 1}, {0, 0, 0, 0}, {1.5, 1.5, 1.5, 1.5}},
        TestParams{4, 0, 0.5, 0, 1, 2, {1, 1, 1, 1}, {0, 0, 0, 0}, {1.5, 1.5, 1.5, 1.5}},
        TestParams{4, 0, 1, 0, 8, 8, {1, 1, 1, 1}, {0, 0, 0, 0}, {0.75, 0.75, 0.75, 0.75}},
        TestParams{2, 0, 0.5, 0, 0.5, 1, {1, 1, 1, 1}, {0, 0, 0, 0}, {1.5, 1.5, 1.5, 1.5}})
);

int main(int argc, char **argv) {
    MppTest mppTest = MppTestBuilder(argc, argv).WithPPM().WithScreenLogging();
    return mppTest.RUN_ALL_MPP_TESTS();
}