#ifndef MPP_PROCTRANSFER_HPP
#define MPP_PROCTRANSFER_HPP

#include "TestEnvironment.hpp"
#include "PMLHandler.hpp"

struct TestParams {
    int nrCellsX;
    double LeftBnd;
    double RightBnd;
    double BottomBnd;
    double TopBnd;
    int ratio;
    vector<int> bndFlags;
    vector<int> pmlBndFlags;
    vector<double> PMLPaddingLength;
};


class FWITestProcTransfer2 : public TestWithParam<TestParams> {
    void initialize(double LeftBnd, double RightBnd, double BottomBnd, double TopBnd,
                    int dimX, int dimY, int bndTop, int bndBottom, int bndLeft, int bndRight,
                    std::shared_ptr<RectangularGeometry2D> &Recti,
                    std::shared_ptr<Meshes> &meshesi,
                    std::shared_ptr<DGDiscretization> &Disci,
                    std::shared_ptr<Vector> &Veci,
                    std::string &olapName) {
        Recti = std::make_shared<RectangularGeometry2D>(std::pair<double, double>({LeftBnd, RightBnd}),
                                                        std::pair<double, double>({BottomBnd, TopBnd}),
                                                        dimX, dimY,
                                                        std::vector<int>{bndLeft, bndRight, bndBottom, bndTop},
                                                        "Rect");
        meshesi = MeshesCreator().
                WithCoarseGeometry(Recti).
                WithOverlap(olapName).
                WithDistribute(distName).
                CreateShared();
        Disci = std::make_shared<DGDiscretization>(*meshesi, 1, 3);
        Veci = std::make_shared<Vector>(0.0, *Disci);

    }

public:
    std::string distName = "RCB";
    std::string olapNameMember = "CellsOverlap_dG1";
    std::string olapNameTest = "FullOverlap";

    std::shared_ptr<RectangularGeometry2D> RectFull;
    std::shared_ptr<Meshes> meshesFull;
    std::shared_ptr<DGDiscretization> DiscFull;
    std::shared_ptr<Vector> VecFull;


    std::shared_ptr<RectangularGeometry2D> Rect;
    std::shared_ptr<Meshes> meshes;
    std::shared_ptr<DGDiscretization> Disc;
    std::shared_ptr<Vector> Vec;
    std::shared_ptr<DGDiscretization> DiscMat;
    std::shared_ptr<Vector> VecMat;
    PMLHandler PML;
    PMLHandler PMLFull;

    FWITestProcTransfer2() {

        double LeftBnd = GetParam().LeftBnd;
        double RightBnd = GetParam().RightBnd;
        double BottomBnd = GetParam().BottomBnd;
        double TopBnd = GetParam().TopBnd;
        int nrCellsX = GetParam().nrCellsX;
        int nrCellsY = GetParam().ratio * nrCellsX;
        int dimX = nrCellsX + 1;
        int dimY = nrCellsY + 1;
        int bndLeft = GetParam().bndFlags[0];
        int bndRight = GetParam().bndFlags[1];
        int bndBottom = GetParam().bndFlags[2];
        int bndTop = GetParam().bndFlags[3];
        LevelPair lev = {0, -1};
        initialize(LeftBnd, RightBnd, BottomBnd, TopBnd, dimX, dimY, bndTop, bndBottom, bndLeft, bndRight,
                   Rect, meshes, Disc, Vec, olapNameMember);
        DiscMat = std::make_shared<DGDiscretization>(*meshes, 0, 3);
        VecMat = std::make_shared<Vector>(0.0, *DiscMat);
        PMLConf Conf;
        Conf.pmlBndFlags = GetParam().pmlBndFlags;
        Conf.SetPMLPaddingLength(GetParam().PMLPaddingLength);
        PML.Initialize(Vec->GetMesh(), Conf, 1, 0, 3);
        initialize(LeftBnd, RightBnd, BottomBnd, TopBnd, dimX, dimY, bndTop, bndBottom, bndLeft, bndRight,
                   RectFull, meshesFull, DiscFull, VecFull, olapNameTest);
        PMLFull.Initialize(VecFull->GetMesh(), Conf, 1, 0, 3);
    }

    std::pair<int, double> TestPMLTransfer(const std::unordered_map<Point, std::vector<std::pair<Point, short>>>
                                           &BoundaryCellToTransfer, const Vector &VecFuP) {
        double d = 0;
        for (const auto &PointAndTransferVector: BoundaryCellToTransfer) {
            const auto &SendPoint = PointAndTransferVector.first;
            for (const auto &Transfer: PointAndTransferVector.second) {
                const auto &ReceivePoint = Transfer.first;
                short receiveProc = Transfer.second;
                int i = VecFuP.Id(ReceivePoint);
                if (VecFuP.find_procset(ReceivePoint) == VecFuP.procsets_end()) continue;
//                std::cout << " ReceivePointProc: " << " " << (VecFullPML->find_procset(ReceivePoint).master()) << endl;
                if (receiveProc != (VecFuP.find_procset(ReceivePoint).master())) {
                    return {1, infty};
                }
                for (int k = 0; k < VecFuP.Dof(i); ++k) {
                    double val = VecFuP(ReceivePoint, k) - VecFuP(SendPoint, k);
                    d += val * val;
                }
            }
        }
        return {0, d};
    }
};

#endif //MPP_PROCTRANSFER_HPP
