#include "../gtests/googletest/include/gtest/gtest.hpp"
#include "../src/RegularisationStrategy.hpp"
#include <istd::ostream>
#include <cmath>
using Residuals= std::vector<double>;
using Stops= std::vector<double>;
using Iterations= std::vector<unsigned int>;
using Thetas = std::vector<double>;
auto roundToCompare(std::vector<double> input,int digit) -> std::vector<double> {
    int factor = 1;
    for (int i=0; i< digit; i++) factor *= 10;
    for (int i =0; i< input.size(); i++){
        input[i] = std::floor(input[i]*factor);
    }
    return input;
}
TEST(Regularisation, Thetas) {
    RegularisationStrategyWinkler regu0(0.9,0.999);
    auto res0=Residuals{
            10370.4,8107.48,3070.46,2736.03,2282.11,2049.53,
            1793.53,2042.81,1560,1445.95};
    auto inner0=Iterations{1u, 2u, 2u, 3u, 2u, 5u, 4u, 4u, 8u, 12u};
    auto thetasExpected =Thetas{999, 781, 890,801,
                                867,781,912, 821,
                                739,869};
    auto thetaResult =Thetas{};
    for (int i = 0; i<res0.size(); i++){
        regu0.addResNorm(res0[i]);
        thetaResult.push_back(regu0.calcTheta());
        regu0.addInnerIter(inner0[i]);
    }
    EXPECT_EQ(roundToCompare(thetaResult,3),thetasExpected); //
}
TEST(Regularisation, maxInner) {
    RegularisationStrategyWinkler regu0(0.9,0.999);
    auto inner0=Iterations{1u, 2u, 2u, 3u, 2u, 5u, 4u, 4u, 8u, 12u};
    auto maxExpected= Iterations {1, 2, 3, 4, 5, 5, 7, 9, 8, 12};
    auto maxResult = Iterations{};
    for (int i = 0; i<inner0.size(); i++){
        maxResult.push_back(regu0.calcMaxInner());
        regu0.addInnerIter(inner0[i]);
    }
    EXPECT_EQ(maxResult,maxExpected); //
}
TEST(Regularisation, innerStop) {
    RegularisationStrategyWinkler regu0(0.9,0.999);
    auto res0=Residuals{10370.4,8107.48,3070.46,2736.03,2282.11,2049.53,
                        1793.53,2042.81,1560,1445.95};
    auto thetas =Thetas{0.999, 0.78179, 0.890895,0.801805,
                                0.86787,0.781083,0.912433,0.82119,
                                0.739071,0.869535};
    auto innerStopExpected = Stops{103600,63383, 27354, 21937,
                           19805, 16008, 16364, 16775,
                           11529, 12573};
    auto innerStopResult = Stops{};
    for (int i = 0; i<innerStopExpected.size(); i++){
        regu0.addResNorm(res0[i]);
        regu0.addTheta(thetas[i]);
        innerStopResult.push_back(regu0.calcInnerStopp());
    }
    EXPECT_EQ(roundToCompare(innerStopResult,1),innerStopExpected); //
}
