#ifndef _DGASSEMBLE_H_
#define _DGASSEMBLE_H_
const int MaxQuadraturePointsDG = 36;
const int MaxNodalPointsDG = 25;
class DGAssemble : public Assemble {
protected:
    Plot& plot;
    Scalar gamma;
public:
    DGAssemble (Plot& _plot) : plot(_plot), gamma(1) {}
    double Residual(const Vector&, Vector&) const { Exit("Not implemented!"); return 0.0; }
    virtual const char* Name() const { return "DGAssemble"; }
    virtual string DiscName() const = 0;
    virtual double Energy (const Vector& u, double x0, double x1) const = 0;
    virtual double Energy (const Vector& u) const = 0;
    virtual void SetInitialValue (Vector& u) const = 0;
    virtual void PrintStep (double, const Vector&, int) = 0;
    void VtkPlotting (const Vector&, int) {}
    void VtkPlotting (const Vector&, int, double) {}
    virtual void InitialStep (double t, Vector& u, bool = false) {
        u = 0;
        PrintStep(0,u,t);
        VtkPlotting(u,0);
    }
    virtual void FinishStep (Vector& u, double t, double dt,
                             int step, int plot_step = 1) {
        if(step%plot_step == 0){
            PrintStep(step,u,t);
            VtkPlotting(u,step);
        }
    }
    virtual void PrintMatrixInfo(Matrix& A, int diagonal = 1) = 0;
    virtual void FluxMatrix (Matrix&) const = 0;
    virtual void FluxMatrixOnly (Matrix&) const {Exit("not implemwnted!");}
    virtual void MassMatrix (Matrix&) const = 0;
    virtual void RHS (double, Vector&) const = 0;
    virtual bool RHS (double) const = 0;
    void SetGamma (Scalar g) { gamma = g; }
    void PrintInfo (const Vector& u) {
        mout << endl << " " << Name() << endl;
        Assemble::PrintInfo(u);
    }
    cell find_cell (const Mesh& M, const face& f, const Point& z) const {
        cell c = M.find_cell(z);
        if (c == M.cells_end())
            c = M.find_overlap_cell(z);
        if (c == M.overlap_end()) {
            Exit("Error: no cell found");
            pout << "z " << z << endl;
            pout << "Cells: " << M.Cells::size() << endl << M.Cells::ref();
            pout << "OverlapCells: " << M.OverlapCells::size() << endl
                 << M.OverlapCells::ref();
            Exit("Error: no cell found");
        }
        return c;
    }

    cell find_nb_cell(const VectorMatrixBase &A, const cell &c, int f) const {
        face ff = A.GetMesh().find_face(c.Face(f));
        if (c() != ff.Right())
            return find_cell(A.GetMesh(), ff, ff.Right());
        else
            return find_cell(A.GetMesh(), ff, ff.Left());
    }

    bool check_if_bnd(const VectorMatrixBase &A, const cell &c, int f) const {
        face ff = A.GetMesh().find_face(c.Face(f));
        if ((A.GetMesh().find_bnd_face(c.Face(f)) != A.GetMesh().bnd_faces_end())
            && ((ff.Right() == Infty) || (ff.Left() == Infty)))
            // To exclude periodic bnd
            return true;
        else
            return false;
    }

    int find_nb_face_id(const VectorMatrixBase &A,
                        const Point &f_c, const cell &cf) const {
        for (int f1 = 0; f1 < cf.Faces(); ++f1)
            if (f_c == cf.Face(f1)) return f1;
        Exit("Error: no face found");
        return 0;
    }

    int bnd_id(const VectorMatrixBase &A, const cell &c, int f) const {
        bnd_face fff = A.GetMesh().find_bnd_face(c.Face(f));
        return fff.Part();
    }

    virtual ~DGAssemble() {}
};
class DGTSDiscretization  : public Discretization {
    const Shape* S_tri[5];
    const Shape* S_quad[6];
    const Shape* S_int[5];
//    const Shape* S_tet[2];
//    const Shape* S_hex[4];
//    const Shape* T_tri[5];
//    const Shape* T_quad[5];
//    const Shape* T_tet[2];
//    const Shape* T_hex[2];
    DoF& D;
    int numL;
public:
    DGTSDiscretization (DoF* dof,int _numL =-1) :
        Discretization(dof,"DG_VA"), D(*dof),numL(_numL) {
        S_int[0] = Get_Shape("P0int", "Qint1", "Qpoint1");
	S_int[1] = Get_Shape("P1int", "Qint2", "Qpoint1");
	S_int[2] = Get_Shape("P2int", "Qint3", "Qpoint1");
	S_int[3] = Get_Shape("P3int", "Qint4", "Qpoint1");
	S_int[4] = Get_Shape("P4int", "Qint5", "Qpoint1");
        S_tri[0] = Get_Shape("P0tri");
        S_tri[1] = Get_Shape("P1tri");
        S_tri[2] = Get_Shape("P2tri");
        S_tri[3] = Get_Shape("P3tri");
        S_tri[4] = Get_Shape("P4tri");
        // Overkill Quadrature!
        S_quad[0] = Get_Shape("P0quad");
        S_quad[1] = Get_Shape("P1quad");
        S_quad[2] = Get_Shape("P2quad");
        S_quad[3] = Get_Shape("P3quad");
        S_quad[4] = Get_Shape("P4quad");
        S_quad[5] = Get_Shape("P5quad");
//        S_tet[1] = Get_Shape("P1tet");
//        S_tet[2] = Get_Shape("P2tet");

//        S_hex[1] = Get_Shape("P1hex");
//        S_hex[2] = Get_Shape("P2hex");
//        S_hex[3] = Get_Shape("P3hex");
    }

    const Shape& GetCellShape (const cell& c, int p, int component=0) const {
        if (p > 5) Exit("DGVAcousticShapes::GetShape: Not implemented");
        switch (c.Corners()) {
        case 2: return *S_int[p];
        case 3: return *S_tri[p];
        case 4:
            switch (c.dim()) {
            case 2: return *S_quad[p];
            default: Exit("not implemented!")
//            case 3: return *S_tet[p];
            }
//        case 8: return *S_hex[p];
        }
        Exit("Not implemented!");
        Shape* null=nullptr;
        return *null;
    }
    string DiscName () const { return "DGTSDiscretization"; }
    string Name_Disc_dG () const { return "DGTSDiscretization"; }
    ~DGTSDiscretization() {
        for (int i =0; i< 5; ++i) delete S_tri[i];
        for (int i =0; i< 6; ++i) delete S_quad[i];
//        for (int i =0; i< 4; ++i) delete S_tet[i];
//        for (int i =0; i< 4; ++i) delete S_hex[i];
//        for (int i =0; i< 5; ++i) delete T_tri[i];
//        for (int i =0; i< 5; ++i) delete T_quad[i];
//        for (int i =0; i< 2; ++i) delete T_tet[i];
//        for (int i =0; i< 2; ++i) delete T_hex[i];
  }
};

#endif
