//
// Created by cr on 14.07.21.
//

#ifndef FWI_DGVACOUSTICPLOTTER_HPP
#define FWI_DGVACOUSTICPLOTTER_HPP

#include "Vector.hpp"
#include <vector>
#include "LagrangeDiscretization.hpp"
#include "DGElement.hpp"
#include "Plotting.hpp"
#include "ctools.hpp"
#include "DGViscoAcousticProblem.hpp"
#include "DGAcousticElement.hpp"
#include "DGVAcousticMatrixAssembler.hpp"
#include <unordered_map>
struct plotOptions {
    bool plot_rho = true;
    bool plot_velocity = false;
    bool plot_kappa_0 = true;
    bool plot_kappa_l = false;
    bool plotV = false;
    int plotTStep = -1;

public:
    plotOptions() = default;

    explicit plotOptions(bool plot_rho, bool plot_velocity, bool plot_kappa_0, bool plot_kappa_l) :
            plot_rho(plot_rho), plot_velocity(plot_velocity), plot_kappa_0(plot_kappa_0), plot_kappa_l(plot_kappa_l) {}
};

class DGVAcousticPlotter {
    int verbose = 1;
    plotOptions opt;
    int numL = -1;
    const std::unordered_map<Point, DGAcousticElement *> *EleCache;
public:
    explicit DGVAcousticPlotter(int numL, const std::unordered_map<Point,
            DGAcousticElement *> *EleCache = nullptr);

    void SetElementCache(const std::unordered_map<Point, DGAcousticElement *> *EleCache_) { EleCache = EleCache_; }

    bool PlotTStep(int step) const {
        return !(opt.plotTStep <= 0 || step % opt.plotTStep != 0);
    }

    void plotWaveVector(const Vector &u, int step, const char *name = nullptr) const;

    void plotMatVec(const Vector &u, const DGVAcousticMaterial &Mat) const;

    void plotProcs(const Vector &u) const;
};
#endif //FWI_DGVACOUSTICPLOTTER_HPP
