//
// Created by cr on 22.08.22.
//

#ifndef FWI_DGVACOUSTICRHSASSEMBLER_HPP
#define FWI_DGVACOUSTICRHSASSEMBLER_HPP

#include "m++.hpp"
#include "Point.hpp"
#include "Cell.hpp"
#include "DGViscoAcousticProblem.hpp"
#include "DGAcousticElement.hpp"

class DGVAcousticRHSAssembler {
protected:
    const std::unordered_map<Point, DGAcousticElement *> *EleCache{};
public:
    DGVAcousticRHSAssembler() = default;

    void SetElementCache(const std::unordered_map<Point, DGAcousticElement *> *_EleCache);

    virtual void RHS(double t, Vector &b) const = 0;

    virtual void Initialize(Vector &u) const = 0;

    ~DGVAcousticRHSAssembler() = default;

};

class DGVAcousticRHSAssemblerDefault : public DGVAcousticRHSAssembler {
    const ProblemVAcoustic &problem;
    mutable std::unordered_map<Point, std::list<std::unique_ptr<cell>>> pointSrcCells;
    bool pointSrcCellsAssembled = false;
    bool matInRHS = true;
    DGAcousticElement *getElem(const VectorMatrixBase &base, const Cell &C) const;

    void destroyElement(DGAcousticElement *elem) const;

    void assemblePointSrcRHSPart(double t, Vector &b) const;


    void assembleRHSPart(double t, Vector &b, const cell &c,
                         const DGAcousticElement *elem) const;

    void assembleExactRHSPart(double t, Vector &b, const cell &c,
                              const DGAcousticElement *elem,
                              bool fwd) const;

    void initializePointSrcCache(Vector &b) const;

public:
    DGVAcousticRHSAssemblerDefault(const ProblemVAcoustic &problem) : DGVAcousticRHSAssembler(),
                                                                      problem(problem) {}

    void Initialize(Vector &u) const override;

    void RHS(double t, Vector &b) const override;

    ~DGVAcousticRHSAssemblerDefault() {}

};

class DGVAcousticRHSAssemblerWavefield : public DGVAcousticRHSAssembler {
    const Vectors &U;
    mutable int step = 1;
    mutable bool increase = false;
public:
    DGVAcousticRHSAssemblerWavefield(const Vectors &U) : DGVAcousticRHSAssembler(), U(U) {}

    void RHS(double t, Vector &b) const override {
        b = U[step];
        if (increase) {
            increase = false;
            step++;
        } else {
            increase = true;
        }
    }

    void Initialize(Vector &) const override {}

    ~DGVAcousticRHSAssemblerWavefield() {}

};


#endif //FWI_DGVACOUSTICRHSASSEMBLER_HPP
