#include "DGViscoAcousticProblem.hpp"
//
// Created by cr on 12.07.21.
//


double FWIDeltaFunctionGauss_TS::operator()(const Point &z) const {
  double delta = dist(z, c);
  if (delta >= r) return 0;
  delta /= r;
  return centeredAndNormalizedValue(delta) * (1.0 / (r * r));
}

FWIDeltaFunctionGauss_TS::FWIDeltaFunctionGauss_TS(int dim, const Point &center, double radius,
                                                   int _exponent)
    : c(center), r(radius), exponent(_exponent) {
  switch (exponent) {
    case 1:
      normconst = 0.466512050417577;
      break;
    case 2:
      normconst = 0.117917361026638;
      break;
    default: Exit("wrong exponent")
  }
}

VectorField FWIDeltaFunctionGauss_TS::gradient(const Point &z) const {
  double delta = dist(z, c);
  if (delta >= r) return zero;
  delta /= r;
  double ret = -(exponent * 4.0) / (pow(r, 4) * pow(1.0 / (1 - delta * delta), 2))
               * centeredAndNormalizedValue(delta);
  return VectorField((z[0] - c[0]) * ret, (z[1] - c[1]) * ret);
}

void
ProblemVAcoustic::Force(VectorField &V_rhs, RVector &Force, double t, const Cell &C, const Point &z,
                        int q) const {
  V_rhs = ForceV(t, C, z, q);
  for (int n = 0; n < Force.size(); ++n) {
    Force[n] = ForceP_i(t, C, z, n, q);
  }
}

double ProblemVAcoustic::Kappa(const Cell &C, const Point &x) const {
    double kappa = 0;
    for (int i = 0; i <= numL; ++i) {
        kappa += Kappa_i(C, x, i);
    }
    return kappa;

}





VAcoustic_Riemann_osc::VAcoustic_Riemann_osc()
    : ProblemVAcoustic(0, 2), c(sqrt(kappa / rho)),
      direc(1.0, 0.0), A_osc(0.5), k_osc(10), width(0.125), shift(2.5), kappa(1),
      rho(1) {
    Config::Get("k_osc", k_osc);
    Config::Get("A_osc", A_osc);
    Config::Get("width", width);
    Config::Get("shift", shift);
    Config::Get("kappa", kappa);
    Config::Get("rho", rho);
    c = sqrt(kappa / rho);
}

double VAcoustic_Riemann_osc::Kappa(const Cell &C, const Point &x) const {
  if (cos(Pi * k_osc * C()[0]) > 0)
    return kappa + A_osc;
  else
    return kappa - A_osc;
  return kappa + A_osc * cos(Pi * k_osc * x[0]);
}

double VAcoustic_Riemann_osc::ut(double t, const Cell &C, const Point &x, int i) const {
  double y = A(x[0]);
  switch (i) {
    case 2:
      return y;
    default:
      return 0;
  }
}

double VAcoustic_Riemann_osc::_ut(double t, const Cell &C, const Point &x, int i) const {
  double s = x[0] * direc[0] + x[1] * direc[1] - c * t - 1;
  double s2 = x[0] * direc[0] + x[1] * direc[1] + c * t - 9;
  double y = exp(-20 * s * s);
  double y2 = exp(-20 * s2 * s2);
  double kinv = 1 / sqrt(Kappa(C, x));
  switch (i) {
    case 0:
      return kinv * (direc[0] * c * y - direc[0] * c * y2);
    case 1:
      return kinv * (direc[1] * c * y - direc[1] * c * y2);
    case 2:
      return -y - y2;
    default:
      return 0;
  }
}
