#ifndef _DGVISCOACOUSTIC_H_
#define _DGVISCOACOUSTIC_H_

#include "ctools.hpp"
#include "../../FWI/src/FWIHelpers.hpp"  // TODO: move PointInCellDGTS into M++-Library
#include <fstream>
#include "DGDiscretization.hpp"
#include <Plotting.hpp>
#include <LagrangeDiscretization.hpp>
#include "DGElement.hpp"

class FWIDeltaFunctionGauss_TS {
  Point c;
  double r;
  double normconst;
  int exponent;
public:
  FWIDeltaFunctionGauss_TS(int dim,
                           const Point &center,
                           double radius,
                           int _exponent = 1);

  virtual double operator()(const Point &z) const;

  VectorField gradient(const Point &z) const;

  double centeredAndNormalizedValue(double t) const {

    return exp(-1.0 / (1 - t * t))
           / normconst; // 0.466512050417577 war das was ich berechnet hatte mit MATLAB (heute)
  }
};

class ProblemVAcoustic {
protected:
  std::string meshname;
  int dim;
  int numL;
  int nrPointSrc;
  std::vector<Point> sourceLocations;
public:
  ProblemVAcoustic(int _numL = -1, int d = 3) :
      meshname("none"), dim(d), numL(_numL), nrPointSrc(0) {
    Config::Get("Mesh", meshname, -1);
  }

  virtual ~ProblemVAcoustic() {}

  int Dim() const { return dim; }

  int NumDamping() const { return numL; }

  virtual int getNrPointSrc() const { return nrPointSrc; }

  virtual Point getPointSrcLoc(int nr) const { return sourceLocations[nr]; }

  virtual bool HasExactSolution() const { return false; }

  virtual bool RHS(double) const { return false; }

  virtual void Force(VectorField &V_rhs,
                     RVector &Force,
                     double t,
                     const Cell &C,
                     const Point &z,
                     int q = -1) const;

  virtual double ForceP_i(double,
                          const Cell &,
                          const Point &,
                          int /*i*/,
                          int /*q*/= -1) const {
    Exit("specify force in inherited class")
    return 0;
  }

  virtual VectorField ForceV(double, const Cell &, const Point &, int /*q*/= -1) const {
    Exit("specify force in inherited class")

    return zero;
  }

  virtual double Dirichlet_i(double t, const Cell &C, const Point &z, int i) const {
    return ut(t, C, z, dim + i);
  }

  virtual double Neumann(double t, const Cell &C, const Point z,
                         const VectorField &N) const {
    VectorField V(ut(t, C, z, 0), ut(t, C, z, 1));
    return V * N;
  }

  virtual double ut(double t, const Point &, int) const {
    Exit("Not Implemented!");
    return 0;
  }

  virtual double ut(double t, const Cell &C, const Point &x, int i) const {
    return ut(t, x, i);
  }

  virtual double u0(const Point &x, int i) const { return ut(0, x, i); }

  virtual double u0(const Cell &C, const Point &x, int i) const {
    return ut(0,C,x,i);
  }

  virtual double Kappa(const Cell &C, const Point &x) const;

  virtual double Kappa_i(const Cell &C, const Point &x, int i) const {
    Exit("not implemented");
    return 0.0;
  }

  virtual double invKappa_i(const Cell &C, const Point &x, int i) const {
    return 1.0 / Kappa_i(C, x, i);
  }

  virtual double Velocity(const Point &x) const {
    Exit("Velocity no modelparameter in this case.");
    return 0.0;
  }

  virtual double Tau_i(const Cell &c, const Point &x, int i) const {
    return Tau_i(c, x, i);
  }

  virtual double invTau_i(const Cell &C, const Point &x, int i) const {
    return 1.0 / Tau_i(C, x, i);
  }

  virtual double Rho(const Cell &C, const Point &x) const {
    Exit("notimplemented");
    return 0.0;
  }

  virtual const char *Name() const {
    Exit("not implemented");
    return "";
  }

  virtual const char *MeshName() const { return meshname.c_str(); }
  virtual double alpha(const Cell &,
                       const Point &) const {
    Exit("not implemented here!");
    return 0.;
  }
  virtual double alpha1() const {
    Exit("not implemented here!");
    return 0.;
  }
  virtual double alpha2() const {
    Exit("not implemented here!");
    return 0.;
  }
};


class VAcoustic_Riemann_osc : public ProblemVAcoustic {
  double kappa;
  double rho;
  double c;
  Point direc;
  double A_osc;
  double k_osc;
  double width;
  double shift;
public:
  VAcoustic_Riemann_osc();

  double Kappa(const Cell &C, const Point &x) const;

  bool RHS(double) const { return false; }

  const char *Name() const { return "Acoustic_Riemann_osc"; }

  double A0(double s) const {
    if (s * s >= width) return 0;
    return exp(-1 / (width - s * s)) / exp(-1 / width);
  }

  double A(double s) const { return A0(s - shift); }

  //double A (double s) const { return A0(s); }
  double ut(double t, const Cell &C, const Point &x, int i) const;

  double _ut(double t, const Cell &C, const Point &x, int i) const;

  double ForceP_i(double t, const Cell &C, const Point &x, int i, int q = -1) const {
    return 0.0;
  }

  VectorField ForceV(double t, const cell &c, const Point &z, int q = -1) const {
    return zero;
  }

  double Rho(const Cell &C, const Point &x) const {
    return rho;
  }
};

class VAcoustic_Benchmark_Cosine : public ProblemVAcoustic {
  Point normal;
  double normNormal;
  double f;
  double c;
  double kappa;
  double rho;
public:
  VAcoustic_Benchmark_Cosine(int dim) : ProblemVAcoustic(0, dim),
                                        normal(Point(1, 0, 0)) {
    kappa = 1;
    rho = 1;
    c = sqrt(kappa / rho);
    f = 1;
    f *= 2 * Pi;
    if (dim == 2) {
      normal[0] = 1;
      normal[1] = 0;
      normNormal = norm(normal);
      normal /= normNormal;
    }
  }

  bool HasExactSolution() const { return true; }

  const char *Name() const { return "Cosine wave"; }

  double ut(double t, const Point &x, int i) const {
    double nx = x * normal;
    if (i < dim) {
      return c * normal[i] * cos(f * (nx + c * t));
    } else {
      return kappa * cos(f * (nx + c * t));
    }
  }

  bool RHS(double) const { return true; }

  VectorField ForceV(double, const Cell &, const Point &, int = -1) const {
    return zero;
  }

  double ForceP_i(double, const cell &c, const Point &, int i, int = -1) const {
    switch (i) {
      case 0:
        return 0.0;
      default: Exit("Wrong call")
    }
    return 0.0;
  }

  double Kappa_i(const Cell &C, const Point &x, int i) const {
    if (i != 0) Exit("no wai");
    return kappa;
    Exit("Not implemented");
    return 0.0;
  }

  virtual double Rho(const Cell &C, const Point &x) const { return rho; }

  double Tau_i(const Cell &c, const Point &x, int i) const {
    Exit("CIAO")
    return 0.0;
  }
};

class VAcoustic_Benchmark_Constant1 : public ProblemVAcoustic {
  double rho;
  double kappa1;
  double kappa0;
  double tau1;
public:
  VAcoustic_Benchmark_Constant1()
      : ProblemVAcoustic(1, 2), rho(1.0), kappa1(1.0), kappa0(1.0), tau1(1.0) {
    Config::Get("tau1", tau1);
    Config::Get("kappa0", kappa0);
    Config::Get("kappa1", kappa1);
    Config::Get("rho", rho);
  }

  bool HasExactSolution() const { return true; }

  bool RHS(double) const { return true; }

  VectorField ForceV(double t, const cell &c, const Point &z, int q = -1) const {
    return zero;
  }

  double ForceP_i(double t, const cell &c, const Point &x, int i, int q = -1) const {
    switch (i) {
      case 0:
        return 0.0;
      case 1:
        return 0.0;
      default: Exit("Wrong call")
    }
    return 0.0;
  }

  const char *Name() const { return "Constant p0=3 ,p1=0, v = (1,2)"; }

  double ut(double t, const Point &x, int i) const {
    switch (i) {
      case 0:
        return 1;
      case 1:
        return 2;
      case 2:
        return 3;
      case 3:
        return 0;
      default:
        return 0;
    }
  }

  double Rho(const Cell &C, const Point &x) const {
    return rho;
  }

  double Tau_i(const Cell &c, const Point &x, int i) const {
    return tau1;
  }

  double Kappa_i(const Cell &C, const Point &x, int i) const {
    switch (i) {
      case 0:
        return kappa0;
      case 1:
        return kappa1;
    }
    Exit("Not implemented");
    return 0.0;
  }
};

class Acoustic_Benchmark_Constant1 : public ProblemVAcoustic {
  double rho;
  double kappa1;
  double kappa0;
  double tau1;
public:
  Acoustic_Benchmark_Constant1()
      : ProblemVAcoustic(0, 1), rho(1.0), kappa1(1.0), kappa0(1.0), tau1(1.0) {
    Config::Get("tau1", tau1);
    Config::Get("kappa0", kappa0);
    Config::Get("kappa1", kappa1);
    Config::Get("rho", rho);
  }

  bool HasExactSolution() const { return true; }

  bool RHS(double) const { return true; }

  VectorField ForceV(double t, const cell &c, const Point &z, int q = -1) const {
    return zero;
  }

  double ForceP_i(double t, const cell &c, const Point &x, int i, int q = -1) const {
    switch (i) {
      case 0:
        return 0.0;
      case 1:
        return 0.0;
      default: Exit("Wrong call")
    }
    return 0.0;
  }

  const char *Name() const { return "Constant p0=3 ,p1=0, v = (1,2)"; }

  double ut(double t, const Point &x, int i) const {
    switch (i) {
      case 0:
        return 1;
      case 1:
        return 2;
    }
    Exit("wrong turn");
    return -infty;
  }

  double Rho(const Cell &C, const Point &x) const {
    return rho;
  }

  double Tau_i(const Cell &c, const Point &x, int i) const {
    return tau1;
  }

  double Kappa_i(const Cell &C, const Point &x, int i) const {
    switch (i) {
      case 0:
        return kappa0;
      case 1:
        return kappa1;
    }
    Exit("Not implemented");
    return 0.0;
  }
};

class VAcoustic_Benchmark_Constant2 : public ProblemVAcoustic {
  double rho;
  double kappa1;
  double kappa0;
  double tau1;
public:
  VAcoustic_Benchmark_Constant2()
      : ProblemVAcoustic(1, 2), rho(1.0), kappa1(1.0), kappa0(1.0), tau1(1.0) {
    Config::Get("tau1", tau1);
    Config::Get("kappa0", kappa0);
    Config::Get("kappa1", kappa1);
    Config::Get("rho", rho);
    mout << "rho: " << rho << "kappa0: " << kappa0 << "kappa1: " << kappa1 << "tau1: "
         << tau1 << endl;
  }

  bool HasExactSolution() const { return true; }

  bool RHS(double) const { return true; }

  VectorField ForceV(double t, const Cell &C, const Point &z, int q = -1) const {
    return zero;
  }

  double ForceP_i(double t, const Cell &C, const Point &x, int i, int q = -1) const {
    switch (i) {
      case 0:
        return 0.0;
      case 1:
        return +1.0 / (Kappa_i(C, x, 1) * Tau_i(C, x, 1));
      default: Exit("Wrong call")
    }
    return 0.0;
  }

  const char *Name() const { return "Constant p0=1 ,p1=1, v = (x,-y)"; }

  double ut(double t, const Point &x, int i) const {
    switch (i) {
      case 0:
        return x[0];
      case 1:
        return -x[1];
      case 2:
        return 1.0;
      case 3:
        return 1.0;
      default:
        return 0;
    }
  }

  double Rho(const Cell &C, const Point &x) const {
    return rho;
  }

  double Tau_i(const Cell &c, const Point &x, int i) const {
    return tau1;
  }

  double Kappa_i(const Cell &C, const Point &x, int i) const {
    switch (i) {
      case 0:
        return kappa0;
      case 1:
        return kappa1;
    }
    Exit("Not implemented");
    return 0.0;
  }
};

class VAcoustic_Benchmark_Linear1 : public ProblemVAcoustic {
  double rho;
  double kappa1;
  double kappa0;
  double tau1;
public:
  VAcoustic_Benchmark_Linear1()
      : ProblemVAcoustic(1, 2), tau1(1), kappa0(1), kappa1(1), rho(1) {
    Config::Get("tau1", tau1);
    Config::Get("kappa0", kappa0);
    Config::Get("kappa1", kappa1);
    Config::Get("rho", rho);
  }

  bool HasExactSolution() const { return true; }

  bool RHS(double) const { return true; }

  VectorField ForceV(double t, const Cell &C, const Point &z, int q = -1) const {
    return zero;
  }

  double ForceP_i(double t, const Cell &C, const Point &x, int i, int q = -1) const {
    switch (i) {
      case 0:
        return 0.0;
      case 1:
        return (t + Tau_i(C, x, i)) / (Kappa_i(C, x, i) * Tau_i(C, x, i));
      default: Exit("Wrong call");
        return 0.0;
    }
  }

  const char *Name() const { return "Linear1 p0=0 ,p1=t, v =(0,0)"; }

  double ut(double t, const Point &x, int i) const {
    switch (i) {
      case 0:
        return 0;
      case 1:
        return 0;
      case 2:
        return 0;
      case 3:
        return t;
      default:
        return 0;
    }
  }

  double Rho(const Cell &C, const Point &x) const {
    return rho;
  }

  double Tau_i(const Cell &c, const Point &x, int i) const {
    return tau1;
  }

  double Kappa_i(const Cell &C, const Point &x, int i) const {
    switch (i) {
      case 0:
        return kappa0;
      case 1:
        return kappa1;
    }
    Exit("Not implemented");
    return 0.0;
  }
};

class VAcoustic_Benchmark_Linear1_1D : public ProblemVAcoustic {
  double rho;
  double kappa1;
  double kappa0;
  double tau1;
public:
  VAcoustic_Benchmark_Linear1_1D()
      : ProblemVAcoustic(1, 1), tau1(1), kappa0(1), kappa1(1), rho(1) {
    Config::Get("tau1", tau1);
    Config::Get("kappa0", kappa0);
    Config::Get("kappa1", kappa1);
    Config::Get("rho", rho);
  }

  bool HasExactSolution() const { return true; }

  bool RHS(double) const { return true; }

  VectorField ForceV(double t, const cell &c, const Point &z, int q = -1) const {
    return zero;
  }

  double ForceP_i(double t, const Cell &C, const Point &x, int i, int q = -1) const {
    switch (i) {
      case 0:
        return 0.0;
      case 1:
        return (t + Tau_i(C, x, i)) / (Kappa_i(C, x, i) * Tau_i(C, x, i));
      default: Exit("Wrong call");
        return 0.0;
    }
  }

  const char *Name() const { return "Linear1 p0=0 ,p1=t, v =(0,0)"; }

  double ut(double t, const Point &x, int i) const {
    switch (i) {
      case 0:
        return 0;
      case 1:
        return 0;
      case 2:
        return 0;
      case 3:
        return t;
      default:
        return 0;
    }
  }

  double Rho(const Cell &C, const Point &x) const {
    return rho;
  }

  double Tau_i(const Cell &c, const Point &x, int i) const {
    return tau1;
  }

  double Kappa_i(const Cell &C, const Point &x, int i) const {
    switch (i) {
      case 0:
        return kappa0;
      case 1:
        return kappa1;
    }
    Exit("Not implemented");
    return 0.0;
  }
};

class VAcoustic_Benchmark_Linear2 : public ProblemVAcoustic {
  double rho;
  double kappa1;
  double kappa0;
  double tau1;
public:
  VAcoustic_Benchmark_Linear2()
      : ProblemVAcoustic(1, 2), tau1(1), kappa0(1), kappa1(1) {
    Config::Get("tau1", tau1);
    Config::Get("kappa0", kappa0);
    Config::Get("kappa1", kappa1);
    Config::Get("rho", rho);
  }

  bool HasExactSolution() const { return true; }

  bool RHS(double) const { return true; }

  VectorField ForceV(double t, const Cell &C, const Point &z, int q = -1) const {
    return zero;
  }

  double ForceP_i(double t, const Cell &C, const Point &x, int i, int q = -1) const {
    switch (i) {
      case 0:
        return 0.0;
      case 1:
        return -(1.0 / Kappa(C, x));
      default: Exit("Wrong call")
    }
    return 0.0;
  }

  const char *Name() const { return "Linear2 p0=t ,p1=0, v = (0,0)"; }

  double ut(double t, const Cell &C, const Point &x, int i) const {
    switch (i) {
      case 0:
        return x[0] / Kappa_i(C, x, 0);
      case 1:
        return 0;
      case 2:
        return t;
      case 3:
        return 0;
      default:
        return 0;
    }
  }

  double Rho(const Cell &C, const Point &x) const {
    return rho;
  }

  double Tau_i(const Cell &c, const Point &x, int i) const {
    return tau1;
  }

  double Kappa_i(const Cell &C, const Point &x, int i) const {
    switch (i) {
      case 0:
        return kappa0;
      case 1:
        return kappa1;
    }
    Exit("Not implemented");
    return 0.0;
  }
};

class VAcoustic_Benchmark_Linear3 : public ProblemVAcoustic {
  double rho;
  double kappa1;
  double kappa0;
  double tau1;
public:
  VAcoustic_Benchmark_Linear3()
      : ProblemVAcoustic(1, 2), tau1(1), kappa0(1), kappa1(1), rho(1.0) {
    Config::Get("tau1", tau1);
    Config::Get("kappa0", kappa0);
    Config::Get("kappa1", kappa1);
    Config::Get("rho", rho);
  }

  bool HasExactSolution() const { return true; }

  bool RHS(double) const { return true; }

  VectorField ForceV(double t, const Cell &C, const Point &z, int q = -1) const {
    return zero;
  }

  double ForceP_i(double t, const Cell &C, const Point &x, int i, int q = -1) const {
    switch (i) {
      case 0:
        return 0.0;
      case 1:
        return -(1.0 / Kappa_i(C, x, 0))
               + (t + Tau_i(C, x, 1)) / (Kappa_i(C, x, 0) * Tau_i(C, x, 1));
      default: Exit("Wrong call")
    }
    return 0.0;
  }

  const char *Name() const { return "Constant p0=t ,p1=t, v = (x/kappa0,0)"; }

  double ut(double t, const Cell &C, const Point &x, int i) const {
    switch (i) {
      case 0:
        return x[0] / Kappa_i(C, x, 0);
      case 1:
        return 0;
      case 2:
        return t;
      case 3:
        return t;
      default:
        return 0;
    }
  }

  double Rho(const Cell &C, const Point &x) const {
    return rho;
  }

  double Tau_i(const Cell &c, const Point &x, int i) const {
    return tau1;
  }

  double Kappa_i(const Cell &C, const Point &x, int i) const {
    switch (i) {
      case 0:
        return kappa0;
      case 1:
        return kappa1;
    }
    Exit("Not implemented");
    return 0.0;
  }
};

class VAcoustic_Benchmark_ExpP1 : public ProblemVAcoustic {
  double rho;
  double kappa1;
  double kappa0;
  double tau1;
public:
  VAcoustic_Benchmark_ExpP1()
      : ProblemVAcoustic(1, 2), tau1(1), kappa0(1), kappa1(1), rho(1) {
    Config::Get("tau1", tau1);
    Config::Get("kappa0", kappa0);
    Config::Get("kappa1", kappa1);
    Config::Get("rho", rho);
  }

  bool HasExactSolution() const { return true; }

  bool RHS(double) const { return true; }

  double ForceP_i(double t, const Cell &C, const Point &x, int i, int q = -1) const {
    switch (i) {
      case 0:
        return ForceP0(t, C, x);
      case 1:
        return ForceP1(t, x);
      default: Exit("wrong number of Kappas");
    }
    return 0.0;
  }

  VectorField ForceV(double t, const Cell &C, const Point &z, int q = -1) const {
    return zero;
  }

  double ForceP0(double t, const Cell &C, const Point &x) const {
    return (1.0 / Kappa_i(C, x, 0)) * 2 * t;
  }

  double ForceP1(double t, const Point &x) const {
    return 0.0;
  }

  const char *Name() const { return "Constant p0=t*t ,p1=exp(-t/tau), v = (x,-y)"; }

  double ut(double t, const Cell &C, const Point &x, int i) const {
    switch (i) {
      case 0:
        return x[0];
      case 1:
        return -x[1];
      case 2:
        return t * t;
      case 3:
        return exp(-t / Tau_i(C, x, 1));
      default:
        return 0;
    }
  }

  double Rho(const Cell &C, const Point &x) const {
    return rho;
  }

  double Tau_i(const Cell &c, const Point &x, int i) const {
    return tau1;
  }

  double Kappa_i(const Cell &C, const Point &x, int i) const {
    switch (i) {
      case 0:
        return kappa0;
      case 1:
        return kappa1;
    }
    Exit("Not implemented");
    return 0.0;
  }
};

class VAcoustic_Benchmark_ExpP1_2 : public ProblemVAcoustic {
  double rho;
  double kappa1;
  double kappa0;
  double tau1;
public:
  VAcoustic_Benchmark_ExpP1_2()
      : ProblemVAcoustic(1, 2), tau1(1), kappa0(1), kappa1(1) {
    Config::Get("tau1", tau1);
    Config::Get("rho", rho);
    Config::Get("kappa0", kappa0);
    Config::Get("kappa1", kappa1);
  }

  bool HasExactSolution() const { return true; }

  bool RHS(double) const { return true; }

  VectorField ForceV(double t, const cell &c, const Point &z, int q = -1) const {
    return zero;
  }

  double ForceP_i(double t, const cell &c, const Point &x, int i, int q = -1) const {
    switch (i) {
      case 0:
        return ForceP0(t, x);
      case 1:
        return ForceP1(t, x);
      default: Exit("wrong number of Kappas");
    }
    Exit("Not implemented");
    return 0.0;
  }

  double ForceP0(double t, const Point &x) const {
    return 0.0;
  }

  double ForceP1(double t, const Point &x) const {
    return 0.0;
  }

  const char *Name() const { return "Constant p0=0.0 ,p1=exp(-t/tau), v = (0,0)"; }

  double ut(double t, const Cell &C, const Point &x, int i) const {
    switch (i) {
      case 0:
        return 0.0;
      case 1:
        return 0.0;
      case 2:
        return 0.0;
      case 3:
        return exp(-t / Tau_i(C, x, 1));
      default:
        return 0;
    }
  }

  double Rho(const Cell &C, const Point &x) const {
    return rho;
  }

  double Tau_i(const Cell &c, const Point &x, int i) const {
    return tau1;
  }

  double Kappa_i(const Cell &C, const Point &x, int i) const {
    switch (i) {
      case 0:
        return kappa0;
      case 1:
        return kappa1;
    }
    Exit("Not implemented");
    return 0.0;
  }
};

class VAcoustic_Riemann : public ProblemVAcoustic {
  double rho;
  double kappa1;
  double kappa0;
  double tau1;
public:
  VAcoustic_Riemann(int dim)
      : ProblemVAcoustic(0, dim), tau1(1), kappa0(1), kappa1(1), rho(1.0) {
    Config::Get("tau1", tau1);
    Config::Get("rho", rho);
    Config::Get("kappa0", kappa0);
    Config::Get("kappa1", kappa1);
  }

  bool HasExactSolution() const { return false; }

  VectorField ForceV(double t, const cell &c, const Point &z, int q = -1) const {
    return zero;
  }

  double ForceP_i(double t, const cell &c, const Point &x, int i, int q = -1) const {
    switch (i) {
      case 0:
        return ForceP0(t, x);
      case 1:
        return ForceP1(t, x);
      default: Exit("wrong number of Kappas");
    }
    return 0.0;
  }

  double ForceP0(double t, const Point &x) const {
    return 0;
  }

  double ForceP1(double t, const Point &x) const {
    return 0.0;
  }

  const char *Name() const { return "VAcoustic RiemannTest"; }

  double ut(double t, const Point &x, int i) const {
    double s = x[0] - t - 1;
    double y = exp(-20 * s * s);
    if (dim == 1) {
      if (i == 0) {
        return y;
      } else if (i == 1) {
        return -y;
      } else {
        Exit("ERROR");
      }
    } else if (dim == 2) {
      switch (i) {
        case 0:
          return y;
        case 2:
          return -y;
        default:
          return 0.0;
      }
    } else {
      Exit("ERROR");
      return -infty;
    }
  }

  double Rho(const Cell &C, const Point &x) const {
    return rho;
  }

  double Tau_i(const Cell &c, const Point &x, int i) const {
    return tau1;
  }

  double Kappa_i(const Cell &C, const Point &x, int i) const {
    switch (i) {
      case 0:
        return kappa0;
      case 1:
        return kappa1;
    }
    Exit("Not implemented");
    return 0.0;
  }
};

class VAcoustic_Spherical : public ProblemVAcoustic {
  double kappa1;
  double kappa0;
  double tau1;
  double rho;
  double width;
  Point Pmid;
  double aa;
public:
  VAcoustic_Spherical()
      : ProblemVAcoustic(1, 2), Pmid(2, 2, 0), tau1(1), kappa0(1), kappa1(1), rho(1.0),
        width(0.25) {
    Config::Get("tau1", tau1);
    Config::Get("rho", rho);
    Config::Get("kappa0", kappa0);
    Config::Get("kappa1", kappa1);
    aa = 0.25;
  }

  double A0(double s) const {
    if (s < width) return pow(cos(s * 0.5 * M_PI / width), 6.0);
    else return 0.0;
  }

  double A(const Point &x) const {
    double r = dist(Pmid, x);
    return A0(r * r);
  }

  VectorField ForceV(double t, const cell &c, const Point &z, int q = -1) const {
    return zero;
  }

  double ForceP_i(double t, const cell &c, const Point &x, int i, int q = -1) const {
    switch (i) {
      case 0:
        return ForceP0(t, x);
      case 1:
        return ForceP1(t, x);
      default: Exit("wrong number of Kappas");
    }
    return 0.0;
  }

  double ForceP0(double t, const Point &x) const {
    return 0;
  }

  double ForceP1(double t, const Point &x) const {
    return 0.0;
  }

  const char *Name() const { return "VAcoustic Spherical Wave"; }

  double ut(double t, const Point &x, int i) const {
    if (t > 0) return 0.0;
    const double tmp = A(x);
    switch (i) {
      case 0 :
        return (x[0] - Pmid[0]) * tmp;
      case 1 :
        return (x[1] - Pmid[1]) * tmp;
      default:
        return 0.0;
    }
  }

  double Rho(const Cell &C, const Point &x) const {
    return rho;
  }

  double Tau_i(const Cell &c, const Point &x, int i) const {
    return tau1;
  }

  double Kappa_i(const Cell &C, const Point &x, int i) const {
    switch (i) {
      case 0:
        return kappa0;
      case 1:
        return kappa1;
    }
    Exit("Not implemented");
    return 0.0;
  }
};

class VAcoustic_Ricker1D : public ProblemVAcoustic {
  double kappa1;
  double kappa0;
  double tau1;
  double rho;
  double omega_p;
  double T;
  double aa;
public:
  VAcoustic_Ricker1D()
      : ProblemVAcoustic(1, 2), omega_p(20), T(0.5), tau1(1), kappa0(1), kappa1(1),
        rho(1.0) {
    Config::Get("tau1", tau1);
    Config::Get("rho", rho);
    Config::Get("kappa0", kappa0);
    Config::Get("kappa1", kappa1);
    aa = 100;
  }

  VectorField ForceV(double t, const Cell &C, const Point &z, int q = -1) const {
    return zero;
  }

  double ForceP_i(double t, const cell &c, const Point &x, int i, int q = -1) const {
    switch (i) {
      case 0:
        return ForceP0(t, x);
      case 1:
        return ForceP1(t, x);
      default: Exit("wrong number of Kappas");
    }
    return 0.0;
  }

  double ForceP0(double t, const Point &x) const {
    return 0.0;
  }

  double ForceP1(double t, const Point &x) const {
    return 0.0;
  }

  const char *Name() const { return "VAcoustic Ricker Wavelet"; }

  double ut(double t, const Point &x, int i) const {
    if (t == 0 && i == 2) {
      double factor = omega_p * omega_p * ((x[0] - 2.5) * (x[0] - 2.5));
      return (1.0 - 0.5 * factor) * exp(-0.5 * factor);
    } else return 0;
  }

  double Rho(const Cell &C, const Point &x) const {
    return rho;
  }

  double Tau_i(const Cell &c, const Point &x, int i) const {
    return tau1;
  }

  double Kappa_i(const Cell &C, const Point &x, int i) const {
    switch (i) {
      case 0:
        return kappa0;
      case 1:
        return kappa1;
    }
    Exit("Not implemented");
    return 0.0;
  }
};

class VAcousticWaveC2Benchmark : public ProblemVAcoustic {
  int verbose;
  double centerx, centery, radius, f, factor, exponent;
  int dim;
  std::string meshname;
  int n_eta;
  double v_bg, rho_bg, tau_p_bg;
  double v_perturb, rho_perturb, tau_p_perturb;
  double size_perturb;
  double x_perturb;
  double z_perturb_v, z_perturb_rho, z_perturb_tau_p;
  FWIDeltaFunctionGauss_TS kernel;

public:
  VAcousticWaveC2Benchmark() : ProblemVAcoustic(1, 2),
                               v_bg(3500), rho_bg(2000), tau_p_bg(0.02),
                               size_perturb(200),
                               x_perturb(900), z_perturb_v(600), z_perturb_rho(900),
                               z_perturb_tau_p(1200), v_perturb(1.05),
                               rho_perturb(1.05),
                               tau_p_perturb(4), centerx(601), centery(870),
                               radius(50), f(25), factor(100000), exponent(2),
                               kernel(2, Point(centerx, centery), radius, exponent) {
  }

  const char *Name() const {
    return "VAcousticWaveC2Benchmark";
  }

  double ut(double t, const Point &x, int i) const {
    return 0.0;
  }

  double ut(double t, const Cell &C, const Point &x, int i) const {
    return 0.0;
  }

  bool HasExactSolution() const {
    return false;
  }

  bool isInPerturb(const Point &x, int i) const {
    if (x[0] > x_perturb && x[0] < x_perturb + size_perturb) {
      switch (i) {
        case 0:
          if (x[1] > z_perturb_v && x[1] < z_perturb_v + size_perturb) {
            return true;
          } else { return false; }

        case 1:
          if (x[1] > z_perturb_rho && x[1] < z_perturb_rho + size_perturb) {
            return true;
          } else { return false; }
        case 2:
          if (x[1] > z_perturb_tau_p && x[1] < z_perturb_tau_p + size_perturb) {
            return true;
          } else { return false; }
      }
    }
    return false;
  }

  VectorField ForceV(double t, const cell &c, const Point &x, int q = -1) const {
    if (t > 2.0 / f) {
      return zero;
    }
    double dx = 0.0;
    double dy = 0.0;
    double delta = 0.0;
    dx = centerx - x[0];
    dy = centery - x[1];
    delta = sqrt(dx * dx + dy * dy);
    if (delta < radius) {
//            delta /= radius;
      return factor * Ricker(f, t);
    } else { return zero; }
  }

  double derivativeRicker(double f, double t) const {
    double product = Pi * Pi * f * f * (t - 1.0 / f) * (t - 1.0 / f);
    double product_dt = 2 * Pi * Pi * f * f * (t - 1.0 / f);
    return (2 * product - 3) * exp(-product) * product_dt;
  }

  double Ricker(double f, double t) const {
    double product = Pi * Pi * f * f * (t - 1.0 / f) * (t - 1.0 / f);
    return (1.0 - 2.0 * product) * exp(-product);
  }

  bool RHS(double) const { return true; }

  double ForceP_i(double t, const Cell &C, const Point &x, int i, int q = -1) const {
    if (t > 2.0 / f) {
      return 0.0;
    }
    if (i == 0) {
      return (1.0 / Kappa_i(C, x, 0)) * ForceP0(t, x);
    } else {
      return 0.0;
    }
  }

  double ForceP0(double t, const Point &x) const {

    double dx = 0.0;
    double dy = 0.0;
    double delta = 0.0;
    dx = centerx - x[0];
    dy = centery - x[1];
    delta = sqrt(dx * dx + dy * dy);
    if (delta < radius) {
//            delta /= radius;
      return factor * kernel(x) * derivativeRicker(f, t);
    } else { return 0.0; }
  }

  double Kappa_i(const Cell &C, const Point &x, int i) const {
    double v = Velocity(x);
    double Kappa0 = Rho(C, x) * v * v / (1 + 0.5 * Tau_p(C, x));
    switch (i) {
      case 0:
        return Kappa0;
      case 1:
        return Tau_p(C, x) * Kappa0;
      case 99:
        //HACK !!! for plotting only
        return Tau_p(C, x);
      default: Exit("wrong call")
    }
    return 0.0;
  }

  double Rho(const Cell &C, const Point &x) const {
    if (isInPerturb(x, 1)) {
      return rho_bg * rho_perturb;
    } else {
      return rho_bg;
    }
  }

  double Tau_p(const Cell &C, const Point &x) const {
    if (isInPerturb(x, 2)) {
      return tau_p_bg * tau_p_perturb;
    } else {
      return tau_p_bg;
    }
  }

  double Tau_i(const Cell &c, const Point &x, int i) const {
    return 1.0 / (2.0 * Pi * f);
  }

  double Velocity(const Point &x) const {
    if (isInPerturb(x, 0)) {
      return v_bg * v_perturb;
    } else {
      return v_bg;
    }
  }
};

class VAcoustic_HorizontalReflector : public ProblemVAcoustic {
  double kappa0;
  double kappa1;
  double tau1;
  double rho;
  double centert;
  double centerx, centery, radius;
  double factor;
  double layer_min_y, layer_max_y;
  double background_velocity;
  double layer_velocity;
public:
  VAcoustic_HorizontalReflector() :
      ProblemVAcoustic(1, 2), centerx(2.0), centery(2.5), radius(0.1), factor(1.0),
      layer_min_y(1.4), layer_max_y(1.8),
      background_velocity(1.0), layer_velocity(1.0), tau1(1), kappa0(1), kappa1(1),
      rho(1) {
    Config::Get("source_center_x", centerx);
    Config::Get("source_center_y", centery);
    Config::Get("source_radius", radius);
    Config::Get("source_factor", factor);
    Config::Get("layer_min_y", layer_min_y);
    Config::Get("layer_max_y", layer_max_y);
    Config::Get("layer_velocity", layer_velocity);
    Config::Get("background_velocity", background_velocity);
    centert = 0.05;
    Config::Get("rho", rho);
    Config::Get("kappa0", kappa0);
    Config::Get("kappa1", kappa1);
    Config::Get("tau1", tau1);
  }

  double ut(double t, const Cell &C, const Point &x, int i) const {
    return 0.0;
  }

  const char *Name() const { return "HorizontalReflector"; }

  bool RHS(double) const {
    return true;
  }

  double PulseFunction1D(double s) const {
    if (s > 1)
      return 0.0;
    else {
      return exp(-6.0 * s * s);
    };
  }

  double Kappa_i(const Cell &C, const Point &x, int i) const {
    switch (i) {
      case 0:
        return Kappa0(x);
      case 1:
        return Kappa1(x);
      default: Exit("Wrong number of Kappas");
    }
    return 0.0;
  }

  double Kappa0(const Point &x) const {
    double v = background_velocity + layerOffset(x[1]);
    return v * kappa0;
  }

  double Kappa1(const Point &x) const {
    double v = background_velocity + layerOffset(x[1]);
    return v * kappa1;
  }

  double ForceP_i(double t, const cell &c, const Point &x, int i, int q = -1) const {
    switch (i) {
      case 0:
        return ForceP0(t, x);
      case 1:
        return 0.0;
      default: Exit("wrong number of Kappas");
    }
    return 0.0;
  }

  double ForceP0(double t, const Point &x) const {
    double dt = t - centert;
    double dx = centerx - x[0];
    double dy = centery - x[1];
    double delta = sqrt(dt * dt + dx * dx + dy * dy);
    if (delta < radius) {
      delta /= radius;
      return factor * PulseFunction1D(delta);
    } else return 0.0;
  }

  int Deg() const { return 1000; }

  double layerProfile(double y) const {
    return -sin((y - layer_min_y) / (layer_max_y - layer_min_y) * 2 * M_PI);
  }

  double layerOffset(double y) const {
    if (y >= layer_min_y && y <= layer_max_y)
      return layerProfile(y);
    else
      return 0.0;
  }

  double Tau_i(const Cell &c, const Point &x, int i) const {
    return tau1;
  }

  double Rho(const Cell &C, const Point &x) const {
    return rho;
  }

  VectorField ForceV(double t, const cell &c, const Point &z, int q = -1) const {
    return zero;
  }
};

class AcousticSinCos2d_RHS : public ProblemVAcoustic {
  double kappa1;
  double kappa0;
  double tau1;
  double rho;
  double omega;
public:
  AcousticSinCos2d_RHS() : ProblemVAcoustic(0, 2), tau1(1), kappa0(1), kappa1(1),
                           rho(1) {
    omega = 1.0 / 4.0;
    Config::Get("omega", omega);
    omega *= 2 * M_PI;
    Config::Get("tau1", tau1);
    Config::Get("rho", rho);
    Config::Get("kappa0", kappa0);
    Config::Get("kappa1", kappa1);
  }

  const char *Name() const { return "AcousticSinCos2d_RHS p,q = TODO"; }

  bool HasExactSolution() const { return true; }

  bool RHS(double) const { return true; }

  double ut(double t, const Point &x, int i) const {
    switch (i) {
      case 0:
        return Velocity(t, x)[0];
      case 1:
        return Velocity(t, x)[1];
      case 2:
        return Pressure0(t, x);
    }
    Exit("Not implemented");
    return 0.0;
  }

  double ForceP_i(double t, const cell &c, const Point &x, int i, int q = -1) const {
    switch (i) {
      case 0:
        return ForceP0(t, x);
      default: Exit("Wrong number of Kappas");
    }
    return 0.0;
  }

  VectorField ForceV(double t, const cell &c, const Point &z, int q = -1) const {
    return zero;
  }

  double Pressure0(double t, const Point &x) const {
    return sin(omega * x[0]) * sin(omega * x[1]);
  }

  VectorField Velocity(double t, const Point &x) const {
    return VectorField((t - 1) * omega * sin(omega * x[1]) * cos(omega * x[0]),
                       (t - 1) * omega * sin(omega * x[0]) * cos(omega * x[1]), 0.0);
  }

  int Deg() const { return 1000; }

  double ForceP0(double t, const Point &x) const {
    return 2 * (t - 1) * omega * omega * sin(omega * x[0]) * sin(omega * x[1]);
  }

  int SourceDeg() const { return 1000; }

  double Rho(const Cell &C, const Point &x) const {
    return rho;
  }

  double Tau_i(const Cell &c, const Point &x, int i) const {
    return tau1;
  }

  double Kappa_i(const Cell &C, const Point &x, int i) const {
    switch (i) {
      case 0:
        return kappa0;
      case 1:
        return kappa1;
    }
    Exit("Not implemented");
    return 0.0;
  }
};

class VAcousticSinCos2d_RHS : public ProblemVAcoustic {
  double kappa1;
  double tau1;
  double rho;
  double omega;
  double kappa0;
public:
  VAcousticSinCos2d_RHS() : ProblemVAcoustic(1, 2), kappa0(1), rho(1), tau1(1),
                            kappa1(1) {
    omega = 1.0 / 4.0;
    Config::Get("omega", omega);
    omega *= 2 * M_PI;
    Config::Get("kappa1", kappa1);
    Config::Get("rho", rho);
    Config::Get("kappa0", kappa0);
    Config::Get("tau1", tau1);
    mout << "Kappa1: " << kappa1 << " Kappa0: " << kappa0 << " Rho: " << rho
         << " tau1: " << tau1 << endl;
  }

  const char *Name() const { return "VAcousticSinCos2d_RHS p,q = TODO"; }

  bool HasExactSolution() const { return true; }

  bool RHS(double) const { return true; }

  double ut(double t, const Point &x, int i) const {
    switch (i) {
      case 0:
        return Velocity(t, x)[0];
      case 1:
        return Velocity(t, x)[1];
      case 2:
        return Pressure0(t, x);
      case 3:
        return Pressure1(t, x);
    }
    Exit("Not implemented");
    return 0.0;
  }

  VectorField ForceV(double t, const cell &c, const Point &z, int q = -1) const {
    return zero;
  }

  double ForceP_i(double t, const Cell &C, const Point &x, int i, int q = -1) const {
    switch (i) {
      case 0:
        return ForceP0(t, x);
      case 1:
        return ForceP1(t, C, x);
      default: Exit("Wrong number of Kappas");
    }
    return 0.0;
  }

  double Pressure1(double t, const Point &x) const {
    return 0.7 * sin(omega * x[0]) * sin(omega * x[1]);
  }

  double Pressure0(double t, const Point &x) const {
    return 0.3 * sin(omega * x[0]) * sin(omega * x[1]);
  }

  VectorField Velocity(double t, const Point &x) const {
    return VectorField((t - 1) * omega * sin(omega * x[1]) * cos(omega * x[0]),
                       (t - 1) * omega * sin(omega * x[0]) * cos(omega * x[1]), 0.0);
  }

  int Deg() const { return 1000; }

  double ForceP0(double t, const Point &x) const {
    return 2 * (t - 1) * omega * omega * sin(omega * x[0]) * sin(omega * x[1]);
  }

  double ForceP1(double t, const Cell &C, const Point &x) const {
    return 2 * (t - 1) * omega * omega * sin(omega * x[0]) * sin(omega * x[1])
           + 1.0 / (Tau_i(C, x, 1) * Kappa_i(C, x, 1)) * Pressure1(t, x);
  }

  int SourceDeg() const { return 1000; }

  double Rho(const Cell &, const Point &) const {
    return rho;
  }

  double Tau_i(const Cell &, const Point &, int) const {
    return tau1;
  }

  double Kappa_i(const Cell &, const Point &, int i) const {
    switch (i) {
      case 0:
        return kappa0;
      case 1:
        return kappa1;
      default: Exit("wrong kappa!")
    }
    return 0.0;
  }
};

static ProblemVAcoustic *GetProblemVAcoustic(const std::string &name) {
  int dim = 2;
  Config::Get("dim", dim);
  if (name == "Acoustic_Cosine") {
    return new VAcoustic_Benchmark_Cosine(dim);
  }
  if (name == "Riemann_osc") return new VAcoustic_Riemann_osc();
  if (name == "Benchmark_Linear1")
    return new VAcoustic_Benchmark_Linear1();
  if (name == "Benchmark_Linear1")
    return new VAcoustic_Benchmark_Linear1();
  if (name == "Benchmark_Linear2")
    return new VAcoustic_Benchmark_Linear2();
  if (name == "Benchmark_Linear3")
    return new VAcoustic_Benchmark_Linear3();
  if (name == "Benchmark_Constant1")
    return new VAcoustic_Benchmark_Constant1();
  if (name == "Benchmark_Constant1_Acoustic")
    return new Acoustic_Benchmark_Constant1();
  if (name == "Benchmark_Constant2")
    return new VAcoustic_Benchmark_Constant2();
  if (name == "Benchmark_ExpP1")
    return new VAcoustic_Benchmark_ExpP1();
  if (name == "Benchmark_ExpP1_2")
    return new VAcoustic_Benchmark_ExpP1_2();
  if (name == "Riemann")
    return new VAcoustic_Riemann(dim);
  if (name == "Spherical")
    return new VAcoustic_Spherical();
  if (name == "HorizontalReflector")
    return new VAcoustic_HorizontalReflector();
  if (name == "Ricker1D")
    return new VAcoustic_Ricker1D();
  if (name == "AcousticSinCos2d_RHS")
    return new AcousticSinCos2d_RHS();
  if (name == "VAcousticSinCos2d_RHS")
    return new VAcousticSinCos2d_RHS();
  if (name == "VAcousticWaveC2Benchmark")
    return new VAcousticWaveC2Benchmark();
  else Exit("Error: no wave found!");
  return nullptr;
}

static ProblemVAcoustic *GetProblem(const std::string &name, ProblemVAcoustic) {
  int dim = 2;
  Config::Get("dim", dim);
  if (name == "Acoustic_Cosine") {
    return new VAcoustic_Benchmark_Cosine(dim);
  }
  if (name == "Riemann_osc") return new VAcoustic_Riemann_osc();
  if (name == "Benchmark_Linear1")
    return new VAcoustic_Benchmark_Linear1();
  if (name == "Benchmark_Linear1")
    return new VAcoustic_Benchmark_Linear1();
  if (name == "Benchmark_Linear2")
    return new VAcoustic_Benchmark_Linear2();
  if (name == "Benchmark_Linear3")
    return new VAcoustic_Benchmark_Linear3();
  if (name == "Benchmark_Constant1")
    return new VAcoustic_Benchmark_Constant1();
  if (name == "Benchmark_Constant1_Acoustic")
    return new Acoustic_Benchmark_Constant1();
  if (name == "Benchmark_Constant2")
    return new VAcoustic_Benchmark_Constant2();
  if (name == "Benchmark_ExpP1")
    return new VAcoustic_Benchmark_ExpP1();
  if (name == "Benchmark_ExpP1_2")
    return new VAcoustic_Benchmark_ExpP1_2();
  if (name == "Riemann")
    return new VAcoustic_Riemann(dim);
  if (name == "Spherical")
    return new VAcoustic_Spherical();
  if (name == "HorizontalReflector")
    return new VAcoustic_HorizontalReflector();
  if (name == "Ricker1D")
    return new VAcoustic_Ricker1D();
  if (name == "AcousticSinCos2d_RHS")
    return new AcousticSinCos2d_RHS();
  if (name == "VAcousticSinCos2d_RHS")
    return new VAcousticSinCos2d_RHS();
  if (name == "VAcousticWaveC2Benchmark")
    return new VAcousticWaveC2Benchmark();
  else Exit("Error: no wave found!");
  return nullptr;
}


#endif
