NonmigratedFiles// $Header: https://svn.math.kit.edu/svn/M++/trunk/DGWave/src/DGElasticity.h 588 2016-02-03 16:28:12Z ziegler $

#ifndef _DGELASTICITY_H_
#define _DGELASTICITY_H_

#include "m++.hpp"
#include "DGMatrixGraph.hpp"
#include <fstream>

class DGElasticityDoF : public DoF {
    short cell_std_deg;
     std::unordered_map<Point,vector<short>> cell_deg;
 public:
    DGElasticityDoF (short cell_std_deg_) : DoF(0,false),
    cell_std_deg(cell_std_deg_) {}
    int get_celldeg (const Point& z, int ud = 0) const {
        return cell_std_deg;
        if (cell_deg.find(z) != cell_deg.end()) {
            conststd::vector<short>& degs = cell_deg.find(z)->second;
            if (degs[ud] > 0) return degs[ud];
        }
        return cell_std_deg;
    }
    void set_cell_deg (const Point& z, int ud, short deg) {
       std::vector<short>* c_deg = NULL;
        if (cell_deg.find(z) != cell_deg.end())
            c_deg = &cell_deg[z];
        else {
            c_deg = &cell_deg[z];
            c_deg->resize(1,-1);
        }
        (*c_deg)[ud] = deg;
    }
    int CellDoFs (const cell& c) const {
    int m = 9;
        if ( c.dim() == 2 ) m = 5;
        int c_deg = get_celldeg(c(),0);
        switch (c.Type()) {
            case TRIANGLE:      return m*(c_deg+1)*(c_deg+2)/2;
            case TETRAHEDRON:   Exit("not implemented");
            case QUADRILATERAL: return m*(c_deg+1)*(c_deg+1);
            case HEXAHEDRON:    return m*(c_deg+1)*(c_deg+1)*(c_deg+1);
        }
        Exit("not implemented");
    }
    void NodalDoFs (const cell& c,std::vector<short>& z) const {
    z.resize(1);
        z[0] = CellDoFs(c);
    }
    int NodalPoints (const cell& c) const { return 1; }
    void NodalPoints (const cell& c,std::vector<Point>& z) const {
        z.resize(1);
        z[0] = c();
    }
    std::string Name () const { return "DGElasticDoF"; }
    int TypeDoF (int) const { return CELL; }
};


DGElasticityDoF* GetDGElasticityDoF (int cell_std_deg = -1) {
    if(cell_std_deg == -1)
    {
        Config::Get("deg", cell_std_deg);
//        if (cell_std_deg==1 || cell_std_deg==0) Exit("weird flags for FWI")
    }
    return new DGElasticityDoF(cell_std_deg);
}

class DGElasticityDiscretization  : public IDiscretization {
    const Shape* S_tri[5];
    const Shape* S_quad[6];
    const Shape* S_tet[4];
    const Shape* S_hex[4];
    const Shape* T_tri[5];
    const Shape* T_quad[5];
    const Shape* T_tet[2];
    const Shape* T_hex[2];
    DGElasticityDoF& D;
public:
    DGElasticityDiscretization (DGElasticityDoF* dof) :
        IDiscretization(dof,"DG"), D(*dof) {
        S_tri[0] = Get_Shape("P0tri");
        S_tri[1] = Get_Shape("P1tri");
        S_tri[2] = Get_Shape("P2tri");
        S_tri[3] = Get_Shape("P3tri");
        S_tri[4] = Get_Shape("P4tri");
	
        T_tri[1] = Get_Shape("RT0tri");
	
        S_quad[0] = Get_Shape("P0quad");
        S_quad[1] = Get_Shape("P1quad");
        S_quad[2] = Get_Shape("P2quad");
        S_quad[3] = Get_Shape("P3quad");
        S_quad[4] = Get_Shape("P4quad");
        S_quad[5] = Get_Shape("P5quad");
        S_tet[1] = Get_Shape("P1tet");
        S_tet[2] = Get_Shape("P2tet");
	
        S_hex[1] = Get_Shape("P1hex");
        S_hex[2] = Get_Shape("P2hex");
        S_hex[3] = Get_Shape("P3hex");
	
        T_quad[1] = Get_Shape("RT0quad");
        T_tet[1] = Get_Shape("RT0tet");
        T_hex[1] = Get_Shape("RT0hex");
    }
    int get_celldeg (const Point& z, int component = 0) const { 
	return D.get_celldeg(z,component); 
    }
    const Shape& GetCellShape (const cell& c, int p, int component=0) const {
        if (p > 5) Exit("DGElasticityShapes::GetShape: Not implemented");
        switch (c.Corners()) {
            case 3: return *S_tri[p];
            case 4:
                switch (c.dim()) {
                    case 2: return *S_quad[p];
                    case 3: return *S_tet[p];
                }
            case 8: return *S_hex[p];
        }
        Exit("DGElasticityShapes::GetShape: Not implemented!");
    }
//    int CellDoFs (const cell& c) const { return D.CellDoFs(c); }
    std::string DiscName () const { return "DGElasticityDiscretization"; }
    std::string Name_Disc_dG () const { return "DGElasticityDiscretization"; }
};

class DGElasticityElement : public Element {
    const Shape& S;
    const ScalarShapeValues& value;
    ShapeGradient gradient;
    VectorField velocity[MaxQuadraturePoints][MaxNodalPoints];
    Tensor stress[MaxQuadraturePoints][MaxNodalPoints];
    Tensor strain[MaxQuadraturePoints][MaxNodalPoints];
    VectorField divstress[MaxQuadraturePoints][MaxNodalPoints]; 
    int deg;
    const cell& C;
    int dim;
public:
    Point GlobalToLocal (const Point& z) const {
        switch (C.Type()) {
            case TRIANGLE: {
                Tensor T;
                T[0][0] = C.Corner(1)[0]-C.Corner(0)[0];
                T[0][1] = C.Corner(2)[0]-C.Corner(0)[0];
                T[1][0] = C.Corner(1)[1]-C.Corner(0)[1];
                T[1][1] = C.Corner(2)[1]-C.Corner(0)[1];
                T[0][2] = 0;
                T[1][2] = 0;
                T[2][0] = 0;
                T[2][1] = 0;
                T[2][2] = 1;
                Tensor IT = Invert(T);
                return IT * (z - C.Corner(0));
            }
            
            case TETRAHEDRON:
                Exit("not implemented");
                
            case QUADRILATERAL: {
                Tensor T;
                T[0][0] = C.Corner(1)[0]-C.Corner(0)[0];
                T[0][1] = C.Corner(3)[0]-C.Corner(0)[0];
                T[1][0] = C.Corner(1)[1]-C.Corner(0)[1];
                T[1][1] = C.Corner(3)[1]-C.Corner(0)[1];
                T[0][2] = 0;
                T[1][2] = 0;
                T[2][0] = 0;
                T[2][1] = 0;
                T[2][2] = 1;
                Tensor IT = Invert(T);
                return IT * (z - C.Corner(0));
            }
                
            case HEXAHEDRON:
                Exit("not implemented");
                
        }
    }
    void NodalPoints (Point* z) const { S.NodalPoints(C,z); }
    int get_deg() { return deg; }
    DGElasticityElement(const DGElasticityDiscretization &disc, const VectorMatrixBase &g, const cell &c) :
            Element(g, c, disc.GetCellShape(c, disc.get_celldeg(c())).GetQuad()),
            S(disc.GetCellShape(c, disc.get_celldeg(c()))), value(S.values()),
            C(c), dim(c.dim()),
            deg(disc.get_celldeg(c())) {
        for (int q=0; q<nQ(); ++q) {
            for (int i=0; i<S.size(); ++i)
                gradient[q][i] = T[q] * S.LocalGradient(q,i);
            for (int k=0; k<dim; ++k) {
                VectorField V = zero;
                V[k] = 1.0;
                for (int i=0; i<S.size(); ++i) {
                    //velocity[q][i+k*m] = zero;
                    //mout << "value[" << q << "][" << i << "]: " << value[q][i] << endl << endl;
                    velocity[q][i+k*S.size()][k] = value[q][i] * V[k];
                    Tensor Dv = Zero;
                    Dv[k] = gradient[q][i];
                    strain[q][i+k*S.size()] = sym(Dv);
                }
            }
            for (int k=0; k<dim; ++k) {
                for (int i=0; i<S.size(); ++i) {
                    stress[q][i+k*S.size()] = Zero;
                    stress[q][i+k*S.size()][k][k] = value[q][i];
                    divstress[q][i+k*S.size()] = zero;
                    divstress[q][i+k*S.size()][k] = gradient[q][i][k];
                    if (dim == 2) {
                        stress[q][i+2*S.size()] = Zero;
                        stress[q][i+2*S.size()][0][1] = value[q][i]; //0.5*value[q][i];
                        stress[q][i+2*S.size()][1][0] = value[q][i]; //0.5*value[q][i];
                        divstress[q][i+2*S.size()][0] = gradient[q][i][1]; //0.5*gradient[q][i][1];
                        divstress[q][i+2*S.size()][1] = gradient[q][i][0]; //0.5*gradient[q][i][0];
                    }
                    if (dim == 3) {
                       stress[q][i+3*S.size()] = Zero;
                       stress[q][i+3*S.size()][0][1] = value[q][i];
                       stress[q][i+3*S.size()][1][0] = value[q][i];
                       stress[q][i+4*S.size()] = Zero;
                       stress[q][i+4*S.size()][0][2] = value[q][i];
                       stress[q][i+4*S.size()][2][0] = value[q][i];
                       stress[q][i+5*S.size()] = Zero;
                       stress[q][i+5*S.size()][1][3] = value[q][i];
                       stress[q][i+5*S.size()][3][1] = value[q][i];
                       //divstress[q][i+3*S.size()][0] = gradient[q][i][1]; //0.5*gradient[q][i][1];
                       //divstress[q][i+3*S.size()][1] = gradient[q][i][0]; //0.5*gradient[q][i][0];
                    }
                }
            }
        }
    }
    int shape_size () const { return S.size(); }
    int dimension () const {
        if (dim==3) return 9 * shape_size();
        if (dim==2) return 5 * shape_size();
    }
    int u_dimension () const {
        if (dim==3) return 9;
        if (dim==2) return 5;
    }
    int V_dimension () const { return dim * shape_size(); }
    int S_dimension () const {
        if (dim==3) return 6 * shape_size();
        if (dim==2) return 3 * shape_size();
    }
    VectorField Velocity (int q, int i) {
        return velocity[q][i];
    }
    Tensor Stress (int q, int i) {
        return stress[q][i];
    }
    VectorField divStress (int q, int i) {
        return divstress[q][i];
    }
    Tensor Strain (int q, int i) {
        return strain[q][i];
    }
    VectorField divStress(int q, const Vector& u) {
        VectorField S_div = zero;
        for (int i=0; i<S_dimension(); ++i) {
            S_div += u(r(0),i) * divStress(q,i);
        }
        return S_div;
        
    }
    VectorField Velocity (int q, const Vector& u) {
        VectorField V = zero;
        for (int i=0; i<V_dimension(); ++i) {
            //mout << "######################" << endl << endl;
            V += u(r(0),S_dimension()+i) * velocity[q][i];
            //mout << "u(r(0)," << S_dimension()+i << "): " << u(r(0),S_dimension()+i) << endl << endl;
            //mout << "q:" << endl << qPoint[q] << endl << endl;
            //mout << "velocity[" << q << "][" << i << "]:" << endl << velocity[q][i] << endl << endl;
            //mout << "V_" << i << endl << V << endl << endl;
        }
        //mout << "V" << endl << V << endl;
        //if(q == 0)
        //    mout << "q: " << q << endl << "u*" << endl << u << endl;
        return V;
    }
    Tensor Strain (int q, const Vector& u) {
        Tensor E = Zero;
        /*for (int i=0; i<V_dimension(); ++i)
        mout << "i " << i << " u " << u(r(0),S_dimension()+i)
         << " V " << velocity[q][i] << " E " << endl
         << strain[q][i];*/
        for (int i=0; i<V_dimension(); ++i) {
            E += u(r(0),S_dimension()+i) * strain[q][i];
            mout << "strain[" << q << "][" << i << "]: " << endl << strain[q][i] << endl << endl;
        }
        return E;
    }
    Tensor Stress (int q, const Vector& u) {
        Tensor S = Zero;
        for (int i=0; i<S_dimension(); ++i) {
//            mout << "u(r(0)," << i << "): " << u(r(0),i) << endl << endl;
//            S += u(r(0),i) * stress[q][i];
//            mout << "####################################" << endl << endl;
//            mout << "u(r(0)," << i << "): " << u(r(0),i) << endl << endl;
//            mout << "stress[" << q << "][" << i <<"]:" << endl << stress[q][i] << endl;
//            mout << "Hier: " << endl << stress[q][i] << endl << u(r(0),i) << endl << endl;
//            mout << "S_: " << i << endl << S << endl << endl;
        }
        //mout << "S" << endl << S << endl;
        /*if(q == 1) {
            mout << "S" << endl << S << endl;
            mout << "q: " << q << endl << "u*" << endl << u << endl;
        }*/
        return S;
    }
    Scalar SVValue (const Point& lz, const Vector& u, int k) const { 
        Scalar v = 0;
        for (int i=0; i<S.size(); ++i)
            v += u(C(),k*S.size()+i) * S(lz,i);
        return v;
    }
};

class DGElasticityFaceElement : public FaceElement {
    VectorField velocity[MaxQuadraturePoints][MaxNodalPoints];
    Tensor stress[MaxQuadraturePoints][MaxNodalPoints];
    int d;
    int dim;
    const Shape& S;
    const cell& c;
    face f;
    int fid;
    Point qTangent[MaxQuadraturePoints];
    ShapeValue value;
    double ww;
public:
    DGElasticityFaceElement(const DGElasticityDiscretization &D, const VectorMatrixBase &g,
                            const cell &C, int f_id) :
            FaceElement(g, C, f_id, D.GetCellShape(C, D.get_celldeg(C())).GetFaceQuad()),
            S(D.GetCellShape(C, D.get_celldeg(C()))),
            c(C), f(g.GetMesh().find_face(c.Face(f_id))),
            fid(f_id), dim(C.dim()) {
        d = dim*S.size();
        ww = 0;
        for (int q=0; q<nQ(); ++q) ww += QWeight(q);
        ww = c.LocalFaceArea(fid) / ww;
        for (int q=0; q<nQ(); ++q) {
            qTangent[q] = c.FaceCorner(f_id,1) - c.FaceCorner(f_id,0);
            double w = norm(qTangent[q]);
            qTangent[q] *= (1/w);
        }
        for (int q=0; q<nQ(); ++q) {
            for (int i=0; i<S.size(); ++i) 
                value[q][i] = S(QLocal(q),i);
            for (int k=0; k<dim; ++k) {
                VectorField V = zero;
                V[k] = 1.0;
                for (int i=0; i<S.size(); ++i) {
                    velocity[q][i+k*S.size()] = zero;
                    velocity[q][i+k*S.size()][k] = value[q][i] * V[k];
                }
            }
            for (int k=0; k<dim; ++k) {
                for (int i=0; i<S.size(); ++i) {
                    stress[q][i+k*S.size()] = Zero;
                    stress[q][i+k*S.size()][k][k] = value[q][i];
                    if (dim == 2) {
                        stress[q][i+2*S.size()] = Zero;
                        stress[q][i+2*S.size()][0][1] = value[q][i];
                        stress[q][i+2*S.size()][1][0] = value[q][i];
                    }
                    if (dim == 3) {
                        stress[q][i+3*S.size()] = Zero;
                        stress[q][i+3*S.size()][0][1] = value[q][i];
                        stress[q][i+3*S.size()][1][0] = value[q][i];
                        stress[q][i+4*S.size()] = Zero;
                        stress[q][i+4*S.size()][0][2] = value[q][i];
                        stress[q][i+4*S.size()][2][0] = value[q][i];
                        stress[q][i+5*S.size()] = Zero;
                        stress[q][i+5*S.size()][1][3] = value[q][i];
                        stress[q][i+5*S.size()][3][1] = value[q][i];
                    }
                }
            }            
        }
    }
    const Point& LocalFaceCorner (int i) const {
        return c.LocalCorner(c.facecorner(fid,i));
    }
    Point Local (const Point& local) const {
        switch (c.FaceCorners(fid)) {
        case 2: return (1-local[0]) * LocalFaceCorner(0)
                + local[0] * LocalFaceCorner(1);
        case 3: return (1-local[0]-local[1]) * LocalFaceCorner(0)
                + local[0] * LocalFaceCorner(1)
                + local[1] * LocalFaceCorner(2);
        case 4: return (1-local[0])*(1-local[1]) * LocalFaceCorner(0)
                + local[0]*(1-local[1]) * LocalFaceCorner(1)
                + local[0]*local[1] * LocalFaceCorner(2)
                + (1-local[0])*local[1] * LocalFaceCorner(3);
        }
    }    
    const Point& QTangent (int q) const     { return qTangent[q]; }
    int shape_size () const                 { return S.size(); }
    int dimension () const {
        if (dim==3) return 9 * shape_size();
        if (dim==2) return 5 * shape_size();
    }
    int V_dimension () const { return dim * shape_size(); }
    int S_dimension () const {
        if (dim==3) return 6 * shape_size();
        if (dim==2) return 3 * shape_size();
    }
    VectorField Velocity (int q, int i) {
        //mout << i << ": " << value[q][i] << endl << endl;
        //mout << i << ": " << endl << velocity[q][i] << endl << endl;
        return velocity[q][i];
    }
    Tensor Stress (int q, int i) {
        return stress[q][i];
    }
};

class ProblemElasticity {
    const DGElasticityDiscretization& disc;
    int verbose;
protected:
    bool exact;
    int dim;
    double lambda;
    double mu;
    double rho;
    double c_S;
    double c_P;
    std::string meshname;
public:
    ProblemElasticity (const DGElasticityDiscretization& _disc, int d=3) :
    disc(_disc), dim(d), verbose(0), exact(false), mu(1), rho(1), lambda(1), meshname("none") {
        Config::Get("ProblemVerbose", verbose);
        Config::Get("ProblemExact", exact);
        Config::Get("lambda", lambda);
        Config::Get("mu", mu);
        Config::Get("rho", rho);
        c_P = sqrt((2.0 * mu + lambda) / rho);
        c_S = sqrt(mu / rho);
        Config::Get("Mesh", meshname);
    }
    virtual const char* Name () const = 0;
    virtual const char* MeshName () const { return meshname.c_str(); }
    virtual bool Exact () const { return exact; }
    virtual bool RHS(double) const { return false; }
    virtual double ut (double, const Point&, int) const { return 0; }
    virtual double utb (double, const Point&, int) const { return 0; }
    virtual double ut_dual (double &,const cell&, const Point&, int, const Point&, const Point&) const { return 0; }
    virtual double u0 (const Point& x, int i) const { return ut(0,x,i); }
    virtual double Lambda (const Point& x) const { return lambda; }
    virtual double Mu (const Point& x) const { return mu; }
    virtual double Rho (const Point& x) const { return rho; }
    double C_s (const Point& x) const { return sqrt(Mu(x)/Rho(x)); }
    double C_p (const Point& x) const { return sqrt((2*Mu(x)+Lambda(x))/Rho(x)); }
    virtual VectorField Dirichlet (double t, const Point& z) const {
        return VectorField(ut(t,z,3),ut(t,z,4),0.0);
    }
    virtual VectorField Neumann (double t, const Point z,
                            const VectorField& N) const {

	Tensor S(ut(t,z,0),ut(t,z,2),0.0,
		 ut(t,z,2),ut(t,z,1),0.0,
		 0.0,0.0,0.0);
        return S * N;
    }
    virtual void get_u_int (double t, Vector& u) const {
        u = 0;
	Point z[MaxNodalPointsDG];
        std::string discName = disc.Name_Disc_dG();
// 	mout << discName << endl;
        if (discName == "dGvector_elasticity_FV") {
            for (cell c=u.cells(); c!=u.cells_end(); ++c) {                
                row r = u.find_row(c());
                DGElasticityElement Elem(disc,u,c);
                for (int i=0; i<Elem.u_dimension();++i) 
		    for (int j=0; j<c.Corners(); ++j)
			u(r)[i] += ut(t,c[j],i) / c.Corners();
            }
        } else {
            for (cell c=u.cells(); c!=u.cells_end(); ++c) {
                row r = u.find_row(c());
                DGElasticityElement Elem(disc,u,c);
		Elem.NodalPoints(z);
                for (int i=0; i<Elem.u_dimension();++i) 
                    for (int j=0; j<Elem.shape_size();++j) {
                        u(r)[i*Elem.shape_size()+j] = ut(t,z[j],i);
			
//			double U = ut(t,z[j],i);
//			if (abs(U) > 0) 
//			    mout << OUT(j) << OUT(z[j]) << OUT(U) << endl; 

		    }
            }
        }
        u.Accumulate();
	// mout << OUT(u);
    }                
    virtual void get_u_int (Vector& u) const {
        u = 0;
	Point z[MaxNodalPointsDG];
        std::string discName = disc.Name_Disc_dG();
        if(discName == "dGvector_elasticity_FV") {
            mout << "FV Interpolation" << endl;
            const Quadrature* Q;
            if(dim == 3) Q = &GetQuadrature("Qtet11");
            else Q = &GetQuadrature("Qtri7");
           std::vector<double> qWeight;
           std::vector<Point> qPoint;
            qWeight.resize((*Q).size());
            qPoint.resize((*Q).size());                
            for (cell c=u.cells(); c!=u.cells_end(); ++c) {                
                for (int q=0; q<(*Q).size(); ++q) {
                    qWeight[q] = (c.GetTransformation((*Q).QPoint(q))).Det() * (*Q).Weight(q);
                    qPoint[q] = c[(*Q).QPoint(q)]; 
                }
                DGElasticityElement Elem(disc,u,c);
                row r = u.find_row(c());
                for(int i=0;i<Elem.dimension();++i) {
                    double tmp = 0.0;
                    for (int q=0; q<(*Q).size(); ++q) {
                        tmp += qWeight[q]*u0(qPoint[q],i);
                        //mout << u0(qPoint[q],i) << endl << endl;
                    }
                    u(r)[i] = tmp/Elem.Area();
                }
            }
        } else {
            mout << "Lagrange Interpolation" << endl;
            // Lagrange Interpolation
            for (cell c=u.cells(); c!=u.cells_end(); ++c) {
                row r = u.find_row(c());
                DGElasticityElement Elem(disc,u,c);
		Elem.NodalPoints(z);
                int NN = 0;
                if(dim == 3) NN = 9; else NN = 5;
                for(int iN=0;iN<NN;++iN) {
                    for (int i=0;i<Elem.shape_size();++i) {
                        u(r,iN*Elem.shape_size()+i) = u0(z[iN*Elem.shape_size()+i],iN);
                    }
                }
            }
        }
        u.Accumulate();
    }
};

class Elastic_Riemann  : public ProblemElasticity {
    double c;
    Point direc;
    double A_osc;
    double k_osc;
    double width;
    double shift;
    double lambda2;
    double mu2;
    double jump;
 public:
    Elastic_Riemann (const DGElasticityDiscretization& _disc)
        : ProblemElasticity(_disc,2), c(1),
	direc(1.0,0.0), A_osc(0.5), k_osc(10), width(0.125), shift(2.5), lambda2(lambda), mu2(mu), jump(2) {
        Config::Get("k_osc", k_osc);
        Config::Get("A_osc", A_osc);
        Config::Get("width", width);
        Config::Get("shift", shift);
        Config::Get("jump", jump);
        Config::Get("lambda2", lambda2);
        Config::Get("mu2", mu2);
    }
    double Lambda (const Point& x) const {
	if (x[0] < jump) return lambda;
	else return lambda2;
    }
    double Mu (const Point& x) const {
	if (x[0] < jump) return mu;
	else return mu2;
    }
    bool RHS(double) const { return false; }
    bool Exact() const { return true; }
    const char* Name () const { return "Elastic_Riemann"; }
    double A0 (double s) const {
        if (s*s >= width) return 0;
        return exp(-1/(width-s*s)) / exp(-1/width);
    }
    double A (double s) const { return A0(s+shift); }
    double ut (double t, const Point& x, int i) const {
	double y = A(t-x[0]/c_S);
        switch (i) {
        case 2: return -mu * y;
        case 4: return c_S * y;
        default: return 0;
        }
    }
};

class Hat_2D : public ProblemElasticity {
    double mu;
    double lambda;
    double rho;
public:
    Hat_2D (const DGElasticityDiscretization& _disc) : ProblemElasticity(_disc,2) {}
    const char* Name () const { return "Hap_2D"; }
    double ut (double t, const Point& x, int i) const {
        Point Pmid = Point(0.5,1.0);
        //Point Pmid = Point(1.0,0.5);
        //Point Pmid = Point(0.5,0.5);
        double rr;
        switch (i) {
            case 4:
                rr = pow(0.5,3) - norm(Pmid-x);
                if(norm(Pmid-x)<0.05)
                    return rr*4;
                else
                    return 0.0;
        }
    }
};

class Linear : public ProblemElasticity {
public:
    Linear (const DGElasticityDiscretization& _disc) : ProblemElasticity(_disc,2) {}
    const char* Name () const { return "Linear"; }
    bool RHS(double) const { return true; }
    bool Exact() const { return true; }
    double ut (double t, const Point& x, int i) const {
	if (i==0) return 2*x[0];
	if (i==3) return 3*x[0];
	if (i==4) return x[1];
	return 0.0;
    }
    double utb (double t, const Point& z, int i) const {
	if (i==0) return -3;
	if (i==1) return -1;
	if (i==3) return -2;
	return 0.0;
    }
};

class Spherical_Wave_2D : public ProblemElasticity {
    double aa;
    double rad;
    Point Pmid;
    int linear_weight;
public:
    Spherical_Wave_2D (const DGElasticityDiscretization& _disc) : ProblemElasticity(_disc,2) {
        aa = 1.0;//0.25;
        rad = 2.;
        Config::Get("ProblemMid", Pmid);
        Config::Get("linear_weight", linear_weight);
    }
    const char* Name () const { return "Spherical_Wave_2D"; }
    virtual double ut (double t, const Point& x, int i) const {
        Point Pmid = Point(2, 2);
        double r = dist(Pmid,x);
        switch (i) {
            case 4 :
                if (r < 1.0/2.0)
                {
                    return aa * pow(cos(2.0*Pi*r - 2.0*Pi*t) + 1.0,2.0); 
                }
                else 
                {
                    return 0.0;
                } 
            default: return 0.0;
        }
    }
    virtual double ut_dual (double& t, const Point& x, int i, const Point& a, const Point& b) const {
        if ( x[0]<a[0] || x[1]<a[1] || x[0]>b[0] || x[1]>b[1] ) return 0.0;
        switch (linear_weight){
            case 1: {
                Point ba = b-a;
                if (i<=2) return 0; //S
                else return 1.0/(ba[0]*ba[1]); //V
            }
            default: Exit("Not implemented");
        }
    }
};

class Div_free_test : public ProblemElasticity {
    double aa;
    double rad;
    Point Pmid;
    int linear_weight;
    double omega;
public:
    Div_free_test (const DGElasticityDiscretization& _disc) : ProblemElasticity(_disc,2) {
        aa = 1.0;//0.25;
        rad = 2.;
        Config::Get("ProblemMid", Pmid);
        Config::Get("linear_weight", linear_weight);
        omega = M_PI;
    }
    const char* Name () const { return "Div_free_test"; }
    virtual double ut (double t, const Point& z, int i) const {
        double x = z[0];
        double y = z[1];
        Point Pmid = Point(2.0, 2.0);
        if (x*x+y*y >= 1) return 0.0;
        switch (i) {
            case 0 :
                    return -1.0/2.0 * exp(1.0/(x * x + y * y - 1.0)) * ((2.0 * y * y)/pow(x * x + y * y - 1.0,2) - 1);
            case 1 :
                    return -1.0/2.0 * exp(1.0/(x * x + y * y - 1.0)) * ((2.0 * x * x)/pow(x * x + y * y - 1.0,2) - 1);
            case 2 :
                return x * y/pow(-x * x - y * y + 1.0,2) * exp(-1.0/(1.0 - x * x - y * y));
            case 3 : return 0;
            case 4 : return 0;
            default : return 0.0;
        }
    }
    virtual double ut_dual (double& t, const Point& x, int i, const Point& a, const Point& b) const {
        if ( x[0]<a[0] || x[1]<a[1] || x[0]>b[0] || x[1]>b[1] ) return 0.0;
        switch (linear_weight){
            case 1: {
                Point ba = b-a;
                if (i<=2) return 0; //S
                else return 1.0/(ba[0]*ba[1]); //V
            }
            default: Exit("Not implemented");
        }
    }
};

class Plane_Wave_2D_Test : public ProblemElasticity {
    double mu;
    double lambda;
    double rho;
public:
    Plane_Wave_2D_Test (const DGElasticityDiscretization& _disc) : ProblemElasticity(_disc,2) {
        mu = 1.0;
        lambda = 2.0;
        rho = 1.0;
        Config::Get("lambda", lambda);
        Config::Get("mu", mu);
        Config::Get("rho", rho);
    }
    const char* Name () const { return "Plane_Wave_2D_Test"; }
    virtual double ut (double t, const Point& x, int i) const {
        switch (i) {
            case 0: return cos(c_P * t - x[0]) * (2.0 * mu + lambda);
            case 1: return cos(c_P * t - x[0]) * lambda;
            case 2: return cos(c_S * t - x[0]) * mu;
            case 3: return cos(c_P * t - x[0]) * c_P;
            case 4: return cos(c_S * t - x[0]) * c_S;
            default: return 0.0;
        }
    }
};

class Plane_Wave_2D : public ProblemElasticity {
    double mu;
    double lambda;
    double rho;
public:
    Plane_Wave_2D (const DGElasticityDiscretization& _disc) : ProblemElasticity(_disc,2) {
        mu = 1.0;
        lambda = 2.0;
        rho = 1.0;
        Config::Get("lambda", lambda);
        Config::Get("mu", mu);
        Config::Get("rho", rho);
    }
    const char* Name () const { return "Plane_Wave_2D"; }
    virtual double ut (double t, const Point& x, int i) const {
        switch (i) {
            case 0: return (c_P * t - x[0]) * (2.0 * mu + lambda);
            case 1: return (c_P * t - x[0]) * lambda;
            case 2: return (c_S * t - x[0]) * mu;
            case 3: return (c_P * t - x[0]) * c_P;
            case 4: return (c_S * t - x[0]) * c_S;
            default: return 0.0;
        }
    }
    virtual double u0 (const Point& x, int i) const {
        return ut(0,x,i);
    }
};

class Plane_Wave_2D_2 : public ProblemElasticity {
    double aa;
    double bb;
public:
    Plane_Wave_2D_2 (const DGElasticityDiscretization& _disc) 
	: ProblemElasticity(_disc,2) {
        aa = 2.0;
        bb = 2.0;
        Config::Get("aa", aa);
        Config::Get("bb", bb);
    }
    const char* Name () const { return "Plane_Wave_2D_2"; }
    double ut (double t, const Point& x, int i) const {
        Point E0 = Point(bb,-aa);
        double w = sqrt(aa*aa + bb*bb);
        E0 /= w;
        switch (i) {
            case 0 : return sin(w*Pi*t)*sin(2.*Pi*(x[0]-bb)/aa + Pi);
            case 1 : return 0;
            case 2 : return cos(w*Pi*t)*cos(2.*Pi*((x[0]-bb)/aa) + Pi ) + 1.0;
            default : return 0.0;
        }
    }
};

class ElasticBenchmark_Constant_V : public ProblemElasticity {
public:
    ElasticBenchmark_Constant_V (const DGElasticityDiscretization& _disc) : ProblemElasticity(_disc,2) {}
    const char* Name () const { return "v = (1,2)"; }
    bool Exact() const { return true; }
    double ut (double t, const Point& x, int i) const {
        switch (i) {
            case 3 : return 1;
            case 4 : return 2;
            default : return 0;
        }
    }
    bool RHS(double) const { return false; }
};

class ElasticBenchmark_Constant_S : public ProblemElasticity {
public:
    ElasticBenchmark_Constant_S (const DGElasticityDiscretization& _disc) : ProblemElasticity(_disc,2) {}
    const char* Name () const { return "S = (1,3,2)"; }
    bool Exact() const { return true; }
    double ut (double t, const Point& x, int i) const {
        switch (i) {
            case 0 : return 1;
            case 1 : return 3;
            case 2 : return 2;
            default : return 0;
        }
    }
    bool RHS(double) const { return false; }
};

class ElasticBenchmark_Constant_P_div_free : public ProblemElasticity {
public:
    ElasticBenchmark_Constant_P_div_free (const DGElasticityDiscretization& _disc)
	: ProblemElasticity(_disc,2) {}
    bool Exact() const { return true; }
    const char* Name () const { return "S = (x,-x-y;-x-y,y)"; }
    double ut (double t, const Point& x, int i) const {
        switch (i) {
            case 1 : return x[1];
            case 2 : return -x[0];
            default : return 0;
        }
    }
};

class ElasticBenchmark_Quadratic : public ProblemElasticity {
public:
    ElasticBenchmark_Quadratic (const DGElasticityDiscretization& _disc) : ProblemElasticity(_disc,2) {
        mu = 0.5;
        lambda = 1.0;
        rho = 1.0;
        c_P = sqrt((2.0*mu + lambda) / rho);
        c_S = sqrt(mu / rho);
        // v = ( t*x+0.75*t^2, x*y) sigma = (t^2 + t * x + 0.5 * x^2, 0.5 * t^2 + 2 * t * x, 0.5*y*t)
    }
    const char* Name () const { return "Q2 Polynom"; }
    double ut (double t, const Point& z, int i) const {
        double x = z[0];
        double y = z[1];
//         mout << OUT(z) << endl;
        switch (i) {
            case 0 :
            {
//                 mout << OUT(t*t + t * x + 0.5 * x*x) << endl;
                return t*t + t * x + 0.5 * x*x;
            }
            case 1 : return 0.5 * t*t + 2 * t * x;
            case 2 : return 0.5*y*t;
            case 3 : return t*x+0.75*t*t;
            case 4 : return x*y;
            default : return 0.0;
        }
    }
};

class ElasticBenchmark_SinCosRHS : public ProblemElasticity{
    double omegat;
    double omega_x;
public:
    ElasticBenchmark_SinCosRHS(const DGElasticityDiscretization& _disc) :  ProblemElasticity(_disc,2)
    {
        mu = 0.5;
        lambda = 1.0;
        rho = 1.0;
        c_P = sqrt((2.0*mu + lambda) / rho);
        c_S = sqrt(mu / rho);
        omegat = M_PI;
        omega_x = M_PI;
        // v = (sin(omega_t * t)*cos(omega_x * y),cos(omega_t * t))
        //sigma = (cos(omega_x * x),sin(omega_t * omega_x * t * y), 0.0)
        //v_RHS = (omega_t*cos(omega_t*t)*cos(omega_x*y) + 1.0*omega_x*sin(omega_x*x), -1.0*omega_t*omega_x*t*cos(omega_t*omega_x*t*y) - omega_t*sin(omega_t*t))
        //s_RHS = (0, omega_t*omega_x*y*cos(omega_t*omega_x*t*y), 0.5*omega_x*sin(omega_t*t)*sin(omega_x*y))

    }
    const char* Name () const { return "sin - cos"; }
    bool RHS(double) const
    {
        return true;
        
    }
    double ut (double t, const Point& z, int i) const 
    {
        double x = z[0];
        double y = z[1];
        switch (i) {
            case 0 : return cos(omega_x * x);
            case 1 : return sin(omegat * omega_x * t * y);
            case 2 : return 0.0;
            case 3 : return sin(omegat * t)*cos(omega_x * y);
            case 4 : return cos(omegat * t);
            default : return 0.0;
        }
    }
    double utb (double t, const Point& z, int i) const 
    {
        double x = z[0];
        double y = z[1];
        switch (i) 
        {
            case 0 : return 0.0;
            case 1 : return  omegat*omega_x*y*cos(omegat*omega_x*t*y);
            case 2 : return  0.5*omega_x*sin(omegat*t)*sin(omega_x*y);
            case 3 : return  omegat*cos(omegat*t)*cos(omega_x*y) + 1.0*omega_x*sin(omega_x*x);
            case 4 : return - 1.0*omegat*omega_x*t*cos(omegat*omega_x*t*y) - omegat*sin(omegat*t);
            default : return 0.0;
        }
        
    }
};
class ElasticBenchmark_RHS : public ProblemElasticity{
public:
    ElasticBenchmark_RHS(const DGElasticityDiscretization& _disc) :  ProblemElasticity(_disc,2)
    {
        mu = 0.5;
        lambda = 1.0;
        rho = 1.0;
        c_P = sqrt((2.0*mu + lambda) / rho);
        c_S = sqrt(mu / rho);

    }
    const char* Name () const { return "s_xx = x*x"; }
    bool RHS(double) const
    {
        return true;
        
    }
    double ut (double t, const Point& z, int i) const 
    {
        double x = z[0];
        double y = z[1];
        switch (i) {
            case 0 : return x*x;
            case 1 : return 0.0;
            case 2 : return 0.0;
            case 3 : 
		{	
		return t;
		}
            case 4 : return 0.0;
            default : return 0.0;
        }
    }
    double utb (double t, const Point& z, int i) const 
    {
        double x = z[0];
        double y = z[1];
        switch (i) 
        {
            case 0 : return 0.0;
            case 1 : return 0.0;
            case 2 : return 0.0;
            case 3 : return -2.0*x + 1;
            case 4 : return 0.0;
            default : return 0.0;

        }
    }
};

class ElasticBenchmark_RHS_2 : public ProblemElasticity{
public:
    ElasticBenchmark_RHS_2(const DGElasticityDiscretization& _disc) :  ProblemElasticity(_disc,2)
    {
        mu = 0.5;
        lambda = 1.0;
        rho = 1.0;
        c_P = sqrt((2.0*mu + lambda) / rho);
        c_S = sqrt(mu / rho);

    }
    const char* Name () const { return "s_xx = x*x, v_y = x*y"; }
    bool RHS(double) const
    {
        return true;
        
    }
    double ut (double t, const Point& z, int i) const 
    {
        double x = z[0];
        double y = z[1];
        switch (i) {
            case 0 : return x*x;
            case 1 : return 0;
            case 2 : return 0.0;
            case 3 : 
		{	
		return t+x*y;
		}
            case 4 :
                {	
		return x*y;
		}
            default : return 0.0;
        }
    }
    double utb (double t, const Point& z, int i) const 
    {
        double x = z[0];
        double y = z[1];
        switch (i) 
        {
            case 0 : return -(x+2.0*y);
            case 1 : return -(2.0*x+y);
            case 2 : return -(x+y)/2.0;
            case 3 : return -2.0*x + 1;
            case 4 : return 0.0;
            default : return 0.0;

        }
        
    }
};

class ElasticBenchmark_RHS_3 : public ProblemElasticity{
public:
    ElasticBenchmark_RHS_3(const DGElasticityDiscretization& _disc) :  ProblemElasticity(_disc,2)
    {
        mu = 0.5;
        lambda = 1.0;
        rho = 1.0;
        c_P = sqrt((2.0*mu + lambda) / rho);
        c_S = sqrt(mu / rho);

    }
    const char* Name () const { return "v(x,y) = x"; }
    bool RHS(double) const { return true; }
    double ut (double t, const Point& z, int i) const {
        double x = z[0];
        double y = z[1];
        switch (i) {
            case 0 : return 0;
            case 1 : return 0;
            case 2 : return 0.0;
            case 3 : 
		{	
		return x;
		}
            case 4 :
                {	
		return 0.0;
		}
            default : return 0.0;
        }
    }
    double utb (double t, const Point& z, int i) const {
        double x = z[0];
        double y = z[1];
        switch (i) 
        {
            case 0 : return -2.0;
            case 1 : return -1.0;
            case 2 : return 0.0;
            case 3 : return 0.0;
            case 4 : return 0.0;
            default : return 0.0;

        }
        
    }
};

class DGElasticityAssembleMatrix : public DGAssemble {
protected:
    const DGElasticityDiscretization& disc;
    const ProblemElasticity& problem;
    double flux_alpha;
    Point GF_a;
    Point GF_b;
public:
    DGElasticityAssembleMatrix (const DGElasticityDiscretization& D, 
                                const ProblemElasticity& P,
                                Plot& _plot) :
    DGAssemble(_plot), disc(D), problem(P), flux_alpha(1) {
        Config::Get("flux_alpha", flux_alpha);
        //Config::Get("gamma", gamma);
        Config::Get("roi_min", GF_a);
        Config::Get("roi_max", GF_b);
    }
    const char* Name() const { return "DGAssembleMatrix"; }
    void MassMatrix (Matrix& M) const {
        M = 0;
        for (cell c=M.cells(); c!=M.cells_end(); ++c) {
            DGElasticityElement elem(disc,M,c);
            RowEntries_dG M_c(M,c,c);
            for (int q=0; q<elem.nQ(); ++q) {
                double w  = elem.QWeight(q);
                double rho = problem.Rho(elem.QPoint(q));
                double mu = problem.Mu(elem.QPoint(q));
                double lambda = problem.Lambda(elem.QPoint(q));
                double s = 1 / (2.0 * mu);
                double ss = lambda / (2.0 * mu * (2.0 * lambda + 2.0 * mu));
                for (int i=0; i<elem.S_dimension(); ++i) {
                    Tensor S_i = elem.Stress(q,i);
                    for (int j=0; j<elem.S_dimension(); ++j) {
                        Tensor S_j = elem.Stress(q,j);
                        //M_c(i,j) += w * Frobenius(S_i,S_j);
                        M_c(i,j) += w * (s * Frobenius(S_i, S_j) 
					 - ss * trace(S_i) * trace(S_j));
                    }
                }
                for (int i=0; i<elem.V_dimension(); ++i) {
                    VectorField V_i = elem.Velocity(q,i);
                    for (int j=0; j<elem.V_dimension(); ++j) {
                        VectorField V_j = elem.Velocity(q,j);
                        M_c(elem.S_dimension()+i,elem.S_dimension()+j) 
			    += w * rho * V_i * V_j;
                    }
                }
            }
        }
	return;
	SparseMatrix SpM(M);
	RMatrix SmM(SpM);
	mout << OUT(SmM);
    }

    Scalar VFlux_c (const VectorField& V, const VectorField& TN) const {
        return TN * V;
    }

    Scalar VFlux_cf (const VectorField& V, const VectorField& TN, short bnd) const {
        if (bnd == 1) return - (TN * V);             // Dirichlet bnd condition
        else          return TN * V;              
    }

    Scalar SFlux_c (const Tensor& S, const VectorField& TN,
		    const VectorField& N) const {
        VectorField NS = S * N;
        return NS * TN;
    }

    Scalar SFlux_cf (const Tensor& S, const VectorField& TN,
		     const VectorField& N, short bnd) const {
        VectorField NS = S * N;
	Scalar TNNS = NS * TN;
        if ((bnd == 0) || (bnd == 2)) return -TNNS;    // Neumann bnd condition
        else                          return TNNS;
    }

    void FluxMatrix (Matrix& A) const {
        A = 0;
	const Mesh& M = A.GetMesh();
        for (cell c=A.cells(); c!=A.cells_end(); ++c) {
            DGElasticityElement elem(disc,A,c);
            RowEntries_dG A_c(A,c,c);
            for (int q=0; q<elem.nQ(); ++q) {
                double w  = elem.QWeight(q);
                for (int i=0; i<elem.S_dimension(); ++i) {
                    VectorField divS_i = elem.divStress(q,i);
                    for (int j=0; j<elem.V_dimension(); ++j) {
                        VectorField V_j = elem.Velocity(q,j);
                        A_c(elem.S_dimension()+j,i) += w * V_j * divS_i;
                    }
                }
                for (int i=0; i<elem.V_dimension(); ++i) {
                    Tensor E_i = elem.Strain(q,i);
                    for (int j=0; j<elem.S_dimension(); ++j) {
                        Tensor S_j = elem.Stress(q,j);
                        A_c(j,elem.S_dimension()+i) += w * Frobenius(S_j,E_i);
                    }
                }
            }
	    for (int f=0; f<c.Faces(); ++f) {
                cell cf = c;
                int f1 = f;
                int idf = -1;
                if (check_if_bnd(A,c,f)){
                    idf = bnd_id(A,c,f);
                } else {
		    cf = find_nb_cell(A,c,f);
		    f1 = find_nb_face_id(A,c.Face(f),cf);
		}
                DGElasticityFaceElement felem(disc,A,c,f);
                DGElasticityFaceElement felem_1(disc,A,cf,f1);
                RowEntries_dG A_cf(A,c,cf);
                double c_s = problem.C_s(c());
                double c_p = problem.C_p(c());
                double mu = problem.Mu(c());
                double lambda = problem.Lambda(c());
                double c_s_f = problem.C_s(cf());
                double c_p_f = problem.C_p(cf());
                double mu_f = problem.Mu(cf());
                double lambda_f = problem.Lambda(cf());
                double alpha1 = ((2.0 * mu_f + lambda_f) * c_p)
		    / (c_p * (2.0 * mu_f + lambda_f) + c_p_f * (2.0 * mu + lambda));
                double alpha2 = ((2.0 * mu_f + lambda_f) * (2.0 * mu + lambda))
		    / (c_p * (2.0 * mu_f + lambda_f) + c_p_f * (2.0 * mu + lambda));
                double alpha3 = (mu_f * c_s)
		    / (c_s * mu_f + c_s_f * mu);
                double alpha4 = (mu_f * mu)
		    / (c_s * mu_f + c_s_f * mu);
                double alpha5 = (c_p_f * c_p)
		    / (c_p * (2.0 * mu_f + lambda_f) + c_p_f * (2.0 * mu + lambda));
                double alpha6 = (c_p_f * (2.0 * mu + lambda))
		    / (c_p * (2.0 * mu_f + lambda_f) + c_p_f * (2.0 * mu + lambda));
                double alpha7 = (c_s_f * c_s)
		    / (c_s * mu_f + c_s_f * mu);
                double alpha8 = (c_s_f * mu)
		    / (c_s * mu_f + c_s_f * mu);
                for (int q=0; q<felem.nQ(); ++q) {
                    double w  = felem.QWeight(q);
                    VectorField Nq = felem.QNormal(q);
                    VectorField Tq = felem.QTangent(q);
		    Point z = felem.QPoint(q);
		    for (int i=0; i<felem.S_dimension(); ++i) {
                        Tensor S_i = felem.Stress(q,i);
                        for (int j=0; j<felem.S_dimension(); ++j) {
                            Tensor S_j = felem.Stress(q,j);
                            A_c(i,j) -= w * (alpha5 * SFlux_c(S_i, Nq, Nq) * SFlux_c(S_j, Nq, Nq)
                                            + alpha7 * SFlux_c(S_i, Tq, Nq) * SFlux_c(S_j, Tq, Nq));
			}
                        for (int j=0; j<felem_1.S_dimension(); ++j) {
                            Tensor S_j = felem_1.Stress(q,j);
                            A_cf(i,j) += w * (alpha5 * SFlux_c(S_i, Nq, Nq) * SFlux_cf(S_j, Nq, Nq, idf)
                                              + alpha7 * SFlux_c(S_i, Tq, Nq) * SFlux_cf(S_j, Tq, Nq, idf));
                        }
                        for (int j=0; j<felem.V_dimension(); ++j) {
                            VectorField V_j = felem.Velocity(q,j);
                            A_c(i,felem.S_dimension()+j) -= w * (alpha1 * SFlux_c(S_i, Nq, Nq) * VFlux_c(V_j, Nq)
                                                                 + alpha3 * SFlux_c(S_i, Tq, Nq) * VFlux_c(V_j, Tq));
			}
                        for (int j=0; j<felem_1.V_dimension(); ++j) {
                            VectorField V_j = felem_1.Velocity(q,j);
                            A_cf(i,felem_1.S_dimension()+j) += w * (alpha1 * SFlux_c(S_i, Nq, Nq) * VFlux_cf(V_j, Nq, idf)
								    + alpha3 * SFlux_c(S_i, Tq, Nq) * VFlux_cf(V_j, Tq, idf));
                        }
                    }
                    for (int i=0; i<felem.V_dimension(); ++i) {
                        VectorField V_i = felem.Velocity(q,i);
                        for (int j=0; j<felem.S_dimension(); ++j) {
                            Tensor S_j = felem.Stress(q,j);
                            A_c(felem.S_dimension()+i,j) -= w * (alpha6 * VFlux_c(V_i, Nq) * SFlux_c(S_j, Nq, Nq)
                                                                 + alpha8 * VFlux_c(V_i, Tq) * SFlux_c(S_j, Tq, Nq));
			}
                        for (int j=0; j<felem_1.S_dimension(); ++j) {
                            Tensor S_j = felem_1.Stress(q,j);
                            A_cf(felem.S_dimension()+i,j) += w * (alpha6 * VFlux_c(V_i, Nq) * SFlux_cf(S_j, Nq, Nq, idf)
                                                                  + alpha8 * VFlux_c(V_i, Tq) * SFlux_cf(S_j, Tq, Nq, idf));
                        }
                        for (int j=0; j<felem.V_dimension(); ++j) {
                            VectorField V_j = felem.Velocity(q,j);
                            A_c(felem.S_dimension()+i,felem.S_dimension()+j)
                                    -= w * (alpha2 * VFlux_c(V_i, Nq) * VFlux_c(V_j, Nq)
                                            + alpha4 * VFlux_c(V_i, Tq) * VFlux_c(V_j, Tq));
			}
                        for (int j=0; j<felem_1.V_dimension(); ++j) {
                            VectorField V_j = felem_1.Velocity(q,j);
                            A_cf(felem.S_dimension()+i,felem_1.S_dimension()+j)
                                    += w * (alpha2 * VFlux_c(V_i, Nq) * VFlux_cf(V_j, Nq, idf)
                                            + alpha4 * VFlux_c(V_i, Tq) * VFlux_cf(V_j, Tq, idf));
                        }
                    }
                }
            }
        }
	return;
	SparseMatrix SpA(A);
        RMatrix SmA(SpA);
        mout << OUT(SmA);
    }
    bool RHS (double t) const { return problem.RHS(t); }
    void RHS (double t, Vector& b) const {
        b = 0.0;
        if (problem.Exact()){
	    for (cell c = b.cells(); c != b.cells_end(); ++c) {
		for (int f=0; f<c.Faces(); ++f) {
                    if(!check_if_bnd(b,c,f)) continue;
		    int idf = bnd_id(b,c,f);
                    DGElasticityFaceElement felem(disc,b,c,f);
                    double c_s = problem.C_s(c());
                    double c_p = problem.C_p(c());
                    double mu = problem.Mu(c());
                    double lambda = problem.Lambda(c());
                    double c_s_f = c_s;
                    double c_p_f = c_p;
                    double mu_f = mu;
                    double lambda_f = lambda;
                    double alpha1 = ((2.0 * mu_f + lambda_f) * c_p)
                        / (c_p * (2.0 * mu_f + lambda_f) + c_p_f * (2.0 * mu + lambda));
                    double alpha2 = ((2.0 * mu_f + lambda_f) * (2.0 * mu + lambda))
                        / (c_p * (2.0 * mu_f + lambda_f) + c_p_f * (2.0 * mu + lambda));
                    double alpha3 = (mu_f * c_s)
                        / (c_s * mu_f + c_s_f * mu);
                    double alpha4 = (mu_f * mu)
                        / (c_s * mu_f + c_s_f * mu);
                    double alpha5 = (c_p_f * c_p)
                        / (c_p * (2.0 * mu_f + lambda_f) + c_p_f * (2.0 * mu + lambda));
                    double alpha6 = (c_p_f * (2.0 * mu + lambda))
                        / (c_p * (2.0 * mu_f + lambda_f) + c_p_f * (2.0 * mu + lambda));
                    double alpha7 = (c_s_f * c_s)
                        / (c_s * mu_f + c_s_f * mu);
                    double alpha8 = (c_s_f * mu)
                        / (c_s * mu_f + c_s_f * mu);
                    for (int q=0; q<felem.nQ(); ++q) {
                        double w  = felem.QWeight(q);
                        VectorField Nq = felem.QNormal(q);
                        VectorField Tq = felem.QTangent(q);
                        Point z = felem.QPoint(q);
			if (idf == 1) {
			    VectorField V_bnd = problem.Dirichlet(t,z);
			    for (int i=0; i<felem.V_dimension(); ++i) {
				VectorField V_i = felem.Velocity(q,i);
				b(c(),felem.S_dimension()+i)  +=
				    2 * w * (alpha2 * VFlux_c(V_i, Nq) * VFlux_c(V_bnd, Nq)
					     + alpha4 * VFlux_c(V_i, Tq) * VFlux_c(V_bnd, Tq));
			    }
			    for (int j=0; j<felem.S_dimension(); ++j) {
				Tensor S_j = felem.Stress(q,j);
				b(c(),j) +=
				    2 * w * (alpha1 * SFlux_c(S_j, Nq, Nq) * VFlux_c(V_bnd, Nq)
					     + alpha3 * SFlux_c(S_j, Tq, Nq) * VFlux_c(V_bnd, Tq));
			    }
			} else if (idf == 2) {
			    Tensor S_bnd(problem.ut(t,z,0),problem.ut(t,z,2),problem.ut(t,z,2),problem.ut(t,z,1));
			    for (int i=0; i<felem.V_dimension(); ++i) {
				VectorField V_i = felem.Velocity(q,i);
				b(c(),felem.S_dimension()+i) +=
				    2 * w * (alpha6 * VFlux_c(V_i, Nq) * SFlux_c(S_bnd, Nq, Nq)
					 + alpha8 * VFlux_c(V_i, Tq) * SFlux_c(S_bnd, Tq, Nq));
			    }
			    for (int j=0; j<felem.S_dimension(); ++j) {
				Tensor S_i = felem.Stress(q,j);
				b(c(),j) +=
				    2 * w * (alpha5 * SFlux_c(S_i, Nq, Nq) * SFlux_c(S_bnd, Nq, Nq)
					 + alpha7 * SFlux_c(S_i, Tq, Nq) * SFlux_c(S_bnd, Tq, Nq));
			    }
			}
                    }
                }
            }
        }
        if (problem.RHS(t)) {
            for (cell c = b.cells(); c != b.cells_end(); ++c) {
                DGElasticityElement elem(disc,b,c);
                for (int q=0; q<elem.nQ(); ++q) {
                    double w  = elem.QWeight(q);
                    Point z = elem.QPoint(q);
                    VectorField V_rhs(problem.utb(t,z,3),problem.utb(t,z,4));
                    Tensor S_rhs(problem.utb(t,z,0),problem.utb(t,z,2),problem.utb(t,z,2),problem.utb(t,z,1));
                    for (int i=0; i<elem.S_dimension(); ++i) {
                        Tensor S_i = elem.Stress(q,i);
                        b(c(),i) += w * Frobenius(S_i,S_rhs);
                    }
                    for (int i=0; i<elem.V_dimension(); ++i) {
                        VectorField V_i = elem.Velocity(q,i);
                        b(c(),elem.S_dimension()+i) += w * V_i * V_rhs;
                    }
                }
            }
        }
	//mout << OUT(b) << endl;
        b.MakeAdditive();
    }
    
    cell find_cell (const Mesh& M, const face& f, const Point& z) const {
        cell c = M.find_cell(z);
        if (c == M.cells_end())
            c = M.find_overlap_cell(z);
        if (c == M.overlap_end()) {
            Exit("Error: no cell found");
            pout << "z " << z << endl;
            pout << "Cells: " << M.Cells::size() << endl << M.Cells::ref();
            pout << "OverlapCells: " << M.OverlapCells::size() << endl 
                 << M.OverlapCells::ref();
            Exit("Error: no cell found");
        }
        return c;
    }

    int bnd_id(const VectorMatrixBase &A, const cell &c, int f) const {
        bnd_face fff = A.GetMesh().find_bnd_face(c.Face(f));
        return fff.Part();
    }

    int find_nb_face_id(const VectorMatrixBase &A,
                        const Point &f_c, const cell &cf) const {
        int f1 = 0;
        for (; f1 < cf.Faces(); ++f1)
            if (f_c == cf.Face(f1)) return f1;
        Exit("Error: no face found");
    }

    cell find_nb_cell(const VectorMatrixBase &A, const cell &c, int f) const {
        face ff = A.GetMesh().find_face(c.Face(f));
//	mout << OUT(ff);
        if (c() != ff.Right())
            return find_cell(A.GetMesh(), ff, ff.Right());
        else
            return find_cell(A.GetMesh(), ff, ff.Left());
    }

    bool check_if_bnd(const VectorMatrixBase &A, const cell &c, int f) const {
        face ff = A.GetMesh().find_face(c.Face(f));
        if ((A.GetMesh().find_bnd_face(c.Face(f)) != A.GetMesh().bnd_faces_end())
            && ((ff.Right() == Infty) || (ff.Left() == Infty)))// To exclude periodic bnd
            return true;
        else
            return false;
    }

    void Jacobi(const Vector &u, Matrix &B) const {
        MassMatrix(B);
        Matrix A(u);
        FluxMatrix(A);
        B += (-gamma) * A;
    }

    void AssembleTransfer (TransferMatrix& TM) const {
        TM = 0;
        const VectorMatrixBase &cg = TM.CoarseMatrixGraph();
        const VectorMatrixBase &fg = TM.FineMatrixGraph();
        const Mesh &mesh = fg.GetMesh();
        for (cell c = cg.cells(); c != cg.cells_end(); ++c)
            for (int k = 0; k < c.Children(); ++k)
                TM(c(), c.Child(k))[0] = 1;
    }
};

class DGElasticityAssemble : public DGElasticityAssembleMatrix {
    double flux_alpha;
    double X0;
    double X1;
    
    std::fstream f;
    Point Pobs;
public:
    DGElasticityAssemble (const DGElasticityDiscretization& _disc,
                          const ProblemElasticity& _prob,
                          Plot& _plot) :
        DGElasticityAssembleMatrix(_disc,_prob,_plot), X0(8), X1(10) {
        Config::Get("X0", X0);
        Config::Get("X1", X1);

        f.open("Pkt.dat", ios::out);
        Config::Get("Pobs", Pobs);
    }
    const char* Name() const { return "DGElasticityAssemble"; }
    void PrintInfo (const Vector& u) {
        DGAssemble::PrintInfo(u);
        vout(1) << " Mesh            " << problem.MeshName() << endl
                << " Problem         " << problem.Name() << endl << endl;
    }
    std::string DiscName() const { return disc.DiscName(); }
    double Energy (const Vector& u, double x0, double x1) const {
        Scalar e = 0;
        for (cell c=u.cells(); c!=u.cells_end(); ++c) {
            double x = c()[0];
            if (x<x0) continue;
            if (x>x1) continue;
            DGElasticityElement elem(disc,u,c);
            for (int q=0; q<elem.nQ(); ++q) {
                double w  = elem.QWeight(q);
                double rho = problem.Rho(elem.QPoint(q));
                double mu = problem.Mu(elem.QPoint(q));
                double lambda = problem.Lambda(elem.QPoint(q));
                double s = 1 / (2.0 * mu);
                double ss = lambda / (2.0 * mu * (2.0 * lambda + 2.0 * mu));
                VectorField V = elem.Velocity(q,u);
                Tensor S = elem.Stress(q,u);
                Scalar tS = trace(S);
                //e += w * (Frobenius(S,S) + rho * V * V);
                e += w * (s * Frobenius(S,S) - ss * tS * tS + rho * V * V);
            }
        }
        return PPM->Sum(abs(e));
    }
    double Energy (const Vector& u) const {
        Scalar e = 0;
// 	mout << OUT(u) << endl;
        for (cell c=u.cells(); c!=u.cells_end(); ++c) {
            DGElasticityElement elem(disc,u,c);
//             mout << OUT(c()) << endl;

            for (int q=0; q<elem.nQ(); ++q) {
                double w  = elem.QWeight(q);
                double rho = problem.Rho(elem.QPoint(q));
                double mu = problem.Mu(elem.QPoint(q));
                double lambda = problem.Lambda(elem.QPoint(q));
                double s = 1 / (2.0 * mu);
                double ss = lambda / (2.0 * mu * (2.0 * lambda + 2.0 * mu));
//		mout << OUT(w) << OUT(rho) << OUT(mu) << OUT(lambda) << endl;

		VectorField V = elem.Velocity(q,u);
                Tensor S = elem.Stress(q,u);
                Scalar tS = trace(S);
// 		mout << OUT(V) << OUT(S) << OUT(tS) << endl;
                //e += w * (Frobenius(S,S) + rho * V * V);
                e += w * (s * Frobenius(S,S) - ss * tS * tS + rho * V * V);
            }
    //    Exit("2121")
	}
        return PPM->Sum(abs(e));
    }

    double Norm (const Vector& u) const { return sqrt(Energy(u)); }

    double dist (const Vector& u, const Vector& v) const { 
        Vector w(u);
        w -= v;
        return sqrt(Energy(w));
    }

    void SetInitialValue (Vector& u) const {
        problem.get_u_int(0,u);
    }

    void PrintStep (double t, const Vector& u, int step) {
	// mout << OUT(u);
        Vector u_tmp(u);
        //mout << "u_tmp: " << endl << u_tmp << endl << endl;
        mout << "Step" << step << " t=" << t
             << ": E(u) = " << Energy(u);
//           << " Local Energy(u) = " << Energy(u,X0,X1);
        if (problem.Exact()) {
	    problem.get_u_int(t,u_tmp);
	    u_tmp -= u;
            mout << " Err " << sqrt(Energy(u_tmp));
	}
        u_tmp = 0;
        //string discName = disc.Name_Disc_dG();

//	mout << OUT(u_tmp);

        for (cell c=u_tmp.cells(); c!=u_tmp.cells_end(); ++c) {
            row r = u.find_row(c());
            DGElasticityElement elem(disc,u_tmp,c);
            VectorField V = zero;
            Tensor S = Zero;
            double a = 0;
            for (int q=0; q<elem.nQ(); ++q) {
                double w  = elem.QWeight(q);
                V += w * elem.Velocity(q,u);
                S += Scalar(w) * elem.Stress(q,u);
                a += w;
            }
            V *= (1/a);
            S *= (1/a);
            u_tmp(r)[0] = V[0];
            u_tmp(r)[1] = V[1];
            //u_tmp(r)[2] = trace(S);
            //u_tmp(r)[3] = norm(dev(S));
        }
        mout << " |V| = " << norm(u_tmp) << endl; // " " << Date();
    }

    void InitialStep (double t, Vector& u) {
        SetInitialValue(u);
        //mout << "u_0" << endl << u << endl;
        PrintStep(t,u,0);
        VtkPlotting(t,u,0);
    }

    void FinishStep (Vector& u, double t, double dt,
                     int step, int plot_step = 1) {
        if (step%plot_step == 0) {
            PrintStep(t,u,step);
            VtkPlotting(t,u,step);
        }
        if (false) {
            for (cell c=u.cells(); c!=u.cells_end(); ++c) {
                row r = u.find_row(c());
                if(c.dim()==2){
                    double lambda2;
                    Point Pa = c.Corner(1)-c.Corner(0);
                    Point Pb = c.Corner(2)-c.Corner(0);
                    Point Pc = Pobs - c.Corner(0);
                    if (Pa[0]==0 ) {
                        lambda2 = Pc[0]/Pb[0];
                    } else {
                        lambda2 = (Pc[0]/Pa[0]-Pc[1]/Pa[1])/(Pb[0]/Pa[0]-Pb[1]/Pa[1]);
                    }
                    if ( -1.0e-9<lambda2 && lambda2<1+1.0e-9) {
                        double lambda1 = (Pc[1]-lambda2*Pb[1])/Pa[1];
                        //(Pc[0]/Pb[0]-Pc[1]/Pb[1])/(Pa[0]/Pb[0]-Pa[1]/Pb[1]);
                        //mout<<OUT(Pa)<<OUT(Pb)<<OUT(Pc)<<lambda1<<endl<<lambda2<<endl;
                        if( -1.0e-9<lambda1 && lambda1<1+1.0e-9) {
                            if ( lambda1+lambda2 < 1+1.0e-9 ) {
                                int shift = 0;
                                int NN = 5;
                                f<< setw(10) << left 
                                << t << setw(15) 
                                << (1-lambda1-lambda2)*u(r)[0]+ lambda1*u(r)[1]+ lambda2*u(r)[2] << setw(15)
                                << (1-lambda1-lambda2)*u(r)[3]+ lambda1*u(r)[4]+ lambda2*u(r)[5] << setw(15)
                                << (1-lambda1-lambda2)*u(r)[6]+ lambda1*u(r)[7]+ lambda2*u(r)[8] << setw(15)
                                << (1-lambda1-lambda2)*u(r)[9]+ lambda1*u(r)[10]+ lambda2*u(r)[11] << setw(15)
                                << (1-lambda1-lambda2)*u(r)[9]+ lambda1*u(r)[10]+ lambda2*u(r)[11] << setw(15)
                                << (1-lambda1-lambda2)*u(r)[12]+lambda1*u(r)[13]+lambda2*u(r)[14] << endl;
                            }
                        }
                    }
                }
            }
        }
        
        
	return;
        Vector u_int(u);
        problem.get_u_int(t,u_int);
        //mout << "u" << endl << u << endl;
        //mout << "u_int" << endl << u_int;
        u_int -= u;
        mout << "Energy(u-u_int) " << Energy(u_int) << endl;
    }

    //double Pdist (const Point& x, const Point& y) const { return norm(x-y); }


    void VtkPlotting (double t, const Vector& u, int step) {
        const Mesh& Mp = plot.GetMesh();
        const Mesh& Mu = u.GetMesh();
        if (Mp.Cells::size() != Mu.Cells::size()) return;
	vout(2) << "  => VtkPlotting result of step " << step << ".\n";
	char filename[128];
	string discName = disc.Name_Disc_dG();
        Vector u_tmp(u);
        Vector u_tmp2(u);
        Vector u_tmp3(u);
        for (cell c=u_tmp.cells(); c!=u_tmp.cells_end(); ++c) {
            DGElasticityElement elem(disc,u_tmp,c);
            VectorField V = zero;
            VectorField S_div = zero;
            Tensor S = Zero;
            double a = 0;
            for (int q=0; q<elem.nQ(); ++q) {
                double w  = elem.QWeight(q);
                S_div += w * elem.divStress(q,u);
                V += w * elem.Velocity(q,u);
                S += Scalar(w) * elem.Stress(q,u);
                a += w;
            }
            S_div *= (1.0/a);
            V *= (1/a);
            S *= (1/a);
            row r = u.find_row(c());
            u_tmp(r)[0] = V[0];
            u_tmp(r)[1] = V[1];
            u_tmp(r)[2] = trace(S);
            u_tmp(r)[3] = norm(dev(S));
            u_tmp(r)[4] = norm(V);
            u_tmp2(r)[0] = PPM->proc();
            u_tmp2(r)[1] = S[0][0];
            u_tmp2(r)[2] = S[1][1];
            u_tmp2(r)[3] = S_div[0];
            u_tmp2(r)[4] = S_div[1];
	    if (t==0) {
		u_tmp3(r)[0] = problem.Lambda(c());
		u_tmp3(r)[1] = problem.Mu(c());
	    }
        }
        plot.celldata(u_tmp,5);
        NumberName("V0",filename,step);
        plot.vtk_celldata(filename,0);
        NumberName("V1",filename,step);
        plot.vtk_celldata(filename,1);
        NumberName("trace_S",filename,step);
        plot.vtk_celldata(filename,2);
        NumberName("norm_devS",filename,step);
        plot.vtk_celldata(filename,3);
        NumberName("norm_V",filename,step);
        plot.vtk_celldata(filename,4);
	plot.celldata(u_tmp2,5);
	NumberName("S_0",filename,step);
	plot.vtk_celldata(filename,1);
	NumberName("S_1",filename,step);
	plot.vtk_celldata(filename,2);
	NumberName("Sdiv_0",filename,step);
        plot.vtk_celldata(filename,3);
        NumberName("Sdiv_1",filename,step);
        plot.vtk_celldata(filename,4);
	if (t > 0) return;
        plot.vtk_celldata("load",0);
	plot.celldata(u_tmp3,2);
	plot.vtk_celldata("lambda",0);
	plot.vtk_celldata("mu",1);
    }
    
    void PrintMatrixInfo(Matrix& A, int diagonal = 1) {
        for (cell c=A.cells(); c!=A.cells_end(); ++c) {
            DGElasticityElement elem(disc,A,c);
            RowEntries_dG A_c(A,c,c);
            for (int i=0;i<elem.dimension(); ++i) {
                for (int j=0; j<elem.dimension(); ++j) {
                    mout << A_c(i,j) << " ";
                }
                mout << endl;
            }
            mout << endl;
            if(diagonal) continue;

            for (int f=0; f<c.Faces(); ++f) {
                cell cf = c;
                if(check_if_bnd(A,c,f))
                    continue;
                else
                    cf = find_nb_cell(A,c,f);
                RowEntries_dG A_cf(A,c,cf);
                for (int i=0;i<elem.dimension(); ++i) {
                    for (int j=0; j<elem.dimension(); ++j) {
                        mout << A_cf(i,j) << " ";
                    }
                    mout << endl;
                }
                mout << endl;
            }
            mout << "------------------------" << endl;
        }
    }

    void Interpolate (const Vector& u, Vector& U) const {
        U = 0;
        int NN = 0;
        if (u.dim() == 3) NN = 9; else NN = 5;
        for (cell c=u.cells(); c!=u.cells_end(); ++c) {
            DGElasticityElement Elem(disc,u,c);
            row r = u.find_row(c());
            for (int i=0; i<c.Children(); ++i) {
                cell fc = U.find_cell(c.Child(i));
                row fr = U.find_row(fc());
		DGElasticityElement Elem_fc(disc,u,fc);
		Point z[MaxNodalPointsDG];
		Elem_fc.NodalPoints(z);
                for (int iN=0;iN<NN;++iN){
                    for (int i=0; i<Elem.shape_size(); ++i) {
                        Scalar s = Elem.SVValue(Elem.GlobalToLocal(z[i]),u,iN);
                        U(fr,iN*Elem.shape_size()+i) = s;
                    }
                }
            }
        }
        U.Accumulate();
    }
    double Goal_Functional (double t, const Vector& u) const {
        Scalar e = 0.0;
        Scalar eq = 0.0;
        for (cell c=u.cells(); c!=u.cells_end(); ++c){
            DGElasticityElement elem(disc,u,c);
            row r = u.find_row(c());
            for (int q=0; q<elem.nQ(); ++q) {
                Point QP = elem.QPoint(q);
                if ((GF_a[0]>QP[0]) || (QP[0]>GF_b[0])) continue;
                if ((GF_a[1]>QP[1]) || (QP[1]>GF_b[1])) continue;
                double w  = elem.QWeight(q);
                
                Tensor j_S;
                for (int i=0; i<elem.u_dimension()-c.dim(); ++i) {
                    if ( i<c.dim() ) {
                        j_S[i][i] =  problem.ut_dual(t,c,QP,i,GF_a,GF_b);
                    } else {
                        j_S[0][1] =  problem.ut_dual(t,c,QP,2,GF_a,GF_b);
                        j_S[1][0] =  problem.ut_dual(t,c,QP,2,GF_a,GF_b);
                        if ( c.dim() == 3 ) Exit("not implemented")
                    }
                }
                VectorField j_V;
                for (int i=0; i<c.dim(); ++i) {
                    j_V[i] = problem.ut_dual(t,c,QP,elem.u_dimension()
					     -c.dim()-1+i,GF_a,GF_b);
                }
                
                Tensor S = elem.Stress(q,u);
                VectorField V = elem.Velocity(q,u);
                
                e += w * (Frobenius(j_S,S) + j_V*V);
                eq += 0.5 * w * (Frobenius(S,S) + V*V);
            }
        }
        mout<<"GoalFunctional(quadratic): " 
	    <<scientific<<setprecision(12)<< abs(PPM->Sum(eq)) <<endl;
        return abs(PPM->Sum(e));
    }
};
ProblemElasticity* GetProblemElasticity(const DGElasticityDiscretization&_disc,
                                        const std::string& name ) {
    if (name == "PlaneWave2D")                              return new Plane_Wave_2D(_disc);
    else if (name == "ElasticRiemann")                      return new Elastic_Riemann(_disc);
    else if (name == "PlaneWave2DTest")                     return new Plane_Wave_2D_Test(_disc);
    else if (name == "SphericalWave2D")                     return new Spherical_Wave_2D(_disc);
    else if (name == "PlaneWave2D_1")                       return new Plane_Wave_2D_2(_disc);
    else if (name == "Hat2D")                               return new Hat_2D(_disc);
    else if (name == "Linear")                              return new Linear(_disc);
    else if (name == "ElasticBenchmarkConstantV")           return new ElasticBenchmark_Constant_V(_disc);
    else if (name == "ElasticBenchmarkConstantS")           return new ElasticBenchmark_Constant_S(_disc);
    else if (name == "ElasticBenchmarkConstantPdivfree")    return new ElasticBenchmark_Constant_P_div_free(_disc);
    else if (name == "ElasticBenchmarkQuadratic")           return new ElasticBenchmark_Quadratic(_disc);
    else if (name == "ElasticBenchmarkRHS")                 return new ElasticBenchmark_RHS(_disc);
    else if (name == "ElasticBenchmarkRHS_2")               return new ElasticBenchmark_RHS_2(_disc);
    else if (name == "ElasticBenchmarkRHS_3")               return new ElasticBenchmark_RHS_3(_disc);
    else if (name == "ElasticBenchmark_SinCosRHS")          return new ElasticBenchmark_SinCosRHS(_disc);
    else if (name == "Div_free_test" )                      return new Div_free_test(_disc);
    else Exit("Error: no wave found!");
}
#endif
