#ifndef _DGTIMEINTEGRATOR_H_
#define _DGTIMEINTEGRATOR_H_

class DG_TimeIntegrator { 
    Solver& S;
    Solver& S2;
    Solver& S3;
    Solver& IM;
    int plot_tStep;
    int rkorder;
    int Kmin;
    int Kmax;
    double Keps;
    Scalar gamma;
    Scalar gamma0;
    int SSum;
    int MSum;
    int KSum;
    int KPre;
    Time TPre;
    Time TComp;
    int verbose;
public:
    DG_TimeIntegrator (Solver& _S, Solver& _S2, Solver& _S3, Solver& _IM) :
        S(_S), S2(_S2), S3(_S3), IM(_IM), plot_tStep(1), rkorder(1),
        Kmin(2), Kmax(10), Keps(0.001),
        SSum(0), MSum(0), KSum(0), KPre(0), gamma(1), verbose(0) {
        Config::Get("plot_tStep", plot_tStep);
        Config::Get("rkorder", rkorder);
        Config::Get("Kmin", Kmin);
        Config::Get("Kmax", Kmax);
        Config::Get("Keps", Keps);
        Config::Get("TimeVerbose", verbose);
        double g;
        Config::Get("gamma", g);
        gamma0 = gamma = g;
    }
    void checkOrder(conststd::vector<double>& c, conststd::vector<double>& b, const RMatrix A_rk) {
        if (c.size()!= b.size() || A_rk.cols() !=A_rk.rows() || c.size() != A_rk.cols()){
            mout << "c.size: " << c.size() << " b.size(): " << b.size() << " A_rk.cols(): " << A_rk.cols() << "A_rk.rows(): " << A_rk.rows() << endl;
            Exit("Error in the Method!")
        }
        double tmp=0.0;
        int order =0;
        for (int i =0; i<b.size();++i){
            tmp += b[i];
        }
        if ((tmp-1.0)> Eps){
            Exit("rk not consistent!")
        }else{
            order =1;
            tmp=0.0;
        }
        for (int i=0; i<b.size();++i){
            tmp += b[i]*c[i];
        }
        if (abs(tmp-0.5) > Eps){
            mout << ("rk is of order" + std::to_string(order))<< endl;
            return;
        }else{
            order =2;
            tmp=0.0;
        }
        double tmp2 = 0.0;
        for (int i=0; i<b.size();++i){
            tmp += b[i]*c[i]*c[i];
            for (int j=0; j<b.size();++j){
                tmp2 +=b[i]*A_rk[i][j]*c[j];
            }
        }
        if (((abs(tmp-1.0/3.0))> Eps) || (abs(tmp2-1.0/6.0)) > Eps){
            mout << ("rk is of order" + std::to_string(order)) << endl;
            return;
        }else{
            order =3;
            tmp=0.0;
            tmp2=0.0;
        }
        double tmp3 =0.0;
        double tmp4 = 0.0;
        for (int i=0; i<b.size();++i){
            tmp += b[i]*c[i]*c[i]*c[i];
            for (int j=0; j<b.size();++j){
                tmp2 += b[i]*c[i]*A_rk[i][j]*c[j];
                tmp3 += b[i]*c[j]*A_rk[i][j]*c[j];
                for(int k=0; k<b.size();++k){
                    tmp4 +=b[i]*A_rk[i][j]*A_rk[j][k]*c[k];
                }
            }
        }
        if (((abs(tmp-1.0/4.0))> Eps) || (abs(tmp2-1.0/8.0)) > Eps|| (abs(tmp3-1.0/12.0)) > Eps|| (abs(tmp4-1.0/24.0)) > Eps){
            mout << ("rk is of order" + std::to_string(order)) << endl;
            return;
        }else{
            mout <<"rk is of order 4!" << endl;
            return;
        }



    }
    void eRK (const DGAssemble& assemble, conststd::vector<double>& c, conststd::vector<double>& b, const RMatrix A_rk,const double t,
              double dt, const Vector &u,const Matrix& A, Vector& du,std::vector<Vector>& RHS ) {
        if (t-dt<Eps) checkOrder(c,b,A_rk);
        for (int i =0; i< c.size(); ++i){
            assemble.RHS(t-(1-c[i])*dt,RHS[i]);
        }
        int stages = b.size();
        vector <Vector> k(stages,u);
        Vector Ak(u);
        Vector tmp(u);
        du=0.0;
        for (int stage=0; stage < stages; ++stage ){
            k[stage] = u;
            for (int j =0; j < stage; ++j){
                tmp = k[j];
                tmp *= A_rk[stage][j];
                tmp *= dt;
                k[stage]+= tmp;
            }
            Ak = A * k[stage];
            Ak += RHS[stage];
            k[stage] = IM*Ak;
            du += b[stage]*k[stage];
        }
        du *=dt;
    }
    void diRK(const DGAssemble& assemble ,conststd::vector<double>& c,conststd::vector<double>& b, const RMatrix A_rk,const double t,
              double dt,const Vector &u,const Matrix& M, const Matrix& A, Vector& du,std::vector<Vector>& RHS,
              Matrix& B,Matrix& B2,Matrix& B3){
        int stages = c.size();
        if (t-dt<Eps) checkOrder(c,b,A_rk);
        for (int i =0; i< c.size(); ++i){
            assemble.RHS(t-(1-c[i])*dt,RHS[i]);
        }
        vector <Vector> k(stages,u);
        Vector Ak(u);
        if (abs(t-dt)<Eps){
            for (int stage=0; stage < stages; ++stage){
                double factor = A_rk[stage][stage]*dt;
                if(stage == 0) {
                    B = M;
                    B += (-factor) * A;
                    S(B,1);
                }else if(stage == 1){
                    B2 = M;
                    B2 += (-factor) * A;
                    S2(B2,1);
                }else if(stage == 2){
                    B3 = M;
                    B3 += (-factor) *A;
                    S3(B3,1);
                }
                else Exit ("diagonal implicit runge kutta only for maximal 3 stages!")
            }
        }
        Vector tmp(u);
        du=0.0;
        for (int stage=0; stage < stages; ++stage){
            k[stage] = u;
            for (int j=0; j <stage;++j){
                tmp = k[j];
                tmp *= A_rk[stage][j];
                tmp *= dt;
                k[stage]+= tmp;
            }
            Ak = A*k[stage];
            Ak += RHS[stage];
            if (stage ==0) k[stage] = S*Ak;
            if (stage ==1) k[stage] = S2*Ak;
            if (stage ==2) k[stage] = S3*Ak;
            du += b[stage]*k[stage];
        }
        du*=dt;
    }
    void Euler (double dt, const Vector& u, const Matrix& A,
                Vector& du, const Vector& RHS) {
        Vector Au(u);
        Au = A * u;
        Au += RHS;
        du = IM * Au;
        du *= dt;
        MSum += 1;
    }
    void Euler_2 (double dt,const double t,const DGAssemble& assemble, const Vector& u, const Matrix& A, Vector& du) {
        Vector du2(du);
       std::vector<double> c(1,0.0);
       std::vector<double> b(1,1.0);
        RMatrix A_rk(1,1);
       std::vector<Vector> RHS_V(1,u);
        A_rk[0][0] = 0;
        eRK(assemble,c,b,A_rk,t,dt,u,A,du,RHS_V);
        MSum += 1;
    }
    void Heun_2 (double dt,const double t,const DGAssemble& assemble, const Vector& u, const Matrix& A, Vector& du) {
        Vector du2(du);
       std::vector<double> c(2,0.0);
        c[0] = 0;
        c[1] = 1;
       std::vector<double> b(2,1.0);
        b[0] = 0.5;
        b[1] = 0.5;
        RMatrix A_rk(2,2);
       std::vector<Vector> RHS_V(2,u);
        A_rk[1][0] = 1;
        eRK(assemble,c,b,A_rk,t,dt,u,A,du,RHS_V);
        MSum += 2;
    }
    void Heun (double dt, const Vector& u, const Matrix& A,
               Vector& du, const Vector& RHS, const Vector& RHS1) {
        Vector u1(u);
        Vector u2(u);
        Vector Au(u);
        Au = A * u;
        Au += RHS;
        u1 = IM * Au;
        u2 = dt * u1;
        u2 += u;
        Au = A * u2;
        Au += RHS1;
        u2 = IM * Au;
        u2 += u1;
        du = (0.5*dt) * u2;
        MSum += 2;
    }
    void Runge (double dt, const Vector& u, const Matrix& A, Vector& du) {
        const double w3 = 1.0/6.0;
        Vector u1(u);
        Vector u2(u);
        Vector u3(u);
        Vector Au(u);
        Au = A * u;
        u1 = IM * Au;

        u2 = 0.5*dt*u1;
        u2 += u;
        Au = A * u2;
        u2 = IM * Au;

        u3 = 2.0*dt*u2;
        u3 -= dt*u1;
        u3 += u;
        Au = A * u3;
        u3 = IM * Au;

        u3 += 4.0*u2;
        u3 += u1;

        du = (w3*dt) * u3;
        MSum += 3;
    }
    void Kutta (double dt, const Vector& u, const Matrix& A, Vector& du) {
        const double w4 = 1.0/6.0;
        Vector u1(u);
        Vector u2(u);
        Vector u3(u);
        Vector u4(u);
        Vector Au(u);

        Au = A * u;
        u1 = IM * Au;

        u2 = 0.5*dt*u1;
        u2 += u;
        Au = A * u2;
        u2 = IM * Au;

        u3 = 0.5*dt*u2;
        u3 += u;
        Au = A * u3;
        u3 = IM * Au;

        u4 = dt * u3;
        u4 += u;
        Au = A * u4;
        u4 = IM * Au;

        u4 += 2.0 * u3;
        u4 += 2.0 * u2;
        u4 += u1;
        du = (w4*dt) * u4;

        MSum += 4;
    }
    void Kutta_2 (double dt,const double t,const DGAssemble& assemble, const Vector& u, const Matrix& A, Vector& du) {
       std::vector<double> c(4,0.0);
        c[0] = 0;
        c[1] = 0.5;
        c[2] = 0.5;
        c[3] =1;
       std::vector<double> b(4,0);
        b[0]=1.0/6.0;
        b[1]=1.0/3.0;
        b[2]=1.0/3.0;
        b[3]=1.0/6.0;
        RMatrix A_rk(4,4);
        A_rk[1][0] = 0.5;
        A_rk[2][1] = 0.5;
        A_rk[3][2] = 1;
       std::vector<Vector> RHS_V(4,u);
        eRK(assemble,c,b,A_rk,t,dt,u,A,du,RHS_V);
        MSum += 4;
    }
    void Kutta (double dt, const Vector& u, const Matrix& A, Vector& du,
                const Vector& RHS1, const Vector& RHS2, const Vector& RHS3) {
        const double w4 = 1.0/6.0;
        Vector u1(u);
        Vector u2(u);
        Vector u3(u);
        Vector u4(u);
        Vector Au(u);

        Au = A * u;
        Au += RHS1;
        u1 = IM * Au;

        u2 = 0.5*dt*u1;
        u2 += u;
        Au = A * u2;
        Au += RHS2;
        u2 = IM * Au;

        u3 = 0.5*dt*u2;
        u3 += u;
        Au = A * u3;
        Au += RHS2;
        u3 = IM * Au;

        u4 = dt * u3;
        u4 += u;
        Au = A * u4;
        Au += RHS3;
        u4 = IM * Au;

        u4 += 2.0 * u3;
        u4 += 2.0 * u2;
        u4 += u1;
        du = (w4*dt) * u4;

        MSum += 4;
    }
    void Pade2 (double dt, const Vector& u, const Matrix& A, Vector& du) {
        Vector Au(u);
        Au = A * u;
        du = S * Au;
        KPre += S.Iter();
        du *= dt;
        ++KSum;
    }
    void Pade2 (double dt, const Vector& u, const Matrix& A,
                Vector& du, const Vector& RHS) {
        Vector Au(u);
        Au = A * u;
        Au += RHS;
        du = S * Au;
        KPre += S.Iter();
        du *= dt;
        ++KSum;
    }
    void Pade2_2 (double dt,const double t,const DGAssemble& assemble, const Vector& u, const Matrix& A,const Matrix& M, Vector& du,
                  Matrix& B,Matrix& B2, Matrix& B3) {
       std::vector<double> c(1,0.0);
        c[0] = 0.5;
       std::vector<double> b(1,0);
        b[0]=1;
        RMatrix A_rk(1,1);
        A_rk[0][0] = 0.5;
       std::vector<Vector> RHS_V(1,u);
        diRK(assemble,c,b,A_rk,t,dt,u,M,A,du,RHS_V,B,B2,B3);
        ++KSum;
    }
    void ImpEuler (double dt,const double t,const DGAssemble& assemble, const Vector& u, const Matrix& A,const Matrix& M, Vector& du,
                  Matrix& B,Matrix& B2, Matrix& B3) {
       std::vector<double> c(1,0.0);
        c[0] = 1;
       std::vector<double> b(1,0);
        b[0]=1;
        RMatrix A_rk(1,1);
        A_rk[0][0] = 1;
       std::vector<Vector> RHS_V(1,u);
        diRK(assemble,c,b,A_rk,t,dt,u,M,A,du,RHS_V,B,B2,B3);
        ++KSum;
    }
    void diRKorder3 (double dt,const double t,const DGAssemble& assemble, const Vector& u, const Matrix& A,const Matrix& M, Vector& du,
                     Matrix& B,Matrix& B2, Matrix& B3){
       std::vector<double> c(2,0.0);
        c[0] = 0.5 + sqrt(3)/6.0;
        c[1] = 0.5 - sqrt(3)/6.0;
       std::vector<double> b(2,0);
        b[0]=0.5;
        b[1]=0.5;
        RMatrix A_rk(2,2);
        A_rk[0][0] = c[0];
        A_rk[1][0] = -sqrt(3.0)/3.0;
        A_rk[1][1] = c[0];
       std::vector<Vector> RHS_V(2,u);
    }
    void diRKtest (double dt,const double t,const DGAssemble& assemble, const Vector& u, const Matrix& A,const Matrix& M, Vector& du,
                   Matrix& B,Matrix& B2, Matrix& B3) {
       std::vector<double> c(2,0.0);
        c[0] = 0.25;
        c[1] = 0.75;
       std::vector<double> b(2,0);
        b[0]=0.5;
        b[1]=0.5;
        RMatrix A_rk(2,2);
        A_rk[0][0] = 0.25;
        A_rk[1][0] = 0.5;
        A_rk[1][1] = 0.25;
       std::vector<Vector> RHS_V(2,u);
        diRK(assemble,c,b,A_rk,t,dt,u,M,A,du,RHS_V,B,B2,B3);
        ++KSum;
    }
    void Pade2ex (double dt, const Vector& u, const Matrix& A,
                  const Matrix& B, const Matrix& B2, Vector& du,
                  const Vector& RHS, const Vector& RHS2, const Vector& RHS3) {
        Vector Au(u);
        Au = A * u;
        Vector u2(u);
        Vector Au2(Au);
        Vector du0(du);
        Au += RHS;
        du0 *= (1/dt);
        Au -= B * du0;
        du0 += S * Au;
        KPre += S.Iter();
        du0 *= dt;
        ++KSum;
        Au2 += RHS2;
        Vector du2(du0);
        du2 *= (1/dt);
        Au2 -= B2 * du2;
        du2 += S2 * Au2;
        du2 *= (0.5*dt);
        u2 += du2;
        Au2 = A * u2;
        Au2 += RHS3;
        du = du2;
        du *= (2/dt);
        Au2 -= B2 * du;
        du += S2 * Au2;
        du *= (0.5*dt);
        du += du2;

        du *= (4.0/3.0);
        du -= (1.0/3.0) * du0;
    }
    void Pade3 (double dt, const Vector& u,
                const Matrix& M, const Matrix& A, Vector& du) {
        Vector Au(u);
        Vector Mu(u);
        Au = A * u;
        du = S * Au;
        KPre += S.Iter();
        Mu = M * du;
        Au = A * du;
        Au *= (dt *iUnit /sqrt(60));
        Mu -= Au;
        du = S2 * Mu;
        KPre += S2.Iter();
        Mu = M * du;
        Au = A * du;
        Au *= (dt *iUnit /sqrt(60));
        Mu += Au;
        du = S3 * Mu;
        KPre += S3.Iter();
        du *= dt;
        KSum += 3;
    }
    
    void Pade3 (double dt, const Vector& u,
                const Matrix& M, const Matrix& A, Vector& du,
                const Vector& RHS, const Vector& RHS2, const Vector& RHS3) {
        Vector Au(u);
        Vector T(u);
        Vector MT(u);
        Vector AT(u);

        Vector T1(u);
        Vector T2(u);
        Vector T11(u);
        Vector T22(u);
        
        Au = A * u;

        T = Au;
        T += 5./18. * RHS;
        T += 8./18. * RHS2;
        T += 5./18. * RHS3;
        T1 = S * T;

        T = Au;
        T += 5./6. * RHS3;
        T2 = S * T;

        MT = M * T1;
        AT = A * T2;
        AT *= (-dt *iUnit /sqrt(60));

        MT += AT;
        T11 = S2 * MT;

        //  ------------------------

        T = Au;
        T += 5./6. * RHS;
        T1 = S * T;

        T = Au;
        T += 5./6. * RHS;
        T -= 4./6. * RHS2;
        T += 5./6. * RHS3;
        T2 = S * T;

        MT = M * T1;
        AT = A * T2;
        AT *= (-dt *iUnit /sqrt(60));

        MT += AT;
        T22 = S2 * MT;

        //  ------------------------

        MT = M * T11;
        AT = A * T22;
        AT *= (dt *iUnit /sqrt(60));

        MT += AT;

        du = S3 * MT;

        KPre += 4 * S.Iter();
        KPre += 2 * S2.Iter();
        KPre += S3.Iter();
        du *= dt;
        KSum += 7;
    }
    void ExponentialIntegrator (double dt, const Vector& u,
                                const Matrix& M, const Matrix& A, Vector& du){
       std::vector<Vector> v(Kmax+1,u);
       std::vector<Vector> w(Kmax+1,u);
        Vector Sw(u);
        RMatrix h(Kmax+1,Kmax);
        RVector s(Kmax);
        RVector s_old(Kmax);
        s_old = Scalar(0.0);
        h = Scalar(0.0);
        v[0] = u;
        w[0] = M * u;
        double beta = sqrt(real(v[0]*w[0]));
        v[0] *= 1/beta;
        w[0] *= 1/beta;
        s_old[0] = beta;
        S.SetReduction(Keps);
        int k = 0;
        for (; k<Kmax; ++k) {
            Sw = S * w[k];
            KPre += S.Iter();
            Sw *= gamma/dt;
            v[k+1] = Sw;
            for (int j=0; j<=k; ++j) {
                h[j][k] = Sw * w[j];
                v[k+1] -= h[j][k] * v[j];
            }
            w[k+1] = M * v[k+1];
            h[k+1][k] = sqrt(real(v[k+1]*w[k+1]));
            v[k+1] *= (1.0/h[k+1][k]);
            w[k+1] *= (1.0/h[k+1][k]);
            RMatrix S_k(k+1,k+1);
            for (int j=0; j<=k; ++j)
                for (int l=0; l<=k; ++l)
                    S_k[j][l] = h[j][l];
            S_k.invert();
            for (int j=0; j<=k; ++j)
                S_k[j][j] -= dt/gamma;
            S_k *= Scalar(-1.0);
            S_k.Exp();
            for (int j=0; j<=k; ++j)
                s[j] = beta * S_k[j][0];
            //	    mout << "S_k " << s << endl << S_k;
            //	    mout << "S_k " << s << endl;
            for (int j=0; j<=k; ++j)
                s_old[j] -= s[j];
            double norm_u = s.norm();
            double delta = s_old.norm() / norm_u;
            for (int j=0; j<=k; ++j)
                s_old[j] = s[j];
            if (k==0) continue;
            double eta = 1+norm_u;
            if (delta < 1) eta = min(eta,delta*norm_u/(1-delta));
            vout(2) << "eta(" << k << ") = " << eta << endl;
            S.SetReduction(0.1*Keps/(Keps+eta));
            if (eta < Keps) {
                KSum += k;
                break;
            }
        }
        du = 0.0;
        for (int j=0; j<=k; ++j)
            du += s[j] * v[j];
        //	s *= (1/beta);
        //	mout << "s " << s << endl;
        du -= u;
        if (k==Kmax) du[0] = 2e10;
    }
    void Arnoldi (double dt, const Vector& u,
                  const Matrix& M, const Matrix& A, Vector& du, int kphi) {
       std::vector<Vector> v(Kmax+1,u);
        Vector Mv(u);
        Vector Av(u);
        RMatrix h(Kmax+1,Kmax);
        RVector s(Kmax);
        RVector s_old(Kmax);
        s_old = Scalar(0.0);
        h = Scalar(0.0);
        v[0] = u;
        Mv = M * v[0];
        double beta = sqrt(real(v[0]*Mv));
        if (beta < Eps) {
            du = 0;
            return;
        }
        v[0] *= 1/beta;
        s_old[0] = beta;
        int k = 0;
        for (; k<Kmax; ++k) {
            Av = A * v[k];
            v[k+1] = IM * Av;
            for (int j=0; j<=k; ++j) {
                h[j][k] = Av * v[j];
                v[k+1] -= h[j][k] * v[j];
            }
            Mv = M * v[k+1];
            h[k+1][k] = sqrt(real(v[k+1]*Mv));
            v[k+1] *= (1.0/h[k+1][k]);
            RMatrix H_k(k+1,k+1);
            for (int j=0; j<=k; ++j)
                for (int l=0; l<=k; ++l)
                    H_k[j][l] = h[j][l] * dt;
            if (kphi == 0) H_k.Exp();
            if (kphi == 1) H_k.Phi1();
            if (kphi == 2) H_k.Phi2();
            for (int j=0; j<=k; ++j)
                s[j] = beta * H_k[j][0];
            for (int j=0; j<=k; ++j)
                s_old[j] -= s[j];
            double norm_u = s.norm();
            double delta = s_old.norm() / norm_u;
            for (int j=0; j<=k; ++j)
                s_old[j] = s[j];
            if (k==0) continue;
            double eta = 1+norm_u;
            if (delta < 1) eta = min(eta,delta*norm_u/(1-delta));
            vout(2) << "eta(" << k << ") = " << eta << endl;
            if (eta < Keps) break;
        }
        du = 0.0;
        for (int j=0; j<=k; ++j)
            du += s[j] * v[j];
        MSum += k+1;
        if (k==Kmax) du[0] = 2e10;
    }

    void Arnoldi2 (double dt, const Vector& u,
                   const Matrix& M, const Matrix& A, Vector& du,int phik) { // Arnoldi/SAI procedure
       std::vector<Vector> v(Kmax+1,u);
       std::vector<Vector> w(Kmax+1,u);
        Vector Sw(u);
        RMatrix h(Kmax+1,Kmax);
        RVector s(Kmax);
        RVector s_old(Kmax);
        s_old = Scalar(0.0);
        h = Scalar(0.0);
        v[0] = u;
        w[0] = M * u;
        double beta = sqrt(real(v[0]*w[0]));
        v[0] *= 1/beta;
        w[0] *= 1/beta;
        s_old[0] = beta;
        S.SetReduction(Keps);
        int k = 0;
        for (; k<Kmax; ++k) {
            Sw = S * w[k];
            KPre += S.Iter();
            Sw *= gamma/dt;
            v[k+1] = Sw;
            for (int j=0; j<=k; ++j) {
                h[j][k] = Sw * w[j];
                v[k+1] -= h[j][k] * v[j];
            }
            w[k+1] = M * v[k+1];
            h[k+1][k] = sqrt(real(v[k+1]*w[k+1]));
            v[k+1] *= (1.0/h[k+1][k]);
            w[k+1] *= (1.0/h[k+1][k]);
            RMatrix S_k(k+1,k+1);
            for (int j=0; j<=k; ++j)
                for (int l=0; l<=k; ++l)
                    S_k[j][l] = h[j][l];
            S_k.invert();
            for (int j=0; j<=k; ++j)
                S_k[j][j] -= dt/gamma;
            S_k *= Scalar(-1.0);
            if (phik==0) S_k.Exp();
            if (phik==1) S_k.Phi1();
            if (phik==2) S_k.Phi2();
            for (int j=0; j<=k; ++j)
                s[j] = beta * S_k[j][0];
            //	    mout << "S_k " << s << endl << S_k;
            //	    mout << "S_k " << s << endl;
            for (int j=0; j<=k; ++j)
                s_old[j] -= s[j];
            double norm_u = s.norm();
            double delta = s_old.norm() / norm_u;
            for (int j=0; j<=k; ++j)
                s_old[j] = s[j];
            if (k==0) continue;
            double eta = 1+norm_u;
            if (delta < 1) eta = min(eta,delta*norm_u/(1-delta));
            vout(2) << "eta(" << k << ") = " << eta << endl;
            S.SetReduction(Keps/(Keps+eta));
            if (eta < Keps) {
                KSum += k;
                break;
            }
        }
        du = 0.0;
        for (int j=0; j<=k; ++j)
            du += s[j] * v[j];
        //	s *= (1/beta);
        //	mout << "s " << s << endl;
        if (k==Kmax) du[0] = 2e10;
    }

    void operator () (DGAssemble& assemble, TimeSeries& TS, Vector& u) {
        Date Start;
        Matrix M(u);
        Matrix B(u);
        Matrix B2(u);
        Matrix B3(u);
        assemble.MassMatrix(M);
        IM(M);
        Matrix A(u);
        Vector RHS(u);
        Vector RHS2(u);
        Vector RHS3(u);
        RHS = 0.;
        RHS2 = 0.;
        RHS3 = 0.;
        double du_norm =0.0;
        assemble.FluxMatrix(A);
        double T = TS.LastTStep();
        double t = TS.FirstTStep();
        int step = 0;
        assemble.InitialStep(t,u);
//        mout << OUT(u) << endl;
        double dt_old = 0;
        Vector du(u);
        SSum = MSum = KSum = KPre = 0;
        Date CStart;
        TPre = CStart - Start;
        while (t<T) {
            double t_old = t;
            t = TS.NextTStep(t);
            double dt = t - t_old;
            switch (rkorder) {
            case 1: {
                if (dt_old == 0) mout << "explicit Euler " << endl;
                assemble.RHS(t-dt,RHS);
                Euler(dt,u,A,du,RHS);
                //Vector Au = A * u;
                //mout << OUT(u) << OUT(Au) << OUT(RHS) << OUT(du);
                break;
            }
            case 12: {
                if (dt_old == 0) mout << "explicit Euler " << endl;
                Euler_2(dt,t,assemble,u,A,du);
                break;
            }
            case 2: {
                if (dt_old == 0) mout << "explicit Heun " << endl;
                assemble.RHS(t-dt,RHS);
                assemble.RHS(t,RHS2);
                Heun(dt,u,A,du,RHS,RHS2);
                //Vector Au = A * u;
                //mout << OUT(u) << OUT(Au) << OUT(RHS) << OUT(du);
                break;
            }
            case 22: {
                if (dt_old == 0) mout << "explicit Heun " << endl;

                Heun_2(dt,t,assemble,u,A,du);
                //Vector Au = A * u;
                //mout << OUT(u) << OUT(Au) << OUT(RHS) << OUT(du);
                break;
            }
            case 3:
                if (dt_old == 0) {
                    if (assemble.RHS(-1) == true)
                        Exit("RHS not implemented in Exponential Integrator");
                }
                mout << "explicit Runge " << endl;
                Runge(dt,u,A,du);
                break;
            case 4:
                if (dt_old == 0) mout << "classical Runge-Kutta" << endl;
                assemble.RHS(t-dt,RHS);
                assemble.RHS(t-0.5*dt,RHS2);
                assemble.RHS(t,RHS3);
                Kutta(dt,u,A,du,RHS,RHS2,RHS3);
                break;
            case 42:
                if (dt_old == 0){
                    mout << "classical Runge-Kutta" << endl;
                }
                Kutta_2(dt,t,assemble,u,A,du);
                break;
            case 0:
                if (dt_old == 0) {
                    if (assemble.RHS(-1) == true)
                        Exit("RHS not implemented in Exponential Integrator");
                    B = M;
                    gamma = gamma0 * dt;
                    B += (-gamma) * A;
                    mout << "Exponential Integrator gamma = " << gamma <<endl;
                    S(B,1);
                }
                ExponentialIntegrator(dt,u,M,A,du);
                break;
            case 1000:
                if (dt_old == 0) {
                    if (assemble.RHS(-1) == true)
                        Exit("RHS not implemented in Arnoldi");
                    mout << "Arnoldi homogeneous" << endl;
                }
                Arnoldi(dt,u,M,A,du,0);
                du -= u;
                break;
            case 1001:
            {
                if(dt_old ==0){
                    mout << "Arnoldi with exponential euler " << endl; // cf. Hochbruck Skript aijs can be ignored for linear problems
                }
                assemble.RHS(t-dt,RHS);
                Vector tmp(u);
                tmp = A*u;
                tmp +=RHS;
                RHS = IM*tmp;
                Arnoldi(dt,RHS,M,A,du,1);
                du *= dt;
                break;
            }
            case 1002:{
                if(dt_old < 1e-13){
                    mout << "Arnoldi with exponential midpoint" << endl; // cf. Hochbruck Skript aijs can be ignored for linear problems
                }
                assemble.RHS(t-0.5*dt,RHS);
                Vector tmp(u);
                tmp = A*u;
                tmp +=RHS;
                RHS = IM*tmp;
                Arnoldi(dt,RHS,M,A,du,1);
                du *= dt;
                break;
            }
            case 1003:{
                if (dt_old < 1e-13) {
                    mout << "Arnoldi with exponential trapezoid rule " << endl; // cf. Hochbruck Skript aijs can be ignored for linear problems
                }
                Vector tmp(RHS);
                Vector tmp2(RHS);
//                Vector Aupf0(RHS);
                assemble.RHS(t-dt,RHS);
                assemble.RHS(t,RHS2);
                tmp = A*u;
                tmp += RHS;
                du = IM * tmp;
                tmp = A*du;
                tmp *= dt;
                tmp += RHS2;
                tmp -= RHS;
                tmp2 = IM * tmp;
                Arnoldi(dt,tmp2,M,A,tmp,2);
                du += tmp;
                du *= dt;
                break;
            }
            case 2000:
                if (dt_old < 1e-13) {
                    if (assemble.RHS(-1) == true)
                        Exit("RHS not implemented in Arnoldi2");
                    B = M;
                    gamma = gamma0 * dt;
                    B += (-gamma) * A;
                    S(B,1);
                    mout << "Arnoldi shift and invert with  gamma = " << gamma <<endl; // cf. Hochbruck Skript aijs can be ignored for linear problems
                }
                Arnoldi2(dt,u,M,A,du,0);
                du -= u;
                break;
            case 2001:{
                if (dt_old < 1e-13) {

                    B = M;
                    gamma = gamma0 * dt;
                    B += (-gamma) * A;
                    S(B,1);
                    mout << "Arnoldi shift and invert with  gamma = " << gamma <<endl; // cf. Hochbruck Skript aijs can be ignored for linear problems
                    mout << "Exponential Euler " << gamma <<endl;
                }
                assemble.RHS(t-dt,RHS);
                Vector tmp(u);
                tmp = A*u;
                tmp +=RHS;
                RHS = IM*tmp;
                Arnoldi2(dt,RHS,M,A,du,1);
                du *= dt;
                break;
            }
            case 2002:{
                if (dt_old <1e-13) {

                    B = M;
                    gamma = gamma0 * dt;
                    B += (-gamma) * A;
                    S(B,1);
                    mout << "Arnoldi shift and invert with  gamma = " << gamma <<endl;
                    mout << "Exponential Midpoint " << gamma <<endl; // cf. Hochbruck Skript aijs can be ignored for linear problems
                }
                assemble.RHS(t-0.5*dt,RHS);
                Vector tmp(u);
                tmp = A*u;
                tmp +=RHS;
                RHS = IM*tmp;
                Arnoldi2(dt,RHS,M,A,du,1);
                du *= dt;
                break;
            }
            case 2003:{
                if (dt_old == 0) {
                    B = M;
                    gamma = gamma0 * dt;
                    B += (-gamma) * A;
                    S(B,1);
                    mout << "Arnoldi shift and invert with  gamma = " << gamma <<endl;
                    mout << "Exponential Trapezoidal " << gamma <<endl; // cf. Hochbruck Skript aijs can be ignored for linear problems
                }
                Vector tmp(RHS);
                Vector tmp2(RHS);
                assemble.RHS(t-dt,RHS);
                assemble.RHS(t,RHS2);
                tmp = A*u;
                tmp += RHS;
                du = IM * tmp;
                tmp = A*du;
                tmp *= dt;
                tmp += RHS2;
                tmp -= RHS;
                tmp2 = IM * tmp;
                Arnoldi2(dt,tmp2,M,A,tmp,2);
                du += tmp;
                du *= dt;
                break;
            }
            case -2:
                if (abs(dt-dt_old) > 1e-13){
                    B = M;
                    gamma = 0.5*dt;
                    assemble.SetGamma(gamma);
                    B += (-gamma) * A;
                    mout << "Implicit Midpoint Rule: gamma = " << gamma <<endl;
                    S(B,1);
                }
                assemble.RHS(t-0.5*dt,RHS);
                Pade2(dt,u,A,du,RHS);
                break;
            case -22: {
                Pade2_2(dt,t,assemble,u,A,M,du,B,B2,B3);
                break;
            }
            case -12: {
                ImpEuler(dt,t,assemble,u,A,M,du,B,B2,B3);
                break;
            }
            case -1337: {
                diRKtest(dt,t,assemble,u,A,M,du,B,B2,B3);
                break;
            }
            case -3:
                if (abs(dt-dt_old) > 1e-13) {
                    const Scalar gamma = 4.0 - 2*pow(2.0/(1.0+sqrt(5.0)),1.0/3.0)
                            + pow(2.0,2.0/3.0) * pow(1.0+sqrt(5),1.0/3.0);
                    const Scalar alpha = 4.0 + (1.0-iUnit*sqrt(3.0))*pow(2.0/(1.0+sqrt(5)),1.0/3.0)
                            - (1.0+iUnit*sqrt(3.0)) * pow(0.5*(1.0+sqrt(5)),1.0/3.0);
                    B = M;
                    B += Scalar(-dt/gamma) * A;
                    B2 = M;
                    B2 += Scalar(-dt/alpha) * A;
                    B3 = M;
                    B3 += Scalar(-dt/conj(alpha)) * A;
                    mout << "Gauss (s=3)" << endl;
                    S(B,1);
                    S2(B2,1);
                    S3(B3,1);
                }
                assemble.RHS(t+(0.5-sqrt(15.)/10.)*dt,RHS);
                assemble.RHS(t+0.5*dt,RHS2);
                assemble.RHS(t+(0.5+sqrt(15.)/10.)*dt,RHS3);
                Pade3(dt,u,M,A,du,RHS,RHS2,RHS3);
                break;
            case -4:
                if (abs(dt-dt_old) > 1e-13) {
                    B = M;
                    gamma = 0.5*dt;
                    assemble.SetGamma(gamma);
                    B += (-gamma) * A;
                    B2 = M;
                    gamma = 0.25*dt;
                    assemble.SetGamma(gamma);
                    B2 += (-gamma) * A;
                    mout << "Extrapolated Implicit Midpoint Rule: gamma = " << gamma <<endl;
                    S(B,1);
                    S2(B2,1);
                }
                assemble.RHS(t-0.5*dt,RHS);
                assemble.RHS(t-0.75*dt,RHS2);
                assemble.RHS(t-0.25*dt,RHS3);
                Pade2ex(dt,u,A,B,B2,du,RHS,RHS2,RHS3);
                break;
            }
            u += du;
//            mout << OUT(u) << endl;
            ++SSum;
                        if (u.norm() > 1e10) {
                            u = Scalar(0.0);
                            mout << "divergence at t = " << t << endl;
                            return;
                        }
            assemble.FinishStep(u,t,dt,++step,plot_tStep);
            dt_old = dt;
        }
        TComp = Date() - CStart;
    }

    int S_Sum () const { return SSum; }
    int M_Sum () const { return MSum; }
    int K_Sum () const { return KSum; }
    int K_Pre () const { return KPre; }
    Time T_Comp () const { return TComp; }
    Time T_Pre () const { return TPre; }
};

#endif
