#ifndef _DGVISCOACOUSTICX_H_
#define _DGVISCOACOUSTICX_H_

#include "m++.hpp"
#include "DGMatrixGraph.hpp"
#include "DGTSHelper.hpp"
#include <math.h>
#include <fstream>
#include <stdlib.h>
#include <time.h>
#include <stdio.h>

class DGVAcousticXDoF : public DoF {
    short cell_std_deg;
     std::unordered_map<Point,vector<short>> cell_deg;
    int n_eta;
public:
    DGVAcousticXDoF(int _n_eta,short cell_std_deg_) : DoF(0,false),n_eta(_n_eta),
        cell_std_deg(cell_std_deg_) {}
    int get_n_eta () const{ return n_eta; }
    int get_celldeg (const Point& z, int ud = 0) const {
        return cell_std_deg;
        if (cell_deg.find(z) != cell_deg.end()) {
            conststd::vector<short>& degs = cell_deg.find(z)->second;
            if (degs[ud] > 0) return degs[ud];
        }
        return cell_std_deg;
    }
    void set_cell_deg (const Point& z, int ud, short deg) {
       std::vector<short>* c_deg = NULL;
        if (cell_deg.find(z) != cell_deg.end())
            c_deg = &cell_deg[z];
        else {
            c_deg = &cell_deg[z];
            c_deg->resize(1,-1);
        }
        (*c_deg)[ud] = deg;
    }
    int CellDoFs (const cell& c) const {
        int m = c.dim() + 1 + n_eta;
        int c_deg = get_celldeg(c(),0);
        switch (c.Type()) {
        case TRIANGLE:      return m*(c_deg+1)*(c_deg+2)/2;
        case TETRAHEDRON:   Exit("not implemented");
        case QUADRILATERAL: return m*(c_deg+1)*(c_deg+1);
        case HEXAHEDRON:    return m*(c_deg+1)*(c_deg+1)*(c_deg+1);
        }
        Exit("not implemented");
    }
    void NodalDoFs (const cell& c,std::vector<short>& z) const {
        z.resize(1);
        z[0] = CellDoFs(c);
    }
    int NodalPoints (const cell& c) const { return 1; }
    void NodalPoints (const cell& c,std::vector<Point>& z) const {
        z.resize(1);
        z[0] = c();
    }
    std::string Name () const { return "DGVAcousticDoF"; }
    int TypeDoF (int) const { return CELL; }
};
int GetN_EtaX() {
    std::string name="No Wave";
    Config::Get("Problem", name);
    if (name == "Spherical") return 1;
    else Exit(" no nr of kappas defined");
}
DGVAcousticXDoF* GetDGVAcousticXDoF (int cell_std_deg = -1) {
    if(cell_std_deg == -1){
        Config::Get("deg", cell_std_deg);
    }
    int n_eta = GetN_EtaX();
    return new DGVAcousticXDoF(n_eta,cell_std_deg);
}

class DGVAcousticXDiscretization  : public IDiscretization {
    const Shape* S_tri[5];
    const Shape* S_quad[6];
    const Shape* S_tet[4];
    const Shape* S_hex[4];
    const Shape* T_tri[5];
    const Shape* T_quad[5];
    const Shape* T_tet[2];
    const Shape* T_hex[2];
    DGVAcousticXDoF& D;
public:
    DGVAcousticXDiscretization (DGVAcousticXDoF* dof) :
        IDiscretization(dof,"DG"), D(*dof) {
        S_tri[0] = Get_Shape("P0tri");
        S_tri[1] = Get_Shape("P1tri");
        S_tri[2] = Get_Shape("P2tri");
        S_tri[3] = Get_Shape("P3tri");
        S_tri[4] = Get_Shape("P4tri");

        T_tri[1] = Get_Shape("RT0tri");
        // Overkill Quadrature!
        S_quad[0] = Get_Shape("P0quad");
        S_quad[1] = Get_Shape("P1quad","Qquad4","Qint7");
        S_quad[2] = Get_Shape("P2quad");
        S_quad[3] = Get_Shape("P3quad");
        S_quad[4] = Get_Shape("P4quad","Qquad25","Qint7");
        S_quad[5] = Get_Shape("P5quad","Qquad36","Qint7");
        S_tet[1] = Get_Shape("P1tet");
        S_tet[2] = Get_Shape("P2tet");

        S_hex[1] = Get_Shape("P1hex");
        S_hex[2] = Get_Shape("P2hex");
        S_hex[3] = Get_Shape("P3hex");

        T_quad[1] = Get_Shape("RT0quad");
        T_tet[1] = Get_Shape("RT0tet");
        T_hex[1] = Get_Shape("RT0hex");
    }
    int get_celldeg (const Point& z, int component = 0) const {
        return D.get_celldeg(z,component);
    }
    int get_n_eta () const{
        return D.get_n_eta();
    }
    const Shape& GetCellShape (const cell& c, int p, int component=0) const {
        if (p > 5) Exit("DGVAcousticShapes::GetShape: Not implemented");
        switch (c.Corners()) {
        case 3: return *S_tri[p];
        case 4:
            switch (c.dim()) {
            case 2: return *S_quad[p];
            case 3: return *S_tet[p];
            }
        case 8: return *S_hex[p];
        }
        Exit("DGVAcousticShapes::GetShape: Not implemented!");
    }
    int CellDoFs (const cell& c) const { return D.CellDoFs(c); }
    std::string DiscName () const { return "DGVAcousticXDiscretization"; }
    std::string Name_Disc_dG () const { return "DGVAcousticXDiscretization"; }
};

class DGVAcousticXElement : public Element {
    int dim;
    const Shape& S;
    //    ShapeValue value;
    VectorShapeValuesDG gradient;
    VectorShapeValuesDG velocity;
    ScalarShapeValuesDG divergence;
    const ScalarShapeValues& value;
    const cell& C;
    const row& r (int i) const { return (*this)[i]; }
    int deg;
    Transformation T[MaxQuadraturePointsDG];
    int n_eta;
public:
    Point GlobalToLocal (const Point& z) const {
        switch (C.Type()) {
        case TRIANGLE: {
            Tensor T;
            T[0][0] = C.Corner(1)[0]-C.Corner(0)[0];
            T[0][1] = C.Corner(2)[0]-C.Corner(0)[0];
            T[1][0] = C.Corner(1)[1]-C.Corner(0)[1];
            T[1][1] = C.Corner(2)[1]-C.Corner(0)[1];
            T[0][2] = 0;
            T[1][2] = 0;
            T[2][0] = 0;
            T[2][1] = 0;
            T[2][2] = 1;
            Tensor IT = Invert(T);
            return IT * (z - C.Corner(0));
        }
        case TETRAHEDRON:
            Exit("not implemented");
        case QUADRILATERAL: {
            Tensor T;
            T[0][0] = C.Corner(1)[0]-C.Corner(0)[0];
            T[0][1] = C.Corner(3)[0]-C.Corner(0)[0];
            T[1][0] = C.Corner(1)[1]-C.Corner(0)[1];
            T[1][1] = C.Corner(3)[1]-C.Corner(0)[1];
            T[0][2] = 0;
            T[1][2] = 0;
            T[2][0] = 0;
            T[2][1] = 0;
            T[2][2] = 1;
            Tensor IT = Invert(T);
            return IT * (z - C.Corner(0));
        }
        case HEXAHEDRON:
            Exit("not implemented");
        }
    }
    void NodalPoints (Point* z) const { S.NodalPoints(C,z); }
    int get_deg() { return deg; }
    DGVAcousticXElement(const DGVAcousticXDiscretization &disc,
                        const VectorMatrixBase &g, const cell &c) :
            Element(g, c, disc.GetCellShape(c, disc.get_celldeg(c())).GetQuad()),
            S(disc.GetCellShape(c, disc.get_celldeg(c()))), value(S.values()),
            C(c), dim(c.dim()), n_eta(disc.get_n_eta()), gradient(nQ(), S.size()), velocity(nQ(), dim * S.size()),
            divergence(nQ(), dim * S.size()),
            //        C(c), dim(c.dim()),n_eta(disc.get_n_eta()),
            deg(disc.get_celldeg(c())) {
        for (int q=0; q<Q.size(); ++q) {
            T[q] = c.GetTransformation(Q.QPoint(q));
            for (int i=0; i<S.size(); ++i) {
                gradient[q][i] = T[q] * S.LocalGradient(q,i);
            }
            for (int k=0; k<dim; ++k) {
                VectorField V = zero;
                V[k] = 1.0;
                for (int i=0; i<S.size(); ++i) {
                    divergence[q][i+k*S.size()]=gradient[q][i][k];
                    velocity[q][i+k*S.size()] = value[q][i] * V;
                }
            }
        }
    }
    //    int nQ () const { return Q.size(); }
    double QWeight (int q) const { return qWeight[q]; }
    const Point& QPoint (int q) const { return qPoint[q]; }
    double Area () const {
        double a = 0;
        for (int q=0; q<nQ(); ++q)
            a += qWeight[q];
        return a;
    }
    int shape_size () const { return S.size(); }
    int V_dimension () const { return dim * shape_size(); }
    int P_dimension () const { return shape_size(); }
    int dimension () const { return (dim+1+n_eta) * shape_size(); }
    int u_dimension () const { return dim+1+n_eta; }
    int eta_dimension () const { return n_eta;}
    
    VectorField Gradpressure (int q, int i) const  { return gradient[q][i]; }
    double Divergence (int q, int i) const  { return divergence[q][i]; }
    double Pressure (int q, int i) const { return value[q][i]; }
    double Eta (int q, int i) const { return value[q][i]; }
    VectorField Velocity (int q, int i) const { return velocity[q][i]; }
    
    VectorField Velocity (int q, const Vector& u) const  {
        VectorField V = zero;
        for (int i=0; i<V_dimension(); ++i)
            V += u(r(0),i) * velocity[q][i];
        return V;
    }
    double Pressure (int q, const Vector& u) {
        double p = 0;
        for (int i=0; i<shape_size(); ++i)
            p += u(r(0),V_dimension()+i)*value[q][i];
        return p;
    }
    Scalar SVValue (const Point& lz, const Vector& u, int k) const {
        Scalar v = 0;
        for (int i=0; i<S.size(); ++i)
            v += u(C(),k*S.size()+i) * S(lz,i);
        return v;
    }
};

class DGVAcousticXFaceElement : public FaceElement {
    int dim;
    const Shape& S;
    VectorShapeValuesDG velocity;
    ShapeValue value;
    int n_eta;
public:
    DGVAcousticXFaceElement(const DGVAcousticXDiscretization &D,
                            const VectorMatrixBase &g,
                            const cell &C, int f_id) :
        FaceElement(g,C,f_id,D.GetCellShape(C,D.get_celldeg(C())).GetFaceQuad()),
        //        S(D.GetCellShape(C,D.get_celldeg(C()))), dim(C.dim()),n_eta(D.get_n_eta()) {
        S(D.GetCellShape(C,D.get_celldeg(C()))), dim(C.dim()),n_eta(D.get_n_eta()),velocity(nQ(),dim*S.size()) {
        for (int q=0; q<nQ(); ++q) {
            for (int i=0; i<S.size(); ++i)
                value[q][i] = S(QLocal(q),i);
            for (int k=0; k<dim; ++k) {
                VectorField V = zero;
                V[k] = 1.0;
                for (int i=0; i<S.size(); ++i) {
                    velocity[q][i+k*S.size()] = value[q][i] * V;
                }
            }
        }
    }
    int eta_dimension() const { return n_eta;}
    int shape_size () const { return S.size(); }
    int dimension () const { return (dim+1+n_eta) * shape_size(); }
    int V_dimension () const { return dim * shape_size(); }
    VectorField Velocity (int q, int i) const  { return velocity[q][i]; }
    double Pressure (int q, int i) const { return value[q][i]; }
    double Eta (int q, int i) const { return value[q][i]; }
};

class ProblemVAcousticX {
protected:
    const DGVAcousticXDiscretization& disc;
    int verbose;
    double rho;
    double kappa,tau;
    int dim;
    std::string meshname;
    int n_eta;
public:
    ProblemVAcousticX (const DGVAcousticXDiscretization& _disc, int d = 3) :
        disc(_disc), dim(d), verbose(0), tau(1),
        rho(1.0), kappa(1.0), meshname("none"),n_eta(disc.get_n_eta()) {
        Config::Get("ProblemVerbose", verbose);
        Config::Get("rho", rho);
        Config::Get("kappa", kappa);
        Config::Get("tau", tau);
        Config::Get("Mesh", meshname);
    }
    virtual bool Exact () const { return false; }
    virtual bool RHS(double) const { return false; }
    virtual double ForceP (double t, const Point& z) const { return 0; }
    virtual VectorField ForceV (double t, const Point& z) const {
        return zero;
    }
    virtual double Dirichlet_i (double t, const Point& z, int i ) const {
        return ut(t,z,dim+i);
    }
    virtual double Neumann (double t, const Point z,
                            const VectorField& N) const {
        VectorField V(ut(t,z,0),ut(t,z,1));
        return V * N;
    }
    virtual double ut (double, const Point&, int) const { return 0; }
    virtual double ut_dual (double, const Point&, int,
                            const Point&, const Point&) const { return 0; }
    virtual double u0 (const Point& x, int i) const { return ut(0,x,i); }
    virtual double ut (double t, const cell& c, const Point& x, int i) const { return ut(t,x,i); }
    virtual double u0 (const cell& c, const Point& x, int i) const { return ut(0,c,x,i); }
    virtual double Kappa (const Point& x, int i=-1 ) const {
        if (i == -1) return kappa;
        if (i ==  0) return 0.9*kappa;
        else         return 0.1*kappa/n_eta;
    }
    virtual double Tau (const Point& x, int i=1 ) const { return tau; }
    virtual double Rho (const Point& x) const { return rho; }
    virtual void get_u_int (double t, Vector& u) const {
        u = 0.0;
        for (cell c=u.cells(); c!=u.cells_end(); ++c) {
            DGVAcousticXElement Elem(disc,u,c);
            if (disc.get_celldeg(c()) == 0) {
                for (int i=0; i<Elem.u_dimension();++i)
                    u(c(),i) = ut(t,c,c(),i);
            } else {
                row r = u.find_row(c());
                DGVAcousticXElement Elem(disc,u,c);
                Point z[MaxNodalPointsDG];
                Elem.NodalPoints(z);
                for (int i=0; i<Elem.u_dimension();++i)
                    for (int j=0; j<Elem.shape_size();++j)
                        u(r)[i*Elem.shape_size()+j] = ut(t,c,z[j],i);
            }
        }
        u.Accumulate();
        // mout << OUT(u);
    }
    virtual void get_u_int (Vector& u) const {
        u = 0.0;
        for (cell c=u.cells(); c!=u.cells_end(); ++c) {
            DGVAcousticXElement Elem(disc,u,c);
            if (disc.get_celldeg(c()) == 0) {
                for (int i=0; i<Elem.u_dimension();++i)
                    u(c(),i) = u0(c,c(),i);
            } else {
                row r = u.find_row(c());
                DGVAcousticXElement Elem(disc,u,c);
                Point z[MaxNodalPointsDG];
                Elem.NodalPoints(z);
                for (int i=0; i<Elem.u_dimension();++i)
                    for (int j=0; j<Elem.shape_size();++j)
                        u(r)[i*Elem.shape_size()+j] = u0(c,z[j],i);
            }
        }
        u.Accumulate();
    }
    virtual const char* Name () const = 0;
    virtual const char* MeshName () const { return meshname.c_str(); }
};

class VAcousticX_Spherical : public ProblemVAcousticX {
    Point Pmid;
    double width;
public:
    VAcousticX_Spherical (const DGVAcousticXDiscretization& _disc)
        : ProblemVAcousticX(_disc,2), Pmid(2,2), width(0.25) { }

    const char* Name () const { return "VAcousticX Spherical Wave"; }

    double A0 (double s) const {
        if (s<width) return pow(cos(s*0.5*M_PI/width),6.0);
        //if (s*s < width) return exp(-1/(width-s*s) + 1/width);
        else return 0.0;
    }
    double A (const Point& x) const {
        double r = dist(Pmid,x);
        return A0(r*r);
    }
    double ut (double t, const Point& x, int i) const {
        if (t > 0) return 0.0;
        const double tmp = A(x);
        switch (i) {
            case 0 : return (x[0]-Pmid[0])*tmp;
            case 1 : return (x[1]-Pmid[1])*tmp;
            default: return 0.0;
        }
    }
    double ut_dual (double t, const Point& x, int i, const Point& Ga, const Point& Gb) const {
        Point ba = Gb-Ga;
        switch (i) {
            case 0 : return 1.0/(ba[0]*ba[1]);
            //case 1 : return 1.0/(ba[0]*ba[1]);
            default: return 0.0;
        }
    }

};

ProblemVAcousticX* GetProblemVAcousticX (const DGVAcousticXDiscretization&_disc,
                                       const std::string& name ) {
    if (name == "Spherical") return new VAcousticX_Spherical(_disc);
    else Exit("Error: no wave found!");
}

class DGVAcousticXAssembleMatrix : public DGAssemble {
protected:
    const DGVAcousticXDiscretization& disc;
    const ProblemVAcousticX& problem;
    double flux_alpha;
    std::string problem_name;
    Point GF_a;
    Point GF_b;
     std::unordered_map<Point,DGVAcousticXElement*> EleCache;
    bool upwind;
public:
    DGVAcousticXAssembleMatrix (const DGVAcousticXDiscretization& D,
                               const ProblemVAcousticX& P, Plot& _plot) :
        DGAssemble(_plot), disc(D), problem(P), flux_alpha(1),
        GF_a(0.0,0.0), GF_b(1.0,1.0), upwind(true) {
        Config::Get("flux_alpha", flux_alpha);
        Config::Get("Problem", problem_name);
        Config::Get("roi_min", GF_a);
        Config::Get("roi_max", GF_b);
        Config::Get("upwind", upwind);
    }
    ~DGVAcousticXAssembleMatrix (){
        for (auto it = EleCache.begin(); it!=EleCache.end(); ++it){
            delete it->second;
        }
    }

    void PrintInfo(const Vector &u) {
        Assemble::PrintInfo(u);
        vout(1) << " Mesh            " << problem.MeshName() << endl
                << " Problem         " << problem.Name() << endl << endl;
    }

    void initializeEleCache(const VectorMatrixBase &mgr) {
        for (auto it = EleCache.begin(); it != EleCache.end(); ++it)
            delete it->second;
        EleCache.clear();
        for (cell c = mgr.cells(); c != mgr.cells_end(); ++c)
            EleCache[c()] = new DGVAcousticXElement(disc, mgr, c);
    }

    DGVAcousticXElement *getElem(const DGVAcousticXDiscretization &disc,
                                 const VectorMatrixBase &g, const cell &c) const {
        if (EleCache.size() == 0)
            return new DGVAcousticXElement(disc, g, c);
        else {
            auto p = EleCache.find(c());
            return p->second;
        }
    }

    void destroyElement(DGVAcousticXElement *elem) const {
        if (EleCache.size() == 0) delete elem;
    }

    const char* Name() const { return "DGAssembleMatrix"; }
    void MassMatrix (Matrix& M) const {
        M = 0;
        for (cell c=M.cells(); c!=M.cells_end(); ++c) {
            DGVAcousticXElement* elem = getElem(disc,M,c);
            RowEntries_dG M_c(M,c,c);
            int nV = elem->V_dimension();
            int nP = elem->shape_size();
            int eta_dim = elem->eta_dimension();
            
            double rho = problem.Rho(c());
           std::vector<double> kappaInv(1+eta_dim);
           std::vector<double> tau(1+eta_dim);
            kappaInv[0] = 1.0/problem.Kappa(c(),-1);
            for (int i=1; i<1+eta_dim; i++) {
                kappaInv[i] = 1.0/problem.Kappa(c(),i);
                tau[i] = problem.Tau(c(),i);
            }
            
            for (int q=0; q<elem->nQ(); ++q) {
                double w  = elem->QWeight(q);
                Point z = elem->QPoint (q);
                for (int i=0;i<nV; ++i) {
                    VectorField V_i = elem->Velocity(q,i);
                    for (int j=0;j<nV; ++j) {
                        VectorField V_j = elem->Velocity(q,j);
                        M_c(i,j) += w * rho * (V_i * V_j);
                    }
                }
                for (int i=0;i<nP; ++i) {
                    double P_i = elem->Pressure(q,i);
                    for (int j=0;j<nP; ++j) {
                        double P_j = elem->Pressure(q,j);
                        M_c(i+nV,j+nV) += w * kappaInv[0] * P_i * P_j;
                    }
                }
                for (int n=1; n<=eta_dim; ++n) {
                    for (int i=0;i<nP; ++i) {
                        double Eta_i = elem->Eta(q,i);
                        for (int j=0;j<nP; ++j) {
                            double Eta_j = elem->Eta(q,j);
                            M_c(i+nV+n*nP,j+nV+n*nP) += w * tau[n]*kappaInv[n] * Eta_i * Eta_j;
                        }
                    }
                }
            }
            destroyElement(elem);
        }
        
        return;
        SparseMatrix SpM(M);
        RMatrix SmM(SpM);
        mout << OUT(SmM);
    }
    void FluxMatrix (Matrix& A) const {
        A = 0;
        const Mesh& M = A.GetMesh();
        for (cell c=A.cells(); c!=A.cells_end(); ++c) {
            DGVAcousticXElement* elem = getElem(disc,A,c);
            int nV = elem->V_dimension();
            int nP = elem->shape_size();
            int eta_dim = elem->eta_dimension();
            RowEntries_dG A_c(A,c,c);
            
           std::vector<double> kappaInv(1+eta_dim);
            kappaInv[0] = 1.0/problem.Kappa(c(),-1);
            for (int i=1; i<1+eta_dim; i++)
                kappaInv[i] = 1.0/problem.Kappa(c(),i);
            
            for (int q=0; q<elem->nQ(); ++q) {
                double w  = elem->QWeight(q);
                Point z = elem->QPoint (q);
                
                for (int i=0; i<nV; ++i) {
                    VectorField V_i = elem->Velocity(q,i);
                    for (int j=0; j<nP; ++j) {
                        VectorField GradP_j = elem->Gradpressure(q,j);
                            A_c(i,nV+j) += w * (V_i * GradP_j);
                    }
                }
                for (int i=0; i<nP; ++i) {
                    double P_i = elem->Pressure(q,i);
                    for (int j=0;j<nV; ++j) {
                        double Div_j = elem->Divergence(q,j);
                        A_c(nV+i,j) += w * P_i * Div_j;
                    }
                    for (int n=1; n<=eta_dim;++n)
                        for (int j=0;j<nP; ++j) {
                            double Eta_j = elem->Eta(q,j);
                            A_c(nV+i,j+nV+n*nP) += w * kappaInv[0] * P_i * Eta_j;
                        }
                }
                for (int n=1; n<=eta_dim;++n) {
                    for (int i=0; i<nP; ++i) {
                        double Eta_i = elem->Eta(q,i);
                        for (int j=0;j<nV; ++j) {
                            double Div_j = elem->Divergence(q,j);
                            A_c(i+nV+n*nP,j) -= w * Eta_i * Div_j;
                        }
                        for (int j=0; j<nP; ++j) {
                            double Eta_j = elem->Eta(q,j);
                            A_c(i+nV+n*nP,j+nV+n*nP) -= w * kappaInv[n] * Eta_i * Eta_j;
                        }
                    }
                }
            }
            for (int f=0; f<c.Faces(); ++f) {
                DGVAcousticXFaceElement felem(disc,A,c,f);
                int nV = felem.V_dimension();
                int nP = felem.shape_size();
                if (check_if_bnd(A,c,f)) {
                    BFParts bnd(A.GetMesh(),c);
                    
                    double c_K = sqrt(problem.Kappa(c(),-1) * problem.Rho(c()));
                    double alpha =  1.0 / c_K;
                    if (!upwind) alpha = 0;
                    
                    for (int q=0; q<felem.nQ(); ++q) {
                        double w  = felem.QWeight(q);
                        VectorField N = felem.QNormal(q);
                        Point z = felem.QPoint(q);
                        
                        if (bnd[f] == 2 || bnd[f] == 0) {
                            for (int j=0; j<nV; ++j) {
                                Scalar VN_j = felem.Velocity(q,j) * N;
                                for (int i=0; i<nV; ++i) {
                                    Scalar VN_i = felem.Velocity(q,i) * N;
                                    A_c(i,j) -= w * c_K * VN_i * VN_j;
                                }
                                for (int i=0; i<felem.shape_size(); ++i) {
                                    double P_i = felem.Pressure(q,i);
                                    A_c(nV+i,j) -= w * VN_j * P_i;
                                }
                            }
                        } else if (bnd[f] == 1) {
                            for (int j=0; j < nP; ++j) {
                                double P_j = felem.Pressure(q,j);
                                for (int i=0; i<nV; ++i) {
                                    Scalar VN_i = felem.Velocity(q,i) * N;
                                    A_c(i,nV+j) -= w  * VN_i * P_j;
                                }
                                for (int i=0; i < nP; ++i) {
                                    double P_i = felem.Pressure(q,i);
                                    A_c(nV+i,nV+j) -= w * alpha * P_i * P_j;
                                }
                            }
                        }
                    }
                } else {
                    cell cf = find_nb_cell(A,c,f);
                    int f1 = find_nb_face_id(A,c.Face(f),cf);
                    DGVAcousticXFaceElement felem_1(disc,A,cf,f1);
                    int nV_1 = felem_1.V_dimension();
                    int nP_1 = felem_1.shape_size();
                    RowEntries_dG A_cf(A,c,cf);
                    
                    double c_K  = sqrt(problem.Kappa(c(),-1) * problem.Rho(c()));
                    double c_Kf = sqrt(problem.Kappa(cf(),-1) * problem.Rho(cf()));
                    double alpha1 =  1.0 / ( c_K +  c_Kf);
                    double alpha2 =  c_Kf * c_K / ( c_K +  c_Kf);
                    double alpha3 = c_Kf  / ( c_K +  c_Kf);
                    double alpha4 = c_K  / (c_K +  c_Kf);
                    if (!upwind) {
                        alpha1 =  alpha2 = 0;
                        alpha3 = alpha4 = 0.5;
                    }
                    
                    for (int q=0; q<felem.nQ(); ++q) {
                        double w  = felem.QWeight(q);
                        VectorField N = felem.QNormal(q);
                        Point z = felem.QPoint(q);
                        
                        for (int i=0; i < nV; ++i) {
                            Scalar VN_i = felem.Velocity(q,i) * N;
                            for (int j=0; j< nV; ++j) {
                                Scalar VN_j = felem.Velocity(q,j) * N;
                                A_c(i,j) -= w * alpha2 * VN_i * VN_j;
                            }
                            for (int j=0; j < nV_1; ++j) {
                                Scalar V1N_j = felem_1.Velocity(q,j) * N;
                                A_cf(i,j) += w * alpha2 * VN_i * V1N_j;
                            }
                            for (int j=0; j<nP; ++j) {
                                double P_j = felem.Pressure(q,j);
                                A_c(i,nV+j) -= w * alpha4 * P_j * VN_i;
                            }
                            for (int j=0; j<nP_1; ++j) {
                                double P1_j = felem_1.Pressure(q,j);
                                A_cf(i,nV_1+j) += w * alpha4 * P1_j * VN_i;
                            }
                        }
                        for (int i=0; i < nP; ++i) {
                            double P_i = felem.Pressure(q,i);
                            for (int j=0; j < nP; ++j) {
                                double P_j = felem.Pressure(q,j);
                                A_c(nV+i,nV+j) -= w * alpha1 * P_i * P_j;
                            }
                            for (int j=0; j < nP_1; ++j) {
                                double P1_j = felem_1.Pressure(q,j);
                                A_cf(nV+i,nV_1+j) += w * alpha1 * P_i * P1_j;
                            }
                            for (int j=0; j < nV; ++j) {
                                Scalar VN_j = felem.Velocity(q,j) * N;
                                A_c(nV+i,j) -= w * alpha3 * VN_j * P_i;
                            }
                            for (int j=0; j < nV_1; ++j) {
                                Scalar V1N_j = felem_1.Velocity(q,j) * N;
                                A_cf(nV+i,j) += w * alpha3 * V1N_j * P_i;
                            }
                        }
                    }
                }
            }
            destroyElement(elem);
        }
        
        return;
        SparseMatrix SpA(A);
        RMatrix SmA(SpA);
        mout << OUT(SmA);
    }
    bool RHS (double t) const { return problem.RHS(t); }
    void RHS (double t, Vector& b) const {
        b = 0.0;
        for (cell c = b.cells(); c != b.cells_end(); ++c) {
            DGVAcousticXElement* elem= getElem(disc,b,c);
            if (problem.RHS(t)) {
                for (int q=0; q<elem->nQ(); ++q) {
                    double w  = elem->QWeight(q);
                    int nV = elem->V_dimension();
                    int nP = elem->shape_size();
                    Point z = elem->QPoint(q);
                    VectorField V_rhs = problem.ForceV(t,z);
                    for (int i=0; i<elem->V_dimension(); ++i) {
                        VectorField V_i = elem->Velocity(q,i);
                        b(c(),i) += w * V_i * V_rhs;
                    }
                    for (int i=0; i<elem->shape_size(); ++i) {
                        double P_i = elem->Pressure(q,i);
                        b(c(),nV+i) += w * P_i * problem.ForceP(t,z);
                    }
                }
            }
            if (problem.Exact()){Exit("check");
                for (int f=0; f<c.Faces(); ++f) {
                    if(!check_if_bnd(b,c,f)) continue;
                    DGVAcousticXFaceElement felem(disc,b,c,f);
                    int nV = felem.V_dimension();
                    int nP = felem.shape_size();
                    BFParts bnd(b.GetMesh(),c);
                    double c_K   = sqrt(problem.Kappa(c()) * problem.Rho(c()));
                    double alpha =  1 / c_K;
                    for (int q=0; q<felem.nQ(); ++q) {
                        double w  = felem.QWeight(q);
                        VectorField N = felem.QNormal(q);
                        Point z = felem.QPoint(q);
                        if (bnd[f] == 2) {
                            Scalar VN_bnd = problem.Neumann(t,z,N);
                            for (int i=0;i<nV; ++i) {
                                Scalar VN_i = felem.Velocity(q,i) * N;
                                b(c(),i) += w * c_K * VN_i * VN_bnd;
                            }
                            for (int i=0;i<nP; ++i) {
                                double P_i = felem.Pressure(q,i);
                                b(c(),nV+i) +=  w * P_i * VN_bnd;
                            }
                        } else if (bnd[f] == 1) {
                            for (int i=0; i < nV; ++i) {
                                Scalar VN_i = felem.Velocity(q,i) * N;
                                b(c(),i) += w * VN_i * problem.Dirichlet_i(t,z,0);

                            }
                            for (int i=0;i<nP;++i) {
                                double P_i = felem.Pressure(q,i);
                                b(c(),nV+i) += w * alpha * P_i * problem.Dirichlet_i(t,z,0);
                            }
                        }
                    }
                }
            }
            destroyElement(elem);
        }

    }

    cell find_cell (const Mesh& M, const face& f, const Point& z) const {
        cell c = M.find_cell(z);
        if (c == M.cells_end())
            c = M.find_overlap_cell(z);
        if (c == M.overlap_end()) {
            Exit("Error: no cell found");
            pout << "z " << z << endl;
            pout << "Cells: " << M.Cells::size() << endl << M.Cells::ref();
            pout << "OverlapCells: " << M.OverlapCells::size() << endl
                 << M.OverlapCells::ref();
            Exit("Error: no cell found");
        }
        return c;
    }

    cell find_nb_cell(const VectorMatrixBase &A, const cell &c, int f) const {
        face ff = A.GetMesh().find_face(c.Face(f));
        if (c() != ff.Right())
            return find_cell(A.GetMesh(), ff, ff.Right());
        else
            return find_cell(A.GetMesh(), ff, ff.Left());
    }

    bool check_if_bnd(const VectorMatrixBase &A, const cell &c, int f) const {
        face ff = A.GetMesh().find_face(c.Face(f));
        if ((A.GetMesh().find_bnd_face(c.Face(f)) != A.GetMesh().bnd_faces_end())
            && ((ff.Right() == Infty) || (ff.Left() == Infty)))
            // To exclude periodic bnd
            return true;
        else
            return false;
    }

    int find_nb_face_id(const VectorMatrixBase &A,
                        const Point &f_c, const cell &cf) const {
        for (int f1 = 0; f1 < cf.Faces(); ++f1)
            if (f_c == cf.Face(f1)) return f1;
        Exit("Error: no face found");
    }

    int bnd_id(const VectorMatrixBase &A, const cell &c, int f) const {
        bnd_face fff = A.GetMesh().find_bnd_face(c.Face(f));
        return fff.Part();
    }

    void Jacobi(const Vector &u, Matrix &B) const {
        MassMatrix(B);
        Matrix A(u);
        FluxMatrix(A);
        B += (-gamma) * A;
    }
    void AssembleTransfer (TransferMatrix& TM) const {
        TM = 0;
        const VectorMatrixBase &cg = TM.CoarseMatrixGraph();
        const VectorMatrixBase &fg = TM.FineMatrixGraph();
        const Mesh &mesh = fg.GetMesh();
        for (cell c = cg.cells(); c != cg.cells_end(); ++c)
            for (int k = 0; k < c.Children(); ++k)
                TM(c(), c.Child(k))[0] = 1;
    }
};

class DGVAcousticXAssemble : public DGVAcousticXAssembleMatrix {
    double flux_alpha;
    double X0;
    double X1;
    std::fstream f;
    Point Pobs;
public:
    DGVAcousticXAssemble (const DGVAcousticXDiscretization& _disc, const ProblemVAcousticX& _prob, Plot& _plot) :
        DGVAcousticXAssembleMatrix(_disc,_prob,_plot), X0(8), X1(10) {
        Config::Get("X0", X0);
        Config::Get("X1", X1);
        f.open("Pkt.dat", ios::out);
        Config::Get("Pobs", Pobs);
    }
    const char* Name() const { return "DGVAcousticXAssemble"; }
    std::string DiscName() const { return disc.DiscName(); }
    virtual void get_u_int_assemble (double t, Vector& u) const {
        u=0.0;
        std::string discName = disc.Name_Disc_dG();
        if (discName == "dGvector_acoustic_FV") {
            for (cell c=u.cells(); c!=u.cells_end(); ++c) {
                row r = u.find_row(c());
                DGVAcousticXElement Elem(disc,u,c);
                for (int i=0; i<Elem.u_dimension();++i)
                    u(r)[i] = problem.ut(t,c(),i);
            }
        } else {
            for (cell c=u.cells(); c!=u.cells_end(); ++c) {
                row r = u.find_row(c());
                DGVAcousticXElement Elem(disc,u,c);
                Point z[MaxNodalPointsDG];
                Elem.NodalPoints(z);
                for (int i=0; i<Elem.u_dimension();++i)
                    for (int j=0; j<Elem.shape_size();++j)
                        u(r)[i*Elem.shape_size()+j] =problem.ut(t,z[j],i);
            }
        }
        u.Accumulate();
    }
    double Energy_local (const Vector& u,
                         DGVAcousticXElement& elem,const cell& c) const {
        Scalar e = 0;
        for (int q=0; q<elem.nQ(); ++q) {
            double w  = elem.QWeight(q);
            Point z = elem.QPoint(q);
            double rho = problem.Rho(c());
            double kappaInv = 1.0/problem.Kappa(c());
            VectorField V = elem.Velocity(q,u);
            double P=elem.Pressure(q,u);
            e += w* (rho*V*V + kappaInv*P*P);;
        }
        return e;
    }
    double Energy (const Vector& u, double x0, double x1) const {
        Scalar e = 0;
        for (cell c=u.cells(); c!=u.cells_end(); ++c) {
            double x = c()[0];
            if (x<x0) continue;
            if (x>x1) continue;
            DGVAcousticXElement* elem= getElem(disc,u,c);
            e += Energy_local(u,*elem,c);
            destroyElement(elem);
        }
        return 0.5*PPM->Sum(e);
    }
    double Energy (const Vector& u) const {
        Scalar e = 0;
        for (cell c=u.cells(); c!=u.cells_end(); ++c) {
            DGVAcousticXElement* elem= getElem(disc,u,c);
            e += Energy_local(u,*elem,c);
            destroyElement(elem);
        }
        return 0.5*PPM->Sum(e);
    }
    double Norm (const Vector& u) const { return sqrt(Energy(u)); }
    virtual double L2Error (double t, const Vector& u) const {
        //        mout << OUT(u) << endl;
        double err = 0;
        for (cell c = u.cells(); c != u.cells_end(); ++c) {
            DGVAcousticXElement* elem= getElem(disc,u,c);
            for (int q=0; q<elem->nQ(); ++q) {
                double w = elem->QWeight(q);
                Point z = elem->QPoint (q);
                VectorField V = elem->Velocity(q,u)
                        - VectorField(problem.ut(t,z,0),problem.ut(t,z,1));
                double P = elem->Pressure(q,u)- problem.ut(t,z,c.dim());
                err += w * (V*V + P*P);
            }
            destroyElement(elem);
        }
        return sqrt(PPM->Sum (err));
    }
    double dist (const Vector& u, const Vector& v) const {
        Vector w(u);
        w -= v;
        return sqrt(Energy(w));
    }
    void SetInitialValue (Vector& u) const { problem.get_u_int(u); }

    void PrintStep (double t, const Vector& u, int step) {
        std::cout<< std::fixed;
        std::cout.precision(5);
        Vector u_tmp(u);
        mout << "Step(" << step << "), t("
             << t
             << "): E(u) = " << Energy(u);
        //<< " Local Energy(u) = " << Energy(u,X0,X1);
        //             std::cout << std::scientific;
        // mout << " L_2 Error = " << L2Error(u,t) << endl;
        //        << " L_inf = " << Linf(u);
        if (problem.Exact()) {
            problem.get_u_int(t,u_tmp);
            u_tmp -= u;
            mout << " IntErr " << sqrt(2.0*Energy(u_tmp));
            mout << " L2Err " << L2Error(t,u);
        }
        //mout<< endl;
        //return;

        u_tmp = 0;
        Vector u_int(u);
        problem.get_u_int(t,u_int);
        double max_P = -infty;
        double min_P =  infty;
        double max_P1 =0;
        for (cell c=u_tmp.cells(); c!=u_tmp.cells_end(); ++c) {
            row r = u.find_row(c());
            DGVAcousticXElement* elem= getElem(disc,u_tmp,c);
            VectorField V = zero;
            double P = 0;
            double P1 =0;
            double a = 0;
            for (int q=0; q<elem->nQ(); ++q) {
                double w  = elem->QWeight(q);
                V += w * elem->Velocity(q,u);
                P += w * elem->Pressure(q,u);
                P1 += w * elem->Pressure(q,u_int);
                a += w;
            }
            V *=  (1.0/a);
            P *= (1.0/a);
            P1 *= (1.0/a);

            if (max_P < P)
                max_P = P;
            if (min_P > P)
                min_P = P;

            if(max_P1 < P1)
                max_P1 = P1;

            u_tmp(r)[0] = V[0];
            u_tmp(r)[1] = V[1];
            u_tmp(r)[2] = P;
            destroyElement(elem);
        }
        max_P = PPM->Max(max_P);
        min_P = PPM->Min(min_P);
        if (abs(min_P) < 1e-10) min_P = 0;

        mout << " max(P) = " << max_P
             << " min(P) = " << min_P
             <<endl;

        return;
        mout << " Norm(V) = " << norm(u_tmp)
             <<" max(P) = "<<max_P<<" max(P) = "<<max_P1<<endl;
    }

    void InitialStep (double t, Vector& u) {
        SetInitialValue(u);
        PrintStep(t,u,0);
        VtkPlotting(t,u,0);
    }

    void FinishStep (Vector& u, double t, double dt, int step,
                     int plot_step = 1) {
        if (step%plot_step == 0  ) {
            PrintStep(t,u,step);
            VtkPlotting(t,u,step);
        }
        if (false) {
            for (cell c=u.cells(); c!=u.cells_end(); ++c) {
                row r = u.find_row(c());
                if(c.dim()==2){
                    double lambda2;
                    Point Pa = c.Corner(1)-c.Corner(0);
                    Point Pb = c.Corner(2)-c.Corner(0);
                    Point Pc = Pobs - c.Corner(0);
                    if (Pa[0]==0 ) {
                        lambda2 = Pc[0]/Pb[0];
                    } else {
                        lambda2 = (Pc[0]/Pa[0]-Pc[1]/Pa[1])/(Pb[0]/Pa[0]-Pb[1]/Pa[1]);
                    }
                    double lambda1 = (Pc[1]-lambda2*Pb[1])/Pa[1];
                    if( -1.0e-12<lambda1 && -1.0e-12<lambda2 && lambda1+lambda2<1+1.0e-12) {
                        const Shape& shape(disc.GetCellShape(c,disc.get_celldeg(c())));
                        int NN = r.n()/3;
                       std::vector<double> Dat(4);
                        Dat[0] = t;
                        for(int p=0; p<3;++p)
                            for(int i=0; i<NN;++i)
                                Dat[p+1] += shape(Point(lambda1,lambda2),i) * u(r)[p*NN+i];
                        f<< setw(10) << left << Dat[0] <<setw(15)<<Dat[1] <<setw(15)<<Dat[2] <<setw(15)<<Dat[3] << endl;
                    }
                } else Exit("not Implemented");
            }
        }

        return;
        Vector u_int(u);
        problem.get_u_int(t,u_int);
        u_int -= u;
        mout << "Energy(u-u_int) " << Energy(u_int) << endl;
    }

    void VtkPlotting_cell (double t, const Vector& u,
                           char* filename, int step, std::string discName) const {
        Vector u_tmp(u);
        Vector u_tmp2(u);
        Vector u_tmp3(u);
        Vector u_int(u);
        Vector u_int_tmp(u);
        u_tmp = 0.0;
        u_tmp2 = 0.0;
        u_tmp3 = 0.0;
        u_int = 0.0;
        u_int_tmp = 0.0;

        if (problem.Exact())
            problem.get_u_int(t,u_int);
        double tmp=0;
        const Mesh& Mu = u.GetMesh();
        double rho_max = -infty;
        double rho_min = infty;
        double kappa_max = -infty;
        double kappa_min = infty;
        double c_max = -infty;
        double c_min = infty;
        for (cell c=u_tmp.cells(); c!=u_tmp.cells_end(); ++c) {
            DGVAcousticXElement* elem=getElem(disc,u_tmp,c);
            VectorField V = zero;
            double P=0;
            double P_int=0;
            double a = 0;
            for (int q=0; q<elem->nQ(); ++q) {
                double w  = elem->QWeight(q);
                V += w * elem->Velocity(q,u);
                P += w * elem->Pressure(q,u);
                P_int += w * elem->Pressure(q,u_int);
                a += w;
            }
            V *=  (1.0/a);
            P *= (1.0/a);
            P_int *= (1.0/a);
            row r = u.find_row(c());
            u_tmp(r)[0] = V[0];
            u_tmp(r)[1] = V[1];
            u_tmp(r)[2] = P;
            u_int_tmp(r)[1] = P_int;
            if (step == 0) {
                u_tmp2(r)[0] = problem.Rho(c());
                for (int n = 0; n<1+elem->eta_dimension(); ++n)
                    u_tmp2(r)[1] += problem.Kappa(c(),n);
                u_tmp2(r)[2] = PPM->proc();
                rho_max = max(rho_max,u_tmp2(r)[0]);
                rho_min = min(rho_min,u_tmp2(r)[0]);
                //                mout << OUT(u_tmp2) << endl;
                kappa_max = max(kappa_max,u_tmp2(r)[1]);
                kappa_min = min(kappa_min,u_tmp2(r)[1]);
                c_max = max(c_max,sqrt(u_tmp2(r)[1]/u_tmp2(r)[0]));
                c_min = min(c_min,sqrt(u_tmp2(r)[1]/u_tmp2(r)[0]));
            }
            destroyElement(elem);
        }
        plot.celldata(u_tmp,3);
        NumberName("Vx",filename,step);
        plot.vtk_celldata(filename,0);
        NumberName("Vy",filename,step);
        plot.vtk_celldata(filename,1);
        NumberName("P",filename,step);
        plot.vtk_celldata(filename,2);
        if (problem.Exact()) {
            plot.celldata(u_int_tmp,4);
            NumberName("P0_int",filename,step);
            plot.vtk_celldata(filename,0);
            NumberName("P_int",filename,step);
            plot.vtk_celldata(filename,1);
        }
        rho_max = PPM->Max(rho_max);
        rho_min = PPM->Min(rho_min);
        kappa_max = PPM->Max(kappa_max);
        kappa_min = PPM->Min(kappa_min);
        c_max = PPM->Max(c_max);
        c_min = PPM->Min(c_min);
        if (step == 0) {
            plot.celldata(u_tmp2,3);
            plot.vtk_celldata("rho",0);
            plot.vtk_celldata("kappa",1);
            plot.vtk_celldata("load",2);
            vout(1) << "rho in   ["<< rho_min << "," << rho_max << "]" << endl
                    << "kappa in ["<< kappa_min << ","
                    << kappa_max << "]" << endl
                    << "c in ["<< c_min << "," << c_max << "]" << endl;
        }
    }
    void VtkPlotting (double t, const Vector& u, int step) {
        const Mesh& Mp = plot.GetMesh();
        const Mesh& Mu = u.GetMesh();
        if (Mp.Cells::size() != Mu.Cells::size()) return;
        vout(2) << "  => VtkPlotting result of step " << step << ".\n";
        char filename[128];
        std::string discName = disc.Name_Disc_dG();
        VtkPlotting_cell(t,u,filename,step,discName);
    }
    void PrintMatrixInfo (Matrix& A, int diagonal = 1) {
        for (cell c=A.cells(); c!=A.cells_end(); ++c) {
            DGVAcousticXElement* elem= getElem(disc,A,c);
            RowEntries_dG A_c(A,c,c);
            for (int i=0;i<elem->dimension(); ++i) {
                for (int j=0; j<elem->dimension(); ++j) {
                    mout << A_c(i,j) << " ";
                }
                mout << endl;
            }
            mout << endl;
            if(diagonal) continue;

            for (int f=0; f<c.Faces(); ++f) {
                cell cf = c;
                if(check_if_bnd(A,c,f))
                    continue;
                else
                    cf = find_nb_cell(A,c,f);
                RowEntries_dG A_cf(A,c,cf);
                for (int i=0;i<elem->dimension(); ++i) {
                    for (int j=0; j<elem->dimension(); ++j) {
                        mout << A_cf(i,j) << " ";
                    }
                    mout << endl;
                }
                mout << endl;
            }
            mout << "------------------------" << endl;
            destroyElement(elem);
        }
    }
    void Interpolate (const Vector& u, Vector& U) const {
        U = 0;
        int NN = 1 + u.dim();
        for (cell c=u.cells(); c!=u.cells_end(); ++c) {
            DGVAcousticXElement Elem(disc,u,c);
            row r = u.find_row(c());
            for (int i=0; i<c.Children(); ++i) {
                cell fc = U.find_cell(c.Child(i));
                DGVAcousticXElement Elem_fc(disc,u,fc);
                row fr = U.find_row(fc());
                Point z[MaxNodalPointsDG];
                Elem_fc.NodalPoints(z);
                for (int iN=0;iN<NN;++iN){
                    for (int i=0; i<Elem.shape_size(); ++i) {
                        Scalar s = Elem.SVValue(Elem.GlobalToLocal(z[i]),u,iN);
                        U(fr,iN*Elem.shape_size()+i) = s;
                    }
                }
            }
        }
        U.Accumulate();
    }
    double Goal_Functional (double t, const Vector& u) const {
        Scalar e = 0.0;
        Scalar eq = 0.0;
        
        ExchangeBuffer E;
        std::fstream p;
        p.open("pktVerf.dat", ios::out);
        
        for (cell c=u.cells(); c!=u.cells_end(); ++c){
            Scalar Pxx(0.0);
            VectorField Vxx;
            DGVAcousticXElement elem(disc,u,c);
            row r = u.find_row(c());
            for (int q=0; q<elem.nQ(); ++q) {
                Point QP = elem.QPoint(q);
                Pxx += elem.QWeight(q) * elem.Pressure(q,u);
                Vxx += elem.QWeight(q) * elem.Velocity(q,u);
                if ((GF_a[0]>QP[0]) || (QP[0]>GF_b[0])) continue;
                if ((GF_a[1]>QP[1]) || (QP[1]>GF_b[1])) continue;
                double w  = elem.QWeight(q);
                VectorField j_V;
                j_V[0] =  problem.ut_dual(t,QP,0,GF_a,GF_b);
                j_V[1] =  problem.ut_dual(t,QP,1,GF_a,GF_b);
                Scalar j_P = problem.ut_dual(t,QP,2,GF_a,GF_b);
                VectorField V = elem.Velocity(q,u);
                Scalar P = elem.Pressure(q,u);
                e += w * (j_P*P + j_V*V);
                eq += 0.5 * w * (P*P + V*V);
            }
            E.Send(0) << c()[0] << c()[1] << Pxx << Vxx[0] << Vxx[1];
            
            Point c0( 0.5*(c.Corner(0)+c.Corner(1)) );
            Point c1( 0.5*(c.Corner(1)+c.Corner(2)) );
            Point c2( 0.5*(c.Corner(2)+c.Corner(3)) );
            Point c3( 0.5*(c.Corner(3)+c.Corner(0)) );
            
            Point p0(0.5*(c.Corner(0)[0]+c()[0]),0.5*(c.Corner(0)[1]+c()[1]));
            Point p1(0.5*(c.Corner(1)[0]+c()[0]),0.5*(c.Corner(1)[1]+c()[1]));
            Point p2(0.5*(c.Corner(2)[0]+c()[0]),0.5*(c.Corner(2)[1]+c()[1]));
            Point p3(0.5*(c.Corner(3)[0]+c()[0]),0.5*(c.Corner(3)[1]+c()[1]));
            
            //p << setw(15)<<c()[0] <<setw(15)<<c()[1]
                        //<< p0[0] << p0[1]
                        //<< p1[0] << p1[1]
                        //<< p2[0] << p2[1]
                        //<< p3[0] << p3[1]
                    
             p      <<setw(15)<<0.5*(c.Corner(0)+p0)[0] <<setw(15)<<0.5*(c.Corner(0)+p0)[1] << setw(15)<<c()[0] <<setw(15)<<c()[1] <<endl
                    <<setw(15)<<0.5*(c0+p0)[0]          <<setw(15)<<0.5*(c0+p0)[1]          << setw(15)<<c()[0] <<setw(15)<<c()[1] <<endl
                    <<setw(15)<<0.5*(c()+p0)[0]         <<setw(15)<<0.5*(c()+p0)[1]         << setw(15)<<c()[0] <<setw(15)<<c()[1] <<endl
                    <<setw(15)<<0.5*(c3+p0)[0]          <<setw(15)<<0.5*(c3+p0)[1]          << setw(15)<<c()[0] <<setw(15)<<c()[1] <<endl
                    
                    <<setw(15)<<0.5*(c0+p1)[0]          <<setw(15)<<0.5*(c0+p1)[1]          << setw(15)<<c()[0] <<setw(15)<<c()[1] <<endl
                    <<setw(15)<<0.5*(c.Corner(1)+p1)[0] <<setw(15)<<0.5*(c.Corner(1)+p1)[1] << setw(15)<<c()[0] <<setw(15)<<c()[1] <<endl
                    <<setw(15)<<0.5*(c1+p1)[0]          <<setw(15)<<0.5*(c1+p1)[1]          << setw(15)<<c()[0] <<setw(15)<<c()[1] <<endl
                    <<setw(15)<<0.5*(c()+p1)[0]         <<setw(15)<<0.5*(c()+p1)[1]         << setw(15)<<c()[0] <<setw(15)<<c()[1] <<endl
                    
                    <<setw(15)<<0.5*(c()+p2)[0]         <<setw(15)<<0.5*(c()+p2)[1]         << setw(15)<<c()[0] <<setw(15)<<c()[1] <<endl
                    <<setw(15)<<0.5*(c1+p2)[0]          <<setw(15)<<0.5*(c1+p2)[1]          << setw(15)<<c()[0] <<setw(15)<<c()[1] <<endl
                    <<setw(15)<<0.5*(c.Corner(2)+p2)[0] <<setw(15)<<0.5*(c.Corner(2)+p2)[1] << setw(15)<<c()[0] <<setw(15)<<c()[1] <<endl
                    <<setw(15)<<0.5*(c2+p2)[0]          <<setw(15)<<0.5*(c2+p2)[1]          << setw(15)<<c()[0] <<setw(15)<<c()[1] <<endl
                    
                    <<setw(15)<<0.5*(c3+p3)[0]          <<setw(15)<<0.5*(c3+p3)[1]          << setw(15)<<c()[0] <<setw(15)<<c()[1] <<endl
                    <<setw(15)<<0.5*(c()+p3)[0]         <<setw(15)<<0.5*(c()+p3)[1]         << setw(15)<<c()[0] <<setw(15)<<c()[1] <<endl
                    <<setw(15)<<0.5*(c2+p3)[0]          <<setw(15)<<0.5*(c2+p3)[1]          << setw(15)<<c()[0] <<setw(15)<<c()[1] <<endl
                    <<setw(15)<<0.5*(c.Corner(3)+p3)[0] <<setw(15)<<0.5*(c.Corner(3)+p3)[1] << setw(15)<<c()[0] <<setw(15)<<c()[1] <<endl
                    ;
        }
        p.close();
        
        E.Communicate();
        if (PPM->master() ) {
            std::fstream f;
            f.open("test.dat", ios::out);
            for (short q=0; q<PPM->size(); ++q) {
                while (E.Receive(q).size() < E.ReceiveSize(q)) {
                   std::vector<double> Dat(5);
                        E.Receive(q) >>Dat[0] >>Dat[1] >>Dat[2] >>Dat[3] >>Dat[4];
                        f<< setw(15) << left << Dat[0] <<setw(15)<< Dat[1] <<setw(15)<< Dat[2] <<setw(15)<< Dat[3]
                         <<setw(15)<< Dat[4] << endl;
                }
            }
            f.close();
        }
        
        mout<<"GoalFunctional(quadratic): " << abs(PPM->Sum(eq)) <<endl;
        return abs(PPM->Sum(e));
    }
};

#endif
