#ifndef _DGVISCOELASTIC_H_
#define _DGVISCOELASTIC_H_

#include "m++.hpp"
#include "DGMatrixGraph.hpp"
#include "DGTSHelper.hpp"
#include <math.h>
#include <fstream>
#include <stdlib.h>
#include <time.h>
#include <stdio.h>
const Tensor One2d(1,0,0,1);
static inline void HookeTensorCml_loc (double m,double l, Tensor& sigma)
{
    double tr_tmp= trace(sigma);
    sigma *= 2.0*m;
    sigma += l * tr_tmp*One2d;

}
static inline void HookeTensorCmlInv_loc(double m,double l,Tensor& sigma)
{
    int dim = 2;
    double m_inv =  1.0/(4*m);
    double l_inv = - l/(2.0*m*(dim*l+2*m));
    HookeTensorCml_loc(m_inv,l_inv,sigma);
}
class ProblemVElastic {
protected:
    int verbose;
    int dim;
    int numL;
    int stressElements;
    std::string meshname;
public:
    ProblemVElastic (int _numL = -1, int d=2) :
        verbose(0),dim(d), numL(_numL),stressElements(-1),
        meshname("none") {
        Config::Get("ProblemVerbose", verbose);
        Config::Get("Mesh", meshname);
        if (dim == 1) stressElements = 1;
        if (dim == 2) stressElements = 3;
        if (dim == 3) Exit("code needs inspection for 3d, e.g. the Neumann  part");
    }
    int getDim() const{return dim;}
    int getNumL() const{return numL;}
    virtual bool Exact () const { return false; }
    virtual bool RHS(double) const { return false; }
    virtual VectorField ForceV (double, const cell&, const Point&, int =-1) const{
        Exit("implement force")
        return zero;
    }
    virtual Tensor ForceS(double, const cell&, const Point&, int, int =-1) const{
        Exit("implement force")
        return Zero;
    }
    VectorField Neumann (double t, const cell& c,const Point x,
                         const VectorField& N) const{
        Tensor S = Zero;
        for (int i=0; i<=numL; ++i){
            S[0][0] += ut(t,c,x,dim + i*stressElements);
            S[1][1] += ut(t,c,x,dim + 1 + i*stressElements);
            S[0][1] += ut(t,c,x,dim + 2 + i*stressElements);
        }
        S[1][0] =S[0][1];
        return S * N;
    }

    VectorField Dirichlet (double t, const cell& c,const Point x) const{
        VectorField V = zero;
        for (int d=0; d<dim; ++d){
            V[d] = ut(t,c,x,d);
        }
        return V;
    }
    virtual double ut (double,const cell&, const Point&, int) const { return 0; }
    virtual double ut_dual (double &,const cell& , const Point&, int,
                            const Point&, const Point&) const { return 0; }
    virtual double u0 (const cell& c,const Point& z, int i) const { return ut(0,c,z,i); }

    virtual double Rho (const cell&, const Point& ) const {Exit("not implemented"); return 0.0; }
    double Mu (const cell& c, const Point& z, int var = -1) const{
        if (var == -1){
            double mu = 0;
            for (int i=0; i<=numL;++i)
            {
                mu += Mu_i(c,z,i);
            }
            return mu; // Sum of mus
        }else{
            return Mu_i(c,z,var);
        }
    }
    virtual double Mu_i (const cell&, const Point&, int) const{
        Exit("not implemented"); return 0.0;
    }
    double Lambda (const cell& c,const Point& z, int var =-1) const{
        if (var == -1){
            double lambda=0;
            for (int i=0; i<=numL;++i){
                lambda += Lambda_i(c,z,i);
            }
            return lambda;
        }else{
            return Lambda_i(c,z,var);
        }
    }
    virtual double Lambda_i(const cell&,const Point&, int) const{Exit("not implemented"); return 0.0;}
    virtual double Tau(const cell&, const Point&, int) const{Exit("not implemented"); return 0.0;}
    virtual double C_s (const cell& c,const Point& z) const { return sqrt(Mu(c,z) / Rho(c,z)); }
    virtual double C_p (const cell& c,const Point& z) const { return sqrt((2.0*Mu(c,z) + Lambda(c,z)) / Rho(c,z)); }
    virtual const char* Name () const {Exit("not implented"); return ""; }
    virtual const char* MeshName () const { return meshname.c_str(); }
    virtual ~ProblemVElastic(){ }
};

//static DoF* GetDGVElasticDoF (ProblemVElastic* problem,int cell_std_deg) {
//    int dim = problem->getDim();
//    int m = dim + (1+problem->getNumL())*std::max(1,3*dim-3);
//    DoF* ElasticDoF =GetDoF("cellconstdeg",m,problem->getDim(),cell_std_deg);
//    return ElasticDoF;
//}
//
//static DoF* GetDoF (ProblemVElastic* problem,int cell_std_deg) {
//    int dim = problem->getDim();
//    int m = dim + (1+problem->getNumL())*std::max(1,3*dim-3);
//    DoF* ElasticDoF =GetDoF("cellconstdeg",m,problem->getDim(),cell_std_deg);
//    return ElasticDoF;
//}

class DGVElasticElement : public Element {
    int dim;
    int deg;
    int numL;
    const Shape& S;
    const ShapeValues<double>& value;
    VectorShapeValuesDG gradient;
    VectorShapeValuesDG velocity;
    TensorShapeValuesDG strain;
    TensorShapeValuesDG stress;
    VectorShapeValuesDG divstress;
    const cell C;
    const row& r (int i) const { return (*this)[i]; }
public:
    Point GlobalToLocal (const Point& z) const {
        switch (C.Type()) {
        case TRIANGLE: {
            Tensor T;
            T[0][0] = C.Corner(1)[0]-C.Corner(0)[0];
            T[0][1] = C.Corner(2)[0]-C.Corner(0)[0];
            T[1][0] = C.Corner(1)[1]-C.Corner(0)[1];
            T[1][1] = C.Corner(2)[1]-C.Corner(0)[1];
            T[0][2] = 0;
            T[1][2] = 0;
            T[2][0] = 0;
            T[2][1] = 0;
            T[2][2] = 1;
            Tensor IT = Invert(T);
            return IT * (z - C.Corner(0));
        }
        case TETRAHEDRON:
            Exit("not implemented");
        case QUADRILATERAL: {
            Tensor T;
            T[0][0] = C.Corner(1)[0]-C.Corner(0)[0];
            T[0][1] = C.Corner(3)[0]-C.Corner(0)[0];
            T[1][0] = C.Corner(1)[1]-C.Corner(0)[1];
            T[1][1] = C.Corner(3)[1]-C.Corner(0)[1];
            T[0][2] = 0;
            T[1][2] = 0;
            T[2][0] = 0;
            T[2][1] = 0;
            T[2][2] = 1;
            Tensor IT = Invert(T);
            return IT * (z - C.Corner(0));
        }
        case HEXAHEDRON:
            Exit("not implemented");
        default:
            Exit("not implemented");
        }
        Exit("Not implemented!"); return Origin;
    }
    void NodalPoints (vector<Point> z) const{S.NodalPoints(C, z);}
    int get_deg() { return deg; }
    DGVElasticElement(const DGDiscretization &disc,
                      const VectorMatrixBase &g, const cell &c, int _numL) :
            Element(g, c),
            dim(c.dim()),
            deg(g.GetDoF(0).get_cell_deg(c)),
            numL(_numL),
            S(disc.GetShape(c, g.GetDoF(0).get_cell_deg(c))),
            value(S.values()),
            gradient(nQ(), S.size()),
            velocity(nQ(), c.dim() * S.size()),
            strain(nQ(), c.dim() * S.size()),
            stress(nQ(), std::max(1, 3 * c.dim() - 3) * S.size()),
        divstress(nQ(),std::max(1,3*c.dim()-3)*S.size()),
        C(c){
        for (int q=0; q<nQ(); ++q) {
            for (int i=0; i<S.size(); ++i)
                gradient[q][i] = T[q] * S.LocalGradient(q,i);
            for (int k=0; k<dim; ++k) {
                VectorField V = zero;
                V[k] = 1.0;
                for (int i=0; i<S.size(); ++i) {
                    velocity[q][i+k*S.size()][k] = value[q][i] * V[k];
                    Tensor Dv = Zero;
                    Dv[k] = gradient[q][i];
                    strain[q][i+k*S.size()] = sym(Dv);
                }
            }

            for (int i=0; i<S.size(); ++i) {
                for (int k=0; k<dim; ++k) {
                    stress[q][i+k*S.size()] = Zero;
                    stress[q][i+k*S.size()][k][k] = value[q][i];
                    divstress[q][i+k*S.size()] = zero;
                    divstress[q][i+k*S.size()][k] = gradient[q][i][k];
                }
                if (dim == 2) {
                    stress[q][i+2*S.size()] = Zero;
                    stress[q][i+2*S.size()][0][1] = value[q][i];
                    stress[q][i+2*S.size()][1][0] = value[q][i];
                    divstress[q][i+2*S.size()][0] = gradient[q][i][1];
                    divstress[q][i+2*S.size()][1] = gradient[q][i][0];
                }
                if (dim == 3) {
                    Exit("Check");
                    stress[q][i+3*S.size()] = Zero;
                    stress[q][i+3*S.size()][0][1] = value[q][i];
                    stress[q][i+3*S.size()][1][0] = value[q][i];
                    stress[q][i+4*S.size()] = Zero;
                    stress[q][i+4*S.size()][0][2] = value[q][i];
                    stress[q][i+4*S.size()][2][0] = value[q][i];
                    stress[q][i+5*S.size()] = Zero;
                    stress[q][i+5*S.size()][1][3] = value[q][i];
                    stress[q][i+5*S.size()][3][1] = value[q][i];
                }
            }
        }
    }
    double QWeight (int q) const { return qWeight[q]; }
    const Point& QPoint (int q) const { return qPoint[q]; }
    double Area () const {
        double a = 0;
        for (int q=0; q<nQ(); ++q)
            a += qWeight[q];
        return a;
    }
    int shape_size () const { return S.size(); }
    int dimension () const {return V_dimension()+(1+numL)*S_dimension();}
    int u_dimension () const {return dimension()/shape_size();}
    int V_dimension () const { return dim * shape_size(); }
    int S_dimension () const {
        if (dim==3) return 6 * shape_size();
        if (dim==2) return 3 * shape_size();
        Exit("Not implemented!"); return 0;
    }
    int getNumL () const { return numL; }
    VectorField Velocity (int q, int i) const { return velocity[q][i]; }
    Tensor Stress (int q, int i) { return stress[q][i]; }
    VectorField DivStress (int q, int i) { return divstress[q][i]; }
    Tensor Strain (int q, int i) { return strain[q][i]; }
    VectorField divStress(int q, const Vector& u) {
        VectorField S_div = zero;
        for (int i=0; i<S_dimension(); ++i)
            S_div += u(r(0),i) * divstress[q][i];
        return S_div;

    }
    VectorField Velocity (int q, const Vector& u) const  {
        VectorField V = zero;
        for (int i=0; i<V_dimension(); ++i)
            V += u(r(0),i) * velocity[q][i];
        return V;
    }
    Tensor Stress (int q, const Vector& u) const {
        Tensor S = Zero;
        for (int n=0; n<1+numL; ++n)
            S += Stress_i(q,u,n);
        return S;
    }
    Tensor Stress_i (int q, const Vector& u,int j) const {
        Tensor S = 0;
        for (int i=0; i<S_dimension(); ++i)
            S += u(r(0),V_dimension()+j*S_dimension()+i) * stress[q][i];
        return S;
    }
    Scalar SVValue (const Point& lz, const Vector& u, int k) const {
        Scalar v = 0;
        for (int i=0; i<S.size(); ++i)
            v += u(C(),k*S.size()+i) * S(lz,i);
        return v;
    }

    ~DGVElasticElement(){ }
};

class DGVElasticFaceElement : public FaceElement {
    const cell& c;
    int dim;
    face f;
    int fid;
    Point qTangent[MaxQuadraturePointsDG];
    const Shape& S;
    VectorShapeValuesDG velocity;
    TensorShapeValuesDG stress;
    ShapeValues<double> value;
    int numL;
public:
    DGVElasticFaceElement(const DGDiscretization &disc,
                          const VectorMatrixBase &g,
                          const cell &_c, int f_id, int _numL) :
        FaceElement(g,_c,f_id),
        c(_c),
        S(disc.GetShape(c,g.GetDoF(0).get_cell_deg(c))),
        dim(c.dim()),
        f(g.GetMesh().find_face(c.Face(f_id))),
        fid(f_id),
        numL(_numL),
        velocity(nQ(),dim*S.size()),
        stress(nQ(),std::max(1,3*dim-3)*S.size()) {
//        double ww = 0;
//        for (int q=0; q<nQ(); ++q){
//            ww += QWeight(q);
//        }
//        ww = c.LocalFaceArea(fid) / ww;
        for (int q=0; q<nQ(); ++q) {
            qTangent[q] = c.FaceCorner(f_id,1) - c.FaceCorner(f_id,0);
            double w = norm(qTangent[q]);
            qTangent[q] *= (1/w);
        }
        for (int q=0; q<nQ(); ++q) {
            for (int i=0; i<S.size(); ++i)
                value[q][i] = S(QLocal(q),i);
            for (int k=0; k<dim; ++k) {
                VectorField V = zero;
                V[k] = 1.0;
                for (int i=0; i<S.size(); ++i) {
                    velocity[q][i+k*S.size()] = zero;
                    velocity[q][i+k*S.size()][k] = value[q][i] * V[k];
                }
            }
            for (int k=0; k<dim; ++k) {
                for (int i=0; i<S.size(); ++i) {
                    stress[q][i+k*S.size()] = Zero;
                    stress[q][i+k*S.size()][k][k] = value[q][i];
                    if (dim == 2) {
                        stress[q][i+2*S.size()] = Zero;
                        stress[q][i+2*S.size()][0][1] = value[q][i];
                        stress[q][i+2*S.size()][1][0] = value[q][i];
                    }
                    if (dim == 3) {
                        stress[q][i+3*S.size()] = Zero;
                        stress[q][i+3*S.size()][0][1] = value[q][i];
                        stress[q][i+3*S.size()][1][0] = value[q][i];
                        stress[q][i+4*S.size()] = Zero;
                        stress[q][i+4*S.size()][0][2] = value[q][i];
                        stress[q][i+4*S.size()][2][0] = value[q][i];
                        stress[q][i+5*S.size()] = Zero;
                        stress[q][i+5*S.size()][1][3] = value[q][i];
                        stress[q][i+5*S.size()][3][1] = value[q][i];
                    }
                }
            }
        }
    }
    const Point& QTangent (int q) const { return qTangent[q]; }
    int shape_size () const { return S.size(); }
    int dimension () const { return V_dimension()+S_dimension(); }
    int V_dimension () const { return dim * shape_size(); }
    int S_dimension () const { return std::max(1,3*dim-3)* shape_size();
                               if (dim==3) return 6 * shape_size();
                               if (dim==2) return 3 * shape_size();
                             }

    VectorField Velocity (int q, int i) const  { return velocity[q][i]; }
    Tensor Stress (int q, int i) { return stress[q][i]; }
};

class ProblemVElastic_old : public ProblemVElastic {
protected:
    int verbose;
    double rho, mu, lambda, tau;
    std::string meshname;
public:
    ProblemVElastic_old (int numL = 1,int dim = 2) :ProblemVElastic(numL,dim),
        verbose(0),
        rho(1.0), mu(1.0), lambda(1.0), tau(1.0),
        meshname("none") {
        Config::Get("ProblemVerbose", verbose);
        Config::Get("lambda", lambda);
        Config::Get("mu", mu);
        Config::Get("rho", rho);
        Config::Get("tau", tau);
        Config::Get("Mesh", meshname);
    }
    double Mu (const Point& x, int var = -1) const {
        if (var == -1) return mu;
        if (var ==  1) return mu+tau;
        else         return mu;
    }
    double Lambda (const Point& x) const { return lambda; }
    virtual double Rho (const cell&, const Point& x) const { return rho; }
    virtual double Mu (const cell& c, const Point& x, int var = -1) const {
        if (var == -1) return mu;
        if (var ==  1) return mu+tau;
        else         return mu;
    }
    virtual double Mu(const cell&,Point& x) const{return mu;}
    virtual double Lambda (const cell&,const Point& x, int var =-1) const{return lambda;}
    virtual double Tau(const cell&, const Point&, int =-1) const{return tau;}
    virtual double C_s (const cell& c,const Point& z) const { return sqrt(Mu(c,z) / Rho(c,z)); }
    virtual double C_p (const cell& c,const Point& z) const { return sqrt((2.0*Mu(c,z) + Lambda(c,z)) / Rho(c,z)); }

};

class VElastic_Spherical_test : public ProblemVElastic {
    Point Pmid;
    int linear_weight;
    double width;
    double mu;
    double rho;
    double lambda;
    double tau;
public:
    VElastic_Spherical_test () :
        ProblemVElastic(1,2), Pmid(0.5,1.0), width(0.25), linear_weight(1), mu(1),rho(1),lambda(1), tau(1) {
        Config::Get("ProblemVerbose", verbose);
        Config::Get("lambda", lambda);
        Config::Get("mu", mu);
        Config::Get("rho", rho);
        Config::Get("tau", tau);
        Config::Get("Mesh", meshname);
    }

    const char* Name () const { return "VElastic Spherical Wave"; }

    double A0 (double s) const {
        if (s<width) return pow(cos(s*0.5*M_PI/width),6.0);
        else return 0.0;
    }
    double A (const Point& x) const {
        double r = dist(Pmid,x);
        return A0(r*r);
    }
    double ut (double t, const cell& c,const Point& x, int i) const {
        if (t > 0) return 0.0;
        const double tmp = A(x);
        switch (i) {
        case 0 : return (x[0]-Pmid[0])*tmp;
        case 1 : return (x[1]-Pmid[1])*tmp;
        default: return 0.0;
        }
    }
    virtual double ut_dual (double & t,const cell& c, const Point& x, int i,
                            const Point& a, const Point& b) const {
        if ( x[0]<a[0] || x[1]<a[1] || x[0]>b[0] || x[1]>b[1] ) return 0.0;
        switch (linear_weight){
        case 1: {
            Point ba = b-a;
            if (i!=0) return 0.0;
            else return 1.0/(ba[0]*ba[1]);
        }
        default: Exit("Not implemented");
        }
        return 0.0;
    }
    double Mu_i(const cell &c, const Point &x, int var) const
    {
        switch(var){
        case 0: return mu;
        case 1: return tau*mu;
        }
        Exit("Not implemented"); return 0.0;
    }
    double Lambda_i(const cell &c, const Point &x, int var) const
    {
        switch(var)
        {
        case 0: return lambda;
        case 1: return tau*lambda;
        }
        Exit("Not implemented"); return 0.0;
    }
    double Rho(const cell &, const Point &x) const
    {
        return rho;
    }
    double Tau(const cell &, const Point &x, int var) const
    {
        return tau;
    }
};
class VElasticConstant : public ProblemVElastic
{
    double lambda0,lambda1,mu0,mu1,rho,tau;
public:
    VElasticConstant() : ProblemVElastic(0,2), lambda0(1),lambda1(2),mu0(1),mu1(2),rho(1),tau(1)
    {
        Config::Get("ProblemVerbose", verbose);
    }
    const char* Name() const
    {
        return "VElasticConstant";
    }
    bool Exact() const
    {
        return true;
    }
    double Mu_i(const cell &c, const Point &x, int var) const
    {
        switch(var){
        case 0: return mu0;
        case 1: return mu1;
        }
        Exit("Not implemented"); return 0.0;
    }
    double Lambda_i(const cell &c, const Point &x, int var) const
    {
        switch(var)
        {
        case 0: return lambda0;
        case 1: return lambda1;
        }
        Exit("Not implemented"); return 0.0;
    }
    double Rho(const cell &, const Point &x) const
    {
        return rho;
    }
    double Tau(const cell &, const Point &x, int var) const
    {
        return tau;
    }
    double ut (double t ,const cell& c, const Point& x , int i) const
    {
        if(i<dim)
        {
            return 1.0;
        } else
        {
            return 1.0;
        }

    }
    VectorField ForceV (double t,const cell&c, const Point& z, int q=-1) const
    {
        return zero;
    }
    Tensor ForceS(double t,const cell&c, const Point& x, int i, int q=-1) const
    {
        if (i == 0)
        {
            return Zero;
        } else if (i == 1)
        {
            Tensor S(1,1,1,1);
            HookeTensorCmlInv_loc(mu1,lambda1,S);
            S*= 1.0/tau;
            return S;
        }
        Exit("Not implemented"); return Zero;
    }
};

static ProblemVElastic* GetProblemVElastic (const std::string& name ) {
    if (name == "Spherical")
        return new VElastic_Spherical_test();
    if(name == "VElasticConstant")
        return new VElasticConstant();
    else Exit("Error: no wave found!");
    return nullptr;
}
static ProblemVElastic* GetProblem (const std::string& name, ProblemVElastic ) {
    if (name == "Spherical")
        return new VElastic_Spherical_test();
    if (name == "VElasticConstant")
        return new VElasticConstant();
    else Exit("Error: no wave found!");
    return nullptr;
}


class DGVElasticAssembleMatrix : public DGAssemble {
protected:
    const DGDiscretization& disc;
    const ProblemVElastic& problem;
    double flux_alpha;
    std::string problem_name;
    Point GF_a;
    Point GF_b;
     std::unordered_map<Point,DGVElasticElement*>* EleCache;
    bool upwind;
    int numL;
    bool clearcache;
public:
    DGVElasticAssembleMatrix (const DGDiscretization& D,
                              const ProblemVElastic& P, Plot& _plot) :
        DGAssemble(_plot), disc(D), problem(P), flux_alpha(1),numL(problem.getNumL()),
        GF_a(0.0,0.0), GF_b(1.0,1.0), upwind(true),clearcache(true),EleCache(nullptr) {
        Config::Get("flux_alpha", flux_alpha);
        Config::Get("Problem", problem_name);
        Config::Get("roi_min", GF_a);
        Config::Get("roi_max", GF_b);
        Config::Get("upwind", upwind);
        Config::Get("clearcache", clearcache);
    }
    ~DGVElasticAssembleMatrix (){
        if(clearcache){
            for (auto it = EleCache->begin(); it!=EleCache->end(); ++it){
                delete it->second;
            }
        }
    }
    void SetElementCache( std::unordered_map<Point,DGVElasticElement*>* _EleCache){
        EleCache = _EleCache;
    }

    void PrintInfo(const Vector &u) {
        IAssemble::PrintInfo(u);
        vout(1) << " Mesh            " << problem.MeshName() << endl
                << " Problem         " << problem.Name() << endl << endl;
    }

    void initializeEleCache(const VectorMatrixBase &mgr) {
        for (auto it = EleCache->begin(); it != EleCache->end(); ++it) {
            delete it->second;
        }
        EleCache->clear();

        for (cell c = mgr.cells(); c != mgr.cells_end(); ++c) {
            (*EleCache)[c()] = new DGVElasticElement(disc, mgr, c, numL);
        }
    }

    DGVElasticElement *getElem(const DGDiscretization &disc,
                               const VectorMatrixBase &g, const cell &c) const {
        if (EleCache == nullptr)
            return new DGVElasticElement(disc, g, c, numL);
        else {
            auto p = EleCache->find(c());
            return p->second;
        }
    }

    void destroyElement(DGVElasticElement *elem) const {
        if (EleCache == nullptr) {
            delete elem;
        }
    }

    const char* Name() const { return "DGAssembleMatrix"; }
    void MassMatrix (Matrix& M) const {
        M = 0;
        for (cell c=M.cells(); c!=M.cells_end(); ++c) {
            DGVElasticElement* elem= getElem(disc,M,c);
            RowEntries_dG M_c(M,c,c);
            const int nV = elem->V_dimension();
            const int nS = elem->S_dimension();
            const double rho = problem.Rho(c,c());
            for (int q=0; q<elem->nQ(); ++q) {
                double w  = elem->QWeight(q);
                for (int i=0; i<nV; ++i) {
                    VectorField V_i = elem->Velocity(q,i);
                    for (int j=0; j<nV; ++j) {
                        VectorField V_j = elem->Velocity(q,j);
                        M_c(i,j) += w * rho * V_i * V_j;
                    }
                }
                for (int nEta = 0; nEta<=elem->getNumL(); ++nEta) {
                    const double lambda = problem.Lambda(c,c(),nEta);
                    const double mu = problem.Mu(c,c(),nEta);
                    const double s = 1 / (2.0 * mu);
                    const double ss = - lambda / (2.0 * mu * (2.0 * lambda + 2.0 * mu));
                    for (int i=0; i<nS; ++i) {
                        Tensor S_i = elem->Stress(q,i);
                        for (int j=0; j<nS; ++j) {
                            Tensor S_j = elem->Stress(q,j);
                            M_c(nV+nEta*nS+i,nV+nEta*nS+j)
                                    += w * (s*Frobenius(S_i,S_j) + ss*trace(S_i)*trace(S_j));
                        }
                    }
                }
            }
            destroyElement(elem);
            /*if (c()[0]==0.0625 && c()[1]==0.0625){
                std::cout<<"Blablabla"<<endl;
                for(int i=0;i<16; i++) {
                    for(int j=0;j<16; j++) {
                        std::cout<<"  "<<M_c(i,j);
                    }
                    std::cout<<endl;
                }
            }*/
        }
        return;
    }
    void FluxMatrix (Matrix& A) const {
        A = 0;
        for (cell c=A.cells(); c!=A.cells_end(); ++c) {
            DGVElasticElement* elem = getElem(disc,A,c);
            const int nV = elem->V_dimension();
            const int nS = elem->S_dimension();
            RowEntries_dG A_c(A,c,c);
            for (int q=0; q<elem->nQ(); ++q) {
                double w  = elem->QWeight(q);

                for (int i=0; i<nV; ++i) {
                    VectorField V_i = elem->Velocity(q,i);
                    for (int nEta = 0; nEta<=numL; ++nEta) {
                        for (int j=0; j<nS; ++j) {
                            VectorField DivS_j = elem->DivStress(q,j);
                            A_c(i,nV+nEta*nS+j) += w * DivS_j * V_i;
                        }
                    }
                }
                for (int nEta = 0; nEta<=numL; ++nEta){
                    for (int i=0; i<nS; ++i) {
                        Tensor S_i = elem->Stress(q,i);
                        for (int j=0; j<nV; ++j) {
                            Tensor E_j = elem->Strain(q,j);
                            A_c(nV+nEta*nS+i,j) += w * Frobenius(E_j,S_i);
                        }
                    }
                }
                for (int nEta = 1; nEta<=numL; ++nEta) {
                    double mu = problem.Mu(c, c(),nEta);
                    double s = 1 / (2.0 * mu);
                    double lambda = problem.Lambda(c, c(),nEta );
                    double ss = - lambda / (2.0 * mu * (2.0 * lambda + 2.0 * mu));
                    double tau = problem.Tau(c, c(),nEta );
                    for (int i=0; i<nS; ++i) {
                        Tensor S_i = elem->Stress(q,i);
                        for (int j=0; j<nS; ++j) {
                            Tensor S_j = elem->Stress(q,j);
                            A_c(nV+nEta*nS+i,nV+nEta*nS+j) -= w / tau * (
                                        s*Frobenius(S_j,S_i) + ss*trace(S_j)*trace(S_i) );
                        }
                    }
                }
            }
            for (int f=0; f<c.Faces(); ++f) {
                DGVElasticFaceElement felem(disc,A,c,f,numL);
                const int nV = felem.V_dimension();
                const int nS = felem.S_dimension();

                if (check_if_bnd(A,c,f)) {
                    BFParts bnd(A.GetMesh(),c);
                    const double c_s   = problem.C_s(c,c());
                    const double c_p   = problem.C_p(c,c());
                    const double rho   = problem.Rho(c,c());
                    for (int q=0; q<felem.nQ(); ++q) {
                        double w  = felem.QWeight(q);
                        VectorField Nq = felem.QNormal(q);
                        VectorField Tq = felem.QTangent(q);
                        if (bnd[f] == 2 || bnd[f] == 0) {
                            const double alpha1 = 1.0 / ( rho*c_p );
                            const double alpha5 = 1.0 / (rho*c_s );
                            for (int i=0; i<nV; ++i) {
                                VectorField V_i = felem.Velocity(q,i);
                                Scalar NV = Nq * V_i;
                                Scalar TV = Tq * V_i;
                                for (int nEta = 0; nEta<=numL; ++nEta) {
                                    for (int j=0; j<nS; ++j) {
                                        Tensor S_j = felem.Stress(q,j);
                                        VectorField NS_j = S_j * Nq;
                                        A_c(i,nV+nEta*nS+j)
                                                -= w* (NV*(NS_j*Nq) + TV*(NS_j*Tq));
                                    }
                                }
                            }
                            for (int nEta = 0; nEta<=numL; ++nEta) {
                                for (int i=0; i<nS; ++i) {
                                    Tensor S_i = felem.Stress(q,i);
                                    VectorField NS = S_i * Nq;
                                    Scalar NSN = NS * Nq;
                                    Scalar NST = NS * Tq;
                                    for (int nEta2 = 0; nEta2<=numL; ++nEta2)
                                    {
                                        for (int j=0; j<nS; ++j) {
                                            Tensor S_j = felem.Stress(q,j);
                                            VectorField NS_j = S_j * Nq;
                                            A_c(nV+nEta*nS+i,nV+nEta2*nS+j)
                                                    -= w * (alpha1 * NSN * (NS_j*Nq)
                                                            + alpha5 * NST * (NS_j*Tq));
                                        }
                                    }
                                }
                            }
                        } else if (bnd[f] == 1) {
                            const double alpha4 = rho*c_p;
                            const double alpha8 = rho*c_s;
                            for (int i=0; i<nV; ++i) {
                                VectorField V_i = felem.Velocity(q,i);
                                Scalar NV = Nq * V_i;
                                Scalar TV = Tq * V_i;
                                for (int j=0; j<nV; ++j) {
                                    VectorField V_j = felem.Velocity(q,j);
                                    A_c(i,j) -= w * (alpha4 * NV * (Nq*V_j)
                                                     + alpha8 * TV * (Tq*V_j));
                                }
                            }
                            for (int nEta = 0; nEta<=numL; ++nEta) {
                                for (int i=0; i<nS; ++i) {
                                    Tensor S_i = felem.Stress(q,i);
                                    VectorField NS = S_i * Nq;
                                    Scalar NSN = NS * Nq;
                                    Scalar NST = NS * Tq;
                                    for (int j=0; j<nV; ++j) {
                                        VectorField V_j = felem.Velocity(q,j);
                                        A_c(nV+nEta*nS+i,j) -= w * ( NSN * (Nq*V_j)
                                                                     + NST * (Tq*V_j));
                                    }
                                }
                            }
                        }
                    }
                } else {
                    cell cf = find_nb_cell(A,c,f);
                    int f1 = find_nb_face_id(A,c.Face(f),cf);
                    DGVElasticFaceElement felem_1(disc,A,cf,f1,numL);
                    RowEntries_dG A_cf(A,c,cf);
                    const double c_s   = problem.C_s(c,c());
                    const double c_p   = problem.C_p(c,c());
                    const double rho   = problem.Rho(c,c());
                    const double c_s_f = problem.C_s(c,cf());
                    const double c_p_f = problem.C_p(c,cf());
                    const double rho_f = problem.Rho(c,cf());
                    const double alpha1 = 1.0 / ( rho*c_p + rho_f*c_p_f );
                    const double alpha2 = rho*c_p * alpha1;
                    const double alpha3 = rho_f*c_p_f * alpha1;
                    const double alpha4 = (rho*c_p*rho_f*c_p_f) * alpha1;

                    const double alpha5 = 1.0 / ( rho*c_s + rho_f*c_s_f );
                    const double alpha6 = rho*c_s * alpha5;
                    const double alpha7 = rho_f*c_s_f * alpha5;
                    const double alpha8 = (rho*c_s*rho_f*c_s_f) * alpha5;
                    for (int q=0; q<felem.nQ(); ++q) {
                        double w  = felem.QWeight(q);
                        VectorField Nq = felem.QNormal(q);
                        VectorField Tq = felem.QTangent(q);
                        for (int i=0; i<nV; ++i) {
                            VectorField V_i = felem.Velocity(q,i);
                            Scalar NV = Nq * V_i;

                            Scalar TV = Tq * V_i;
                            for (int j=0; j<nV; ++j) {
                                VectorField V_j = felem.Velocity(q,j);
                                A_c(i,j) -= w * (alpha4 * NV * (Nq*V_j)
                                                 + alpha8 * TV * (Tq*V_j));
                                VectorField V1_j = felem_1.Velocity(q,j);
                                A_cf(i,j) += w * (alpha4 * NV * (Nq*V1_j)
                                                  + alpha8 * TV * (Tq*V1_j));
                            }
                            for (int nEta = 0; nEta<=numL; ++nEta) {
                                for (int j=0; j<nS; ++j) {
                                    Tensor S_j = felem.Stress(q,j);
                                    VectorField NS_j = S_j * Nq;

                                    A_c(i,nV+nEta*nS+j)
                                            -= w * (alpha2 * NV * (NS_j*Nq)
                                                    + alpha6 * TV * (NS_j*Tq));
                                    Tensor S1_j = felem_1.Stress(q,j);
                                    VectorField NS1_j = S1_j * Nq;
                                    A_cf(i,nV+nEta*nS+j)
                                            += w * (alpha2 * NV * (NS1_j*Nq)
                                                    + alpha6 * TV * (NS1_j*Tq));
                                }
                            }
                        }
                        for (int nEta = 0; nEta<=numL; ++nEta) {
                            for (int i=0; i<nS; ++i) {
                                Tensor S_i = felem.Stress(q,i);
                                VectorField NS = S_i * Nq;
                                Scalar NSN = NS * Nq;
                                Scalar NST = NS * Tq;
                                for (int j=0; j<nV; ++j) {
                                    VectorField V_j = felem.Velocity(q,j);
                                    A_c(nV+nEta*nS+i,j)
                                            -= w * (alpha3 * NSN * (Nq*V_j)
                                                    + alpha7 * NST * (Tq*V_j));
                                    VectorField V1_j = felem_1.Velocity(q,j);
                                    A_cf(nV+nEta*nS+i,j)
                                            += w * (alpha3 * NSN * (Nq*V1_j)
                                                    + alpha7 * NST * (Tq*V1_j));
                                }
                                for (int nEta2 = 0; nEta2<=numL; ++nEta2)
                                    for (int j=0; j<nS; ++j) {
                                        Tensor S_j = felem.Stress(q,j);
                                        VectorField NS_j = S_j * Nq;
                                        A_c(nV+nEta*nS+i,nV+nEta2*nS+j)
                                                -= w * (alpha1 * NSN * (NS_j*Nq)
                                                        + alpha5 * NST * (NS_j*Tq));
                                        Tensor S1_j = felem_1.Stress(q,j);
                                        VectorField NS1_j = S1_j * Nq;
                                        A_cf(nV+nEta*nS+i,nV+nEta2*nS+j)
                                                += w * (alpha1 * NSN * (NS1_j*Nq)
                                                        + alpha5 * NST * (NS1_j*Tq));
                                    }
                            }
                        }
                    }
                }
            }
            destroyElement(elem);
        }
        return;
    }
    bool RHS (double t) const { return problem.RHS(t); }
    void RHS (double t, Vector& b) const {
        b = 0.0;
        for (cell c = b.cells(); c != b.cells_end(); ++c) {
            DGVElasticElement* elem = getElem(disc,b,c);
            int s_dim = numL +1;
            if (problem.RHS(-1)) {
                for (int q  = 0; q<elem->nQ();++q)
                {
                    int nV = elem->V_dimension();
                    int nS = elem->S_dimension();
                    double w = elem->QWeight(q);
                    const Point& z = elem->QPoint(q);
                    VectorField V_rhs = problem.ForceV(t,c,z,q);
                    for (int i= 0; i< elem->V_dimension(); ++i){
                        VectorField V_i = elem->Velocity(q,i);
                        b(c(),i) += w * V_i * V_rhs;
                    }
                    for (int i =0; i < nS; ++i){
                        Tensor S_i = elem->Stress(q,i);
                        for (int n =0; n<s_dim;++n)
                        {
                            Tensor S_rhs = problem.ForceS(t,c,z,n,q);
                            b(c(),nV+n*nS+i) += w * Frobenius(S_i,S_rhs);
                        }
                    }

                }
            }
            if (problem.Exact()){
                //                continue;
                for(int f =0; f<c.Faces();++f)
                {
                    if(!check_if_bnd(b,c,f)) continue;
                    DGVElasticFaceElement felem(disc,b,c,f,problem.getNumL());
                    int nV = felem.V_dimension();
                    int nS = felem.S_dimension();
                    BFParts bnd(b.GetMesh(),c);
                    const double c_s   = problem.C_s(c,c());
                    const double c_p   = problem.C_p(c,c());
                    const double rho   = problem.Rho(c,c());
                    const double alpha1 = 1.0 / ( rho*c_p + rho*c_p );
//                    const double alpha2 = rho*c_p * alpha1;
                    const double alpha5 = 1.0 / ( rho*c_s + rho*c_s );
//                    const double alpha6 = rho*c_s * alpha5;
                    for (int q  = 0; q<felem.nQ();++q)
                    {
                        double w = felem.QWeight(q);
                        VectorField Nq = felem.QNormal(q);
                        VectorField Tq = felem.QTangent(q);
                        Point x = felem.QPoint(q);
                        if (bnd[f] == 2)
                        {
                            VectorField NS_bnd = problem.Neumann(t,c,x,Nq);
                            for (int i=0; i<nV; ++i)
                            {
                                VectorField V_i = felem.Velocity(q,i);
                                Scalar NV = Nq * V_i;
                                Scalar TV = Tq * V_i;
                                b(c(),i) += w * NV * (NS_bnd * Nq);
                                b(c(),i) += w * TV * (NS_bnd * Tq);
                            }
                            for (int nEta = 0; nEta<=numL; ++nEta)
                            {
                                for (int i=0; i<nS; ++i)
                                {
                                    Tensor S_i = felem.Stress(q,i);
                                    VectorField NS = S_i * Nq;
                                    Scalar NSN = NS * Nq;
                                    Scalar NST = NS * Tq;
                                    b(c(),nV+nEta*nS+i) += w * NSN * (NS_bnd * Nq) * 1.0 / (rho * c_p);
                                    b(c(),nV+nEta*nS+i) += w * NST * (NS_bnd * Tq) * 1.0 / ( rho * c_s);
                                }
                            }
                        }else if (bnd[f] == 1)
                        {
                            VectorField V_bnd = problem.Dirichlet(t,c,x);
                            const double alpha4 = rho*c_p;
                            const double alpha8 = rho*c_s;
                            for (int i=0; i<nV; ++i)
                            {
                                VectorField V_i = felem.Velocity(q,i);
                                Scalar NV = Nq * V_i;
                                Scalar TV = Tq * V_i;
                                b(c(),i) += w * (alpha4 * NV * (Nq*V_bnd)
                                                 + alpha8 * TV * (Tq*V_bnd));
                            }
                            for (int nEta = 0; nEta<=numL; ++nEta)
                            {
                                for (int i=0; i<nS; ++i)
                                {
                                    Tensor S_i = felem.Stress(q,i);
                                    VectorField NS = S_i * Nq;
                                    Scalar NSN = NS * Nq;
                                    Scalar NST = NS * Tq;
                                    b(c(),nV+nEta*nS+i) += w * ( NSN * (Nq*V_bnd)
                                                                 + NST * (Tq*V_bnd));
                                }
                            }
                        } else {Exit("unknown bnd flag")}

                    }
                }
            }
        }

    }


    void Jacobi (const Vector& u, Matrix& B) const {
        MassMatrix(B);
        Matrix A(u);
        FluxMatrix(A);
        B += (-gamma) * A;
    }
};

class DGVElasticAssemble : public DGVElasticAssembleMatrix {
    double flux_alpha;
    double X0;
    double X1;
    std::fstream f;
    Point Pobs;
public:
    DGVElasticAssemble (const DGDiscretization& _disc, const ProblemVElastic& _prob, Plot& _plot) :
        DGVElasticAssembleMatrix(_disc,_prob,_plot), X0(8), X1(10) {
        Config::Get("X0", X0);
        Config::Get("X1", X1);
        //f.open("Pkt.dat", ios::out);
        //Config::Get("Pobs", Pobs);
    }
    const char* Name() const { return "DGVElasticAssemble"; }
    std::string DiscName() const { return disc.DiscName(); }
    virtual void get_u_int_assemble (double t, Vector& u) const {
        u=0.0;
        for (cell c=u.cells(); c!=u.cells_end(); ++c) {
            row r = u.find_row(c());
            DGVElasticElement Elem(disc,u,c,numL);
            std::vector<Point> z;
            Elem.NodalPoints(z);
            for (int i=0; i<Elem.u_dimension();++i)
                for (int j=0; j<Elem.shape_size();++j)
                    u(r)[i*Elem.shape_size()+j] =problem.ut(t,c,z[j],i);
        }
        u.Accumulate();
    }
    double Energy_local (const Vector& u, DGVElasticElement& elem,const cell& c) const {
        Scalar e = 0;
        return e;
    }
    double Energy (const Vector& u, double x0, double x1) const {
        Scalar e = 0;
        return 0.5*PPM->Sum(e);
    }
    double Energy (const Vector& u) const {
        Scalar e = 0;
        return 0.5*PPM->Sum(e);
    }
    double Norm (const Vector& u) const { return sqrt(Energy(u)); }
    double L2Error (double t, const Vector& u) const {
        double err = 0;
        return sqrt(PPM->Sum (err));
    }
    double dist (const Vector& u, const Vector& v) const {
        Vector w(u);
        w -= v;
        return sqrt(Energy(w));
    }
    void SetInitialValue (Vector& u) const { get_u_int_assemble(0,u); }

    void PrintStep (double t, const Vector& u, int step) {
        std::cout<< std::fixed;
        std::cout.precision(5);
        Vector u_tmp(u);
        mout << "Step(" << step << "), t("
             << t
             << "): E(u) = " << Energy(u);
        mout<< endl;
        return;
    }
    void InitialStep (double t, Vector& u) {
        SetInitialValue(u);
        PrintStep(t,u,0);
        VtkPlotting(t,u,0);
    }

    void FinishStep (Vector& u, double t, double dt, int step,
                     int plot_step = 1) {
        if (step%plot_step == 0  ) {
            PrintStep(t,u,step);
            VtkPlotting(t,u,step);
        }
        return;
        Vector u_int(u);
        get_u_int_assemble(t,u_int);
        u_int -= u;
        mout << "Energy(u-u_int) " << Energy(u_int) << endl;
    }

    void VtkPlotting_cell (double t, const Vector& u,
                           char* filename, int step, std::string discName) const {
        Vector u_param(0.0,u);
        Vector u_param2(0.0,u);
        Vector u_param3(0.0,u);
        Vector u_vel(u);
        Vector u_str(u);
        for (cell c=u.cells(); c!=u.cells_end(); ++c) {
            DGVElasticElement* elem = getElem(disc,u,c);
            VectorField V = zero;
            Tensor S = Zero;

            double a = 0;
            for (int q=0; q<elem->nQ(); ++q) {
                double w  = elem->QWeight(q);
                //                S_div += w * elem->divStress(q,u);
                V += w * elem->Velocity(q,u);
                S += Scalar(w) * elem->Stress(q,u);
                a += w;
            }
            S *= (1/a);
            V *=  (1.0/a);
            row r = u.find_row(c());
            u_str(r)[0] = S[0][0];
            u_str(r)[1] = S[1][1];
            u_str(r)[2] = S[1][2];
            u_vel(r)[0] = V[0];
            u_vel(r)[1] = V[1];

            if (step == 0) {
                u_param(r)[0] = PPM->proc();
                u_param(r)[1] = problem.Rho(c,c());
                u_param(r)[2] = problem.Lambda(c,c());
                u_param2(r)[0] = problem.Mu(c,c());
                u_param2(r)[1] = problem.Mu(c,c(),0);
                u_param2(r)[2] = problem.Mu(c,c(),1);
                u_param3(r)[0] = problem.Lambda(c,c(),0);
                u_param3(r)[1] = problem.Lambda(c,c(),1);
//                for (int n = 1; n<=numL; ++n)
//                {
//                    u_param(r)[2+n] = problem.Tau(c,c(),n);
//                    u_param2(r)[n] = problem.Mu(c,c(),n);
//                }
            }
        }
        plot.celldata(u_str,3);
        NumberName("S00",filename,step);
        plot.vtk_celldata(filename,0);
        NumberName("S11",filename,step);
        plot.vtk_celldata(filename,1);
        NumberName("S12",filename,step);
        plot.vtk_celldata(filename,2);


        plot.celldata(u_vel,2);
        NumberName("Vx",filename,step);
        plot.vtk_celldata(filename,0);
        NumberName("Vy",filename,step);
        plot.vtk_celldata(filename,1);

        if (step == 0) {
            plot.celldata(u_param,4);
            plot.vtk_celldata("proc",0);
            plot.vtk_celldata("rho",1);
            plot.vtk_celldata("lambda",2);
            plot.vtk_celldata("tau",3);

            plot.celldata(u_param2,3);
            plot.vtk_celldata("mu",0);
            plot.vtk_celldata("mu_0",1);
            plot.vtk_celldata("mu_1",2);
            plot.celldata(u_param3,2);
//            plot.vtk_celldata("mu",0);
            plot.vtk_celldata("lambda_0",0);
            plot.vtk_celldata("lambda_1",1);
        }
    }
    void VtkPlotting (double t, const Vector& u, int step) {
        const Mesh& Mp = plot.GetMesh();
        const Mesh& Mu = u.GetMesh();
        if (Mp.CellCount() != Mu.CellCount()) return;
        vout(2) << "  => VtkPlotting result of step " << step << ".\n";
        char filename[128];
        VtkPlotting_cell(t,u,filename,step,"bla");
    }
    void PrintMatrixInfo (Matrix& A, int diagonal = 1) {
        for (cell c=A.cells(); c!=A.cells_end(); ++c) {
            DGVElasticElement* elem = getElem(disc,A,c);
            RowEntries_dG A_c(A,c,c);
            for (int i=0;i<elem->dimension(); ++i) {
                for (int j=0; j<elem->dimension(); ++j) {
                    mout << A_c(i,j) << " ";
                }
                mout << endl;
            }
            mout << endl;
            if(diagonal) continue;

            for (int f=0; f<c.Faces(); ++f) {
                cell cf = c;
                if(check_if_bnd(A,c,f)) continue;
                else cf = find_nb_cell(A,c,f);
                RowEntries_dG A_cf(A,c,cf);
                for (int i=0;i<elem->dimension(); ++i) {
                    for (int j=0; j<elem->dimension(); ++j) {
                        mout << A_cf(i,j) << " ";
                    }
                    mout << endl;
                }
                mout << endl;
            }
            mout << "------------------------" << endl;
        }
    }
    void Interpolate (const Vector& u, Vector& U) const {
        U = 0;
        U.Accumulate();
    }
    double Goal_Functional (double t, const Vector& u) const {
        Scalar e = 0.0;
        Scalar eq = 0.0;
        for (cell c=u.cells(); c!=u.cells_end(); ++c){
            if ((GF_a[0]>c()[0]) || (c()[0]>GF_b[0])) continue;
            if ((GF_a[1]>c()[1]) || (c()[1]>GF_b[1])) continue;

            DGVElasticElement elem(disc,u,c,numL);
            row r = u.find_row(c());
            for (int q=0; q<elem.nQ(); ++q) {
                Point QP = elem.QPoint(q);
                double w  = elem.QWeight(q);
                VectorField j_V;
                j_V[0] =  problem.ut_dual(t,c,QP,0,GF_a,GF_b);
                j_V[1] =  problem.ut_dual(t,c,QP,1,GF_a,GF_b);
                VectorField V = elem.Velocity(q,u);
                e += w * (j_V*V);
                eq += 0.5 * w * (V*V);
            }
        }
        mout<<"GoalFunctional(quadratic): " << abs(PPM->Sum(eq)) <<endl;
        mout<<"GoalFunctional(linear)   : " << abs(PPM->Sum(e)) <<endl;
        return abs(PPM->Sum(e));
    }
};
class DGVElasticAssembleAdjoint : public DGVElasticAssemble {
    double T;
public:
    DGVElasticAssembleAdjoint (const DGDiscretization& _disc, const ProblemVElastic& _prob,
                               Plot& _plot) :
        DGVElasticAssemble(_disc,_prob,_plot),T(1.0) {
        Config::Get("T", T, -1);
    }
    const char* Name() const { return "DGVElasticAssembleAdjoint"; }
    void FluxMatrix (Matrix& A) const {
        A = 0;
        for (cell c=A.cells(); c!=A.cells_end(); ++c) {
            DGVElasticElement* elem = getElem(disc,A,c);
            const int nV = elem->V_dimension();
            const int nS = elem->S_dimension();
            RowEntries_dG A_c(A,c,c);
            for (int q=0; q<elem->nQ(); ++q) {
                double w  = elem->QWeight(q);

                for (int i=0; i<nV; ++i) {
                    VectorField V_i = elem->Velocity(q,i);
                    for (int nEta = 0; nEta<=numL; ++nEta) {
                        for (int j=0; j<nS; ++j) {
                            VectorField DivS_j = elem->DivStress(q,j);
                            A_c(i,nV+nEta*nS+j) -= w * DivS_j * V_i;
                        }
                    }
                }
                for (int nEta = 0; nEta<=numL; ++nEta){
                    for (int i=0; i<nS; ++i) {
                        Tensor S_i = elem->Stress(q,i);
                        for (int j=0; j<nV; ++j) {
                            Tensor E_j = elem->Strain(q,j);
                            A_c(nV+nEta*nS+i,j) -= w * Frobenius(E_j,S_i);
                        }
                    }
                }
                for (int nEta = 1; nEta<=numL; ++nEta) {
                    double mu = problem.Mu(c, c(),nEta);
                    double s = 1 / (2.0 * mu);
                    double lambda = problem.Lambda(c, c(),nEta );
                    double ss = - lambda / (2.0 * mu * (2.0 * lambda + 2.0 * mu));
                    double tau = problem.Tau(c, c(),nEta );
                    for (int i=0; i<nS; ++i) {
                        Tensor S_i = elem->Stress(q,i);
                        for (int j=0; j<nS; ++j) {
                            Tensor S_j = elem->Stress(q,j);
                            A_c(nV+nEta*nS+i,nV+nEta*nS+j) -= w / tau * (
                                        s*Frobenius(S_j,S_i) + ss*trace(S_j)*trace(S_i) );
                        }
                    }
                }
            }
            for (int f=0; f<c.Faces(); ++f) {
                DGVElasticFaceElement felem(disc,A,c,f,numL);
                const int nV = felem.V_dimension();
                const int nS = felem.S_dimension();

                if (check_if_bnd(A,c,f)) {
                    BFParts bnd(A.GetMesh(),c);
                    const double mu_L = problem.Mu(c,c());
                    const double lambda_L = problem.Mu(c,c());
                    const double rho_L   = problem.Rho(c,c());
                    const double Z_PL = sqrt((2*mu_L+lambda_L)*rho_L);
                    const double Z_SL   = sqrt(mu_L*rho_L);
                    for (int q=0; q<felem.nQ(); ++q) {
                        double w  = felem.QWeight(q);
                        VectorField Nq = felem.QNormal(q);
                        VectorField Tq = felem.QTangent(q);
                        if (bnd[f] == 2 || bnd[f] == 0) {
                            const double alpha1 = 1.0 / ( Z_PL );
                            const double alpha5 = 1.0 / ( Z_SL );
                            for (int i=0; i<nV; ++i) {
                                VectorField V_i = felem.Velocity(q,i);
                                Scalar NV = Nq * V_i;
                                Scalar TV = Tq * V_i;
                                for (int nEta = 0; nEta<=numL; ++nEta) {
                                    for (int j=0; j<nS; ++j) {
                                        Tensor S_j = felem.Stress(q,j);
                                        VectorField NS_j = S_j * Nq;
                                        A_c(i,nV+nEta*nS+j)
                                                += w* (NV*(NS_j*Nq) + TV*(NS_j*Tq));
                                    }
                                }
                            }
                            for (int nEta = 0; nEta<=numL; ++nEta) {
                                for (int i=0; i<nS; ++i) {
                                    Tensor S_i = felem.Stress(q,i);
                                    VectorField NS = S_i * Nq;
                                    Scalar NSN = NS * Nq;
                                    Scalar NST = NS * Tq;
                                    for (int nEta2 = 0; nEta2<=numL; ++nEta2)
                                    {
                                        for (int j=0; j<nS; ++j) {
                                            Tensor S_j = felem.Stress(q,j);
                                            VectorField NS_j = S_j * Nq;
                                            A_c(nV+nEta*nS+i,nV+nEta2*nS+j)
                                                    -= w * (alpha1 * NSN * (NS_j*Nq)
                                                            + alpha5 * NST * (NS_j*Tq));
                                        }
                                    }
                                }
                            }
                        } else if (bnd[f] == 1) {
                            const double alpha4 = Z_PL;
                            const double alpha8 = Z_SL;
                            for (int i=0; i<nV; ++i) {
                                VectorField V_i = felem.Velocity(q,i);
                                Scalar NV = Nq * V_i;
                                Scalar TV = Tq * V_i;
                                for (int j=0; j<nV; ++j) {
                                    VectorField V_j = felem.Velocity(q,j);
                                    A_c(i,j) += w * (alpha4 * NV * (Nq*V_j)
                                                     + alpha8 * TV * (Tq*V_j));
                                }
                            }
                            for (int nEta = 0; nEta<=numL; ++nEta) {
                                for (int i=0; i<nS; ++i) {
                                    Tensor S_i = felem.Stress(q,i);
                                    VectorField NS = S_i * Nq;
                                    Scalar NSN = NS * Nq;
                                    Scalar NST = NS * Tq;
                                    for (int j=0; j<nV; ++j) {
                                        VectorField V_j = felem.Velocity(q,j);
                                        A_c(nV+nEta*nS+i,j) -= w * ( NSN * (Nq*V_j)
                                                                     + NST * (Tq*V_j));
                                    }
                                }
                            }
                        }
                    }
                } else {
                    cell cf = find_nb_cell(A,c,f);
                    int f1 = find_nb_face_id(A,c.Face(f),cf);
                    DGVElasticFaceElement felem_1(disc,A,cf,f1,numL);
                    RowEntries_dG A_cf(A,c,cf);
                    const double mu_L = problem.Mu(c,c());
                    const double lambda_L = problem.Lambda(c,c());
                    const double rho_L   = problem.Rho(c,c());
                    const double Z_PL = sqrt((2*mu_L+lambda_L)*rho_L);
                    const double Z_SL   = sqrt(mu_L*rho_L);

                    const double mu_R = problem.Mu(c,cf());
                    const double lambda_R = problem.Lambda(c,cf());
                    const double rho_R   = problem.Rho(c,cf());
                    const double Z_PR = sqrt((2*mu_R+lambda_R)*rho_R);
                    const double Z_SR   = sqrt(mu_R*rho_R);

                    const double alpha1 = 1.0 / ( Z_PL + Z_PR);
                    const double alpha2 = -Z_PL * alpha1;
                    const double alpha3 = -Z_PR * alpha1;
                    const double alpha4 = (Z_PL*Z_PL) * alpha1;

                    const double alpha5 = 1.0 / ( Z_SL + Z_SR );
                    const double alpha6 = -Z_SL * alpha5;
                    const double alpha7 = -Z_SR * alpha5;
                    const double alpha8 = (Z_SL*Z_SR) * alpha5;

                    for (int q=0; q<felem.nQ(); ++q) {
                        double w  = felem.QWeight(q);
                        VectorField Nq = felem.QNormal(q);
                        VectorField Tq = felem.QTangent(q);
                        for (int i=0; i<nV; ++i) {
                            VectorField V_i = felem.Velocity(q,i);
                            Scalar NV = Nq * V_i;

                            Scalar TV = Tq * V_i;
                            for (int j=0; j<nV; ++j) {
                                VectorField V_j = felem.Velocity(q,j);
                                A_c(i,j) -= w * (alpha4 * NV * (Nq*V_j)
                                                 + alpha8 * TV * (Tq*V_j));
                                VectorField V1_j = felem_1.Velocity(q,j);
                                A_cf(i,j) += w * (alpha4 * NV * (Nq*V1_j)
                                                  + alpha8 * TV * (Tq*V1_j));
                            }
                            for (int nEta = 0; nEta<=numL; ++nEta) {
                                for (int j=0; j<nS; ++j) {
                                    Tensor S_j = felem.Stress(q,j);
                                    VectorField NS_j = S_j * Nq;

                                    A_c(i,nV+nEta*nS+j)
                                            -= w * (alpha2 * NV * (NS_j*Nq)
                                                    + alpha6 * TV * (NS_j*Tq));
                                    Tensor S1_j = felem_1.Stress(q,j);
                                    VectorField NS1_j = S1_j * Nq;
                                    A_cf(i,nV+nEta*nS+j)
                                            += w * (alpha2 * NV * (NS1_j*Nq)
                                                    + alpha6 * TV * (NS1_j*Tq));
                                }
                            }
                        }
                        for (int nEta = 0; nEta<=numL; ++nEta) {
                            for (int i=0; i<nS; ++i) {
                                Tensor S_i = felem.Stress(q,i);
                                VectorField NS = S_i * Nq;
                                Scalar NSN = NS * Nq;
                                Scalar NST = NS * Tq;
                                for (int j=0; j<nV; ++j) {
                                    VectorField V_j = felem.Velocity(q,j);
                                    A_c(nV+nEta*nS+i,j)
                                            -= w * (alpha3 * NSN * (Nq*V_j)
                                                    + alpha7 * NST * (Tq*V_j));
                                    VectorField V1_j = felem_1.Velocity(q,j);
                                    A_cf(nV+nEta*nS+i,j)
                                            += w * (alpha3 * NSN * (Nq*V1_j)
                                                    + alpha7 * NST * (Tq*V1_j));
                                }
                                for (int nEta2 = 0; nEta2<=numL; ++nEta2)
                                    for (int j=0; j<nS; ++j) {
                                        Tensor S_j = felem.Stress(q,j);
                                        VectorField NS_j = S_j * Nq;
                                        A_c(nV+nEta*nS+i,nV+nEta2*nS+j)
                                                -= w * (alpha1 * NSN * (NS_j*Nq)
                                                        + alpha5 * NST * (NS_j*Tq));
                                        Tensor S1_j = felem_1.Stress(q,j);
                                        VectorField NS1_j = S1_j * Nq;
                                        A_cf(nV+nEta*nS+i,nV+nEta2*nS+j)
                                                += w * (alpha1 * NSN * (NS1_j*Nq)
                                                        + alpha5 * NST * (NS1_j*Tq));
                                    }
                            }
                        }
                    }
                }
            }
            destroyElement(elem);
        }
        return;
    }
    double L2Error (double t, const Vector& u) const {
        double err = 0;
        return sqrt(PPM->Sum (err));
    }
    void SetInitialValue (Vector& u) const { get_u_int_assemble(T,u); }

    void PrintStep (double t, Vector& u, int step) {
        std::cout<< std::fixed;
        std::cout.precision(5);
        Vector u_tmp(u);
        mout << "Step(" << step << "), t("
             << t
             << "): E(u) = " << Energy(u);
        mout<< endl;
        return;
    }

    void FinishStep (Vector& u, double t, double dt, int step,
                     int plot_step = 1) {
        if (step%plot_step == 0  ) {
            PrintStep(t,u,step);
            VtkPlotting(t,u,step);
        }
        return;
        Vector u_int(u);
        get_u_int_assemble(t,u_int);
        u_int -= u;
        mout << "Energy(u-u_int) " << Energy(u_int) << endl;
    }
    void RHS(double t, Vector &b) const
    {
        b=0.0;
        for (cell c = b.cells(); c != b.cells_end(); ++c)
        {
            DGVElasticElement* elem = getElem(disc,b,c);
            int s_dim = numL +1;
            if (problem.RHS(-1))
            {
                for (int q  = 0; q<elem->nQ();++q)
                {
                    int nV = elem->V_dimension();
                    int nS = elem->S_dimension();
                    double w = elem->QWeight(q);
                    const Point& z = elem->QPoint(q);
                    VectorField V_rhs = problem.ForceV(T-t,c,z,q);
                    for (int i= 0; i< elem->V_dimension(); ++i)
                    {
                        VectorField V_i = elem->Velocity(q,i);
                        b(c(),i)+= w * V_i * V_rhs;
                    }
                    for (int i =0; i < nS; ++i)
                    {
                        Tensor S_i = elem->Stress(q,i);
                        for (int n =0; n<s_dim;++n)
                        {
                            Tensor S_rhs = problem.ForceS(T-t,c,z,n,q);
                            b(c(),nV+n*nS+i)+= w * Frobenius(S_i,S_rhs);
                        }
                    }
                }
            }
            if (problem.Exact()){
               for(int f =0; f<c.Faces();++f)
                {
                    if(!check_if_bnd(b,c,f)) continue;
                    DGVElasticFaceElement felem(disc,b,c,f,problem.getNumL());
                    int nV = felem.V_dimension();
                    int nS = felem.S_dimension();
                    BFParts bnd(b.GetMesh(),c);
                    const double rho = problem.Rho(c,c());
                    const double lambda = problem.Lambda(c,c());
                    const double mu = problem.Mu(c,c());
                    const double Z_PL = sqrt((2*mu+lambda)*rho);
                    const double Z_SL   = sqrt(mu*rho);


                    const double alpha1 = 1.0 /  Z_PL;
                    const double alpha5 = 1.0 /  Z_SL;

                    const double alpha4 = Z_PL;
                    const double alpha8 = Z_SL;

                    for (int q  = 0; q<felem.nQ();++q)
                    {
                        double w = felem.QWeight(q);
                        VectorField Nq = felem.QNormal(q);
                        VectorField Tq = felem.QTangent(q);
                        Point x = felem.QPoint(q);
                        if (bnd[f] == 2)
                        {
                            VectorField NS_bnd = problem.Neumann(T-t,c,x,Nq);
                            for (int i=0; i<nV; ++i)
                            {
                                VectorField V_i = felem.Velocity(q,i);
                                Scalar NV = Nq * V_i;
                                Scalar TV = Tq * V_i;
                                b(c(),i) -= w * (NV * (NS_bnd * Nq)
                                                 + TV * (NS_bnd * Tq));
                            }
                            for (int nEta = 0; nEta<=numL; ++nEta)
                            {
                                for (int i=0; i<nS; ++i)
                                {
                                    Tensor S_i = felem.Stress(q,i);
                                    VectorField NS = S_i * Nq;
                                    Scalar NSN = NS * Nq;
                                    Scalar NST = NS * Tq;
                                    b(c(),nV+nEta*nS+i) += w* ( NSN * (NS_bnd * Nq) * alpha1
                                                                +NST * (NS_bnd * Tq) * alpha5) ;
//                                    b(c(),nV+nEta*nS+i) += w * NST * (NS_bnd * Tq) * alpha5;
                                }
                            }
                        }else if (bnd[f] == 1)
                        {
                            VectorField V_bnd = problem.Dirichlet(T-t,c,x);
                            for (int i=0; i<nV; ++i)
                            {
                                VectorField V_i = felem.Velocity(q,i);
                                Scalar NV = Nq * V_i;
                                Scalar TV = Tq * V_i;
                                b(c(),i) -= w * (alpha4 * NV * (Nq*V_bnd)
                                                 + alpha8 * TV * (Tq*V_bnd));
                            }
                            for (int nEta = 0; nEta<=numL; ++nEta)
                            {
                                for (int i=0; i<nS; ++i)
                                {
                                    Tensor S_i = felem.Stress(q,i);
                                    VectorField NS = S_i * Nq;
                                    Scalar NSN = NS * Nq;
                                    Scalar NST = NS * Tq;
                                    b(c(),nV+nEta*nS+i) += w * ( NSN * (Nq*V_bnd)
                                                                 + NST * (Tq*V_bnd));
                                }
                            }
                        } else {Exit("unknown bnd flag")}
                    }
                }
            }
        }
    }
};
#endif
