#ifndef _TOOLS_H_
#define _TOOLS_H_

#include "m++.hpp"
#include "NonmigratedFiles/DGAcoustic.hpp"
#include "DGViscoAcoustic.hpp"
bool transformPointGlobalToLocal (const Point& P_glob, Point& PT, const cell& CT) {
    switch (CT.Type()) {
    case TRIANGLE: {
        Tensor T;
        T[0][0] = CT.Corner(1)[0]-CT.Corner(0)[0];
        T[0][1] = CT.Corner(2)[0]-CT.Corner(0)[0];
        T[1][0] = CT.Corner(1)[1]-CT.Corner(0)[1];
        T[1][1] = CT.Corner(2)[1]-CT.Corner(0)[1];
        T[0][2] = 0;
        T[1][2] = 0;
        T[2][0] = 0;
        T[2][1] = 0;
        T[2][2] = 1;
        Tensor IT = Invert(T);
        PT = IT * (P_glob - CT.Corner(0));

        if( -1.0e-12<PT[0] && -1.0e-12<PT[1] && PT[0]+PT[1]<1+1.0e-12) return true;
        else return false;
    }
        
    case TETRAHEDRON:
        Exit("not implemented");

    case QUADRILATERAL: {
        Tensor T;
        T[0][0] = CT.Corner(1)[0]-CT.Corner(0)[0];
        T[0][1] = CT.Corner(3)[0]-CT.Corner(0)[0];
        T[1][0] = CT.Corner(1)[1]-CT.Corner(0)[1];
        T[1][1] = CT.Corner(3)[1]-CT.Corner(0)[1];
        T[0][2] = 0;
        T[1][2] = 0;
        T[2][0] = 0;
        T[2][1] = 0;
        T[2][2] = 1;
        Tensor IT = Invert(T);
        PT = IT * (P_glob - CT.Corner(0));
        if( -1.0e-12<PT[0] && -1.0e-12<PT[1] && PT[0]<1+1.0e-12 && PT[1]<1+1.0e-12) return true;
        else  return false;
    }

    case HEXAHEDRON:
        Exit("not implemented");

    }
}

Point transformPointLocaltoGlobal (const Point PF, const cell& CF) {
    switch (CF.Type()) {
    case TRIANGLE: {
        Point P_glob = (1-PF[0]-PF[1])*CF.Corner(0) + PF[0]*CF.Corner(1) + PF[1]*CF.Corner(2);
        return P_glob;
    }
        
    case TETRAHEDRON:
        Exit("not implemented");

    case QUADRILATERAL: {
        Point P_glob = CF.Corner(0) + PF[0]*(CF.Corner(1)-CF.Corner(0)) + PF[1]*(CF.Corner(3)-CF.Corner(0));
        return P_glob;
    }

    case HEXAHEDRON:
        Exit("not implemented");

    }
}


// ----------------------------------------------------
//    InterpolateWave TransferTWave
// ----------------------------------------------------

class InterpolateWave_dG : public Operator {
public:
    InterpolateWave_dG () {}
    virtual void Interpolate(const Vector& f, Vector& c) const{}
    virtual void multiply (Vector& u, const Vector& v) const {}
    virtual void multiply_transpose (Vector& u, const Vector& v) const { }
};

inline constAB<Operator,Vector> 
operator * (const InterpolateWave_dG& T, const Vector& v) {
    return constAB<Operator,Vector>(T,v);
}
inline constAB<Vector,Operator> 
operator * (const Vector& v, const InterpolateWave_dG& T) {
    return constAB<Vector,Operator>(v,T);
}

class NewCellTransfer_2D : public InterpolateWave_dG {
private:
    DGAcousticDiscretization* disc_fine;
    DGAcousticDiscretization* disc_coarse;
    const int k_max; //problemDim
   std::vector<const Shape*> space_Shape;
   std::vector<vector<Point> > z;

public:
    NewCellTransfer_2D (DGAcousticDiscretization* _disc_fine, DGAcousticDiscretization* _disc_coarse) :
        disc_fine(_disc_fine),
        disc_coarse(_disc_coarse),
        k_max(3)
    {
        space_Shape.resize(7);
        /*
        const Shape& P0_space( disc_fine->GetCellShape(0) );
        const Shape& P1_space( disc_fine->GetCellShape(1) );
        const Shape& P2_space( disc_fine->GetCellShape(2) );
        const Shape& P3_space( disc_fine->GetCellShape(3) );
        const Shape& P4_space( disc_fine->GetCellShape(4) );
        const Shape& P5_space( disc_fine->GetCellShape(5) );
        const Shape& P6_space( disc_fine->GetCellShape(6) );
        */
        space_Shape[0] = Get_Shape("P0quad");//&P0_space;
        space_Shape[1] = Get_Shape("P1quad");//&P1_space;
        space_Shape[2] = Get_Shape("P2quad");//&P2_space;
        space_Shape[3] = Get_Shape("P3quad");//&P3_space;
        space_Shape[4] = Get_Shape("P4quad");//&P4_space;
        space_Shape[5] = Get_Shape("P5quad");//&P5_space;
        space_Shape[6] = Get_Shape("P6quad");//&P6_space;
        
        z.resize(7);
        for (int s=0; s<2; ++s){
            //disc_fine->NodalPoints_dg(s,z[s]);
            int m = 1;//k_max;
            z[s].resize(m*(s+1)*(s+1));
            int n = 0;
            switch (s) {
            case 0:
                for (int k=0; k<m; ++k){
                    z[s][n++] = Point(0.5,0.5);
                }
                break;
            case 1:
                for (int k=0; k<m; ++k){
                    z[s][n++] = Point(0.,0.);
                    z[s][n++] = Point(1.,0.);
                    z[s][n++] = Point(1.,1.);
                    z[s][n++] = Point(0.,1.);
                }
                break;
            default:
                for (int k=0; k<m; ++k){
                    double x = 1.0/s;
                    for (int i=0;i<=s;++i)
                        for (int j=0;j<=s;++j)
                            z[s][n++] = Point(j*x , i*x);
                }
                break;
            }
        }
    }
    ~NewCellTransfer_2D () {
    }
    
    bool transformPointLocaltoLocal (const Point PF, const cell& CF, Point& PT, const cell& CT) const {
        if ( CF.Type()!=CT.Type() ) Exit("types are different");
        return transformPointGlobalToLocal (transformPointLocaltoGlobal (PF,CF), PT, CT);
    }
    
    void Interpolate (const Vector& F, Vector& C, std::string s) const {
        mout<<"DEBUGGING " << endl;
        C = 0.;
        if (s == "Res"){
            Exit("TODO");
        }
        else if (s == "Pro"){
            for (cell cf=F.cells(); cf!=F.cells_end(); ++cf){
                row rf = F.find_row(cf());
                int space_deg_f = disc_coarse->get_celldeg(cf());
                
               std::vector<Point> Children(cf.Children());
                for (int i=0; i<cf.Children(); ++i)
                    Children[i] = cf.Child(i);
                
               std::vector<row> row_c (Children.size());
               std::vector<int> space_deg_c (Children.size());
                for (int i = 0; i<Children.size(); ++i){
                    row_c[i] = C.find_row(Children[i]);
                    space_deg_c[i] =  disc_fine->get_celldeg(Children[i]);
                }
                
                for (int i=0; i<Children.size();++i){
                    cell cc = C.find_cell(Children[i]);
                    
//                    mout << cf << "   " << cc <<"    "<<Children[i]<< endl;
                    int idx_c = 0;
                    for (int k=0; k<k_max; ++k){
//                        mout << "   "<< k<< "   " << z[space_deg_c[i]].size()<<endl;
                        for (int sc=0; sc<z[space_deg_c[i]].size(); ++sc){
                            
                            Point z_space(z[space_deg_c[i]][sc]);
                            
                            transformPointLocaltoLocal(z[space_deg_c[i]][sc], cc, z_space, cf);
                            
                            int shift = k * z[space_deg_f].size();
                            for (int sf=0; sf<z[space_deg_f].size(); ++sf){
                                double value_space_shape = (*space_Shape[space_deg_f])(z_space,sf);
                                C(row_c[i])[idx_c] += F(rf)[shift + sf] * value_space_shape;
                            }
                            
                            idx_c++;
                        }
                    }
                }
            }
        }
        
        if (s == "Pro"){
            C.MakeAdditive();
            C.Accumulate();
        }
    }
    void multiply (Vector& u, const Vector& v) const { Interpolate(v,u,"Pro");}
    void multiply_transpose (Vector& u,const Vector& v)const{ Exit("TODO"); }
};
#endif
//class NewCellTransfer_2D_VA : public InterpolateWave_dG {
//private:
//    DGTSDiscretization* disc_fine;
//    DGVAcousticDiscretization* disc_coarse;
//    const int k_max; //problemDim
//   std::vector<const Shape*> space_Shape;
//   std::vector<vector<Point> > z;

//public:
//    NewCellTransfer_2D_VA (DGVAcousticDiscretization* _disc_fine, DGVAcousticDiscretization* _disc_coarse) :
//        disc_fine(_disc_fine),
//        disc_coarse(_disc_coarse),
//        k_max(4)
//    {
//        space_Shape.resize(7);
//        /*
//        const Shape& P0_space( disc_fine->GetCellShape(0) );
//        const Shape& P1_space( disc_fine->GetCellShape(1) );
//        const Shape& P2_space( disc_fine->GetCellShape(2) );
//        const Shape& P3_space( disc_fine->GetCellShape(3) );
//        const Shape& P4_space( disc_fine->GetCellShape(4) );
//        const Shape& P5_space( disc_fine->GetCellShape(5) );
//        const Shape& P6_space( disc_fine->GetCellShape(6) );
//        */
//        space_Shape[0] = Get_Shape("P0quad");//&P0_space;
//        space_Shape[1] = Get_Shape("P1quad");//&P1_space;
//        space_Shape[2] = Get_Shape("P2quad");//&P2_space;
//        space_Shape[3] = Get_Shape("P3quad");//&P3_space;
//        space_Shape[4] = Get_Shape("P4quad");//&P4_space;
//        space_Shape[5] = Get_Shape("P5quad");//&P5_space;
//        space_Shape[6] = Get_Shape("P6quad");//&P6_space;

//        z.resize(7);
//        for (int s=0; s<2; ++s){
//            //disc_fine->NodalPoints_dg(s,z[s]);
//            int m = 1;//k_max;
//            z[s].resize(m*(s+1)*(s+1));
//            int n = 0;
//            switch (s) {
//            case 0:
//                for (int k=0; k<m; ++k){
//                    z[s][n++] = Point(0.5,0.5);
//                }
//                break;
//            case 1:
//                for (int k=0; k<m; ++k){
//                    z[s][n++] = Point(0.,0.);
//                    z[s][n++] = Point(1.,0.);
//                    z[s][n++] = Point(1.,1.);
//                    z[s][n++] = Point(0.,1.);
//                }
//                break;
//            default:
//                for (int k=0; k<m; ++k){
//                    double x = 1.0/s;
//                    for (int i=0;i<=s;++i)
//                        for (int j=0;j<=s;++j)
//                            z[s][n++] = Point(j*x , i*x);
//                }
//                break;
//            }
//        }
//    }
//    ~NewCellTransfer_2D_VA () {
//    }

//    bool transformPointLocaltoLocal (const Point PF, const cell& CF, Point& PT, const cell& CT) const {
//        if ( CF.Type()!=CT.Type() ) Exit("types are different");
//        return transformPointGlobalToLocal (transformPointLocaltoGlobal (PF,CF), PT, CT);
//    }

//    void Interpolate (const Vector& F, Vector& C, std::string s) const {
//        mout<<"DEBUGGING " << endl;
//        C = 0.;
//        if (s == "Res"){
//            Exit("TODO");
//        }
//        else if (s == "Pro"){
//            for (cell cf=F.cells(); cf!=F.cells_end(); ++cf){
//                row rf = F.find_row(cf());
////                int space_deg_f = disc_coarse->get_cell_deg(cf());
//                int space_deg_f = 2;
//               std::vector<Point> Children(cf.Children());
//                for (int i=0; i<cf.Children(); ++i)
//                    Children[i] = cf.Child(i);

//               std::vector<row> row_c (Children.size());
//               std::vector<int> space_deg_c (Children.size());
//                for (int i = 0; i<Children.size(); ++i){
//                    row_c[i] = C.find_row(Children[i]);
//                    space_deg_c[i] =2;
////                    space_deg_c[i] =  disc_fine->get_celldeg(Children[i]);
//                }
//                Exit("WARNING CHANGED STUFF AND DID NOT FIX IT ( the getcelldeg function")
//                for (int i=0; i<Children.size();++i){
//                    cell cc = C.find_cell(Children[i]);

//                    //                    mout << cf << "   " << cc <<"    "<<Children[i]<< endl;
//                    int idx_c = 0;
//                    for (int k=0; k<k_max; ++k){
//                        //                        mout << "   "<< k<< "   " << z[space_deg_c[i]].size()<<endl;
//                        for (int sc=0; sc<z[space_deg_c[i]].size(); ++sc){

//                            Point z_space(z[space_deg_c[i]][sc]);

//                            transformPointLocaltoLocal(z[space_deg_c[i]][sc], cc, z_space, cf);

//                            int shift = k * z[space_deg_f].size();
//                            for (int sf=0; sf<z[space_deg_f].size(); ++sf){
//                                double value_space_shape = (*space_Shape[space_deg_f])(z_space,sf);
//                                C(row_c[i])[idx_c] += F(rf)[shift + sf] * value_space_shape;
//                            }

//                            idx_c++;
//                        }
//                    }
//                }
//            }
//        }

//        if (s == "Pro"){
//            C.MakeAdditive();
//            C.Accumulate();
//        }
//    }
//    void multiply (Vector& u, const Vector& v) const { Interpolate(v,u,"Pro");}
//    void multiply_transpose (Vector& u,const Vector& v)const{ Exit("TODO"); }
//};

//InterpolateWave_dG* GetNewCellTransfer (int dim, DGAcousticDiscretization* disc_fine, DGAcousticDiscretization* disc_coarse) {
//    if (dim == 2) return new NewCellTransfer_2D(disc_fine, disc_coarse);
//    else Exit("no interpolation for this dimension implemented");
//}

//InterpolateWave_dG* GetNewCellTransfer_VA (int dim, DGVAcousticDiscretization* disc_fine, DGVAcousticDiscretization* disc_coarse) {
//    if (dim == 2) return new NewCellTransfer_2D_VA(disc_fine, disc_coarse);
//    else Exit("no interpolation for this dimension implemented");
//}

//#endif
