import os
import re
import numpy as np
import pandas as pd
import matplotlib as mpl
import matplotlib.pyplot as plt

import vacoustic_exact_solution_2D as vaexact
from seismogram_lib import FWISeismogram
def calculateNorms(seismos, t0_err = -np.inf, t1_err=np.inf):
    norms = {}
    norms_time = {}
    for deg in seismos.keys():
        for level in seismos[deg].keys():
            times_dict = seismos[deg][level]
            time_keys = list(times_dict.keys())
            time_keys.sort()
            time_keys.sort(reverse = True)
            for i in range(len(time_keys)):
                seismonorm = seismos[deg][level][time_keys[i]].norm2(t0=t0_err, t1=t1_err)
                if deg not in norms:
                    norms[deg] = {}

                if level not in norms[deg]:
                    norms[deg][level] = {}

                norms[deg][level][i] = seismonorm

                if deg not in norms_time:
                    norms_time[deg] = {}

                if level not in norms_time[deg]:
                    norms_time[deg][level] = {}

                norms_time[deg][level][time_keys[i]] = seismonorm
    return norms, norms_time
def calculateTimeDiffNorms(seismos,t0_err = -np.inf, t1_err=np.inf):
    t_d_n_t = {}
    t_d_n = {}
    f_int = {}
    f_int_t={}
    for deg in seismos.keys():
        for level in seismos[deg].keys():
            times_dict = seismos[deg][level]
            time_keys = list(times_dict.keys())
            time_keys.sort(reverse=True)
            seismo_diff = [seismos[deg][level][time_keys[i]] - seismos[deg][level][time_keys[i+1]] for i in range(len(time_keys) - 1)]
            for i in range(0, len(seismo_diff)):
                diffnorm = seismo_diff[i].norm2(t0=t0_err, t1=t1_err)

                if deg not in t_d_n_t:
                    t_d_n_t[deg] = {}
                if level not in t_d_n_t[deg]:
                    t_d_n_t[deg][level] = {}

                t_d_n_t[deg][level][time_keys[i+1]] = diffnorm

                if deg not in t_d_n:
                    t_d_n[deg] = {}
                if level not in t_d_n[deg]:
                    t_d_n[deg][level] = {}

                t_d_n[deg][level][i+1] = diffnorm
            if deg not in f_int:
                f_int[deg] = {}
            if level not in f_int[deg]:
                f_int[deg][level] = {}
            if deg not in f_int_t:
                f_int_t[deg] = {}
            if level not in f_int_t[deg]:
                f_int_t[deg][level] = {}
            for tk in range(1, len(t_d_n[deg][level])):
                f_int[deg][level][time_keys[tk+1]] = t_d_n_t[deg][level][time_keys[tk]]/t_d_n_t[deg][level][time_keys[tk+1]]
                f_int_t[deg][level][tk] = f_int[deg][level][time_keys[tk+1]]

    return f_int, f_int_t, t_d_n_t, t_d_n

def calculateSpaceDiffNorms_timeIndexing(seismos,t0_err = -np.inf, t1_err=np.inf):
    s_d_n = {}
    s_d_n_t ={}
    g_int = {}
    g_int_t = {}
    for deg in seismos.keys():
        for level in seismos[deg].keys():
            if level+1 in seismos[deg]:
                time_keys = list(seismos[deg][level+1].keys())
                time_keys.sort(reverse=True)
                for tk in range(len(list(time_keys))):
                    if time_keys[tk] in seismos[deg][level]:
                        space_diff = seismos[deg][level][time_keys[tk]] - seismos[deg][level+1][time_keys[tk]]
                        if deg not in s_d_n:
                            s_d_n[deg] = {}
                        if tk not in s_d_n[deg]:
                            s_d_n[deg][tk] = {}
                        if deg not in s_d_n_t:
                            s_d_n_t[deg] = {}
                        if time_keys[tk] not in s_d_n_t[deg]:
                            s_d_n_t[deg][time_keys[tk]] = {}
                        spacediffnorm = space_diff.norm2(t0=t0_err, t1=t1_err)
                        s_d_n_t[deg][time_keys[tk]][level+1] = spacediffnorm
                        s_d_n[deg][tk][level+1] = spacediffnorm
        for level2 in seismos[deg].keys():
                time_keys2 = list(seismos[deg][level2].keys())
                time_keys2.sort(reverse=True)
                for tk2 in range(0,len(time_keys2)):
                    if time_keys2[tk2] in seismos[deg][level2]:
                        if deg not in g_int:
                            g_int[deg] = {}
                        if tk2 not in g_int[deg]:
                            g_int[deg][tk2] = {}

                        if deg not in g_int_t:
                            g_int_t[deg] = {}
                        if time_keys2[tk2] not in g_int_t[deg]:
                            g_int_t[deg][time_keys2[tk2]] = {}

                        for i in range(3, len(s_d_n_t[deg][time_keys2[tk2]])+2):
                            g_int_t[deg][time_keys2[tk2]][i] = s_d_n_t[deg][time_keys2[tk2]][i-1]/s_d_n_t[deg][time_keys2[tk2]][i]
                            g_int[deg][tk2][i] = g_int_t[deg][time_keys2[tk2]][i]
    return g_int, g_int_t ,s_d_n, s_d_n_t

def calculateSpaceDiffNorms(seismograms,t0_err = -np.inf, t1_err=np.inf):
    space_diff_norms = {}
    g = {}
    for deg in seismograms.keys():
        for level in seismograms[deg].keys():
            sortlevel = list(seismograms[deg].keys())
            sortlevel.sort()
            time_key_sorted = list(seismograms[deg][level].keys())
            time_key_sorted.sort()
#            time_key_sorted.sort(reverse=True)
            if level + 1 in seismograms[deg]:
                for tk in range(len(time_key_sorted)):
                    if time_key_sorted[tk] in seismograms[deg][level+1]:
                        space_diff = seismograms[deg][level][time_key_sorted[tk]] - seismograms[deg][level + 1][time_key_sorted[tk]]
                        if deg not in space_diff_norms:
                            space_diff_norms[deg] = {}
                        if tk not in space_diff_norms[deg]:
                            space_diff_norms[deg][tk] = {}
                        space_diff_norms[deg][tk][level+1] = space_diff.norm2(t0=t0_err, t1=t1_err)
        for level in seismograms[deg].keys():
            time_key_sorted2 = list(seismograms[deg][level].keys())
            time_key_sorted2.sort()
            for tk2 in range(len(time_key_sorted2)):
                if deg not in g:
                    g[deg] = {}
                if tk2 not in g[deg]:
                    g[deg][tk2]={}
                for j in range(2,len(space_diff_norms[deg][tk2])+1):
                    g[deg][tk2][j] = space_diff_norms[deg][tk2][j-1]/space_diff_norms[deg][tk2][j]
    print(g)
    print(space_diff_norms)
    return g, space_diff_norms


def readSeismogramsFromPath(seismograms, data_dir, start="s_", dt_min=-np.inf, dt_max=np.inf, shift=False):
    min_level,min_deg = np.inf, np.inf
    max_level,max_deg = -np.inf, -np.inf
    for fn in os.listdir(data_dir):
        if not fn.startswith(start):
            continue
        m = re.match(start + "(\w+)_(\w+)_level_(\d)_dt_([\d\.e-]+)_deg_(\d)_rkorder_([-\d]+)_shot_(\d+)", fn)

        t_kernel, x_kernel, level, dt, deg, rkorder, shot = m.groups()
        dt, deg, level = float(dt), int(deg), int(level)
        max_deg, max_level = max(max_deg, deg), max(max_level, level)
        min_deg, min_level = min(min_deg, deg), min(min_level, level)
#        s_p = vaexact.read_pressure_seismogram(data_dir + fn,state_components_override=True)
        s_p = vaexact.read_pressure_seismogram(data_dir + fn)
        if shift:
            s_p.shift_times(1)
        if deg not in seismograms:
            seismograms[deg] = {}
        if level not in seismograms[deg]:
            seismograms[deg][level] = {}
        if dt_min < dt < dt_max:
            seismograms[deg][level][dt] = s_p

    return min_deg, max_deg, min_level, max_level, seismograms


def extractInfoFromlogs(seismograms,log_dir):
    info = {"deg": [], "dt": [], "level": [], "error": [], "GMRESmin": [], "GMRESmax": [], "GMRESsum": [],
              "GMRESavg": [], "dofs": [], "steps": []}
    keys =[(k1,k2,k3) for k1 in seismograms.keys() for k2 in seismograms[k1].keys() for k3 in seismograms[k1][k2].keys()]
    for (deg,level,dt) in keys:
        ss = "level=" + str(level) + "_deg=" + str(deg) + "_rkorder=-2_numL=5_" + "dt=" + str(dt)
        info["deg"].append(deg)
        info["level"].append(level)
        info["dt"].append(dt)
        found = 0
        for fn2 in os.listdir(log_dir):
            f1 = fn2.find(ss)
            if f1 is not -1:
                found = 1
                file = open(LOG_DIR + fn2)
                S = file.read()
                f2 = re.findall("GMRES: d\((\d+)\)= [\d\.e-]+ rate", S)
                f3 = re.findall("matrixgraph on level " + str(level) + ":" + " (\d+)", S)
                GMRESlist = [int(s) for s in f2]
                info["dofs"].append(f3[0])
                info["GMRESmin"].append(min(GMRESlist))
                info["GMRESmax"].append(max(GMRESlist))
                info["GMRESsum"].append(sum(GMRESlist))
                info["GMRESavg"].append(sum(GMRESlist) / len(GMRESlist))
        if found == 0:
            print(ss + " not found!")
            del info["deg"][-1]
            del info["dt"][-1]
            del info["level"][-1]
    return info


def extractErrorFromlog(fn_, error_dict, deg_max, level_max=-1, dt_mi=-1):
    degs = error_dict["deg"]
    dts = error_dict["dt"]
    error_ = error_dict["error"]
    level_ = error_dict["level"]
    GMRESmin = error_dict["GMRESmin"]
    GMRESmax = error_dict["GMRESmax"]
    GMRESsum = error_dict["GMRESsum"]
    GMRESavg = error_dict["GMRESavg"]
    dofs = error_dict["dofs"]
    for d in range(deg_max + 1):
        error_vec = []
        for i in range(len(degs)):
            if int(degs[i]) == d:
                append = 0
                if level_max > 0:
                    if level_[i] is level_max:
                        append = 1
                if dt_mi >0:
                    if abs(dt_mi-dts[i])<1e-10:
                        append = 1
                if append == 1:
                    error_vec.append((level_[i], error_[i], dts[i],GMRESmin[i],GMRESmax[i],GMRESsum[i],GMRESavg[i],dofs[i]))
        if len(error_vec) < 1:
            for i in range(len(degs)):
                if degs[i] is d:
                    if level_[i] is level_max-1:
                        error_vec.append((level_[i], error_[i], dts[i],GMRESmin[i],GMRESmax[i],GMRESsum[i],GMRESavg[i],dofs[i]))
        if level_max > 0:
            error_vec.sort(key = lambda x : x[2],reverse = True)
            #error_vec = sorted(error_vec, reverse=True)
        if dt_mi > 0:
            error_vec = sorted(error_vec)
        S1 = ""
        S2 = ""
        S3 = ""
        S4 = ""
        S5 = ""
        S6 = ""
        S7 = ""
        S8 = ""
        S9 = ""
        S10 = ""
        S11 = ""
        e_tmp = 0
        cnt = 0
        for tt in error_vec:
            if cnt is 0:
                e_tmp = tt[1]
            S1 += "{:d}".format(tt[0]) + ' '
            S2 += "{:.7f}".format(tt[1]) + ' '
            S3 += "{:.1f}".format(h0 / (2 ** tt[0])) + ' '
            S4 += "{:.5f}".format(tt[2]) + ' '
            #if i > 0:
            #    S5 += "{:.5f}".format(0 if np.isinf(np.log2(e_tmp / tt[1])) else np.log2(e_tmp / tt[1])) + ' '
            #    e_tmp = tt[1]
            cnt = cnt + 1
            S6 += str(d) + ' '
            S7 += str(tt[3]) + ' '
            S8 += str(tt[4]) + ' '
            S9 += str(tt[5]) + ' '
            S10 += "{:.2f}".format(tt[6]) + ' '
            S11 += str(tt[7]) + ' '

        S = 'level ' + S1 + ' \n'
        S += 'deg ' + S6 + '\n'
        S += 'h ' + S3 + '\n'
        S += 'L2error ' + S2 + '\n'
        #S += 'EOC ' + S5 + '\n'
        S += 'dt ' + S4 + '\n'
        S += 'GMRESmin ' + S7 + '\n'
        S += 'GMRESmax ' + S8 + '\n'
        S += 'GMRESsum ' + S9 + '\n'
        S += 'GMRESavg ' + S10 + '\n'
        S += 'dofs ' + S11 + '\n'
        f = open(fn_ + str(d), "w+")
        f.write(S)
        f.close()
        print(S)


def createTimeTable(deg, norms_tt, time_diff_tt, space_diff_tt, f_tt):
    norm_df = pd.DataFrame(data=norms_tt[deg]).transpose()
    time_diff_df = pd.DataFrame(data=time_diff_tt[deg]).transpose()
    f_df = pd.DataFrame(data=f_tt[deg]).transpose()
    space_diff_df = pd.DataFrame(data=space_diff_tt[deg])

    norm_rows = norm_df.axes[0].values
    time_diff_rows = time_diff_df.axes[0].values
    f_rows = f_df.axes[0].values
    space_diff_rows = space_diff_df.axes[0].values

    tmp1 = norm_df.append(time_diff_df)
    tmp2 = tmp1.append(f_df)
    tmp3 = tmp2.append(space_diff_df).round(4)
    tmp3 = tmp3[tmp3.columns[::-1]]
    rownames_tt = []
    colnames_tt = []
    for nr in norm_rows:
        rownames_tt.append("$\|\\vec{s}_{" + str(deg) + ",n," + str(nr) + "}\|_{0,(0,T)}$ ")

    for tdr in time_diff_rows:
        rownames_tt.append("$\|\\vec{s}_{" + str(deg) + "," + str(tdr) + ",j+1}-\\vec{s}_{" + str(deg) + "," + str(
                tdr) + ",j}\|_{0,(0,T)}$")

    for fr in f_rows:
        rownames_tt.append("$f_{" + str(deg) + "," + str(fr) + ",j}$")

    for sdr in space_diff_rows:
        rownames_tt.append("$\|\\vec{s}_{" + str(deg) + "," + str(sdr - 1) + ",j}-\\vec{s}_{" + str(deg) + "," + str(
                sdr) + ",j}\|_{0,(0,T)}$")
    header = "c|"
    for j in range(len(tmp3.axes[1].values)):
        colnames_tt.append("$j="+str(j)+"$")
        header += "c"
    tmp3.set_axis(rownames_tt, axis='index', inplace=True)
    tmp3.set_axis(colnames_tt, axis='columns', inplace=True)
    tmp3.columns.name = "$k=" + str(deg) + "$"
    return header, tmp3


def createSpaceTable(deg, norms_st, time_diff_st, space_diff_st, g_st):
    norm_df = pd.DataFrame(data=norms_st[deg])
    time_diff_df = pd.DataFrame(data=time_diff_st[deg])
    g_df = pd.DataFrame(data=g_st[deg]).transpose()
    space_diff_df = pd.DataFrame(data=space_diff_st[deg]).transpose()
    norm_rows = norm_df.axes[0].values
    time_diff_rows = time_diff_df.axes[0].values
    g_rows = g_df.axes[0].values
    space_diff_rows = space_diff_df.axes[0].values

    tmp1 = norm_df.append(space_diff_df)
    tmp2 = tmp1.append(g_df)
    tmp3 = tmp2.append(time_diff_df).round(4)
    rownames_st = []
    colnames_st = []
    for nr in norm_rows:
        rownames_st.append("$\|\\vec{s}_{" + str(deg) + ",n," + str(nr) + "}\|_{0,(0,T)}$ ")

    for tdr in space_diff_rows:
        rownames_st.append("$\|\\vec{s}_{" + str(deg) + ",n+1," + str(tdr) + "}-\\vec{s}_{" + str(deg) + ",n,"+str(tdr)+"}\|_{0,(0,T)}$")

    for fr in g_rows:
        rownames_st.append("$g_{" + str(deg) + ",n," + str(fr) + "}$")

    for sdr in time_diff_rows:
        rownames_st.append("$\|\\vec{s}_{" + str(deg) + ",n," + str(sdr) + "}-\\vec{s}_{" + str(deg) + ",n," + str(
                sdr+1) + "}\|_{0,(0,T)}$")
    header = "c|"
    for j in range(len(tmp3.axes[1].values)):
        colnames_st.append("$n="+str(j)+"$")
        header += "c"
    tmp3.set_axis(rownames_st, axis='index', inplace=True)
    tmp3.set_axis(colnames_st, axis='columns', inplace=True)
    tmp3.columns.name = "$k="+str(deg) + "$"
    return header, tmp3


def calculate_norms(seismogram_dict, t=(-np.inf,np.inf)):
    deg_list = list(seismogram_dict.keys())
    deg_list.sort()
    size_deg = len(deg_list)
    size_level = -np.inf
    size_time = -np.inf
    for key_deg in seismogram_dict.keys():
        size_level = max(size_level, len(list(seismogram_dict[key_deg])))
        for key_level in seismogram_dict[key_deg].keys():
            size_time = max(size_time, len(list(seismogram_dict[key_deg][key_level])))

    seismogram_nd = np.ndarray(shape=(size_deg, size_level, size_time))
    for deg in deg_list:
        key_level = seismogram_dict[deg].keys()
        level_list = list(key_level)
        level_list.sort()
        for level in level_list:
            key_time = seismogram_dict[deg][level].keys()
            time_list = list(key_time)
            time_list.sort(reverse=True)
            for time in range(len(time_list)):
                seismogram_nd[deg][level][time] = seismogram_dict[deg][level][time_list[time]].norm2(t0=t[0], t1=t[1])

    print(seismogram_nd)
    exit()
if __name__== "__main__":
    #t0 = 0.00512
    #t1 = 0.5
    #h0 = 100
    t0 = -np.inf
    t1 = np.inf
    LOG_DIR = f"/home/cr/data_paper/log/"
    seismograms = {}
    min_deg, min_level = np.inf, np.inf
    max_deg, max_level = -1, -1
    best_deg = -1
    best_level = -1
    best_s_sol, min_dt = None, np.inf
    min_dt = -np.inf
    max_dt = np.inf
    #min_dt = 9e-05
    #max_dt = 0.002561
    #DATA_DIR1 = f"/home/cr/data_edited/pdedata/"
    #DATA_DIR2 = f"/home/cr/data_edited/fh2data/corner/"
    #min_deg, max_deg, min_level, max_level, seismograms = readSeismogramsFromPath(seismograms,DATA_DIR1,start="shifted_",dt_min=min_dt,dt_max=max_dt)
    #min_deg, max_deg, min_level, max_level, seismograms = readSeismogramsFromPath(seismograms,DATA_DIR2,start="s_",dt_min=min_dt,dt_max=max_dt)
    #DATA_DIR1 = "/home/cr/data_paper_newcalcs/fh2/corner/"
    #DATA_DIR2 = "/home/cr/data_paper_newcalcs/pde/corner/"
    DATA_DIR1 = "/home/cr/data_paper_newcalcs/selected_calcs_line_measurements/"
    DATA_DIR1 = "/tmp/rkorder4_2/"

#    DATA_DIR1 = "/home/cr/data_paper_newcalcs/selected_calcs_line_measurements/"

    min_deg, max_deg, min_level, max_level, seismograms = readSeismogramsFromPath(seismograms,DATA_DIR1,start="s_",dt_min=min_dt,dt_max=max_dt)
    #min_deg, max_deg, min_level, max_level, seismograms = readSeismogramsFromPath(seismograms,DATA_DIR2,start="s_",dt_min=min_dt,dt_max=max_dt)
    #calculate_norms(seismograms)

    #info = extractInfoFromlogs(seismograms,LOG_DIR)

    norms, norms_time = calculateNorms(seismograms, t0_err=t0, t1_err=t1)
    print("aa")
    f, f2, time_diff_norms, time_diff_norms_time = calculateTimeDiffNorms(seismograms, t0_err=t0, t1_err=t1)

    #g, g2, space_diff_norms, space_diff_norms_time = calculateSpaceDiffNorms_timeIndexing(seismograms, t0_err=t0, t1_err=t1)
    for deg in norms.keys():
      #  headerspace, spacetable = createSpaceTable(deg, norms, time_diff_norms_time, space_diff_norms, g)
        headertime, timetable = createTimeTable(deg, norms_time, time_diff_norms, space_diff_norms_time, f)
     #   print(spacetable.to_latex(escape=False, na_rep="-", column_format=headerspace))
        print(timetable.to_latex(escape=False, na_rep="-", column_format=headertime))
