import re
import itertools
import functools
import operator
import sys

from pprint import pprint

from math import sqrt
inf = float("inf")
eps = sys.float_info.epsilon

import functools
import operator
import numpy as np
from scipy.spatial import ConvexHull


try:
    from math import isclose
except ImportError:
    def isclose(a,b):
        return (abs(a-b) < eps)
    

def fix_orientation(indices, mesh_points):
    if len(indices) == 4:
        new_indices = sorted(indices, key=lambda i: mesh_points[i].x)
        l, r = new_indices[0:2], new_indices[2:4]
        l.sort(key=lambda i: mesh_points[i].y)
        r.sort(key=lambda i: mesh_points[i].y)
    
        indices = [l[0], r[0], r[1], l[1]]
        
    elif len(indices) == 8:
        z_sorted = sorted(indices, key=lambda i: mesh_points[i].z)
        front, back = z_sorted[0:4], z_sorted[4:8]
        
        y_sorted_front = sorted(front, key=lambda i: mesh_points[i].y)
        y_sorted_back = sorted(back, key=lambda i: mesh_points[i].y)
        bottom_front, top_front = y_sorted_front[0:2], y_sorted_front[2:4]
        bottom_back, top_back = y_sorted_back[0:2], y_sorted_back[2:4]
        
        x_sorted_bottom_front = sorted(bottom_front, key=lambda i: mesh_points[i].x)
        x_sorted_top_front = sorted(top_front, key=lambda i: mesh_points[i].x)
        x_sorted_bottom_back = sorted(bottom_back, key=lambda i: mesh_points[i].x)
        x_sorted_top_back = sorted(top_back, key=lambda i: mesh_points[i].x)
        indices = (x_sorted_bottom_front + x_sorted_top_front[::-1] +
                        x_sorted_bottom_back + x_sorted_top_back[::-1])  
        
    return indices

GeometricTolerance = 1e-10
class MppPoint:
    def __init__(self, x,y,z=None):
        self.x, self.y, self.z = x, y, z

    def __str__(self):
        if (self.z is not None):
            return "{:12.5f}{:12.5f}{:12.5f}".format(self.x, self.y, self.z)
        else:
            return "{:12.5f}{:12.5f}".format(self.x, self.y)

    def __repr__(self):
        return "MppPoint({})".format(str(self))

    def __sub__(self, p):
        if (self.z is not None and p.z is not None):
            return MppPoint(self.x-p.x, self.y-p.y, self.z-p.z)
        else:
            return MppPoint(self.x-p.x, self.y-p.y)
    
    def __lt__(self, p):
        if self.x < p.x - GeometricTolerance: return True
        if self.x > p.x + GeometricTolerance: return False
        if self.y < p.y - GeometricTolerance: return True
        if self.y > p.y + GeometricTolerance: return False
    
        if self.z is None and p.z is not None: return False
        if self.z is not None:
            if self.z < p.z - GeometricTolerance: return True
            if self.z > p.z + GeometricTolerance: return False
        return False
        
    def __eq__(self, p):
        return (not (p < self)) and (not (p > self)) 
    
    def __hash__(self):
        x, y, z = self.x, self.y, self.z
        if z is None : z = 0
        return int(115421.1*x + 124323.3*y + 221313.7*z 
                   + 345453*GeometricTolerance)
        
    def __add__(self, p):
        if self.z is None:
            return MppPoint(self.x + p.x, self.y + p.y)
        else:
            return MppPoint(self.x + p.x, self.y + p.y, self.z + p.z)
        
    def __truediv__(self, alpha):
        if self.z is None:
            return MppPoint(self.x/alpha, self.y/alpha)
        else:
            return MppPoint(self.x/alpha, self.y/alpha, self.z/alpha)
        
    def coords(self):
        if self.z is None: 
            return (self.x, self.y)
        else:
            return (self.x, self.y, self.z)
        
    def norm(self):
        return sqrt(self.x*self.x + self.y*self.y)
    

class MppCell:
    def __init__(self, corners, sd, indices, mesh_points):
        self.mesh_points = mesh_points
        self.corners = corners
        self.subdomain = sd
        self.indices = indices
        self.fix_orientation()
        
    def fix_orientation(self):
        self.indices = fix_orientation(self.indices, self.mesh_points)
    
        #if len(self.indices) == 4:
            #new_indices = sorted(self.indices, key=lambda i: self.mesh_points[i].x)
            #l, r = new_indices[0:2], new_indices[2:4]
            #l.sort(key=lambda i: self.mesh_points[i].y)
            #r.sort(key=lambda i: self.mesh_points[i].y)
        
            #self.indices = [l[0], r[0], r[1], l[1]]
            
        #elif len(self.indices) == 8:
            #z_sorted = sorted(self.indices, key=lambda i: self.mesh_points[i].z)
            #front, back = z_sorted[0:4], z_sorted[4:8]
            
            #y_sorted_front = sorted(front, key=lambda i: self.mesh_points[i].y)
            #y_sorted_back = sorted(back, key=lambda i: self.mesh_points[i].y)
            #bottom_front, top_front = y_sorted_front[0:2], y_sorted_front[2:4]
            #bottom_back, top_back = y_sorted_back[0:2], y_sorted_back[2:4]
            
            #x_sorted_bottom_front = sorted(bottom_front, key=lambda i: self.mesh_points[i].x)
            #x_sorted_top_front = sorted(top_front, key=lambda i: self.mesh_points[i].x)
            #x_sorted_bottom_back = sorted(bottom_back, key=lambda i: self.mesh_points[i].x)
            #x_sorted_top_back = sorted(top_back, key=lambda i: self.mesh_points[i].x)
            #self.indices = (x_sorted_bottom_front + x_sorted_top_front[::-1] +
                            #x_sorted_bottom_back + x_sorted_top_back[::-1])
        
    def __str__(self):
        indices = "".join("{:6d}".format(i) for i in self.indices)
        return "{:6d}{:6d}{}".format(self.corners, self.subdomain, indices)
    
    def __repr__(self):
        return "MppCell({})".format(str(self))
    
    def __eq__(self, c):
        return set(self.indices) == set(c.indices)
    
    def get_faces(self):
        points = [self.mesh_points[i].coords() for i in self.indices]
        hull = ConvexHull(points)
        equations = hull.equations
        def in_cell(coords, tol=1e-10):
            return np.all( (np.dot(equations[:,:-1], coords) + equations[:,-1] ) < -tol)
        
        face_corners = 2
        while True:
            on_bnd = False
            for candidate in itertools.combinations(self.indices, face_corners):
                candidate_center = functools.reduce(operator.add, (self.mesh_points[i] for i in candidate) ) / len(candidate)
                if not in_cell(candidate_center.coords()):
                    on_bnd = True
                    break
            if on_bnd:
                face_corners += 1
            else:
                face_corners -= 1
                break

        faces = []
        for candidate in itertools.combinations(self.indices, face_corners):
            candidate_center = functools.reduce(operator.add, (self.mesh_points[i] for i in candidate) ) / len(candidate)
            if not in_cell(candidate_center.coords()):
                face = MppFace(len(candidate), 0, candidate, self.mesh_points)
                faces += [face]
                
        return faces
        


class MppFace:
    def __init__(self, corners, sd, indices, mesh_points):
        self.mesh_points = mesh_points
        self.corners = corners
        self.subdomain = sd
        self.indices = indices
        self.fix_orientation()
        
    def fix_orientation(self):
        self.indices = fix_orientation(self.indices, self.mesh_points)
        
    def set_bnd_cond(self, val):
        #if callable(val):
            #val = val([self.mesh_points[i].coords() for i in self.indices])
        self.subdomain = val
    
    def center(self):
        corners = (self.mesh_points[i] for i in self.indices)
        return functools.reduce(operator.add, corners) / len(self.indices)
    
    def __str__(self):
        indices = "".join("{:6d}".format(i) for i in self.indices)
        return "{:6d}{:6d}{}".format(self.corners, self.subdomain, indices)
    

class MppMesh:
    def __init__(self):
        self.points = []
        self.cells = []
        self.faces = []
        
    def load(self, filename):
        mode = None
        min_index = 1e10
        with open(filename, "r") as f:
            for line in f:
                line = line.strip().strip(":")
                if not line: continue

                if line in ("POINTS", "CELLS", "FACES","TIMES"):
                    mode = line
                    continue
                            
                if mode == "POINTS":
                    p = MppPoint(*(float(e) for e in line.split()))
                    self.points.append(p)
                elif mode == "CELLS":
                    m = re.match("\s*(\d+)\s+(\d+)\s+(.*)", line)
                    corners, sd, indices = m.groups()
                    indices = [int(i) for i in indices.split()]
                    min_index = min(min_index, min(indices))
                    self.cells.append(MppCell(int(corners), int(sd), indices, self.points))
                elif mode == "FACES":
                    m = re.match("\s*(\d+)\s+([\d-]+)\s+(.*)", line)
                    corners, sd, indices = m.groups()
                    indices = [int(i) for i in indices.split()]
                    self.faces.append(MppFace(int(corners), int(sd), indices, self.points))
        
        #print(min_index)
        #if min_index > 0:
            #for c in self.cells:
                #c.indices = [i-min_index for i in c.indices]
            #for f in self.faces:
                #f.indices = [i-min_index for i in f.indices]
            
    def merge_triangles2squares(self):
        def face_length(f):
            p1, p2 = self.points[f[0]], self.points[f[1]]
            return (p1-p2).norm()
        
        def longest_face(c):
            faces = [(f,face_length(f)) for f in
                                itertools.combinations(c.indices, 2)]
            return set(max(faces, key=lambda x: x[1])[0])
        
        def find_other_triangle(c):
            lf = longest_face(c)
            
            for c2 in self.cells:
                if longest_face(c2) == lf and c2 != c:
                    return c2
            
            raise ValueError("No matching triangle found for '{}'".format(c))
        
        new_cells = []
        cells_done = []
        new_points = self.points.copy()
        for c in self.cells:
            if c in cells_done:
                continue
                    
            c2 = find_other_triangle(c)
            new_indices = list(set(c.indices + c2.indices))
            if c.subdomain != c2.subdomain:
                print("WARNING: subdomains of '{}'"
                      "and '{}' not matching".format(c1, c2))
            new_cells.append(MppCell(4, c.subdomain, new_indices, new_points))
            cells_done += [c, c2]
            
        new_mesh = MppMesh()
        new_mesh.points = new_points
        new_mesh.cells = new_cells
        return new_mesh
    
    def split_squares2triangles(self):
        def face_length(f):
            p1, p2 = self.points[f[0]], self.points[f[1]]
            return (p1-p2).norm()
        
        def longest_face(c):
            faces = [(f,face_length(f)) for f in
                                itertools.combinations(c.indices, 2)]
            return set(max(faces, key=lambda x: x[1])[0])
        
        new_cells = []
        cells_done = []
        new_points = self.points.copy()
        for c in self.cells:
            cell_points = [new_points[i] for i in c.indices]
            cell_points.sort()
            
            new_indices_all = sorted(c.indices, key=lambda i: new_points[i])
            new_indices_T1 = new_indices_all[:-1]
            new_indices_T2 = new_indices_all[1:]
                    
            new_cells.append(MppCell(3, c.subdomain, new_indices_T1, new_points))
            new_cells.append(MppCell(3, c.subdomain, new_indices_T2, new_points))
            
        new_mesh = MppMesh()
        new_mesh.points = new_points
        new_mesh.cells = new_cells
        return new_mesh
    
    def find_face(self, indices):
        indices = set(indices)
        for f in self.faces:
            if indices == set(f.indices):
                return f
        return None
        
    def set_outer_bnd_cond(self, val):
        self.set_bnd_cond(outer_value=val)
    
    def set_inner_bnd_cond(self, val):
        self.set_bnd_cond(inner_value=val)
    
    def set_bnd_cond(self, outer_value=None, inner_value=None):
        all_faces = {}
        face_counter = {}
        
        for c in self.cells:
            cell_faces = c.get_faces()
            for f in cell_faces:
                key = frozenset(f.indices)
                face_counter[key] = (1 if key not in face_counter else face_counter[key]+1)
                all_faces[key] = f
        
        bnd_faces = []
        inner_faces = []
        for key, count in face_counter.items():
            if count == 1:
                bnd_faces.append(all_faces[key])
            else:
                inner_faces.append(all_faces[key])
        
        faces = []
        if outer_value is not None:
            for f in bnd_faces:
                if callable(outer_value):
                    f.set_bnd_cond(outer_value(f))
                else:
                    f.set_bnd_cond(outer_value)
                faces.append(f)
        
        if inner_value is not None:
            for f in inner_faces:
                f.set_bnd_cond(inner_value)
                faces.append(f)
        
        self.faces = faces
    
    def set_bnd_cond_(self, val, where=("inner", "outer")):
        min_x, min_y, max_x, max_y = inf, inf, -inf, -inf
        for c in self.cells:
            xs = [self.points[i].x for i in c.indices]
            ys = [self.points[i].y for i in c.indices]
            min_x, max_x = min(min_x, min(xs)), max(max_x, max(xs))
            min_y, max_y = min(min_y, min(ys)), max(max_y, max(ys))
            
        print("mesh bounding box: [{},{}] x [{},{}]".format(min_x, max_x, min_y, max_y))
            
        def on_bnd(p):
            return ((isclose(p.x, min_x) or isclose(p.x, max_x)) or 
                    (isclose(p.y, min_y) or isclose(p.y, max_y)))
            
        for c in self.cells:
            indices = (c.indices + c.indices)[:len(c.indices)+1]
            for i in range(len(indices)-1):
                i1, i2 = indices[i:i+2]
                p1, p2 = self.points[i1], self.points[i2]
                is_outer_face = (on_bnd(p1) and on_bnd(p2))
                if where == ("outer",) and not is_outer_face:
                    continue
                elif where == ("inner",) and is_outer_face:
                    continue
                
                cell_count = sum((i1 in c.indices) and (i2 in c.indices) for c in self.cells)
                if cell_count == 1:
                    f = self.find_face((i1,i2))
                    if f is None:
                        self.faces.append(MppFace(2, val, [i1,i2], self.points))
                    else:
                        f.set_bnd_cond(val)
    
    def threedified(self, ts, initial_bnd_cond=1337, final_bnd_cond=2337):
        tdm = MppMesh()
        new_points, new_faces = [], []
        for n,t in enumerate(ts):
            for p in self.points:
                np = MppPoint(t, p.x, p.y)
                new_points.append(np)
                
        nP = len(self.points)
        new_cells = []
        for n in range(len(ts)-1):
            for c in self.cells:
                new_indices = [i+n*nP for i in c.indices] + [i+(n+1)*nP for i in c.indices]
                #print(c.indices, new_indices)
                new_cells.append(MppCell(8, c.subdomain, new_indices, new_points))
                
            # bnd values
            for f in self.faces:
                new_indices = [i+n*nP for i in f.indices] + [i+(n+1)*nP for i in f.indices]
                #print(c.indices, new_indices)
                new_faces.append(MppFace(4, f.subdomain, new_indices, new_points))
                
                
        # initial values
        for c in self.cells:
            new_indices = [i for i in c.indices]
            new_faces.append(MppFace(4, initial_bnd_cond, new_indices, new_points))
        
        # final values
        for c in self.cells:
            new_indices = [(len(ts)-1)*len(self.points) + i for i in c.indices]
            new_faces.append(MppFace(4, final_bnd_cond, new_indices, new_points))
        
        new_mesh = MppMesh()
        new_mesh.points = new_points
        new_mesh.cells = new_cells
        new_mesh.faces = new_faces
        return new_mesh
    
    
    def print_info(self):
        print("Points: {}\nCells: {}\nBndFaces: {}".format(len(self.points), len(self.cells), len(self.faces)))
    
    def __str__(self):
        return (
            "POINTS:\n" + "\n".join(str(p) for p in self.points) + "\n"
            "CELLS:\n" + "\n".join(str(c) for c in self.cells) + "\n" +
            (("FACES:\n" + "\n".join(str(f) for f in self.faces)) if self.faces else "") +
            "\n"
            )
