import fwiPlotLib
import pyvista as pv
import os.path
import os
from time import time
LSDFpath = '/home/cr/LSDF/'



class PlotConfig:
    def __init__(self):
        self.name = ''
        self.fn_shots = ''
        self.input_path = '/tmp/'
        self.algo = ''
        self.fn_input_final = ''
        self.fn_input_true = ''
        self.fn_input_initial = ''

        self.path_errors = ''
        self.final = 100

        self.fn_output_final = ''
        self.fn_output_true = ''
        self.fn_output_initial = ''
        self.fn_output_diff = ''

        self.data_name = ''

        self.bounding_box = [(-100000, 100000), (-100000, 100000)]

        self.cmap = 'RdYlBu'

        self.midpoint_iterate = 1.0
        self.vmin_iterate = -1000
        self.vmax_iterate = 1000

        self.midpoint_diff = 0.0
        self.vmin_diff = -1000
        self.vmax_diff = 1000

        self.ticks_middle = []

        self.normalization_value = -100000

        self.normone = None
        self.normtwo = None

        self.plotting = False
        self.plot_iterates = False
        self.errors = True
        self.frequency = ''


    def __init__(self, dict, lsdf=False):
        self.name = dict['name']
        self.fn_shots = dict['fn_shots']
        self.frequency = dict['frequency']

        if lsdf is True:
            self.input_path = LSDFpath + dict['input_path']
        else:
            self.input_path = dict['input_path']
        self.algo = dict['algo']
        self.fn_input_true = self.input_path + dict['fn_input_true']
        self.fn_input_initial = self.input_path + dict['fn_input_initial']

        self.final = dict['final']

        self.pdf_output_path = self.input_path + dict['pdf_path']
        if os.path.exists(self.pdf_output_path) is False:
            os.mkdir(self.pdf_output_path)
        self.data_name = dict['data_name']

        self.fn_output_final = self.pdf_output_path + self.name + self.algo + self.data_name + 'Final.pdf'
        self.fn_output_true = self.pdf_output_path + self.name + self.algo + self.data_name + 'True.pdf'
        self.fn_output_initial = self.pdf_output_path + self.name + self.algo + self.data_name + 'Initial.pdf'
        self.fn_output_diff = self.pdf_output_path + self.name + self.algo + self.data_name + 'Diff.pdf'


        self.bounding_box = dict['bounding_box']

        self.cmap = dict['cmap']

        self.midpoint_iterate = dict['midpoint_iterate']
        self.vmin_iterate = dict['vmin_iterate']
        self.vmax_iterate = dict['vmax_iterate']

        self.midpoint_diff = dict['midpoint_diff']
        self.vmin_diff = dict['vmin_diff']
        self.vmax_diff = dict['vmax_diff']

        self.ticks_middle = dict['ticks_middle']

        self.normalization_value = dict['normalization_value']

        self.plotting = dict['plotting']
        self.plot_iterates = dict['plot_iterates']
        self.path_errors = self.input_path + dict['path_errors']

        if os.path.exists(self.path_errors) is False:
            os.mkdir(self.path_errors)
        self.errors = dict['errors']

        self.normone = fwiPlotLib.MidpointNormalize(midpoint=self.midpoint_iterate,
                                                    vmin=self.vmin_iterate,
                                                    vmax=self.vmax_iterate)
        self.normtwo = fwiPlotLib.MidpointNormalize(midpoint=self.midpoint_diff,
                                                    vmin=self.vmin_diff,
                                                    vmax=self.vmax_diff)
        self.fn_input_final = self.make_iterate_input_fn(self.final)

    def make_iterate_input_fn(self, nr):
        return self.input_fn() + self.fn_shots + '_' + str(nr) + '.vtu'

    def make_iterate_output_fn(self, nr):
        return self.input_fn() + self.fn_shots + '_' + str(nr) + '.pdf'

    def input_fn(self):
        return self.input_path + self.data_name + '__fwi_' + self.algo + self.frequency + '_iterate_m_allshots_'


def getBasicConfig(namestring, algo):
    if namestring == 'C2':
        plot1 = {'name': 'C2', 'algo': algo, 'cmap': 'RdYlBu', 'fn_shots': '5', 'final': 9,
                 'pdf_path': 'pdfs/',
                 'path_errors': 'errors/', 'input_path': '',
                 'data_name': 'VpExp', 'fn_input_true': 'VpExp_fwi_m_real.vtu', 'fn_input_initial': '',
                 'bounding_box': [(600, 1400), (370, 1620)], 'midpoint_iterate': 1.0,
                 'midpoint_diff': 0,
                 'vmin_diff': -0.08, 'vmax_diff': 0.20, 'vmin_iterate': 0.90, 'vmax_iterate': 1.20,
                 'ticks_middle': [],
                 'normalization_value': 1.0 / 3500.0, 'plotting': True, 'plot_iterates': True, 'errors': True,
                 'frequency': ''}
        return plot1
    else:
        exit("srting not known")


def vtuPlotIterates(config, createMapping=False):
    fn_input_final = config.make_iterate_input_fn(config.final)
    fn_input_true = config.fn_input_true
    fn_input_initial = config.fn_input_initial

    fn_output_final = config.fn_output_final
    print(fn_output_final)
    fn_output_true = config.fn_output_true
    fn_output_initial = config.fn_output_initial
    fn_output_diff = config.fn_output_diff

    data_name = config.data_name

    normalization_value = config.normalization_value

    cmap = config.cmap

    vtu_final = pv.read(fn_input_final)
    vtu_true = pv.read(fn_input_true)
    if config.fn_input_initial == config.input_path:
        vtu_initial = pv.read(fn_input_true)
        vtu_initial.cell_data[data_name] = 1.0/normalization_value
    else:
        vtu_initial = pv.read(fn_input_initial)

    # if createMapping is False:
    x_Val_f, x_Val, y_Val, ImageMapping = fwiPlotLib.vtuMapping(vtu_final, config.bounding_box)
    image_true = fwiPlotLib.fillImage(vtu_true.cell_data[data_name] * normalization_value, ImageMapping, x_Val_f,
                                           y_Val, x_Val)
    image_initial = fwiPlotLib.fillImage(vtu_initial.cell_data[data_name] * normalization_value, ImageMapping,
                                          x_Val_f, y_Val, x_Val)
    image_final = fwiPlotLib.fillImage(vtu_final.cell_data[data_name] * normalization_value, ImageMapping,
                                          x_Val_f, y_Val, x_Val)
    # else:
    # image_true = fwiPlotLib.fillImageAndMapping(vtu_true, data_name, config.bounding_box) * normalization_value
    # image_initial = fwiPlotLib.fillImageAndMapping(vtu_initial, data_name,
    #                                                    config.bounding_box) * normalization_value
    # image_final = fwiPlotLib.fillImageAndMapping(vtu_final, data_name, config.bounding_box) * normalization_value

    image_diff = image_true - image_final
    ticks_middle = config.ticks_middle
    vmin = config.vmin_iterate
    vmax = config.vmax_iterate
    if config.plotting is True:
        fwiPlotLib.savePlot(image_true, fn_output_true, vmin, vmax, cmap=cmap, norm=config.normone,
                            ticksmiddle=ticks_middle)

        fwiPlotLib.savePlot(image_initial, fn_output_initial, vmin, vmax, cmap=cmap, norm=config.normone,
                            ticksmiddle=ticks_middle)
        if config.plot_iterates is True:
            s_rel_linf = 'relative error linf\n 0 1 \n'
            s_rel_l2 = 'relative error l2 \n0 1 \n'
            image_initial -= image_true
            norm_l2, norm_l_inf = fwiPlotLib.calculateErrors(image_initial)
            print(norm_l2)
            print(norm_l_inf)
            for i in range(config.final + 1):
                vtu_iterate = pv.read(config.make_iterate_input_fn(i))
                if createMapping is False:
                    image_iterate = fwiPlotLib.fillImage(vtu_iterate.cell_data[data_name] * normalization_value,
                                                     ImageMapping,
                                                     x_Val_f, y_Val, x_Val)
                else:
                    image_iterate = fwiPlotLib.fillImageAndMapping(vtu_iterate,data_name, config.bounding_box) * normalization_value

                fwiPlotLib.savePlot(image_iterate, config.make_iterate_output_fn(i), vmin, vmax, cmap=cmap,
                                    norm=config.normone, ticksmiddle=ticks_middle)
                image_iterate -= image_true
               # fwiPlotLib.savePlot(image_iterate, "test" + str(i), config.vmin_diff, config.vmax_diff, cmap=cmap,
               #                                                  norm=config.normtwo)
                l2, l_inf = fwiPlotLib.calculateErrors(image_iterate)
                print(str(l2))
                s_rel_l2 += str(i + 1) + ' ' + str(l2 / norm_l2) + '\n'
                s_rel_linf += str(i + 1) + ' ' + str(l_inf / norm_l_inf) + '\n'
                # if i in range(config.final):
                #     fwiPlotLib.savePlot(image_iterate, "test" + str(i), config.vmin_diff, config.vmax_diff, cmap=cmap,
                #                         norm=config.normtwo)
            f = open(config.path_errors + 'errorlinf_rel_' + config.name, 'w+')
            f.write(s_rel_linf)
            f.close()
            f = open(config.path_errors + 'errorl2_rel_' + config.name, 'w+')
            f.write(s_rel_l2)
            f.close()
        fwiPlotLib.savePlot(image_final, fn_output_final, vmin, vmax, cmap=cmap, norm=config.normone,
                            ticksmiddle=ticks_middle)

        fwiPlotLib.savePlot(image_diff, fn_output_diff, config.vmin_diff, config.vmax_diff, cmap=cmap,
                            norm=config.normtwo)



    # if config.errors is True:
    #     s_rel_linf = 'relative error linf\n 0 1 \n'
    #     s_rel_l2 = 'relative error l2 \n0 1 \n'
    #     image_initial -= image_true
    #     norm_l2, norm_l_inf = fwiPlotLib.calculateErrors(image_initial)
    #     print(norm_l2)
    #     print(norm_l_inf)
    #     for i in range(config.final+1):
    #         vtu_iterate = pv.read(config.make_iterate_input_fn(i))
    #         print( config.make_iterate_input_fn(i))
    #         if createMapping is False:
    #             image_iterate = fwiPlotLib.fillImage(vtu_iterate.cell_data[data_name] * normalization_value,
    #                                                  ImageMapping,
    #                                                  x_Val_f, y_Val, x_Val)
    #         else:
    #             image_iterate = fwiPlotLib.fillImageAndMapping(vtu_iterate, data_name,
    #                                                            config.bounding_box) * normalization_value
    #         image_iterate -= image_true
    #         l2, l_inf = fwiPlotLib.calculateErrors(image_iterate)
    #         fwiPlotLib.savePlot(image_iterate, "test" + str(i), config.vmin_diff, config.vmax_diff, cmap=cmap,
    #                             norm=config.normtwo)
    #         print(str(l2))
    #         s_rel_l2 += str(i + 1) + ' ' + str(l2 / norm_l2) + '\n'
    #         s_rel_linf += str(i + 1) + ' ' + str(l_inf / norm_l_inf) + '\n'
    #
    #     f = open(config.path_errors + 'errorlinf_rel_' + config.name, 'w+')
    #     f.write(s_rel_linf)
    #     f.close()
    #     f = open(config.path_errors + 'errorl2_rel_' + config.name, 'w+')
    #     f.write(s_rel_l2)
    #     f.close()

# config = PlotConfig(Plot1)
# vtuPlot(config)
