#!/usr/bin/python

import matplotlib as mpl
#mpl.rc('text', usetex=True)
mpl.use('Agg')

from matplotlib import pyplot as plt

from itertools import count as counter
import numpy as np
from pylab import rand

from mpl_toolkits.axes_grid1 import make_axes_locatable

#from renderparam_config import *

import matplotlib.ticker as mticker

#font = {'family' : 'Bitstream Vera Sans', 'size'   : 9}
#mpl.rc('font', **font)


inf = 1e100

class VTKFile:    
    def calcCenter(self, corners):
        center = [0,0]
        for c in corners:
            p = self.Points[c]
            center[0] += p[0]
            center[1] += p[1]
            
        center[0] /= len(corners)
        center[1] /= len(corners)
        
        return round(center[0], self.RoundingDigits), round(center[1], self.RoundingDigits)
    
    
    def getIndex(self, p):
        cx, cy = len(self.CellCentersX), len(self.CellCentersY)
        dx, dy = (self.Bounds[0][1]-self.Bounds[0][0])/cx, (self.Bounds[1][1]-self.Bounds[1][0])/cy

        return int(round((p[0]-self.Bounds[0][0]-dx/2)/dx)),int(round((p[1]-self.Bounds[1][0]-dy/2)/dy))
    
    def min(self):
        return min(self.FirstSliceValues.values())
    
    def max(self):
        return max(self.FirstSliceValues.values())

    
    def __init__(self, filename, bbox=None, space_time=False, rounding_digits=10):
        self.RoundingDigits = rounding_digits
        self.FileName = filename
        
        print("Reading", filename)
        
        self.Points = {}
        self.FirstSlicePoints = {}
        self.FirstSliceCells = {}
        self.FirstSliceValues = {}
        
        self.Bounds = [[inf,-inf], [inf,-inf]]
        self.MinMax = [np.inf, -np.inf]
        
        f = iter(open(self.FileName))
        for l in f:
            if l[:6] == "POINTS":
                count = int(l.split(maxsplit=2)[1])
        
                for i in range(count):
                    line = next(f)
                    
                    if space_time:
                        t,x,y = map(float, line.strip().split(maxsplit=3))
                    else:
                        t = 0
                        x,y,*rest = map(float, line.strip().split(maxsplit=3))
                    
                    if t == 0:
                        self.FirstSlicePoints[i] = (x,y)
                        
                    self.Points[i] = (x,y)
                    
            
            if l[:5] == "CELLS":
                count = int(l.split()[1])
                
                for i in range(count):
                    line = next(f)
                    corners = tuple(map(float, line.split()[1:]))
                    
                    for c in corners:
                        if c in self.FirstSlicePoints:
                            self.FirstSliceCells[i] = self.calcCenter(corners)
        
            if l.strip() == "LOOKUP_TABLE default":
                for i,line in zip(counter(),f):
                    line = float(line.strip())
                    if i in self.FirstSliceCells:
                        self.FirstSliceValues[i] = line

        self.Bounds = [
            [min(p[0] for p in self.FirstSlicePoints.values()), max(p[0] for p in self.FirstSlicePoints.values())],
            [min(p[1] for p in self.FirstSlicePoints.values()), max(p[1] for p in self.FirstSlicePoints.values())],
        ]
                
        self.CellCentersX = set(i[0] for i in self.FirstSliceCells.values())
        self.CellCentersY = set(i[1] for i in self.FirstSliceCells.values())        
        self.Image = np.zeros((len(self.CellCentersY), len(self.CellCentersX)), dtype=float)
        
        for i in self.FirstSliceCells:
            center = self.FirstSliceCells[i]
            index = self.getIndex(center)
            value = self.FirstSliceValues[i]
            self.Image[index[::-1]] = value
            
        if bbox is not None:
            upper_left = self.getIndex([bbox[0], bbox[2]])
            lower_right = self.getIndex([bbox[1], bbox[3]])
            self.Image = self.Image[upper_left[1]:lower_right[1]+1, upper_left[0]:lower_right[0]+1]
        
        self.MinMax = [self.Image.min(), self.Image.max()]

        print("[Min,Max]:", self.MinMax)
        
    
    def __isub__(self, vtk):
        if self.Image.shape != vtk.Image.shape:
            raise ValueError("Dimension mismatch: {} vs {}".format(self.Image.shape, vtk.Image.shape))
        
        self.Image -= vtk.Image
        self.MinMax = [self.Image.min(), self.Image.max()]
        
        print("[Min,Max]:", self.MinMax)
        
        return self
    
        
    def savePlot(self, filename, title="", colorbar=True,show_axis=True, vmin=None, vmax=None):
        if vmin is None: vmin = self.MinMax[0]
        if vmax is None: vmax = self.MinMax[1]
        
        plt.clf()
        plt.cla()
        fig, ax = plt.subplots()
        #plt.locator_params(axis="x",nbins=5)
        #plt.locator_params(axis="y",nbins=5)
        [i.set_linewidth(0.2) for i in plt.gca().spines.values()]

        ax.get_xaxis().set_visible(show_axis)
        ax.get_yaxis().set_visible(show_axis)
        #plt.xticks(rotation=-30)
        #myLocator = mticker.MultipleLocator(5)
        #ax.get_xaxis().set_major_locator(myLocator)
            
        imgplot = plt.imshow(self.Image, interpolation='nearest', vmin=vmin, vmax=vmax)
        imgplot.set_cmap('viridis')
        
        if colorbar:
            #cb = plt.colorbar(imgplot, shrink=.45, pad=.02, aspect=10)
            divider = make_axes_locatable(ax)
            cax = divider.append_axes("top", size="5%", pad=0.1)
            cb = plt.colorbar(imgplot, cax=cax, orientation='horizontal', ticks=[vmin,vmax]) #, norm=mpl.colors.Normalize(vmin=vmax, vmax=vmin))
            cax.xaxis.set_ticks_position("top")
            cb.outline.set_linewidth(0.2)
            
        if title:
            #plt.title(title)
            ax.text(0.05, 0.05, title,
                verticalalignment='bottom', horizontalalignment='left',
                transform=ax.transAxes, color='white', fontsize=14)
            
        fig.set_size_inches(6,4)
        plt.savefig(filename, dpi=300, bbox_inches="tight", pad_inches=0.02)
        
        #plt.plot(x,y, "x")
        #plt.show()
        #print(self.Points)

#plt.contourf(1-img2, origin="image", colors='w', hatches=["xxx", ' '], interpolation="nearest", levels=[-1,0,1])


if __name__ == "__main__":
    import sys
    import argparse
    
    parser = argparse.ArgumentParser(description='Render material parameter from vtk-File')
    parser.add_argument("input_fn", type=str, help="path to input file")
    parser.add_argument("output_fn", type=str, help="path to the output file")
    
    parser.add_argument("-c", action='store_true', help="show color bar (default: 0)")
    parser.add_argument("-x", action='store_true', help="show axis (default: 0)")
    parser.add_argument("-m", metavar="m", type=float, default=None, help="minimal value for color scaling")
    parser.add_argument("-M", type=float, default=None, help="maximal value for color scaling")
    parser.add_argument('-b', metavar=("xmin", "xmax", "ymin", "ymax"), default=None,  nargs=4, help="Strips image to a bounding box")
    parser.add_argument("-r", metavar="ref", type=str, default=None, help="plots input_fn - ref instead of input_fn")
    parser.add_argument("-T", action='store_true', help="interpret file as (2+1)D space-time plot")
    parser.add_argument("-d", metavar="d", type=int, default=4, help="round cell centers to d digits to compensate rounding errors in coordinates")
    parser.add_argument("-p", metavar="p", type=float, default=None, help="reference value")
    
    
    args = parser.parse_args()
    print(args)

    bbox = (None if args.b is None else list(map(float,args.b)))
    vtk = VTKFile(args.input_fn, bbox=bbox, space_time=args.T, rounding_digits=args.d)
    
    if args.r is not None:
        vtk_ref = VTKFile(args.r, bbox=bbox, space_time=args.T, rounding_digits=args.d)
        vtk -= vtk_ref
        
    if args.p is not None:
        vtk *= 1.0 / args.p
    
    vtk.savePlot(args.output_fn, 
                 colorbar=args.c, 
                 show_axis=args.x, 
                 vmin=args.m, vmax=args.M
                 )
    
    #args = parser.parse_args()
    #print(args)

    #bbox = (None if args.b is None else list(map(float,args.b)))
    #vtk = VTKFile(args.input_fn, bbox=bbox, space_time=args.T)
    
    #if args.r is not None:
        #vtk_ref = VTKFile(args.r, bbox=bbox)
        #vtk -= vtk_ref
    
    #vtk.savePlot(args.output_fn, 
                 #colorbar=args.c, 
                 #show_axis=args.x, 
                 #vmin=args.m, vmax=args.M
                 #)
